/*
 * BookmarkPanel.java
 *
 * Bookmark Panel used by ControllerGUI to add/select
 * server resources (URLs) 
 */

package streamredirector.gui;

import streamredirector.*;

import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;

/**
 *
 * @author  Kalpak Kothari
 */
public class BookmarkPanel extends JFrame implements ActionListener {

    // Variables declaration
    private ServerEntry curServerEntry;
    private ResourceEntry curResourceEntry;
    private ControllerGUI controllerGUI;

    // GUI variables
    private JScrollPane treeScrollPane;
    private JTree bookmarkJTree;
    private JScrollPane detailScrollPane;
    private JTextArea detailTextArea;
    private JPanel commandPanel;
    private JButton applyButton;
    private JButton closeButton;

    // choose JTree line styles.  Possible values are
    //"Angled", "Horizontal", and "None" (the default).
    private boolean playWithLineStyle = false;
    private String lineStyle = "Angled";
    
    // variable used for double-click events.  if double-click occurs
    // and nodeIsLeaf==true, we simulate "Apply" button action.
    private boolean nodeIsLeaf = false;

    // End of variables declaration

    /** Creates new form BookmarkPanel */
    public BookmarkPanel() {
	initBookmarks();
        initComponents();
    }

    public BookmarkPanel(ControllerGUI connGUI) {
	controllerGUI = connGUI;
	initBookmarks();
        initComponents();
    }


    /** This method is called from within the constructor to
     * initialize the bookmark database.
     */
    private void initBookmarks() {
	try {
	    ServerDatabaseManager.initialize ("serverDB.dat");
	}
	catch (Exception e) {
	    System.out.println ("Controller::main -> Error creating database manager: " + e);
	}
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     */
    private void initComponents() {
        setTitle("ControllerGUI:: Bookmarks");
        getContentPane().setLayout(new GridBagLayout());

        treeScrollPane = new JScrollPane();	
        detailScrollPane = new JScrollPane();	
        detailTextArea = new JTextArea();
        commandPanel = new JPanel();
        applyButton = new JButton("Apply");
        closeButton = new JButton("Close");

        GridBagConstraints gridBagConstraints5;


	/*** begin bookmarkJTree configuration */

	// Create the nodes
        DefaultMutableTreeNode top = new DefaultMutableTreeNode("URLs");
        createNodes(top);

	// Create a tree that allows one selection at a time
        bookmarkJTree = new JTree(top);
	bookmarkJTree.setFont(new Font ("Dialog", 0, 10));
	bookmarkJTree.getSelectionModel().setSelectionMode
	    (TreeSelectionModel.SINGLE_TREE_SELECTION);

	// Listen for when the selection changes
        bookmarkJTree.addTreeSelectionListener(new TreeSelectionListener() {
		public void valueChanged(TreeSelectionEvent e) {
		    DefaultMutableTreeNode node = (DefaultMutableTreeNode)
			bookmarkJTree.getLastSelectedPathComponent();
		    
		    if (node == null) {
			nodeIsLeaf = false;
			return;
		    }

		    Object nodeInfo = node.getUserObject();
		    if (node.isLeaf()) {
			TreePath path = bookmarkJTree.getSelectionPath();
			System.out.println("Path: " + path);
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)path.getPathComponent(1);
			curServerEntry = ((ServerNode)parent.getUserObject()).serverEntry;
			curResourceEntry = ((ResourceNode)nodeInfo).resourceEntry;
			nodeIsLeaf = true;
			System.out.println("BookmarkPanel::valueChanged() - node is Leaf\n" + curServerEntry + curResourceEntry);
			printBookmarkDetails();
			applyButton.setEnabled(true);
		    } else {
			nodeIsLeaf = false;
			System.out.println("BookmarkPanel::valueChanged() - node not Leaf\n");
		    }
		}
	    });
	
	if (playWithLineStyle) {
            bookmarkJTree.putClientProperty("JTree.lineStyle", lineStyle);
        }


	MouseAdapter treeMouseListener = new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
		    if (e.getClickCount() == 2 && nodeIsLeaf == true) {
			applyButtonActionPerformed();
		    }
		}
	    };
	
	bookmarkJTree.addMouseListener(treeMouseListener);
	
	/* end bookmarkJTree configuration ***/

	// configure JScrollPane for bookmarkJTree
        treeScrollPane.getViewport().setLayout(new FlowLayout());
        treeScrollPane.setBorder(new TitledBorder(null, "Bookmarks", TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));
        treeScrollPane.setViewportBorder(new EtchedBorder());
        treeScrollPane.setToolTipText("Select/Edit Bookmarks");
        treeScrollPane.setAutoscrolls(true);
	treeScrollPane.setMinimumSize(new Dimension(220, 140));
	treeScrollPane.setPreferredSize(new Dimension(220, 140));
        treeScrollPane.getViewport().add(bookmarkJTree);

	// configure detailScrollPane and detailTextArea	
        detailScrollPane.setPreferredSize(new Dimension(220, 110));
        detailScrollPane.setMinimumSize(new Dimension(220, 110));
        detailScrollPane.setBorder(new TitledBorder(null, "Details", TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));
        detailScrollPane.setViewportBorder(new EtchedBorder());
        detailScrollPane.setAutoscrolls(true);
	
        detailTextArea.setLineWrap(true);
	detailTextArea.setWrapStyleWord(true);
        detailTextArea.setEditable(false);
        detailTextArea.setFont(new Font ("Dialog", 0, 10));
        // detailTextArea.setText("bookmark details here");
        detailScrollPane.setViewportView(detailTextArea);

        // configure and add apply/cancel buttons to commandPanel
	applyButton.setToolTipText("Apply selection");
	applyButton.setMnemonic(KeyEvent.VK_A);
	applyButton.setEnabled(false);
	applyButton.setActionCommand("applybutton");
	applyButton.addActionListener(this);

        closeButton.setToolTipText("Close Window");
	closeButton.setMnemonic(KeyEvent.VK_C);
	closeButton.setActionCommand("closebutton");
	closeButton.addActionListener(this);

        commandPanel.add(applyButton);
        commandPanel.add(closeButton);
        
	// add treeScrollPane
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 0;
        gridBagConstraints5.anchor = GridBagConstraints.NORTH;
        getContentPane().add(treeScrollPane, gridBagConstraints5);

	// add detailScrollPane
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 1;
        getContentPane().add(detailScrollPane, gridBagConstraints5);

	// add commandPanel
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 2;
        getContentPane().add(commandPanel, gridBagConstraints5);

	pack();
    }

    private void printBookmarkDetails() {
	detailTextArea.setText("Protocol:" + curServerEntry.getProtocol() + "\n");
	detailTextArea.append("Size:" + 
			      ((curResourceEntry.getSize() == null) ? "Unknown" : curResourceEntry.getSize()) + "\n");
	detailTextArea.append("Description:" + 
			      ((curResourceEntry.getDescription() == null) ? "Unknown" : curResourceEntry.getDescription()) + "\n");
	detailTextArea.append("MIME Type:" + 
			      ((curResourceEntry.getMimeType() == null) ? "Unknown" : curResourceEntry.getMimeType()) + "\n");
    }

    public void actionPerformed(ActionEvent e) {
	// GUI event handler
	if(e.getActionCommand().equals("applybutton")) {
	    applyButtonActionPerformed();
        }
	else if(e.getActionCommand().equals("closebutton")) {
	    this.hide();
	}
    }

    private void applyButtonActionPerformed() {
	// set the server URL in controllerGUI
	String url = "http://" + curServerEntry.getHostName() + 
	    ":" + curServerEntry.getPort() + 
	    curResourceEntry.getName();
	// controllerGUI.setServerURL(url);
	controllerGUI.setServerURLAndConnect(url);
	applyButton.setEnabled(false);
    }

    /** method to create nodes for bookmarkJTree */
    private void createNodes(DefaultMutableTreeNode top) {
        DefaultMutableTreeNode host = null;
        DefaultMutableTreeNode resource = null;
	
	for(Enumeration e = ServerDatabaseManager.getServerList(); e.hasMoreElements(); ) {
	    ServerEntry serverEntry = (ServerEntry)e.nextElement();
	    host = new DefaultMutableTreeNode(new ServerNode(serverEntry));
	    top.add(host);
	    
	    for(Enumeration f = serverEntry.getResourceList().elements(); f.hasMoreElements(); ) {
		ResourceEntry resourceEntry = (ResourceEntry)f.nextElement();
		resource = new DefaultMutableTreeNode(new ResourceNode(resourceEntry));
		host.add(resource);
		System.out.println(resourceEntry);
	    }
	}
    }

    private class ServerNode {
	private String serverName;
	private ServerEntry serverEntry;

	public ServerNode(ServerEntry entry) {
	    serverEntry = entry;
	    serverName = serverEntry.getHostName() + ":" + serverEntry.getPort();
	}
	
	public String toString() {
	    return serverName;
	}
    }

    private class ResourceNode {
	private String resourceName;
	private ResourceEntry resourceEntry;

	public ResourceNode(ResourceEntry entry) {
	    resourceEntry = entry;
	    resourceName = resourceEntry.getName();
	}
	
	public String toString() {
	    return resourceName;
	}
    }

}


