/*
 * ControllerGUI.java
 *
 * Swing GUI frontend to showcase Stream Redirector
 */

package streamredirector.gui;

import streamredirector.*;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;

/**
 *
 * @author  Kalpak Kothari
 */
 
public class ControllerGUI extends JFrame implements ActionListener, LocationUpdateListener
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = true;//Controller.DEBUG;
  
    // Variables declaration

    /** the default virtual space name */
    protected static final String defaultVSpaceName = "wind";
    
    /** the default DSR name */
    protected static final String defaultDSRName = "localhost";

    // Stream Redirector variables
    private final String vspace;
    private final String dsrName;
    private Controller controller;
    private INSTCPSinkProxyDescriptor sinkProxy;
    private INSTCPSinkProxyDescriptor localSinkProxy;
    public INSTCPSinkProxyDescriptor selectedSinkProxy;

    private ServerInfo serverInfo;
    private SinkInfo sinkInfo;
    private Long cookie;
    private ServerCommand resumeCommand;
    
    // GUI Variables
    private JMenuBar mbar;
    private JMenu menuFile;
    private JMenuItem menuItemAbout;
    private JMenuItem menuItemBookmark;
    private JMenuItem menuItemExit;
    private ButtonGroup sinkButtonGroup;
    private JScrollPane statusPanel;
    private JTextArea statusText;
    private JPanel commandPanel;
    private JPanel connectPanel;
    private JPanel serverPanel;
    private JTextField serverURLField;
    private JPanel sinkPanel;
    private JCheckBox currentSinkCheckBox;
    private JButton currentSinkViewButton;
    private JCheckBox otherSinkCheckBox;
    private JButton otherSinkViewButton;
    private JPanel actionPanel;
    private JButton connectButton;
    private JButton resumeButton;
    private JButton moveButton;
    private BookmarkPanel bookmarkPanel = null;
    private SinkWidget sinkWidget = null;
    private boolean checkboxsel = false;
    // End of variables declaration

    /** Creates new form ControllerGUI */
    public ControllerGUI() {
	dsrName = defaultDSRName;
	vspace = defaultVSpaceName;
	initController();
        initComponents();
    }

    public ControllerGUI(String _dsrName, String _vspace) {
	dsrName = _dsrName;
	vspace = _vspace;
	initComponents();
	initController();
    }

    /** This method is called from within the constructor to
     * initialize the Controller.
     */
    private void initController() {
	try {
	    // <HERE> Added by Rafa: this web server is necessary for
	    // dynamic class loading
	    
	    // <HERE> Added by Rafa: this name should be gotten from
	    // the command line
	    String jarfileName = "streamredirector.jar";
	    String serverCodebase = System.getProperty ("java.rmi.server.codebase");
	    
	    // we check if this property has been set in the command
	    // line: if it has, we echo it back, but if it has not, we
	    // start our own webserver and set the codebase to a URL
	    // from where the jarfile containing the classes may be
	    // found

	    if (serverCodebase != null) {
		if (DEBUG)
		    System.out.println ("Property: \"java.rmi.server.codebase\" already set to " + serverCodebase);
	    }
	    else {
		// initialize the local web server: it will be in charge of
		// responding to file requests from the remote sink proxies
		// if classes will be loaded dynamically at run-time; we
		// assume that the jar file "streamredirector.jar" will
		// exist in the current directory
		try {
		    WebServerManager.initialize (".", new String [] {jarfileName});
		}
		catch (Exception e) {
		    System.err.println (e);
		    System.exit (0);
		}
          
		String webServerHost = WebServerManager.getLocalHost();
		int webServerPort = WebServerManager.getLocalPort();
      
		// we set the "java.rmi.server.codebase" property; this
		// causes all object encapsulated by a MarshalledObject to
		// be annotated with this codebase; when a remote JVM needs
		// to dynamically load a class, it will ask for our web
		// server to the "streamredirector.jar" class, which will
		// contain all the classes
    
		System.setProperty ("java.rmi.server.codebase", "http://" + webServerHost + ":" + webServerPort + "/" + jarfileName);
    
		if (DEBUG)
		    System.out.println ("ControllerGUI::initController -> Set codebase property to: " + System.getProperty ("java.rmi.server.codebase"));
	    }
            
	    // start the Controller, passing it a reference to this object
	    // as the listener for location change updates, an object that
	    // interfaces with INS to look for sink proxies and an object
	    // that interfaces with Cricket to figure out the controller's
	    // current location
	    controller = new Controller (this, new INSNetworkResourceManager (dsrName, vspace),
					 new CricketPhysicalLocationManager());
          
	    // starting this class allows us to examine runtime values of
	    // the controller
	    new ClassMonitor (controller);
          
	    // set localSinkProxy to a local sink
	    localSinkProxy = (INSTCPSinkProxyDescriptor) controller.getAnyLocalSink (NetworkResourceManager.REFRESH);
      
      if (DEBUG)
        System.out.println ("ControllerGUI::initController -> FOUND local sink: " + localSinkProxy);
	    // Initialize selectedSinkProxy to local sink.
	    selectedSinkProxy = localSinkProxy;
	}
	catch(Exception e) {
	    if (DEBUG) {
		System.out.println ("ControllerGUI::initController -> Error creating Controller: " + e);
		e.printStackTrace();
	    }
      
	    System.exit(0);
	}
    }

    public void updateLocation (LocationInformation info) {
	statusText.setText("Changed current location to: "  + info.getNewLocation() + "\n");

	currentSinkCheckBox.setToolTipText ("Play resource at current location: "  + info.getNewLocation());
	currentSinkCheckBox.setText ("Local: " + info.getNewLocation());
  
	if (DEBUG)
	    System.out.println ("ControllerGUI::updateLocation -> Notified of location info: " + info);

  localSinkProxy = (INSTCPSinkProxyDescriptor) controller.getAnyLocalSink (NetworkResourceManager.NO_REFRESH);
	sinkProxy = localSinkProxy;
  
	if (sinkProxy == null) {
	    // display warning dialog box
	    /** commented out since this is annoying if cricket updates too frequently
	    JOptionPane.showMessageDialog(this, 
					  "Warning:\n\n" +
					  "No sinks found in current location!\n",
					  "Warning",
					  JOptionPane.WARNING_MESSAGE);
	    */
	    statusText.setText("Warning: No sinks found in current location!!\n");
	    return;
	}
      
	if (DEBUG)
	    System.out.println ("ControllerGUI::updateLocation -> Got local sink: "  + sinkProxy);
    
	// if a file is playing, we must move the connection to a new sink proxy
	if (cookie != null) {
	    sinkInfo = new NullTCPSinkInfo (sinkProxy.getHostName(), sinkProxy.getPort());
  
	    ReturnCode returnCode = ReturnCode.COMMAND_UNINITIALIZED;
        
	    try {
		if (DEBUG)
		    System.out.println ("ControllerGUI::updateLocation -> Moving connection to sink: " + sinkInfo);
        
		// we move the connection to the new sink and remember
		// this new connection's cookie value
		cookie = controller.moveConnection (cookie, sinkInfo, returnCode);

		statusText.append("Moved connection to " + sinkProxy.getLocation() + "\n");

		if (DEBUG)
		    System.out.println ("ControllerGUI::updateLocation -> Return code is: " + returnCode);
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }
    
    /** This method is called from within the constructor to
     *  initialize the GUI.
     */
    private void initComponents() {
	addMenu();

        sinkButtonGroup = new ButtonGroup();
        statusPanel = new JScrollPane();
        statusText = new JTextArea();
        commandPanel = new JPanel();
        connectPanel = new JPanel();
        serverPanel = new JPanel();
        serverURLField = new JTextField();
        sinkPanel = new JPanel();
        currentSinkCheckBox = new JCheckBox();
        currentSinkViewButton = new JButton("Info...");
        otherSinkCheckBox = new JCheckBox();
        otherSinkViewButton = new JButton("Select...");
        actionPanel = new JPanel();
        connectButton = new JButton("Connect");
        resumeButton = new JButton("Resume");
        moveButton = new JButton("Move");
        
        getContentPane().setLayout(new BorderLayout());
        
        setTitle("ControllerGUI");
        addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent evt) {
		    exitForm(evt);
		}
	    });
        
        statusPanel.setViewportBorder(new TitledBorder("Status"));
        statusPanel.setPreferredSize(new Dimension(200, 100));
        statusPanel.setMinimumSize(new Dimension(200, 100));
        statusText.setLineWrap(true);
	statusText.setWrapStyleWord(true);
        statusText.setEditable(false);
	statusText.setFont(new Font ("Dialog", 0, 10));
	statusText.setBackground(new Color(204,204,204));
        statusPanel.setViewportView(statusText);
        
        getContentPane().add(statusPanel, BorderLayout.SOUTH);
        
        commandPanel.setBorder(new EtchedBorder());
        connectPanel.setLayout(new GridBagLayout());
        GridBagConstraints gridBagConstraints1;
        
        connectPanel.setBorder(new MatteBorder(null));
        connectPanel.setOpaque(true);
        serverPanel.setLayout(new BorderLayout());
        
        serverPanel.setBorder(new TitledBorder("Server Resource (URL)"));
        
	serverURLField.setFont(new Font ("Dialog", 0, 10));
        serverURLField.setText("http://india.mit.edu/downloads/foo.mp3");
        serverURLField.setPreferredSize(new Dimension(195, 25));
        serverPanel.add(serverURLField, BorderLayout.CENTER);
        
        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 0;
        gridBagConstraints1.anchor = GridBagConstraints.NORTH;
        connectPanel.add(serverPanel, gridBagConstraints1);
        
        sinkPanel.setLayout(new GridBagLayout());
        sinkPanel.setBorder(new TitledBorder("Sink"));

	currentSinkCheckBox.setPreferredSize(new Dimension(135, 24));
        currentSinkCheckBox.setFont(new Font ("Dialog", 0, 10));
        currentSinkCheckBox.setToolTipText("Play resource at current location: " +
					   ((localSinkProxy == null) ? "none" : localSinkProxy.getLocation()));
        currentSinkCheckBox.setSelected(true);
        currentSinkCheckBox.setText("Local: " + 
				    ((localSinkProxy == null) ? "none" : localSinkProxy.getLocation()));
	currentSinkCheckBox.setActionCommand("currentsinkcheckbox");
	currentSinkCheckBox.addActionListener(this);

        sinkButtonGroup.add(currentSinkCheckBox);

        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 0;
	gridBagConstraints1.fill = GridBagConstraints.BOTH;
        sinkPanel.add(currentSinkCheckBox, gridBagConstraints1);

	currentSinkViewButton.setPreferredSize(new Dimension(60, 24));
        currentSinkViewButton.setFont(new Font ("Dialog", 0, 10));
	currentSinkViewButton.setMargin(new Insets(2,1,2,1));
        currentSinkViewButton.setToolTipText("View current sink information");
        currentSinkViewButton.setVerticalAlignment(SwingConstants.BOTTOM);
	currentSinkViewButton.setMnemonic(KeyEvent.VK_I);
	currentSinkViewButton.setActionCommand("currentsinkbutton");
	currentSinkViewButton.addActionListener(this);

        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 1;
        gridBagConstraints1.gridy = 0;
	gridBagConstraints1.fill = GridBagConstraints.BOTH;
	sinkPanel.add(currentSinkViewButton);

	otherSinkCheckBox.setPreferredSize(new Dimension(135, 24));
        otherSinkCheckBox.setFont(new Font ("Dialog", 0, 10));
        otherSinkCheckBox.setToolTipText("Play resource at location: " + 
					 ((selectedSinkProxy == null) ? "none" : selectedSinkProxy.getLocation()));
        otherSinkCheckBox.setText("Other: " + 
				  ((selectedSinkProxy == null) ? "none" : selectedSinkProxy.getLocation()));
	otherSinkCheckBox.setActionCommand("othersinkcheckbox");
	otherSinkCheckBox.addActionListener(this);
  
        sinkButtonGroup.add(otherSinkCheckBox);

        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 1;
	gridBagConstraints1.fill = GridBagConstraints.BOTH;
        sinkPanel.add(otherSinkCheckBox, gridBagConstraints1);

	otherSinkViewButton.setPreferredSize(new Dimension(60, 24));
        otherSinkViewButton.setFont(new Font ("Dialog", 0, 10));
	otherSinkViewButton.setMargin(new Insets(2,1,2,1));
        otherSinkViewButton.setToolTipText("Choose desired sink");
	otherSinkViewButton.setMnemonic(KeyEvent.VK_S);
	otherSinkViewButton.setActionCommand("othersinkbutton");
	otherSinkViewButton.addActionListener(this);

        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 1;
        gridBagConstraints1.gridy = 1;
	gridBagConstraints1.fill = GridBagConstraints.BOTH;
        sinkPanel.add(otherSinkViewButton, gridBagConstraints1);
        
        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 1;
        gridBagConstraints1.fill = GridBagConstraints.BOTH;
        gridBagConstraints1.anchor = GridBagConstraints.SOUTH;
        connectPanel.add(sinkPanel, gridBagConstraints1);
        
        connectButton.setFont(new Font ("Dialog", 0, 12));
        connectButton.setPreferredSize(new Dimension(70, 25));
	connectButton.setMargin(new Insets(2,2,2,2));
        connectButton.setToolTipText("Connect server resource to sink");
	connectButton.setMnemonic(KeyEvent.VK_C);
	connectButton.setActionCommand("connectbutton");
	connectButton.addActionListener(this);
        actionPanel.add(connectButton);

	resumeButton.setFont(new Font ("Dialog", 0, 12));	
        resumeButton.setPreferredSize(new Dimension(60, 25));
	resumeButton.setMargin(new Insets(2,2,2,2));
        resumeButton.setToolTipText("Resume existing connection");
	resumeButton.setMnemonic(KeyEvent.VK_R);
	resumeButton.setActionCommand("resumebutton");
	resumeButton.setEnabled(false);
	resumeButton.addActionListener(this);
	actionPanel.add(resumeButton);
	
	moveButton.setFont(new Font ("Dialog", 0, 12));
        moveButton.setPreferredSize(new Dimension(55, 25));
	moveButton.setMargin(new Insets(2,2,2,2));
        moveButton.setToolTipText("Move resource from current sink to selected sink");
	moveButton.setMnemonic(KeyEvent.VK_M);
	moveButton.setActionCommand("movebutton");
	moveButton.setEnabled(false);
	moveButton.addActionListener(this);
        actionPanel.add(moveButton);

        gridBagConstraints1 = new GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 2;
        connectPanel.add(actionPanel, gridBagConstraints1);

        commandPanel.add(connectPanel);
        
        getContentPane().add(commandPanel, BorderLayout.NORTH);
        
        pack();
    }

    /** Exit the Application */
    private void exitForm(WindowEvent evt) {
        System.exit(0);
    }

    /* Create and add menu to GUI */
    private void addMenu() {
        mbar = new JMenuBar();

        menuFile = new JMenu("File");
	menuFile.setMnemonic(KeyEvent.VK_F);

        menuItemAbout = new JMenuItem("About", KeyEvent.VK_A);
	menuItemAbout.setActionCommand("itemabout");
	menuItemAbout.addActionListener(this);
	
        menuItemBookmark = new JMenuItem("Bookmarks", KeyEvent.VK_B);
        menuItemBookmark.setToolTipText("Add/Edit Resource List");
        menuItemBookmark.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK));
	menuItemBookmark.setActionCommand("itembookmark");
	menuItemBookmark.addActionListener(this);
  
	menuItemExit = new JMenuItem("Exit", KeyEvent.VK_X);
        menuItemExit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK));
	menuItemExit.setActionCommand("itemexit");
	menuItemExit.addActionListener(this);
  
        menuFile.add(menuItemAbout);
        menuFile.add(menuItemBookmark);
        menuFile.add(menuItemExit);
        mbar.add(menuFile);
	setJMenuBar(mbar);
    }

    private void connectButtonActionCommand() {
	// initiate connection between server and sink
	try {
	    // Added by Kalpak: assign <sinkProxy> to local proxy or
	    // selection from SinkWidget based on checkbox selection
	    if(sinkWidget == null)
		selectedSinkProxy = localSinkProxy;
	    else
		selectedSinkProxy = sinkWidget.getSelectedSinkProxy();
	    sinkProxy = (checkboxsel ? selectedSinkProxy : localSinkProxy);
	    
	    if (sinkProxy == null && connectButton.getText().equalsIgnoreCase ("connect")) {
		statusText.append("No sink proxy selected\n");
		return;
	    }
      
	    if (DEBUG)
		System.out.println ("ControllerGUI::connectButtonActionCommand -> checkboxsel=" + checkboxsel + "; Got sink: "  + sinkProxy);

	    
	    // should only be used for now, for debugging purposes
	    //	serverInfo = new WebProxyServerInfo("beacon.lcs.mit.edu", 80, 
	    //		    new ResourceEntry ("/speech.mp3", "audio/mpeg", null, null));

	    serverInfo = URL2ServerInfo(serverURLField.getText());
	    if(serverInfo == null) {
		return;
	    }
	    // serverInfo = new WebProxyServerInfo("rafaeln.ne.mediaone.net", 80, 
	    // new ResourceEntry ("/~secret/music.mpeg", "video/mpeg", null, null));

	    sinkInfo = new NullTCPSinkInfo(sinkProxy.getHostName(), sinkProxy.getPort());

	    if(DEBUG) {
		// statusText.append("Server: " + serverInfo + "\n");
		// statusText.append("Sink: " + sinkInfo + "\n");
		System.out.println("ControllerGUI::connectButtonActionCommand: \n\t" + 
				   "Server: " + serverInfo + "\n" +
				   "Sink: " + sinkInfo + "\n");
	    }

	    // <HERE> Added by Rafa: close previous connection before
	    // starting new one so it toggles between establishing a
	    // connection or closing it
	    if (cookie != null) {
		resumeCommand = controller.endConnection(cookie);
		
		if (DEBUG)
		    System.out.println ("ControllerGUI::connectButtonActionCommand -> Ended command; got resume command: " + resumeCommand + " with old cookie: " + cookie);
	  
		cookie = null;

		statusText.append("Stopped previous connection.\n");

		connectButton.setText("Connect");
		connectButton.setToolTipText("Connect server resource to sink");
		connectButton.setMnemonic(KeyEvent.VK_C);
		moveButton.setEnabled(false);
		
		// enable resume button
		resumeButton.setEnabled(true);
		return;
	    }
	    else {
		if (DEBUG)
		    System.out.println ("ControllerGUI::connectButtonActionCommand -> Starting new connection");
		cookie = controller.establishConnection(serverInfo, sinkInfo);

// <HERE> Kalpak, this is one of the lines you have to change...
//    controller.sendCommand(cookie, 
//               new HTTPGetCommand(serverInfo.getSelectedResource().getName(), serverInfo.getSelectedResource().getMimeType()));
//
		statusText.append("Started new connection.\n");

		connectButton.setText("Stop");
		connectButton.setToolTipText("Stop current server to sink connection");
		connectButton.setMnemonic(KeyEvent.VK_T);
		moveButton.setEnabled(true);

		// disable resume button
		resumeButton.setEnabled(false);
	    }
	}
	catch(Exception e) {
	    statusText.append("ERROR: Connection failed. See stdout for details.\n");
	    e.printStackTrace();
	}
    }

    private void resumeButtonActionCommand() {
	// resume existing connection between server and sink
	try {
	    // assign <sinkProxy> to local proxy or selection from
	    // SinkWidget based on checkbox selection
	    if(sinkWidget == null)
		selectedSinkProxy = localSinkProxy;
	    else
		selectedSinkProxy = sinkWidget.getSelectedSinkProxy();
	    sinkProxy = (checkboxsel ? selectedSinkProxy : localSinkProxy);
	    
	    if (sinkProxy == null) {
		statusText.append("No sink proxy selected\n");
		return;
	    }
      
	    if (DEBUG)
		System.out.println ("ControllerGUI::resumeButtonActionCommand -> checkboxsel=" + 
				    checkboxsel + "; Got sink: "  + sinkProxy);

	    serverInfo = URL2ServerInfo(serverURLField.getText());
	    if(serverInfo == null) {
        if (DEBUG)
      System.out.println ("ControllerGUI::resumeButtonActionCommand -> serverInfo is null when url field is: " + serverURLField);
    		return;
	    }

	    sinkInfo = new NullTCPSinkInfo(sinkProxy.getHostName(), sinkProxy.getPort());

	    if(DEBUG) {
		// statusText.append("Server: " + serverInfo + "\n");
		// statusText.append("Sink: " + sinkInfo + "\n");
		System.out.println("ControllerGUI::resumeButtonActionCommand: \n\t" + 
				   "Server: " + serverInfo + "\n" +
				   "Sink: " + sinkInfo + "\n");
	    }
      
	    // <HERE> Added by Rafa: close previous connection before
	    // starting new one so it toggles between establishing a
	    // connection or closing it
	    if (resumeCommand != null) {
		cookie = controller.establishConnection(serverInfo, sinkInfo);
    
		if (DEBUG)
		    System.out.println ("ControllerGUI::resumeButtonActionCommand -> Resumed connection with command: " + resumeCommand + " and new cookie: " + cookie);
		
		controller.sendCommand(cookie, resumeCommand);

		statusText.append("Resumed previous connection.\n");

		connectButton.setText("Stop");
		connectButton.setToolTipText("Stop current server to sink connection");
		connectButton.setMnemonic(KeyEvent.VK_T);
		moveButton.setEnabled(true);
		// disable resume button
		resumeButton.setEnabled(false);
		resumeCommand = null;
		return;
	    }
	}
	catch(Exception e) {
	    statusText.append("ERROR: Resume failed. See stdout for details.\n");
	    e.printStackTrace();
	}
    }

    private void moveButtonActionCommand() {
	// move connection from current sink to selected sink
	if(sinkWidget == null)
	    selectedSinkProxy = localSinkProxy;
	else
	    selectedSinkProxy = sinkWidget.getSelectedSinkProxy();
	sinkProxy = (checkboxsel ? selectedSinkProxy : localSinkProxy);
  
	if (DEBUG)
	    System.out.println ("ControllerGUI::moveButtonActionCommand -> checkboxsel=" + checkboxsel + "; Got sink: "  + sinkProxy);

	statusText.append("Moving connection to: " + sinkProxy.getLocation() + "\n");
	if (sinkProxy == null) {
	    statusText.append("No sink proxy selected\n");
	    return;
	}
      
	ReturnCode returnCode = ReturnCode.COMMAND_UNINITIALIZED;
	try {
	    sinkInfo = new NullTCPSinkInfo(sinkProxy.getHostName(), sinkProxy.getPort());
	    cookie = controller.moveConnection(cookie, sinkInfo, returnCode);
	    if(returnCode == ReturnCode.COMMAND_ERROR) {
		statusText.append("Move Failed.\n"); 
	    }
	    else if(returnCode == ReturnCode.COMMAND_OK) {
		statusText.append("Move Succeeded.\n"); 
		connectButton.setText("Stop");
		connectButton.setToolTipText("Stop current server to sink connection");
		connectButton.setMnemonic(KeyEvent.VK_T);
	    }
	    else if(returnCode == ReturnCode.COMMAND_COULD_NOT_BE_RESUMED) { 
		statusText.append("Move failed - Stream could not be resumed since the original connection has already progressed to completion\n");
	    }
	}
	catch(Exception e) {
	    statusText.append("ERROR: Move connection failed. See stdout for details.\n");
	    e.printStackTrace();
	}
    }
    
    public void actionPerformed(ActionEvent e) {
	// GUI event handler
	if(e.getActionCommand().equals("connectbutton")) {
	    if(connectButton.getText().equalsIgnoreCase("connect"))
		statusText.setText("Trying to initiate connection...\n");
	    else
		statusText.setText("Trying to stop connection...\n");
	    connectButtonActionCommand();
        }
	else if(e.getActionCommand().equals("resumebutton")) {
	    statusText.setText("Trying to resume connection...\n");
	    resumeButtonActionCommand();
	}
	else if(e.getActionCommand().equals("movebutton")) {
	    statusText.setText("Trying to move connection...\n");
	    moveButtonActionCommand();
	}
	else if(e.getActionCommand().equals("currentsinkcheckbox")) {
	    checkboxsel = false;
	}
	else if(e.getActionCommand().equals("othersinkcheckbox")) {
	    checkboxsel = true;
	}
	else if(e.getActionCommand().equals("currentsinkbutton")) {
	    // statusText.setText("Current Sink View...\n");
	    if(sinkWidget == null)
		sinkWidget = new SinkWidget(controller, otherSinkCheckBox);
	    // sinkWidget.setSelectedSinkProxy(localSinkProxy);
	    sinkWidget.show();
	}
	else if(e.getActionCommand().equals("othersinkbutton")) {
	    // statusText.setText("View all sinks...\n");
	    if(sinkWidget == null)
		sinkWidget = new SinkWidget(controller, otherSinkCheckBox);
	    sinkWidget.show();
	}
	else if(e.getActionCommand().equals("itembookmark")) {
	    if(bookmarkPanel == null)
		bookmarkPanel = new BookmarkPanel(this);
	    bookmarkPanel.show();
        }
        else if(e.getActionCommand().equals("itemabout")) {
            popupAboutBox();
        }
        else if(e.getActionCommand().equals("itemexit")) {
            System.exit(0); 
        }
    }
    
    /** public method to set the current server URL */
    public void setServerURL(String s) {
	serverURLField.setText(s);
	if (DEBUG)
	    System.out.println("ControllerGUI::serverURLField : " + s);
	
	// added by Rafa: null out the resume command
	//cookie = null;
	//resumeCommand = null;

	// added by Kalpak; end previous connection, then null out
	// cookie and resumeCommand
	if(cookie != null) {
	    try {
		controller.endConnection(cookie);
		statusText.setText("Stopped previous connection.\n");
	    }
	    catch(Exception e) {
		statusText.append("ERROR: Could not stop previous connection. See stdout for details.\n");
		e.printStackTrace();
	    }
	}
		
	if (DEBUG)
	    System.out.println ("ControllerGUI::setServerURL -> Ended command;  with old cookie: " + cookie);
	
	cookie = null;
	resumeCommand = null;
    }

    /** public method to set the current server URL and initiate connection */
    public void setServerURLAndConnect(String s) {
	setServerURL(s);
	connectButtonActionCommand();
    }

    /** Convert URL string to ServerInfo */
    private ServerInfo URL2ServerInfo(String s) {
	URL url;
	ServerInfo serverInfo = null;
	String ext = "mp3";		// assume default extension: mp3
	String mime = "audio/mpeg";	// assume default mimetype : audio/mpeg
	
	try {
	    url = new URL(s);
	}
	catch(MalformedURLException e) {
	    statusText.append("ERROR: Malformed URL. See stdout for details.\n");
	    e.printStackTrace();
	    return null;
	}
	// form MIMEType string based on filename extension 
	int i = s.lastIndexOf(".");
	if(i != -1)
	    ext = s.substring(i+1);
	
	if(ext.equals("mpg") || ext.equals("mpeg"))
	    mime = "video/mpeg";
  else if(ext.equals("avi"))
      mime = "video/x-msvideo";
	else if(!(ext.equals("mp3"))) {
	    statusText.append("ERROR: Unknown file extension.\n");
	    return null;
	}
	
	try {
	    // if no port specified, assume default: 80
//      serverInfo = new WebProxyServerInfo(url.getHost(), 
//            ((url.getPort() == -1) ? 80 : url.getPort()),
//            // JDK1.3 // new ResourceEntry(url.getPath(), 
//            new ResourceEntry(url.getFile(), 
//                  mime, null, null));
    
      // Added by Rafa: now using NativePlayer
      serverInfo = new NativePlayerServerInfo(url.getHost(), 
            ((url.getPort() == -1) ? 80 : url.getPort()),
            // JDK1.3 // new ResourceEntry(url.getPath(), 
            new ResourceEntry(url.getFile(), 
                  mime, null, null));
	}
	catch(NumberFormatException ex) {
	    statusText.append("ERROR: Port text field does not contain valid number.  See stdout for details.\n");
	    ex.printStackTrace();
	    return null;
	}
	return serverInfo;
    }

    private void popupAboutBox() {
        JOptionPane.showMessageDialog(this, 
				      "Stream Redirector Controller\n" +
				      "(C) 2001, MIT LCS\n\n" +
				      "Authors:\n" +
				      "Kalpak Kothari\n   <kalpak@mit.edu>\n" +
				      "Jorge Rafael Nogueras\n   <rafaeln@wind.lcs.mit.edu>\n",
				      "About",
				      JOptionPane.INFORMATION_MESSAGE);
    }

    protected static void printUsage() {
      System.err.println ("Usage:\n\n" +
          "streamredirector.gui.ControllerGUI -v <vspace name> -d <DSR name>\n\n" +
          "\twhere:\n" +
          "\t\t<vspace name> is name of the INS virtual space where all the nodes are\n" +
          "\t\t\t(default value: \"" + defaultVSpaceName + "\")\n" +
          "\t\t<DSR name> is name of the host where the INS' Domain Space Resolver is running\n" +
          "\t\t\t(default value: \"" + defaultDSRName + "\")\n");
    }


    /**
     * @param args the command line arguments
     */
    public static void main(String args[]) {
	String arg_vspaceName = null;
	String arg_DSRName = null;
      
	for (int i = 0; i < args.length; i++) {
	    try {
		if (args [i].equals ("-v")) {
		    if (DEBUG)
			System.out.println ("Read vspace command line parameter: " + args [i+1]);
		    arg_vspaceName = args [++i];
		}
		else if (args [i].equals ("-d")) {
		    if (DEBUG)
			System.out.println ("Read DSR command line parameter: " + args [i+1]);
		    arg_DSRName = args [++i];
		}
	    }
	    catch (Exception e) {
		e.printStackTrace();
		printUsage();
	    }
	}
    
	// if no virtual space name is specified, the default value will be used
	String vspace = (arg_vspaceName != null)? arg_vspaceName: defaultVSpaceName;
      
	// if no DSR is specified, the default will be used
	String dsrName = (arg_DSRName != null)? arg_DSRName: defaultDSRName;

	new ControllerGUI(dsrName, vspace).show();
    }


}
