/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package beaconconfig;

import gnu.getopt.*;
import java.awt.*;
import javax.swing.border.*;
import javax.swing.*;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;
import cricketdaemon.clientlib.sim.*;
import java.io.*;
import java.util.*;

public class BeaconConfig implements Callback
{
    public static Random r = new Random(0);

    SolverPanel solver;
    static int numBeacons = 1;
    static boolean remoteEnable = false;

    // window dimensions in pixels
    private static final int WIN_SIZE     = 600; // always a square window
    private static final int H_WIN_MARGIN = 10;
    //private static final int V_WIN_MARGIN = 151;
    private static final int V_WIN_MARGIN = 129;
    
    // coordinate system params
    private static final int COORD_WIDTH = 300;
    private static final int GRID_STEP = 61; // 61cm is about 2 feet

    static ServerBroker server;

    SketchPanel sketchPanel;
    Controls controlPanel;

    public BeaconConfig(ServerBroker s,
			int winSize, int coordWidth, int gridStep,
			int numBeacons, int closeness, int stddev, 
			int enoughsamples, 
			int streamerSamplingPeriod,
			int streamerDelay)
    {
	server = s;

	//CustomFrame f = new CustomFrame("Beacon Configuration", 410, 513);
	CustomFrame f = new CustomFrame("Beacon Configuration", 
					winSize+H_WIN_MARGIN, winSize+V_WIN_MARGIN);
	Container container = f.getContentPane();
	BorderLayout layout = new BorderLayout();
	container.setLayout(layout);
	f.setBackground(Color.white);

	DrawingArea d = new DrawingArea(winSize, coordWidth, gridStep);
	d.setBorder(createBorder());
	d.setStreamerSamplingPeriod(streamerSamplingPeriod);
	d.setStreamerDelay(streamerDelay);
	container.add(d, "Center");
	
	if(remoteEnable) {
	    NetworkServer ns = new NetworkServer();
	    solver = new RemoteSolverPanel(this, d, closeness, stddev, enoughsamples, ns);
	    sketchPanel = new RemoteSketchPanel(solver, ns, false);
	    controlPanel = new RemoteControls(d, solver, ns, false);
	    ns.addActionListener(sketchPanel);
	    ns.addActionListener(controlPanel);
	    ns.start();
	}
	else {
	    solver = new SolverPanel(this, d, closeness, stddev, enoughsamples);
	    sketchPanel = new SketchPanel(solver);
	    controlPanel = new Controls(d, solver);
	}

	solver.setBorder(createBorder());
	container.add(solver, "South");
	Controls.getResetButton().addActionListener(sketchPanel);

	JPanel cPanel = new JPanel(new BorderLayout(), true);
	cPanel.add(controlPanel, "North");
	cPanel.add(sketchPanel, "South");
	cPanel.setBorder(createBorder());
	container.add(cPanel, "North");

	d.initialize();

	f.validate();
	f.pack();
	f.setVisible(true);

	this.numBeacons = numBeacons;

    }

    public static Border createBorder() {
	Border raisedbevel = BorderFactory.createRaisedBevelBorder();
	Border loweredbevel = BorderFactory.createLoweredBevelBorder();
	
	return
	    BorderFactory.createCompoundBorder(raisedbevel, loweredbevel);
    }


    synchronized public void callback(CricketData data, BitSet mask)
    {
	if(mask.get(Broker.BEACONSHEARDLIST)) {
	    solver.processDistStats(data.getBeaconsHeard());
	}
	if(mask.get(Broker.DEVICECOORD) || mask.get(Broker.POSERR)) {
	    solver.processDeviceCoord(data.getDevicePosition());
	}
    }

    public static void usage()
    {
	System.err.println("Usage: java beaconconfig.BeaconConfig [-n numBeacons][-c closeness][-s stddev_thresh]\n"+
			   "                                      [-g gridStep][-r][-W winSize][-w maxcoord]\n"+
			   "                                      [-e enoughsamples][-d streamerDelay]\n"+
			   "                                      [-t streamerSamplingPeriod]\n");
	System.exit(1);
    }

    public static void main(String args[]) {
	int enoughsamples = 0; // disabled
	int numBeacons = 3;
	int closeness  = 100; // in cm
	int stddev     = 10;  // in cm
	int winSize    = WIN_SIZE;
	int coordWidth = COORD_WIDTH;
	int gridStep   = GRID_STEP;
	int streamerSamplingPeriod = 200;
	int streamerDelay = 3000;


	Getopt g = new Getopt("BeaconConfig", args, "d:c:e:g:n:rs:t:W:w:");

	int c;
	try {
	    while((c = g.getopt()) != -1) {
		switch(c) {
		case 'c':
		    closeness = Integer.parseInt(g.getOptarg());
		    break;
		case 'd':
		    streamerDelay = Integer.parseInt(g.getOptarg());
		    break;
		case 'e':
		    enoughsamples = Integer.parseInt(g.getOptarg());
		    break;
		case 'g':
		    gridStep   = Integer.parseInt(g.getOptarg());
		    break;
		case 'n':
		    numBeacons = Integer.parseInt(g.getOptarg());
		    break;
		case 'r':
		    System.err.println("Remote Enabled:  If GUI hangs, check network "+
				       "connectivity\nand restart application\n");
		    remoteEnable = true;
		    break;
		case 's':
		    stddev = Integer.parseInt(g.getOptarg());
		    break;
		case 't':
		    streamerSamplingPeriod = Integer.parseInt(g.getOptarg());
		    break;
		case 'W':
		    winSize = Integer.parseInt(g.getOptarg());
		    break;
		case 'w':
		    coordWidth = Integer.parseInt(g.getOptarg());
		    break;
		default:
		    usage();
		}
	    }
	}
	catch (Exception e) {
	    usage();
	}
	
	Util.assert(numBeacons >= 3, "must specify numBeacons >= 3");

	ServerBroker server = new ServerBroker();	
	BeaconConfig bc = new BeaconConfig(server, winSize, coordWidth, gridStep,
					   numBeacons, closeness, stddev, 
					   enoughsamples,  
					   streamerSamplingPeriod,
					   streamerDelay);


	BitSet mask = new BitSet();
	mask.set(Broker.BEACONSHEARDLIST);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.POSERR);
	server.register(bc, mask);
	server.start();

	try { server.join(); } catch (InterruptedException ie) { }
    }
}
