/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package beaconconfig;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

class Controls extends JPanel 
    implements ActionListener
{
    private DrawingArea drawArea;
    private SolverControl solver;

    private static JButton reset;
    protected JCheckBox manualOverride;
    protected JCheckBox aggressive;

    protected JCheckBox grid;
    protected JCheckBox annulus;
    protected JCheckBox streamer;

    //JButton addBeacon;

    public Controls(DrawingArea d, SolverControl s)
    {
	setLayout(new GridLayout(2, 3));
	((GridLayout)getLayout()).setHgap(10);

	drawArea = d;
	solver   = s;
	reset = addButton("Reset");
	manualOverride = createCheckBox("Override", false, false);
	aggressive     = createCheckBox("Aggressive", false, false);

	grid     = createCheckBox("Grid", false, true);
	annulus  = createCheckBox("Annulus", false, false);
	streamer = createCheckBox("Streamer", false, false);
    }

    public static JButton getResetButton()
    {
	return reset;
    }

    private JButton addButton(String name)
    {
	JButton result = new JButton(name);
	result.setHorizontalAlignment(JButton.CENTER);
	result.setActionCommand(result.getText());
	result.addActionListener(this);
	add(result);	
	return result;
    }

    public JCheckBox createCheckBox(String label, boolean check, boolean enable)
    {
	JCheckBox result = new JCheckBox(label, check);
	result.setEnabled(enable);
	result.addActionListener(this);
	add(result);
	return result;
    }

    public void enableManualOverride(boolean v)
    {
	if(v != manualOverride.isEnabled())
	    manualOverride.setEnabled(v);
    }

    public void enableTrackingOptions(boolean v)
    {
	if(v != aggressive.isEnabled())
	    aggressive.setEnabled(v);
	if(v != annulus.isEnabled())
	    annulus.setEnabled(v);
	if(v != streamer.isEnabled())
	    streamer.setEnabled(v);
    }

    public void actionPerformed(ActionEvent e) 
    {
	Util.dbgPrintln("Controls::actionPerformed "+e);

	if(e.getActionCommand().equals(reset.getText())) {
	    resetEverything();
	    if(solver != null)
		solver.reset();
	}
	else if(e.getActionCommand().equals(grid.getText())) {
	    if(e.getSource() == null) { // event is from remote
		grid.setSelected(!grid.isSelected());
	    }
	    if(drawArea != null) {
		drawArea.setGrid(grid.isSelected());
	    }
	}
	else if(e.getActionCommand().equals(manualOverride.getText())) {
	    if(e.getSource() == null) // event is from remote
		manualOverride.setSelected(!manualOverride.isSelected());
	    if(solver != null && manualOverride.isSelected()) {
		//Util.dbgPrintln("force tracking!!!!");
		solver.forceTrackingMode();
	    }
	    if(drawArea != null)
		drawArea.setManualOverride(manualOverride.isSelected());
	}
	else if(e.getActionCommand().equals(annulus.getText())) {
	    if(e.getSource() == null) // event is from remote
		annulus.setSelected(!annulus.isSelected());

	    if(drawArea != null)
		drawArea.setAnnulus(annulus.isSelected());

	}
	else if(e.getActionCommand().equals(aggressive.getText())) {
	    if(e.getSource() == null) // event is from remote
		aggressive.setSelected(!aggressive.isSelected());

	    if(solver != null)
		solver.setAggressive(aggressive.isSelected());
	}
	else if(e.getActionCommand().equals(streamer.getText())) {
	    if(e.getSource() == null) // event is from remote
		streamer.setSelected(!streamer.isSelected());

	    if(drawArea != null)
		drawArea.setStreamer(streamer.isSelected());

	}
    }

    public void resetEverything()
    {
	if(drawArea != null) {
	    drawArea.initialize();
	    drawArea.setGrid(false);
	    drawArea.setAnnulus(false);
	    drawArea.setStreamer(false);
	    drawArea.setManualOverride(false);
	}

	manualOverride.setSelected(false);
	aggressive.setSelected(false);
	grid.setSelected(false);	
	annulus.setSelected(false);
	streamer.setSelected(false);

	grid.setEnabled(true);	
	enableManualOverride(false);
	enableTrackingOptions(false);
    }

    public ArrayList getStates()
    {
	ArrayList result = new ArrayList();
	result.add(new Boolean(manualOverride.isSelected()));
	result.add(new Boolean(manualOverride.isEnabled()));
	result.add(new Boolean(aggressive.isSelected()));
	result.add(new Boolean(aggressive.isEnabled()));
	result.add(new Boolean(grid.isSelected()));
	result.add(new Boolean(grid.isEnabled()));
	result.add(new Boolean(annulus.isSelected()));
	result.add(new Boolean(annulus.isEnabled()));
	result.add(new Boolean(streamer.isSelected()));
	result.add(new Boolean(streamer.isEnabled()));
	return result;
    }
}

