/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package beaconconfig;

import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.util.*;

/**
 *
 */
class NetworkServer 
    extends Thread implements RemoteSyncListener
{
    public static final int PORT = 5002;
    public static final String REGISTER_TOKEN = "BeaconConfig";
    
    private ServerSocket s;

    private ArrayList clients;
    private ArrayList actionListeners;

    public void addActionListener(ActionListener a) 
    {
	actionListeners.add(a);
    }

    private class ClientRecord
    {
	Socket s;
	ObjectInputStream ois;
	InputStream is;
	ObjectOutputStream oos;
	OutputStream os;

	public ClientRecord(Socket s)
	    throws IOException
	{
	    this.s = s;

	    s.setTcpNoDelay(true);
	    is = s.getInputStream();
	    ois = new ObjectInputStream(is);
	    os = s.getOutputStream();
	    oos = new ObjectOutputStream(os);
	}

	public void close()
	{
	    try {
		is.close();
		os.close();
	    }
	    catch(IOException e) {
		Util.dbgPrintln("NetworkServer::ClientRecord::close error");
	    }
	}
    }

    NetworkServer() 
    {
	clients = new ArrayList();
	actionListeners = new ArrayList();

	latestStates = new RemoteSyncEvent();

	try {
	    s = new ServerSocket(PORT);
	    s.setSoTimeout(50); // disable blocking
	}
	catch (IOException e) {
	    Util.assert(false, "error opening port "+PORT);
	}
    }
    
    public void run()
    {
	//this should just register each client
	byte[] buf = new byte[64];
	while(true) {
	    try {
		Socket newSock = s.accept(); // should block
		Util.dbgPrintln("accept");
		synchronized(this) {		    
		    clients.add(new ClientRecord(newSock));
		    transmitEvent(latestStates);
		}
		Util.dbgPrintln("done");
	    }
	    catch (InterruptedIOException iio) {
		pollClientsForEvents();
	    }
	    catch (IOException io) {
		io.printStackTrace();
		Util.assert(false,"ServerThread::run() ioexception");
	    }
	}
	//s.close(); never reached 
    }

    /* FIXME: RemoteSyncEvent needs to have states from BOTH SketchPanel and
       Controls */
    private RemoteSyncEvent latestStates;

    public synchronized void remoteSyncEvent(RemoteSyncEvent e)
    {
	Util.dbgPrintln("NetworkServer::remoteSyncEvent");
	latestStates.update(e);
	transmitEvent((EventObject)latestStates.clone());
    }

    private void transmitEvent(EventObject e)
    {
	Util.dbgPrintln("NetworkServer::transmitEvent\n"+e);
	ClientRecord cr = null;
	Iterator it = clients.iterator();

	while(it.hasNext()) {
	    cr = (ClientRecord) it.next();
	    try {
		Util.dbgPrintln("NetworkServer: transmit");		
		cr.oos.writeObject(e);
		cr.oos.flush();
	    }
	    catch (IOException ie) {
		Util.dbgPrintln("NetworkServer: transmit error, closing");		
		if(cr != null) {
		    synchronized(this) {
			cr.close();
			it.remove();
		    }
		}
	    }
	}
    }

    private void pollClientsForEvents()
    {
	Util.dbgPrintln("NetworkServer::pollClientsForEvents");
	ClientRecord cr = null;
	Iterator it = clients.iterator();

	while(it.hasNext()) {
	    cr = (ClientRecord) it.next();
	    try {
		if(cr.is.available() > 0 ) {
		    Object o = cr.ois.readObject();
		    Util.assert(o instanceof EventObject, "NetworkServer::pollClientsForEvents()");
		    dispatchEvent(o);
		}
	    }
	    catch (ClassNotFoundException cnf) {
		Util.assert(false, "NetworkServer::pollClientsForEvents()");
	    }
	    catch(IOException e) {
		Util.dbgPrintln("Client Unreachable");
		if(cr != null) {
		    synchronized(this) {
			cr.close();
			it.remove();
		    }
		}
	    }
	}
    }
    
    private void dispatchEvent(Object o) {
	Iterator it;
	if(o instanceof ActionEvent) {
	    it = actionListeners.iterator();
	    while(it.hasNext())
		((ActionListener) it.next()).actionPerformed((ActionEvent)o);
	}
	else 
	    Util.assert(false, "NetworkServer::dispatchEvent "+o);
    }

}
