/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package beaconconfig;

import java.awt.event.*;
import java.util.*;

class RemoteControls 
    extends Controls implements RemoteSyncListener
{
    /* states of the root BeaconConfigApp */
    private RemoteSyncEvent currentStates;
    private EventListener eventHandler;
    private boolean override;

    public RemoteControls(DrawingArea d, SolverPanel s, 
			  EventListener e, boolean override)
    {
	super(d, s);
	currentStates = new RemoteSyncEvent();
	currentStates.controlPanelStates = super.getStates();
	eventHandler = e;
	this.override = override;
    }

    private void updateRemote()
    {
	if(!override) {
	    currentStates = new RemoteSyncEvent();
	    currentStates.controlPanelStates = super.getStates();
	}
	if(eventHandler instanceof RemoteSyncListener) {
	    // calls handler in NetworkServer
	    ((RemoteSyncListener) eventHandler).remoteSyncEvent(currentStates);
	}
    }

    public void enableManualOverride(boolean v) 
    {
	super.enableManualOverride(v);
	updateRemote();
    }

    public void enableTrackingOptions(boolean v)
    {
	super.enableTrackingOptions(v);
	updateRemote();
    }

    public void actionPerformed(ActionEvent e) 
    {
	Util.dbgPrintln("RemoteControls::actionPerformed "+e.getActionCommand());
	if(!override) { // root app
	    super.actionPerformed(e);
	    currentStates = new RemoteSyncEvent();
	    currentStates.controlPanelStates = super.getStates();
	}
	else {
	    // remote app: restore old states, then wait for ack from
	    // root application
	    setCurrentStates(currentStates);
	}
	if(eventHandler instanceof ActionListener) {
	    // calls handler in NetworkClient
	    ((ActionListener) eventHandler).actionPerformed(e);
	}
	else if(eventHandler instanceof RemoteSyncListener) {
	    // calls handler in NetworkServer
	    ((RemoteSyncListener) eventHandler).remoteSyncEvent(currentStates);
	}
    }

    private void setCurrentStates(RemoteSyncEvent cs)
    {
	Util.dbgPrintln("RemoteControls::setCurrentStates");
	currentStates = cs;
	if(cs.controlPanelStates != null) {
	    Boolean manualOverrideSelected = (Boolean) cs.controlPanelStates.get(0);
	    Boolean manualOverrideEnabled  = (Boolean) cs.controlPanelStates.get(1);
	    Boolean aggressiveSelected     = (Boolean) cs.controlPanelStates.get(2);
	    Boolean aggressiveEnabled      = (Boolean) cs.controlPanelStates.get(3);
	    Boolean gridSelected           = (Boolean) cs.controlPanelStates.get(4);
	    Boolean gridEnabled            = (Boolean) cs.controlPanelStates.get(5);
	    Boolean annulusSelected        = (Boolean) cs.controlPanelStates.get(6);
	    Boolean annulusEnabled         = (Boolean) cs.controlPanelStates.get(7);
	    Boolean streamerSelected       = (Boolean) cs.controlPanelStates.get(8);
	    Boolean streamerEnabled        = (Boolean) cs.controlPanelStates.get(9);

	    manualOverride.setSelected(manualOverrideSelected.booleanValue());
	    aggressive.setSelected(aggressiveSelected.booleanValue());
	    grid.setSelected(gridSelected.booleanValue());
	    annulus.setSelected(annulusSelected.booleanValue());
	    streamer.setSelected(streamerSelected.booleanValue());

	    manualOverride.setEnabled(manualOverrideEnabled.booleanValue());
	    aggressive.setEnabled(aggressiveEnabled.booleanValue());
	    grid.setEnabled(gridEnabled.booleanValue());
	    annulus.setEnabled(annulusEnabled.booleanValue());
	    streamer.setEnabled(streamerEnabled.booleanValue());
	}
	else {
	    manualOverride.setSelected(false);
	    aggressive.setSelected(false);
	    grid.setSelected(false);	
	    annulus.setSelected(false);
	    streamer.setSelected(false);

	    grid.setEnabled(true);	
	    enableManualOverride(false);
	    enableTrackingOptions(false);
	}
    }

    // called by the Network Client in remote app
    public void remoteSyncEvent(RemoteSyncEvent e) {
	Util.dbgPrintln("RemoteControls::remoteSyncEvent");
	setCurrentStates(e);
    }

}
