/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package beaconconfig;

import java.awt.*;
import java.awt.Toolkit;
import java.util.*;
import javax.swing.*;
import cricketdaemon.position.PositionData;
import cricketdaemon.position.PositionSolver;
import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;
import cricketdaemon.UnitConv;

import Jama.*;

/**
 * Implements the intelligence for solving beacon coordinates
 */
public class SolverPanel extends JPanel implements SketchControl, SolverControl
{
    //protected static double K_NOMINAL  = 34449.0/(15625 * 2.54);
    //static double K_NOMINAL  = 1.0;
    private DrawingArea drawArea;

    private int numAnchorBeacons;
    private BeaconRecord[] anchorBeacons;
    private ArrayList[] anchorBeaconsHeardList;
    private PositionSolver posSolver;

    protected JLabel status;
    protected JLabel errmsg;

    private BeaconConfig bc;

    /* conversion constant from Bodhi units to Inches */
    //private final double K_NOMINAL = 0.8680063;

    /* conversion constant from Bodhi units to CM */
    public static final double K_NOMINAL = UnitConv.cricketDistToCM(1);

    /* Msgs to tell user about the state of the algorithm */
    private final String start = " Searching for FIRST anchor beacon";
    private final String second = " Searching for SECOND anchor beacon";
    private final String third = " Searching for THIRD anchor beacon";
    private final String done = " Coordinate Estimation Completed";

    // To determine that the user is steady, all beacons 
    // must have a distance stddev below this threshold.
    // The distance stddev is the std of the distances collected
    // over a sliding window, which is parameter set by
    // cricketdaemon (-S option).
    public static double STDDEV_THRESH = 15.0;

    // Governs how close you need to hold the receiver to a beacon
    // during calibration.  This value is also used to govern the minimum
    // distance you should place each beacon--it's just a heuristic to make
    // sure we don't calibrate the wrong beacon.
    // Want to set this roughly about 5 feet so you can hold the iPAQ
    // about 4 feet under the ceiling.  A smaller value will require the
    // user to hold the iPaQ up.
    //private double CLOSENESS_THRESH = 48.0;
    private double CLOSENESS_THRESH = 100.0;
    //    private final double CLOSENESS_THRESH = 36.0;

    // during calibration, ignore a beaconRecord (a beacon) if we have
    // fewer than ENOUGH_SAMPLE in the sliding time window
    private int ENOUGH_SAMPLE;

    // when true, tracking mode uses the last distance sample for each
    // beacon as the distance estimate.  Otherwise, uses the default
    // distance estimate given by CricketDaemon.
    private boolean AGGRESSIVE = false;

    public SolverPanel(BeaconConfig bc, 
		       DrawingArea da, int closeness, int stddev, int enoughsamples)
    {
	this.bc = bc;

	// override default values
	if(closeness > 0)
	    CLOSENESS_THRESH = closeness;
	if(stddev > 0)
	    STDDEV_THRESH = stddev;
	ENOUGH_SAMPLE = enoughsamples;

	GridLayout grid = new GridLayout(2,1);
	setLayout(grid);
	status = new JLabel(start, JLabel.LEFT);
	errmsg = new JLabel("", JLabel.LEFT);
	errmsg.setForeground(Color.red);
	add(status);
	add(errmsg);
	validate();
	
	drawArea = da;

	anchorBeacons = new BeaconRecord[3];
	anchorBeaconsHeardList = new ArrayList[3];
	//posSolver = new PositionSolver(1, Double.MAX_VALUE);
	posSolver = new PositionSolver(K_NOMINAL, Double.MAX_VALUE);
	uniqueNames = new ArrayList();

	reset();
    }

    public synchronized void reset()
    {
	System.out.println("reset");
	currentPos = null;
	numAnchorBeacons = 0;
	
	// in case the drawing area has zoomed 
	if(drawArea != null) {
	    originX = DrawingArea.minCoord.x + DrawingArea.AREAMARGIN;
	    originY = DrawingArea.minCoord.y + DrawingArea.AREAMARGIN;
	}

	setAggressive(false);

	uniqueNames.clear();
	changeStatus(start);
	errMsg(false, "");
	lastErrMsg = "";

	if(bc != null && bc.controlPanel != null)
	    bc.controlPanel.enableManualOverride(false);
    }

    double[] beacons_x, beacons_y, beacons_z;
    String[] beacons_space;
    int[] beacons_id;
    double[] beacons_dist;

    /* Implements an algorithm that first pin pointsthe locations of
       the first three 'anchor' beacons.  Positions of other beacons are
       configured based on the positions and measured distances at these
       three beacons */
    public synchronized void processDistStats(ArrayList beaconsHeard)
    {
	//System.out.println("processDistStats");
	if(numAnchorBeacons < 3) {
	    if(haveSteadyReading(beaconsHeard)) {
		if(beaconsHeard.size() > BeaconConfig.numBeacons) {
		    // Message: Num beacons heard greater than 
		    // the user specified value
		    int diff = beaconsHeard.size() - BeaconConfig.numBeacons;
		    errMsg(false, " "+diff+" more beacons heard than the specified value");
		}
		else if(beaconsHeard.size() < BeaconConfig.numBeacons) {
		    // Message: Num beacons heard fewer than 
		    // the user specified value
		    int diff = BeaconConfig.numBeacons - beaconsHeard.size();
		    errMsg(false, " "+diff+" fewer beacons heard than the specified value");
		}
		else {
		    updateUniqueNames(beaconsHeard);
		    BeaconRecord anchor = findNearestBeacon(beaconsHeard);
		    if(unheard(anchor)) {
			if(K_NOMINAL*anchor.distStat.mode <= CLOSENESS_THRESH) {
			    addAnchorBeacon(anchor, beaconsHeard);
			}
			else {
			    // Message to signal user to move listener closer
			    // to beacon
			    errMsg(false, " Move listener closer to beacon.");
			}
		    }
		    //else {
			// Message
			// errMsg(false, "beacon already registered");	
		    // }
		}
	    }
	    else { //Message readings not yet steady
		//Util.dbgPrintln("readings not yet steady");
		errMsg(false, " Distance readings unstable");	
	    }
	    //Util.dbgPrintln("numAnchorBeacons = "+numAnchorBeacons);
	}

	else {
	    updateDistances(beaconsHeard);

	    // now tracking processing is done by 
	    // CricketDaemon (via processDeviceCoord)
	    //if(filterBeacons(beaconsHeard))
	    //tracking(beaconsHeard);
	}
    }



    public synchronized void processDeviceCoord(Position pos)
    {
	if(numAnchorBeacons >= 3) {
	    errMsg(false, " Tracking...");

	    if(pos.invalid())  // PositionModule failed to compute a position
		errMsg(false, " Not enough beacons heard or stddev is too high");
	    else {
		if(pendown)
		    drawArea.sketchObject.updateCoord(pos.x, pos.y);
		drawArea.moveCricketListener(pos.x, pos.y, pos.z);
		currentPos = pos;
	    }
	}
    }
    

    private ArrayList uniqueNames;

    /* could use a HashSet here but nah */
    private void updateUniqueNames(ArrayList beaconsHeard)
    {
	if(uniqueNames.size() >= BeaconConfig.numBeacons)
	    return;

	Iterator it = beaconsHeard.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(! uniqueNames.contains(br) )
		uniqueNames.add(br.clone());
	}

	// there can be a race condition where some other app (e.g. previous
	// instance of BeaconConfig) that has pending BeaconTable updates that
	// are happening when ManualOverride is enabled (and triggered) by the
	// user.  Ideally, we want to delay ManualOverride until all pending
	// BeaconTable updates have been finished.  This requires synchronization
	// between CricketDaemon and the apps, which may be difficult to acheive.
	if(uniqueNames.size() >= 3) 
	    bc.controlPanel.enableManualOverride(true);
    }

    private double randCoord(int i) 
    {
	int width = 0;
	if(i == 0)
	    width = (int) Math.round(DrawingArea.maxCoord.x - DrawingArea.minCoord.x);
	else
	    width = (int) Math.round(DrawingArea.maxCoord.y - DrawingArea.minCoord.y);
	    
	return BeaconConfig.r.nextInt(width) - width/2;
    }

    private ArrayList fillKnownBeacons()
    {
	ArrayList result =  new ArrayList();
	Iterator it = uniqueNames.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(br.pos != null) {
		result.add(br);
		it.remove();
	    }
	}
	return result;
    }

    private ArrayList fillRandBeacons(ArrayList result) 
    {
	double x, y;
	int numNames = uniqueNames.size();

	while(!uniqueNames.isEmpty()) {
	    BeaconRecord br = (BeaconRecord) uniqueNames.remove(0);
	    x = randCoord(0);
	    y = randCoord(1);
	    br.pos = new Position(x, y, 0);
	    result.add(br);
	}
	return result;
    }

    /* this needs to be synchronized with the callback handler for Cricket (processDistStat) */
    public synchronized boolean forceTrackingMode()
    {
	if(numAnchorBeacons >= 3)
	    return false; // already in tracking mode

	System.out.println("forceTracking");

	if(uniqueNames.size() < 3) {
	    System.err.println("Critical Error: forceTrackingMode() called with fewer than 3 beacons");
	    System.exit(1);
	}

	ArrayList pos = fillKnownBeacons();

	if(numAnchorBeacons == 2) {
	    // use measured coordinates instead
	    pos = new ArrayList();
	    BeaconRecord anchorA = anchorBeacons[0];
	    BeaconRecord anchorB = anchorBeacons[1];
	    pos.add(anchorA);
	    pos.add(anchorB);

	    ArrayList aList = anchorBeaconsHeardList[0];
	    ArrayList bList = anchorBeaconsHeardList[1];
	    
	    Iterator it = aList.iterator();
	    while(it.hasNext()) {
		BeaconRecord ar = (BeaconRecord) it.next();
		if(!ar.equals(anchorA) && !ar.equals(anchorB)) {
		    BeaconRecord br = (BeaconRecord) bList.get(bList.indexOf(ar));
		    double distAC = K_NOMINAL*Math.sqrt(diffsqr(ar.distStat.mode, anchorA.distStat.mode));
		    double distBC = K_NOMINAL*Math.sqrt(diffsqr(br.distStat.mode, anchorB.distStat.mode));
		    ar.pos = solve2D(anchorA.pos.x, anchorB.pos.x, anchorA.pos.y, distAC, distBC);
		    if(ar.pos != null)
			pos.add(ar);
		    else {
			/* can't solve x, y */
			ar.pos = new Position(randCoord(0), randCoord(1), 0);
			pos.add(ar);
		    }
		}
	    }
	    numAnchorBeacons = 3;
	}
	else if(numAnchorBeacons == 1) {
	    if(uniqueNames.contains(anchorBeacons[0]))
		uniqueNames.remove(anchorBeacons[0]);

	    fillRandBeacons(pos);
	    if(!pos.contains(anchorBeacons[0])) 
		pos.add(anchorBeacons[0]);
	}
	else if(numAnchorBeacons == 0) {
	    fillRandBeacons(pos);
	}

	/* README need to handle this case properly with the remote, etc.
	   // this case is handled by the feature that the ManualOverride button is
	   // disabled unless uniqueNames.size() >= 3.
	if(pos.size() < 3)
	    return false;
	*/

	numAnchorBeacons = 3;
	prepareForTracking(pos, true);
	return true;
    }
    
    /*
    private boolean filterBeacons(ArrayList beacons)
    {
	Iterator it = beacons.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();

	    if ((drawArea.findBeacon(br) == null) || // has not been registered
		(K_NOMINAL*br.distStat.stddev > 3*STDDEV_THRESH) ) {
		//System.out.println("FILTERING:::: "+b);
		it.remove();
	    }
	}
	if(beacons.size() >= 3)
	    return true;
	else {
	    errMsg(false, " Not enough beacons heard or stddev is too high");
	    return false;
	}
    }
    */

    private boolean unheard(BeaconRecord br) {
	for(int i = 0; i < numAnchorBeacons; i++) {
	    if(anchorBeacons[i].equals(br))
		return false;
	}
	return true;
    }

    private boolean haveSteadyReading(ArrayList beaconsHeard) {
	boolean result = true;
	Iterator it = beaconsHeard.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(br.distStat.samples.size() < ENOUGH_SAMPLE) {
		errMsg(false, "Too few samples collected from "+br.space+": dropping it");
		it.remove(); // remove beacon
	    }
	    else {
		if(K_NOMINAL*br.distStat.stddev > STDDEV_THRESH) {
		    result = false;
		    break;
		}
	    }
	}

	//System.out.println("haveSteadyReading returns "+result);
	return result;
    }

    private double originX;
    private double originY;

    private double anchorDistance(int a, int b)
    {
	int i = anchorBeaconsHeardList[a].indexOf(anchorBeacons[b]);
	BeaconRecord ameasureofb = (BeaconRecord) anchorBeaconsHeardList[a].get(i);

	i = anchorBeaconsHeardList[b].indexOf(anchorBeacons[a]);
	BeaconRecord bmeasureofa = (BeaconRecord) anchorBeaconsHeardList[b].get(i);

	//double avgdist = (ameasureofb.distStat.mode + bmeasureofa.distStat.mode)/2;

	double abdist = diffsqr(ameasureofb.distStat.mode, anchorBeacons[a].distStat.mode);
	double badist = diffsqr(bmeasureofa.distStat.mode, anchorBeacons[b].distStat.mode);

	/*
	if(b == 2) {
	    System.out.println(ameasureofb.distStat.mode+" "+
			       bmeasureofa.distStat.mode+" "+
			       abdist+" "+badist);
	}
	*/

	if(abdist < 0 || badist < 0)
	    return -1;
	else
	    return K_NOMINAL*Math.sqrt((abdist+badist)/2);
    }

    private void beep(int times) {
	synchronized(this) {
	    while(times > 0) {
		getToolkit().beep();
		try {
		    wait(300);
		} catch (InterruptedException ie) {
		}
		times--;
	    }
	}
    }

    private String lastErrMsg = "";
    private long lastErrMsgTime = 0;

    // same error message will not be repeated within errMsgPeriod
    private final static long ERRMSGPERIOD = 2000; 

    protected boolean errMsg(boolean beep, String s) {
	long curTime = System.currentTimeMillis();

	if(s.equals(lastErrMsg) && ((curTime - lastErrMsgTime) < ERRMSGPERIOD))
	    return false;

	lastErrMsgTime = curTime;
	lastErrMsg = s;

	if(s.length() > 0) {
	    Util.dbgPrintln(s);
	}
	errmsg.setText(s);
	if(beep)
	    beep(1);

	return true;
    }

    public Position solve2D(double xa, double xb, double ya, double distAC, double distBC)
    {
	if(Math.abs(xb-xa) < 0.1)
	    return null;
	double x = (diffsqr(distAC, distBC) + diffsqr(xb, xa))/(2*(xb-xa));
	double y = Math.sqrt(diffsqr(distAC, (x-xa))) + ya;
	return new Position(x, y, 0);
    }

    private void addAnchorBeacon(BeaconRecord anchor, ArrayList beaconsHeard)
    {
	anchorBeacons[numAnchorBeacons] = anchor;
	anchorBeaconsHeardList[numAnchorBeacons] = beaconsHeard;
	numAnchorBeacons++;

	// toadd: display current distStats
	switch(numAnchorBeacons) {
	case 1:
	    anchorBeacons[0].pos = new Position(originX, originY, anchor.distStat.mode);
	    drawArea.addBeacon(anchorBeacons[0], bc.server);
	    changeStatus(second);
	    errMsg(false, "");
	    beep(2);
	    break;
	case 2:
	    if(compareBeaconsHeardList(anchorBeaconsHeardList[0],
				       anchorBeaconsHeardList[1])) {
		double dist = anchorDistance(0,1);
		double X = dist + originX;
		if(dist > 0) {
		    anchorBeacons[1].pos = new Position(X, originY, anchor.distStat.mode);
		    drawArea.addBeacon(anchorBeacons[1], bc.server);
		    changeStatus(third);
		    errMsg(false, "");
		    beep(3);
		}
		else {
		    // Either the beacon is not directly underneath or the 
		    // error is dominating.  Perhaps moving the beacons apart can help
		    // solve the problem.

		    // Message: anchor beacons 0,1 too close together.
		    errMsg(true, " Beacons "+anchorBeacons[0].space+", "+
			   anchorBeacons[1].space+
			   " are placed too close together");
		    numAnchorBeacons--;
		}
	    }
	    else {
		// Message:  Beacons 0 and 1 hear a different set 
		// of Beacons
		errMsg(true, " Beacons 0, 1 hear a different set of beacons");
		numAnchorBeacons--;
	    }
	    break;
	case 3:
	    if(compareBeaconsHeardList(anchorBeaconsHeardList[0],
				       anchorBeaconsHeardList[2])) {
		double distA = anchorDistance(0,2);
		double distB = anchorDistance(1,2);
		if(distA > 0 && distB > 0) {
		    double xa = anchorBeacons[0].pos.x;

		    double ya = anchorBeacons[0].pos.y;
		    double xb = anchorBeacons[1].pos.x;
		    Util.dbgPrintln(distA+" "+distB+" "+xb+" "+xa);

		    double X = 0;
		    double Y = 0;

		// degeneracy check

		    //if(Math.abs(xb-xa) > CLOSENESS_THRESH) {
			Position xy = solve2D(xa, xb, ya, distA, distB);
			if(xy == null) {
			    errMsg(true, "Divide by zero -- anchor x-coordinates too close together");
			    numAnchorBeacons--;
			    return;
			}
			X = xy.x;
			Y = xy.y;
			/*
		    }
		    else {
			errMsg(true, " Beacons "+anchorBeacons[0].space+", "+
			       anchorBeacons[1].space+
			       " are placed too close together");
			numAnchorBeacons--;
			return;
		    }
			*/

			/* want to move 3rd anchor away from the segment
			 formed by first 2 beacons to avoid degeneracy
			 of 3 co-linear anchors */
		// degeneracy check
		    /*
		    if( Math.abs(Y - ya) < CLOSENESS_THRESH/2 ) {
			errMsg(true, " Move beacon "+anchorBeacons[2].space+" away from line ("+
			       anchorBeacons[0].space+", "+anchorBeacons[1].space+") or pick another anchor.");
			numAnchorBeacons--;
			return;
		    }
		    */

		    //anchorBeacons[2].pos = new Position(X, originY, anchor.distStat.mode);
		    anchorBeacons[2].pos = new Position(X, Y, anchor.distStat.mode);
		    //drawArea.addBeacon(anchorBeacons[2]);

		    ArrayList pos = computeAllBeaconPositions();
		    prepareForTracking(pos, false);

		}
		else if(distA < 0) {
		    // Either the beacon is not directly underneath or the 
		    // error is dominating.  Perhaps moving the beacons apart can help
		    // solve the problem.

		    // Message: anchor beacons 0,2 too close together.
		    Util.dbgPrintln("anchor beacons 0,2 too close together");
		    errMsg(true, " Beacons "+anchorBeacons[0].space+", "+
			   anchorBeacons[2].space+
			   " are placed too close together");
		    numAnchorBeacons--;		    
		} else {
		    Util.dbgPrintln("anchor beacons 1,2 too close together");
		    errMsg(true, " Beacons "+anchorBeacons[1].space+", "+
			   anchorBeacons[2].space+
			   " are placed too close together");
		    numAnchorBeacons--;		    
		}
	    }
	    else {
		// Message:  Beacons 0 and 2 hear a different set of Beacons
		errMsg(true, " Beacons 0, 2 hear a different set of beacons");
		numAnchorBeacons--;
	    }
	    break;
	default:
	    Util.assert(false, "numAnchorBeacons have invalid value of: "+numAnchorBeacons);
	}
    }

    private void prepareForTracking(ArrayList pos, boolean forced) {
	Iterator it = pos.iterator();
	/* compute bounding box */
	double minX = Double.POSITIVE_INFINITY;
	double maxX = Double.NEGATIVE_INFINITY;
	double minY = Double.POSITIVE_INFINITY;
	double maxY = Double.NEGATIVE_INFINITY;
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(br.pos.x < minX)
		minX = br.pos.x;
	    if(br.pos.y < minY)
		minY = br.pos.y;
	    if(br.pos.x > maxX)
		maxX = br.pos.x;
	    if(br.pos.y > maxY)
		maxY = br.pos.y;
	}
	
	double transX = 
	    (DrawingArea.maxCoord.x -  DrawingArea.minCoord.x ) - (maxX - minX);
	//transX = (transX - (maxX-minX))/2;
	transX = transX/2 - (minX - DrawingArea.minCoord.x);
	
	double transY = 
	    (DrawingArea.maxCoord.y -  DrawingArea.minCoord.y ) - (maxY - minY);
	//transY = (transY - (maxY-minY))/2;
	transY = transY/2 - ( minY - DrawingArea.minCoord.y );
	
	/* round to the nearest step */
	//transX = Math.round(transX/drawArea.gridStep)*drawArea.gridStep;
	//transY = Math.round(transY/drawArea.gridStep)*drawArea.gridStep;
	
	drawArea.initialize();  // wipe everything off the screen
	int c = 0;
	it = pos.iterator();

	// perhaps pop a window to ask the user whether to modify
	// CricketDaemon's beacontable

	bc.server.setPositionStddevThresh(STDDEV_THRESH);
	
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();

	    //System.out.println("debug: "+br+"\n"+transX+" "+transY+" "+
	    //originX+" "+originY+" "+maxX+" "+minX+" "+
	    //DrawingArea.maxCoord+" "+
	    //DrawingArea.minCoord );

	    br.pos.x = br.pos.x + transX;
	    br.pos.y = br.pos.y + transY;
	    drawArea.addBeacon(br, bc.server);

	    if(!forced) {
		// BeaconTable can still be modified by drawing the beacon icons
		// see Beacon.mouseReleased()
		System.out.println("registering beacon coords: "+br.space+" "+br.id+" "+
				   br.pos.x+" "+br.pos.y);
		bc.server.setBeaconTableEntry(br.space, br.id.intValue(), 
					      br.pos.x, br.pos.y, 0);
	    }
	}
	
	bc.controlPanel.enableTrackingOptions(true);
	bc.sketchPanel.enableButtons(true);

	changeStatus(done);
	errMsg(false, "");
    }

    protected void changeStatus(String s)
    {
	status.setText(s);
    }

    
    /* returns true if the lists a and b contain the same beacons */
    private boolean compareBeaconsHeardList(ArrayList a, ArrayList b)
    {
	if(a.size() != b.size())
	    return false;

	Iterator ia = a.iterator();
	while(ia.hasNext()) {
	    if(!b.contains(ia.next()))
		return false;
	}
	return true;
    }

    private BeaconRecord findNearestBeacon(ArrayList l)
    {
	double minDist = Double.MAX_VALUE;
	BeaconRecord nearestBeacon = null;

	Iterator it = l.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(br.distStat.mode < minDist) {
		minDist = br.distStat.mode;
		nearestBeacon = br;
	    }
	}
	
	//l.remove(nearestBeacon);
	return nearestBeacon;
    }

    // figure out what to do if user modifies positions on GUI.
    private ArrayList computeAllBeaconPositions()
    {
	double[] x = new double[numAnchorBeacons];
	double[] y = new double[numAnchorBeacons];
	double[] z = new double[numAnchorBeacons];
	String[] space = new String[numAnchorBeacons];
	int[] id = new int[numAnchorBeacons];
	double[] dist = new double[numAnchorBeacons];

	for(int c = 0; c < numAnchorBeacons; c++) {
	    // fixme:  pos is null! need to initialize the position!!
	    x[c] = anchorBeacons[c].pos.x;
	    y[c] = anchorBeacons[c].pos.y;
	    z[c] = anchorBeacons[c].pos.z;
	    space[c] = anchorBeacons[c].space;
	    id[c] = anchorBeacons[c].id.intValue();

	    //System.out.println(space[c]+" "+x[c]+ " "+y[c]+" "+z[c]);
	}

	Iterator it = anchorBeaconsHeardList[0].iterator();
	while(it.hasNext()) {	
	    BeaconRecord a = (BeaconRecord) it.next();
	    //dist[0] = a.distStat.mode;
	    dist[0] = Math.sqrt(diffsqr(a.distStat.mode, z[0]));
	    
	    for(int c = 1; c < numAnchorBeacons; c++) {
		BeaconRecord br = (BeaconRecord) anchorBeaconsHeardList[c].get(anchorBeaconsHeardList[c].indexOf(a));
		//dist[c] = br.distStat.mode;
		dist[c] = Math.sqrt(diffsqr(br.distStat.mode, z[c]));
	    }
		  
	    PositionData pd = 
		//posSolver.positionEstimateVKnown(x, y, z, space, id, dist);
		posSolver.positionEstimate2DVKnown(x, y, space, id, dist);

	    if(pd != null)
		a.pos = new Position(pd.getCoord().x, pd.getCoord().y, pd.getCoord().z);
	    
	    /* config coordinates */
	    //System.out.println(pd);
	}

	return anchorBeaconsHeardList[0];
    }

    public static double diffsqr(double a, double b) 
    {
	return (a+b)*(a-b);
    }

    public void setAggressive(boolean a)
    {
	AGGRESSIVE = a;

	if(bc != null)
	    bc.server.setPositionAggressive(a);
    }

    private void updateDistances(ArrayList beaconsHeard)
    {
	Iterator it = beaconsHeard.iterator();
	while(it.hasNext()) {
	    BeaconRecord reading = (BeaconRecord) it.next();
	    BeaconRecord br = drawArea.findBeacon(reading);
	    if(br != null) 
		br.distStat = reading.distStat;
	}
    }

    // assumes that beaconsHeard array only contains beacons that
    // have been "Registered" during calibration phase
    private void tracking(ArrayList beaconsHeard)
    {
	// tracking
	// now tracking processing is done by CricketDaemon (via processDeviceCoord)
	double[] x = new double[beaconsHeard.size()];
	double[] y = new double[beaconsHeard.size()];
	double[] z = new double[beaconsHeard.size()];
	String[] space = new String[beaconsHeard.size()];
	int[] id = new int[beaconsHeard.size()];
	double[] dist = new double[beaconsHeard.size()];
	

	for(int c = 0; c < beaconsHeard.size(); c++) {
	    BeaconRecord reading = (BeaconRecord) beaconsHeard.get(c);
	    BeaconRecord br = drawArea.findBeacon(reading);

	    x[c] = br.pos.x;
	    y[c] = br.pos.y;
	    z[c] = 0;
	    space[c] = br.space;
	    id[c] = br.id.intValue();
	    dist[c] = reading.distStat.mode;
	    
	    if(AGGRESSIVE) {
		int lastind = reading.distStat.samples.size() - 1;
		dist[c] = ((Sample) reading.distStat.samples.get(lastind)).dist;
	    }
	    else
		dist[c] = reading.distStat.dist;

	    // for drawing annulus around each beacon
	    br.distStat = reading.distStat;
	}
	//System.exit(1);
	
	PositionData pd = posSolver.positionEstimateVKnown(x, y, z, space, id, dist);
	
	if(pd != null) {
	    processDeviceCoord(pd.getCoord());
	}
	else {
	    errMsg(false, " Least Squares failed");
	}
    }

    /* sketch routines */
    private boolean pendown = false;
    private Position currentPos = null;

    public void setRectPenDn(boolean penDn) 
    {
	Util.dbgPrintln("setRectPenDn");
	setPenDn(penDn, DrawingArea.RECTANGLE);
	return;
    }

    public void setCircPenDn(boolean penDn) 
    {
	Util.dbgPrintln("setCircPenDn");
	setPenDn(penDn, DrawingArea.CIRCLE);
	return;
    }

    public void setLinePenDn(boolean penDn) 
    {
	Util.dbgPrintln("setLinePenDn");
	setPenDn(penDn, DrawingArea.PLINE);
	return;
    }

    private void setPenDn(boolean penDn, int type) {
	if(currentPos != null && numAnchorBeacons >= 3) {
	    if(penDn)
		drawArea.newSketchObject(currentPos.x,
					 currentPos.y, 
					 type);
	    else
		drawArea.penUp(currentPos.x,
			       currentPos.y);
	    pendown = penDn;
	}
    }
}
