/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package clientlibexample;


import gnu.getopt.*;
import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;

import java.io.*;
import java.util.*;


/**
 * An example of how to use the CricketDaemon Java clientlib.
 *
 * It simply prints a line containing space, position and orientation 
 * information whenever they are updated by the algorithms running
 * in CricketDaemon.
 *
 * @author Allen Miu
 */
class ClientlibExample implements Callback
{
    /**
     * Handle to cricket
     */
    Broker cricket;

    /**
     * Data structures
     */
    String currentSpace = null;
    Position currentPosition = null;
    double currentOrientation = 0;

    public ClientlibExample()
    {
	/**
	 * Established TCP connection to the CricketDaemon running on
	 * localhost.  The cricket object will trigger a callback whenever
	 * the CricketDaemon decides to push information out to the clients.
	 */
	cricket = new ServerBroker();

	/**
	 * create a bitmask specifying the type of Cricket information 
	 * that should trigger a callback on this object
	 */
	BitSet mask = new BitSet();
	mask.set(Broker.SPACE);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.ORIENTATION);

	/**
	 * register this object's callback with the cricket object
	 */
	cricket.register(this, mask);
	cricket.start(); // runs forever
	System.out.println("ClientlibExample created");
    }

    synchronized public void callback(CricketData data, BitSet mask)
    {
	boolean update = false;

	if(mask.get(Broker.SPACE)) {
	    BeaconRecord cs = data.getCurrentSpaceObject();
	    if(cs != null && cs.uniqueName != null) {
		if(currentSpace == null || !(currentSpace.equals(cs.uniqueName))) {
		    update = true;
		    currentSpace = cs.uniqueName;
		}
	    }
	}

	if(mask.get(Broker.DEVICECOORD)) {
	    Position p = data.getDevicePosition();
	    if(p != null) {
		if(currentPosition == null || !(currentPosition.equals(p))) {
		    update = true;
		    currentPosition = p;
		}
	    }
	}

	if(mask.get(Broker.ORIENTATION)) {
	    double o = data.getOrient();
	    
	    if(Math.abs(o - currentOrientation) > 0.0001) {
		update = true;
		currentOrientation = o;
	    }
	}

	if(update)
	    printState();
    }

    private void printState() 
    {
	System.out.println("space="+currentSpace+" pos="+currentPosition+" ang="+
			   currentOrientation);
    }

    public static void main(String[] args) 
    {
	/* gnu-style getopt to parse cmd line args */
	Getopt g = new Getopt("ClientlibExample", args, "c:");
	int c;
	try {
	    while((c = g.getopt()) != -1) {
		switch(c) {
		case 'c':
		    System.out.println("wassup! "+g.getOptarg());
		    break;
		}
	    }
	}
	catch (Exception e) {
	    //usage();
	}


	ClientlibExample client = new ClientlibExample();
    }
}
