/* geom.c routines to fix the geometry of entities */
#include "acad2ug.h"

#define TRUE 1
#define FALSE 0

/* PlanarTest(verts,num,normal)
 * Tests if a set of vertices define a planar surface by
 * checking that they all satisfy the plane equation of the
 * surface, computed by the Newman and Sproull method.
 */
PlanarTest(verts,num,normal)
VERTEX verts[];
int num;
double *normal;
{
  double distance;
  int i ;
  
  for(i =0 ; i< num; i++) {
    distance = normal[0] * verts[i].x +
               normal[1] * verts[i].y +
               normal[2] * verts[i].z + normal[3];
    if(fabs(distance) > EPSILON )
        return 0;
  }
  return 1;
}


/* FixPlanar()
 * Given a nonplanar surface, try to find the offending vertex if
 * there is only one. Remove its vertices one at a time and
 * check if the remaining vertices define a surface parallel to
 * two of the coordinate axes. In this case, for architectural models
 * it is safe to assume that the surface was originally meant to be 
 * axial. Fix the offending vertex so that it lies on the plane too.
 */
FixPlanar(verts,num,normal,layerindex)
VERTEX verts[];
int num;
double *normal;
int layerindex;
{
   VERTEX newverts[MAXFACEVERT];
   double newnormal[4];
   int i,j,k;
   short fixed = FALSE;
   
   for (i =0; i< num;i++){
      for (j =0,k=0; j< num;j++){
        if ( j== i) continue; 
        newverts[k].x = verts[j].x;
        newverts[k].y = verts[j].y;
        newverts[k++].z = verts[j].z;
      }
      Find_normal(newnormal,newverts,num-1);
      if (fabs(newnormal[0]) == 1.0){
         verts[i].x = verts[(i+1) % num].x;
         fixed = TRUE;
         break;
      } else if (fabs(newnormal[1]) == 1.0){
         verts[i].y = verts[(i+1) % num].y;
         fixed = TRUE;
         break;
      } else if (fabs(newnormal[2]) == 1.0){
         verts[i].z = verts[(i+1) % num].z;
         fixed = TRUE;
         break;
      }
   }
   if (fixed){ 
      normal[0] = newnormal[0];
      normal[1] = newnormal[1];
      normal[2] = newnormal[2];
      normal[3] = newnormal[3];
      fprintf(Errorfile,"Nonplanar face fixed;lineno=%d;layer=%s\n",
                        Lineno,Layertable[layerindex].name);
      for(i=0;i<num;i++)
         fprintf(Errorfile,"%.3f %.3f %.3f\n",verts[i].x,verts[i].y,verts[i].z);
      return 1;
   }
   return 0;
}

Tesselate(verts,num,layerindex)
VERTEX verts[];
int num;
int layerindex;
{
   fprintf(Errorfile,"Nonplanar face should be tesselated;lineno=%d;layer=%s\n",
                                   Lineno,Layertable[layerindex].name);
}

/* OnSameLine(v1,v2,v3)
 * Checks if three vertices lie on the same line.
 */
OnSameLine(v1,v2,v3)
VERTEX v1,v2,v3;
{
    VERTEX a,b;
  
    if (!Find_vec(v1,v2,&a) || !Find_vec(v2,v3,&b))
    /* v1 and v2 or v2 and v3 are duplicates */
        return TRUE ;  
    if ( fabs(Dot(a,b)) == 1.0){
       return TRUE;
    }else 
       return FALSE;
}


/* CheckVertices():
 * Check a set of vertices for common AutoCAD drawing errors such as
 * multiple vertices and vertices on same line.
 */
CheckVertices(verts,number)
VERTEX verts[];
int *number;
{
   int i,j, num = *number;
   /* check for three vertices on the same line */
   for(i=1 ; i < num ; i++){
      if (num <= 2) break;
      if (OnSameLine(verts[i-1],verts[i],verts[(i+1) % num])){
          for(j=i;j< num-1;j++)
             verts[j] = verts[j+1]; 
          num--;
      }
   }
   /* check for duplicate adjacent vertices*/
   for ( i=0; i< num ; i++){
     if (verts[i].x == verts[(i+1)% num].x &&
         verts[i].y == verts[(i+1)% num].y &&
         verts[i].z == verts[(i+1)% num].z ) {
          for(j=i;j< num-1;j++)
             verts[j] = verts[j+1]; 
          num--;
     }
   }
   *number = num;
}
