/**********************************/
/** Author:       Will Leiserson **/
/** Last Updated: 8/9/2001       **/
/** Email:        leisew@rpi.edu **/
/**********************************/

/******************/
/** HEADER FILES **/
/******************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "sug.h"

/**************************/
/** TYPEDEFS AND STRUCTS **/
/**************************/

typedef struct dwPts dwPts;

/** dwPts:
 *  Good for storing the data collected from doors and windows for
 *  door+window jambs.
 */
struct dwPts{
  vert *v[6];
  wire *w[3];
};

/***************************/
/** CONSTANTS AND GLOBALS **/
/***************************/

const double PI = 3.141592653589793238462643383279502884;
const char JUNKLAYER[] = "A_STOOPID";
const int  MAXWIDTH    = 30;

/** Flags **/
bool VERBOSE;
bool MERGE;
bool CLEAN;
bool REMOVEJUNK;
bool DOORSNWINDOWS;

/***************************/
/** FUNCTION DECLARATIONS **/
/***************************/

bool Arguments( int argc, char **argv, char **inFile, char **outFile );
void Usage( );

void Merge( ugfile *uf );
void FilterDW( ugfile *uf );
void Clean( ugfile *uf );
void FilterJunk( ugfile *uf );

void  FixFaces( ugfile *uf );
void  FixPlan( ugfile *uf, edge ***e, int *le, int *me );
void  RemoveEdges( ugfile *uf, edge **e );
void  RemoveDisjoint( ugfile *uf );
dwPts **FilterDoors( ugfile *uf );
void  FilterElevatorsStairs( ugfile *uf );
dwPts **FilterWindows( ugfile *uf, edge **e, int *totalWins );
void  FilterWindowJambs( ugfile *uf, dwPts **winVerts, int numWins );
void  FilterDoorJambs( ugfile *uf, dwPts **doorVerts );

int   FilterStairs( ugfile *uf, wire *f );
dwPts *Correct( ugfile *uf, vert *v, vert *ref1, vert *ref2 );
edge  **EdgeSubset( edge **e, int dir, int *n, int *m );
void  EdgeSort( edge **e, int p, int r );
int   EdgePart( edge **e, int p, int r );
void  VertSort( vert **v, int p, int r );
int   VertPart( vert **v, int p, int r );
int   Intersect( ugfile *uf, edge ***list1, edge ***list2, edge ***e, int i,
		 int j, int *l1, int *m1, int *l2, int *m2, int *le, int *me );

void  Unmark( ugfile *uf, edge **e );
bool  AddLayer( ugfile *uf );
bool  GoodLayer( wire *w );
bool  TestPoint( float x, float y, edge *e );
void  AddEdge( edge ***e, edge *insert, int *l1, int *m1 );
bool  IsBetween( vert *between, vert *v1, vert *v2 );
float Length( float x1, float y1, float x2, float y2 );
bool  SameAngle( vert *v1, vert *v2, vert *v3, vert *v4 );
float WellPlaced( wire *w1, wire *w2 );
float Angle( vert *v1, vert *v2, vert *v3 );
void  Emphasize( ugfile *uf, char *lname );
float sq( float x );

/**********/
/** MAIN **/
/**********/

/** main:
 *  Dejunk is run on a unigraphics formatted MIT file.  It's designed
 *  to seperate out all of the extraneous lines and wires from the
 *  floorplan and put them in a seperate layer.
 */
int main( int argc, char **argv )
{
  ugfile *uf;
  char *inFile, *outFile;

  /** Check the arguments **/
  if( !Arguments( argc, argv, (char**) &inFile, (char**) &outFile ) ) return -1;

  /** Load the file into the data structure **/
  uf = (ugfile*) malloc( sizeof( ugfile ) );
  Load( inFile, uf );
  if( VERBOSE ) fprintf( stderr, "Loaded file %s.\n", inFile );

  /** Begin seperating out wires.  See function definitions for
      specifics on each function. **/
  AddLayer( uf );
  Merge( uf );
  FilterDW( uf );
  Clean( uf );
  FilterJunk( uf );

  /** Write the ugfile into a file (or to the stdout) **/
  if( outFile == NULL ){

    outFile = (char*) calloc( strlen( "stdout" )+1, sizeof( char ) );
    strcpy( outFile, "stdout" );

  }

  Write( outFile, uf );
  if( VERBOSE ) fprintf( stderr, "Done writing to %s.\n", outFile );
  Unload( uf );
  if( VERBOSE ) fprintf( stderr, "Done.\n" );

  return 0;
}

/**************************/
/** FUNCTION DEFINITIONS **/
/**************************/

/** Arguments:
 *  Take the arguments, set flags, and the in/out files.
 */
bool Arguments( int argc, char **argv, char **inFile, char **outFile )
{
  int i = 1;
  char *p;

  VERBOSE       = false;
  MERGE         = true;
  CLEAN         = true;
  REMOVEJUNK    = true;
  DOORSNWINDOWS = true;
  inFile[0]     = NULL;
  outFile[0]    = NULL;

  if( argc < 2 ) Usage( );

  while( i < argc ){
    p = argv[i];

    if( *p == '-' ){
      /** Set flags **/

      ++p;
      if( *p == '\0' ){
	fprintf( stderr, "Error: Bad argument '-', use -h for help.\n" );
      }

      while( *p != '\0' ){
	switch( *p ){
	case 'h':
	  Usage( );
	case 'v':
	  VERBOSE = true;
	  break;
	case 'c':
	  CLEAN = false;
	  break;
	case 'm':
	  MERGE = false;
	  break;
	case 'j':
	  REMOVEJUNK = false;
	  break;
	case 'd':
	  DOORSNWINDOWS = false;
	  break;
	default:
	  fprintf( stderr, "Error: Bad argument '%c', use -h for help.\n", *p );
	  return false;
	  break;
	}
	++p;
      }

    }else{
      /** Set the in and out files **/

      inFile[0] = p;
      ++i;
      if( i < argc ) outFile[0] = argv[i];

      ++i;
      if( i < argc ){
	fprintf( stderr, "Error: Too many arguments.  Use -h for help.\n" );
	return false;
      }
    }

    ++i;
  }

  return true;
}

/** Merge:
 *  Merge vertices that are very close to eachother.
 */
void Merge( ugfile *uf )
{
  vert **vlist;
  int numVerts, i, j, mergedPts;
  float tolerance;

  if( MERGE ){

    tolerance = 0.1;
    mergedPts = 0;

    numVerts = uf->hVert;
    vlist = uf->vertList;
    VertSort( vlist, 0, numVerts - 1 );

    for( i = 0; i < numVerts; ++i ){

      for( j = i + 1; j < numVerts; ++j ){
	if( fabs( vlist[i]->sortValue - vlist[j]->sortValue ) >= tolerance ) break;
	++mergedPts;
      }

    }

    if( VERBOSE ) fprintf( stderr, "Merged %d points.\n", mergedPts );

  }

  return;
}

/** FilterDW:
 *  Filter the doors, windows, door jambs, window jambs, and stairs out of the
 *  plan.  The first run-through hits the doors, windows, and stairs.  The
 *  second run-through picks out the door jambs and window jambs.
 */
void FilterDW( ugfile *uf )
{
  dwPts **doorVerts, **winVerts;
  edge **e;
  int i, numEdges, maxEdges, numWins;

  if( DOORSNWINDOWS ){

    /** Make an array of edges **/
    e = GetEdges( uf, &numEdges, &maxEdges );

    doorVerts = FilterDoors( uf );
    FilterElevatorsStairs( uf );
    winVerts = FilterWindows( uf, e, &numWins );
    FilterWindowJambs( uf, winVerts, numWins );
    FilterDoorJambs( uf, doorVerts );

    for( i = 0; i < numEdges; ++i ) free( e[i] );
    free( e );

    if( VERBOSE ) fprintf( stderr, "Finished doors, windows, etc.\n" );

  }

  return;
}

/** Clean:
 *  Fixes certain features, such as finding intersection points, making faces out
 *  of wires that should be faces, etc.
 */
void Clean( ugfile *uf )
{
  edge **e;
  int i, numEdges, maxEdges;

  if( CLEAN ){

    FixFaces( uf );
    e = GetEdges( uf, &numEdges, &maxEdges );
    FixPlan( uf, &e, &numEdges, &maxEdges );

    for( i = 0; i < numEdges; ++i ) free( e[i] );
    free( e );

    if( VERBOSE ) fprintf( stderr, "Done cleaning.\n" );
  }

  return;
}

/** FilterJunk:
 *  Filter out all of the extraneous lines and individual stairs and such that are
 *  in the plan, but not useful for these purposes.
 */
void FilterJunk( ugfile *uf )
{
  edge **e;
  int i, numEdges, maxEdges;

  if( REMOVEJUNK ){

    e = GetEdges( uf, &numEdges, &maxEdges );
    EdgeSort( e, 0, numEdges-1 );

    RemoveEdges( uf, e );
    RemoveDisjoint( uf );

    for( i = 0; i < numEdges; ++i ) free( e[i] );
    free( e );

    if( VERBOSE ) fprintf( stderr, "Done removing extraneous data.\n" );

  }

  return;
}

/** FilterDoors:
 *  The arcs are already marked, but the line segment is not.  FilterDoors
 *  finds the line segment by checking the vertices at the ends of the arcs.
 */
dwPts **FilterDoors( ugfile *uf )
{
  wire *w1, **doors;
  wnode *wn;
  vert *v1, *v2, *v3;
  int numDoors, maxDoors, i;
  int numVerts, maxVerts;
  dwPts **doorVerts;

  numVerts = 0;
  maxVerts = 64;
  doorVerts = (dwPts**) calloc( maxVerts, sizeof( dwPts* ) );

  numDoors = 0;
  maxDoors = 64;
  doors = (wire**) calloc( maxDoors, sizeof( wire* ) );

  w1 = uf->firstWire;
  do{
    /** Search each wire to see if it's already in the door layer.  If it is,
	check its end vertices for the other part of the door. **/
    if( !strcmp( w1->layer->name, "A_DOOR" ) ){
      v1 = w1->vl->v;
      v2 = w1->vl->prev->v;

      if( numVerts >= maxVerts ){
	maxVerts *= 2;
	doorVerts = (dwPts**) realloc( doorVerts, maxVerts * sizeof( dwPts* ) );
      }

      if( v1->wl->next == v1->wl->prev && v1->wl->next != v1->wl &&
	  !v1->wl->w->isFace && !v1->wl->next->w->isFace && GoodLayer( v1->wl->w ) ){
	/** Save the point not used in the rest of the function, as well as the
	    wire, so that we can locate the door jamb **/
	v3 = v1;
	doorVerts[numVerts] = (dwPts*) malloc( sizeof( dwPts ) );
	doorVerts[numVerts]->w[0] = w1;
	doorVerts[numVerts]->v[0] = v2;
      }else if( v2->wl->next == v2->wl->prev && v2->wl->next != v2->wl &&
		!v2->wl->w->isFace && !v2->wl->next->w->isFace && GoodLayer( v2->wl->w ) ){
	/** Save the point not used in the rest of the function, as well as the
	    wire, so that we can locate the door jamb **/
	v3 = v2;
	doorVerts[numVerts] = (dwPts*) malloc( sizeof( dwPts ) );
	doorVerts[numVerts]->w[0] = w1;
	doorVerts[numVerts]->v[0] = v1;
      }else{
	v3 = NULL;
	while( strcmp( JUNKLAYER, w1->layer->name ) )
	  w1->layer = w1->layer->next;
      }

      if( v3 != NULL ){
	/** If a match was found for this door, move it into the A_DOOR layer and
	    save a reference to it in doorVerts for use in finding door jambs **/
	wn = v3->wl;
	if( wn->w == w1 ) wn = wn->next;

	if( numDoors >= maxDoors ){
	  maxDoors *= 2;
	  doors = (wire**) realloc( doors, maxDoors * sizeof( wire ) );
	}

	doorVerts[numVerts]->w[1] = wn->w;
	if( wn->w->vl->v == v3 ) doorVerts[numVerts]->v[1] = wn->w->vl->prev->v;
	else                     doorVerts[numVerts]->v[1] = wn->w->vl->v;

	doors[numDoors] = wn->w;
	++numVerts;
	++numDoors;
      }
    }

    w1 = w1->next;
  } while( w1 != uf->firstWire );

  /** Go through each of the wires in the list I've compiled and change the layer
      to A_DOOR **/
  for( i = 0; i < numDoors; ++i ){
    w1 = doors[i];
    while( strcmp( w1->layer->name, "A_DOOR" ) )
      w1->layer = w1->layer->next;
  }

  if( VERBOSE ) fprintf( stderr, "  o Found %d doors.\n", numDoors );

  if( numVerts >= maxVerts ){
    doorVerts = (dwPts**) realloc( doorVerts, ( maxVerts + 1 ) * sizeof( dwPts* ) );
  }

  doorVerts[numVerts] = NULL;

  return doorVerts;
}

/** FilterElevatorsStairs:
 *  Move Elevators into a special layer.  Elevators are merely large faces
 *  that sometimes have a large 'X' in them.
 */
void FilterElevatorsStairs( ugfile *uf )
{
  wire *w;
  vert *v1, *v2, *old;
  vnode *vn;
  int i, minElWidth, minElLength, minStWidth;
  int numElevators, numStairs;
  float sides[4], angles[3];
  bool acceptable, isStairs, isThin;

  minElLength  = 45;
  minElWidth   = 40;
  minStWidth   = 10;
  numElevators = 0;
  numStairs    = 0;

  /** Go through all of the wires in uf **/
  w = uf->firstWire;
  do{

    if( w->isFace ){
      acceptable = true;

      /** Make sure w has 4 corners. **/
      vn = w->vl;
      v1 = vn->prev->v;
      i = 0;
      old = NULL;

      do{

	v2 = vn->v;
	assert( v1 != v2 );

	if( i == 4 ) acceptable = false;
	if( acceptable ){
	  /** Copy the length of the line segment and the angle between this
	      and the previous segment into their respective arrays **/

	  assert( i < 4 );
	  sides[i] = Length( v1->x, v1->y, v2->x, v2->y );
	  if( old != NULL ) angles[i-1] = Angle( old, v1, v2 );

	}

	old = v1;
	v1 = v2;

	++i;
	vn = vn->next;
      } while( vn != w->vl );

      if( acceptable && i == 4 ){
	/** It has 4 sides.  Check the size of the angles and sides to make sure
	    it qualifies as an elevator or staircase. **/

	acceptable = false;

	if( sides[0] >= minStWidth && sides[1] >= minStWidth &&
	    sides[2] >= minStWidth && sides[3] >= minStWidth ){

	  acceptable = true;
	  isStairs = false;
	  isThin = false;
	  for( i = 0; i < 4; ++i ){
	    if( sides[i] < minElLength ){
	      if( !isStairs ) isStairs = true;
	      else if( !isThin ) isThin = true;
	      else{
		acceptable = false;
		break;
	      }
	    }
	  }

	}

	if( acceptable ){
	  if( angles[0] == 90 || angles[1] == 90 || angles[2] == 90 ){
	    /** At least two angles must == 90 degrees **/

	    isStairs = false;
	    if( angles[0] != 90 || angles[1] != 90 || angles[2] != 90 ) isStairs = true;
	    i = FilterStairs( uf, w );

	    if( !isStairs && i == 0 ){
	      /** This is an elevator.  It is a rectangle, and there are no individual
		  steps associated with it. **/

	      if( sides[0] >= minElWidth && sides[1] >= minElWidth &&
		  sides[2] >= minElWidth && sides[3] >= minElWidth &&
		  ( ( sides[0] >= minElLength && sides[2] >= minElLength ) ||
		    ( sides[1] >= minElLength && sides[3] >= minElLength ) ) ){

		while( strcmp( w->layer->name, "A_ELEVATOR" ) )
		  w->layer = w->layer->next;
		++numElevators;

	      }

	    }else if( i > 0 ){
	      /** This is a staircase.  There are stairs associated with it. **/

	      while( strcmp( w->layer->name, "A_STAIRCASE" ) )
		w->layer = w->layer->next;
	      ++numStairs;

	    }
	  }
	}
      }
    }

    w = w->next;
  } while( w != uf->firstWire );

  if( VERBOSE ) fprintf( stderr, "  o Found %d elevators, %d staircases (or parts thereof).\n",
			 numElevators, numStairs );

  return;
}

/** FilterStairs:
 *  Pick out stairs that fit within the outline of the staircase, *f.  Mark
 *  them into the junk layer, and return the number of stairs found.  The
 *  reason I mark them as junk is that individual stairs are eye-candy for
 *  people looking at the floor plan.  They have no bearing on how many
 *  stairs are actually present.
 */
int FilterStairs( ugfile *uf, wire *f )
{
  vert *v1, *v2, *v3, *v4, *between;
  wnode *wn1, *wn2;
  wire *w;
  int counter;
  bool elevator;

  elevator = false;
  counter  = 0;
  v1 = f->vl->v;
  v2 = f->vl->next->v;
  v3 = f->vl->next->next->v;
  v4 = f->vl->prev->v;

  assert( v4 == f->vl->next->next->next->v );

  /** First see if there's an 'x' in the box (meaning it's an elevator) **/
  wn1 = v1->wl;
  do{

    wn2 = v3->wl;
    do{
      if( wn1->w == wn2->w && wn1->w != f ) elevator = true;
      else wn2 = wn2->next;
    } while( wn2 != v3->wl && !elevator );

    if( !elevator ) wn1 = wn1->next;
  } while( wn1 != v1->wl && !elevator );

  if( elevator ){
    /** It's an elevator with an 'x' in it.  Some don't have this so if you don't get
	inside this if-statement, it doesn't mean that f isn't an elevator. **/

    while( strcmp( wn1->w->layer->name, JUNKLAYER ) )
      wn1->w->layer = wn1->w->layer->next;

    elevator = false;

    wn1 = v2->wl;
    do{

      wn2 = v4->wl;
      do{
	if( wn1->w == wn2->w && wn1->w != f ) elevator = true;
	else wn2 = wn2->next;
      } while( wn2 != v4->wl && !elevator );

      if( !elevator ) wn1 = wn1->next;
    } while( wn1 != v2->wl && !elevator );

    if( elevator ){
      while( strcmp( wn1->w->layer->name, JUNKLAYER ) )
	wn1->w->layer = wn1->w->layer->next;
    }else{
      fprintf( stderr, "Error: (non-fatal) Found an elevator with half of an 'x'\n" );
    }

    return 0;
  }

  /** Now check for individual stairs **/
  w  = uf->firstWire;

  do{

    if( w->vl->next == w->vl->prev && GoodLayer( w ) ){
      /** Search only those wires that have two vertices **/

      between = w->vl->v;

      if( IsBetween( between, v1, v2 ) || IsBetween( between, v2, v3 ) ||
	  IsBetween( between, v3, v4 ) || IsBetween( between, v4, v1 ) ){

	between = w->vl->prev->v;

	if( IsBetween( between, v1, v2 ) || IsBetween( between, v2, v3 ) ||
	    IsBetween( between, v3, v4 ) || IsBetween( between, v4, v1 ) ){
	  /** It's a stair in this staircase.  Put it in the junk layer and increment
	      the counter. **/

	  while( strcmp( w->layer->name, JUNKLAYER ) )
	    w->layer = w->layer->next;
	  ++counter;

	}
      }
    }

    w = w->next;
  } while( w != uf->firstWire );

  return counter;
}

/** FilterWindows:
 *  Locate and mark all sets of three parallel lines of the same length for which
 *  the distance is the same between adjacent lines.
 */
dwPts **FilterWindows( ugfile *uf, edge **e, int *totalWins )
{
  edge **eRef[3];
  dwPts **winVerts;
  float dist1, dist2;
  int numWins, maxWins;
  int eLength[3];
  int i, j, k, l;
  bool found;

  /** Set up the data structure that will house the windows collected. **/
  numWins  = 0;
  maxWins  = 32;
  winVerts = (dwPts**) calloc( maxWins, sizeof( dwPts* ) );
  for( i = 0; i < maxWins; ++i ) winVerts[i] = NULL;

  /** Get arrays of edges for each direction.  Don't bother with the
      max size because the arrays aren't going to be resized.    **/
  eRef[0] = EdgeSubset( e, HORIZONTAL, &eLength[0], NULL );
  eRef[1] = EdgeSubset( e, VERTICAL,   &eLength[1], NULL );
  eRef[2] = EdgeSubset( e, SLOPED,     &eLength[2], NULL );

  /** Search each of the 3 arrays **/
  for( i = 0; i < 3; ++i ){
    EdgeSort( eRef[i], 0, eLength[i] - 1 );

    /** Check each of the edges in the array. **/
    for( j = 0; j < eLength[i]; ++j ){

      found = false;
      if( numWins >= maxWins ){

	/** Resize the winVerts array if need be **/
	maxWins *= 2;
	winVerts = (dwPts**) realloc( winVerts, maxWins * sizeof( dwPts* ) );
	for( k = numWins; k < maxWins; ++k ) winVerts[k] = NULL;

      }

      /** For each edge, determine if it has already been incorporated in another
	  window, if it is of an acceptable layer, if it has two vertices, and if
	  its wire is not a face. **/
      if( !eRef[i][j]->hasMatch && !eRef[i][j]->w->isFace && GoodLayer( eRef[i][j]->w ) &&
	  eRef[i][j]->w->vl->next == eRef[i][j]->w->vl->prev ){

	if( winVerts[numWins] == NULL )
	  winVerts[numWins] = (dwPts*) malloc( sizeof( dwPts ) );

	winVerts[numWins]->w[0] = eRef[i][j]->w;
	winVerts[numWins]->v[0] = eRef[i][j]->w->vl->v;
	winVerts[numWins]->v[1] = eRef[i][j]->w->vl->prev->v;

	/** Search for possible matches to the line segment already collected. **/
	for( k = j + 1; k < eLength[i] && !found; ++k ){

	  dist1 = WellPlaced( winVerts[numWins]->w[0], eRef[i][k]->w );
	  if( dist1 > MAXWIDTH ) break;

	  /** Go through the same testing as we did for the previous wire + make sure
	      dist1 > 0 **/
	  if( !eRef[i][k]->hasMatch && !eRef[i][k]->w->isFace && GoodLayer( eRef[i][k]->w )
	      && dist1 && eRef[i][k]->w->vl->next == eRef[i][k]->w->vl->prev ){

	    winVerts[numWins]->w[1] = eRef[i][k]->w;
	    winVerts[numWins]->v[2] = eRef[i][k]->w->vl->v;
	    winVerts[numWins]->v[3] = eRef[i][k]->w->vl->prev->v;

	    for( l = k + 1; l < eLength[i] && !found; ++l ){

	      dist2 = WellPlaced( winVerts[numWins]->w[0], eRef[i][l]->w );
	      if( dist2 > MAXWIDTH ) break;

	      /** Go through the riggorous weeding-out process, once more. **/
	      if( !eRef[i][k]->hasMatch && !eRef[i][k]->w->isFace && GoodLayer( eRef[i][k]->w )
		  && 2 * dist1 == dist2 && eRef[i][k]->w->vl->next == eRef[i][k]->w->vl->prev ){

		/** Found a complete window.  Add this wire to the current winVerts
		    and mark all of the used edges as matched. **/
		winVerts[numWins]->w[2] = eRef[i][l]->w;
		winVerts[numWins]->v[4] = eRef[i][l]->w->vl->v;
		winVerts[numWins]->v[5] = eRef[i][l]->w->vl->prev->v;

		eRef[i][j]->hasMatch = true;
		eRef[i][k]->hasMatch = true;
		eRef[i][l]->hasMatch = true;

		++numWins;
		found = true;

	      }
	    }  /** end for <l> **/
	  }
	}      /** end for <k> **/
      }
    }          /** end for <j> **/
  }            /** end for <i> **/

  if( numWins < maxWins ){
    free( winVerts[numWins] );
    winVerts[numWins] = NULL;
  }

  for( i = 0; i < numWins; ++i ){

    /** Move the located windows into the appropriate layer **/
    while( strcmp( winVerts[i]->w[0]->layer->name, "A_WINDOW" ) )
      winVerts[i]->w[0]->layer = winVerts[i]->w[0]->layer->next;
    winVerts[i]->w[1]->layer = winVerts[i]->w[0]->layer;
    winVerts[i]->w[2]->layer = winVerts[i]->w[0]->layer;

  }

  if( VERBOSE ) fprintf( stderr, "  o Found %d windows.\n", numWins );
  *totalWins = numWins;

  return winVerts;
}

/** FilterWindowJambs:
 *  Using the list of dwPts gotten in FilterWindows, find their jambs and mark
 *  them as such.  Also, if such a line segment doesn't exist, try to make one.
 */
void FilterWindowJambs( ugfile *uf, dwPts **winVerts, int numWins )
{
  vert *vList[4], *v1, *v2;
  wire *modWire;
  vnode *vn;
  wnode *wn1, *wn2;
  int i, j, addedWires;

  addedWires = 0;

  for( i = 0; i < numWins; ++i ){

    /** Pick out the vertices on which we will find the common wire. **/
    vList[0] = winVerts[i]->v[0];
    vList[1] = winVerts[i]->v[1];
    vList[2] = winVerts[i]->v[4];

    if( Length( vList[0]->x, vList[0]->y, vList[2]->x, vList[2]->y ) <
	Length( vList[1]->x, vList[1]->y, vList[2]->x, vList[2]->y ) ){

      vList[3] = winVerts[i]->v[5];

    }else{

      vList[2] = winVerts[i]->v[5];
      vList[2] = winVerts[i]->v[4];

    }

    for( j = 0; j < 2; ++j ){

      wn1 = vList[ j ]->wl;
      wn2 = vList[j+2]->wl;

      /** Find the common wire to mark as A_WINDOW_JAMB **/
      do{

	do{
	  if( wn1->w != wn2->w ) wn2 = wn2->next;
	} while( wn1->w != wn2->w && wn2 != vList[j+2]->wl );

	if( wn1->w != wn2->w ) wn1 = wn1->next;
      } while( wn1->w != wn2->w && wn1 != vList[j]->wl );

      v1 = vList[j];
      v2 = vList[j+2];

      if( wn1->w == wn2->w ){

	/** Mark the common wire as A_WINDOW_JAMB **/
	vn = wn1->w->vl;
	while( vn->v != v1 ) vn = vn->next;
	if( vn->prev->v == v2 ){
	  v1 = v2;
	  v2 = vList[j];
	}

	modWire = SplitWire( uf, wn1->w, v1, v2 );
	while( strcmp( modWire->layer->name, "A_WINDOW_JAMB" ) )
	  modWire->layer = modWire->layer->next;

      }else{

	/** Didn't find a common wire.  Add one. **/
	modWire = AddWire( uf, NULL );
	AddVertWire( v1, modWire, NULL, NULL );
	AddVertWire( v2, modWire, NULL, NULL );
	while( strcmp( modWire->layer->name, "A_WINDOW_JAMB" ) )
	  modWire->layer = modWire->layer->next;
	++addedWires;

      }
    }   /** end for <j> **/
  }     /** end for <i> **/

  if( VERBOSE ) fprintf( stderr, "  o Generated %d new wires as window jambs.\n", addedWires );

  return;
}

/** FilterDoorJambs:
 *  Using the vertices gotten while searching for doors, find the door jambs
 *  by searching for the closest vertices.  Sometimes a vertex doesn't exist
 *  where it should.  If this is the case, use some intelligence in creating
 *  a new one.
 */
void FilterDoorJambs( ugfile *uf, dwPts **doorVerts )
{
  dwPts *current, *correction;
  vert **vlist, *v1, *v2, *vref1, *vref2;
  vnode *vn;
  wire *w1, *w2;
  wnode *wn;
  int i, j, k, m, n, dir, create;
  float distance;
  bool found;

  /** Make a copy of the list of vertices in uf so that they can be sorted. **/
  vlist = (vert**) calloc( uf->hVert, sizeof( vert ) );
  for( i = 0; i < uf->hVert; ++i ){
    vlist[i] = uf->vertList[i];
    vlist[i]->sortValue = vlist[i]->x + vlist[i]->y + vlist[i]->z;
  }

  VertSort( vlist, 0, uf->hVert - 1 );

  for( i = 0; i < uf->hVert; ++i ) vlist[i]->location = i;

  create = n = i = 0;
  for( current = doorVerts[i]; current != NULL; ++i ){
    /** Go through all of the vertices at the ends of the doors, searching for
	door jambs **/
    current = doorVerts[i];
    if( current == NULL ) break;
    vref1 = current->w[1]->vl->v;
    vref2 = current->w[1]->vl->next->v;

    assert( vref1 != vref2 );

    for( j = 0; j < 2; ++j ){
      found = false;
      w1 = current->w[j];
      v1 = current->v[j];

      if( v1->wl != v1->wl->next ){
	/** Check if the vertex already exists as a part of the wire containing
	    the door jamb. **/
	wn = v1->wl;
	do{
	  if( GoodLayer( wn->w ) && wn->w != w1 ){
	    assert( wn->w != current->w[0] );
	    assert( wn->w != current->w[1] );
	    vn = wn->w->vl;
	    while( vn->v != v1 ) vn = vn->next;

	    w2 = wn->w;

	    /** Compare the slope of the edge to the flat edge of the door.
		Then make sure the size is appropriate. **/
	    if( vref1->x == vref2->x ){
	      /** Vertical **/

	      if( v1->x == vn->next->v->x && fabs( v1->y - vn->next->v->y ) <= MAXWIDTH ){
		found = true;
		w2 = SplitWire( uf, w2, v1, vn->next->v );
	      }else if( v1->x == vn->prev->v->x && fabs( v1->y - vn->prev->v->y ) <= MAXWIDTH ){
		if( w2->isFace || vn != w2->vl ){
		  found = true;
		  w2 = SplitWire( uf, w2, vn->prev->v, v1 );
		}
	      }
	    }else if( vref1->y == vref2->y ){
	      /** Horizontal **/

	      if( v1->y == vn->next->v->y && fabs( v1->x - vn->next->v->x ) <= MAXWIDTH ){
		found = true;
		w2 = SplitWire( uf, w2, v1, vn->next->v );
	      }else if( v1->y == vn->prev->v->y && fabs( v1->x - vn->prev->v->x ) <= MAXWIDTH ){
		if( w2->isFace || vn != w2->vl ){
		  found = true;
		  w2 = SplitWire( uf, w2, vn->prev->v, v1 );
		}
	      }
	    }else if( v1->x != vn->next->v->x ){
	      /** Sloped **/

	      if( ( v1->y - vn->next->v->y ) / ( v1->x - vn->next->v->x ) ==
		  ( vref1->y - vref2->y ) / ( vref1->x - vref2->x ) ){
		if( Length( v1->x, v1->y, vn->next->v->x, vn->next->v->y ) < MAXWIDTH ){
		  found = true;
		  w2 = SplitWire( uf, w2, v1, vn->next->v );
		}
	      }
	    }else if( v1->x != vn->prev->v->x && w2->isFace ){
	      /** Sloped **/

	      if( ( v1->y - vn->prev->v->y ) / ( v1->x - vn->prev->v->x ) ==
		  ( vref1->y - vref2->y ) / ( vref1->x - vref2->x ) ){
		if( Length( v1->x, v1->y, vn->prev->v->x, vn->prev->v->y ) < MAXWIDTH ){
		  found = true;
		  w2 = SplitWire( uf, w2, vn->prev->v, v1 );
		}
	      }
	    }
	  }
	  wn = wn->next;
	} while( wn != v1->wl && !found );
      }

      if( !found ){
	/** If it hasn't been found, we have to go about it a different way.  Search the
	    list of vertices for points that are reasonably close, and check their wire
	    lists for matches **/
	dir = 1;
	for( k = 0; k < 2 && !found; ++k ){
	  m = v1->location + dir;

	  while( m < uf->hVert && m >= 0 && !found &&
		 abs( v1->sortValue - vlist[m]->sortValue ) <= 1.5 * MAXWIDTH ){
	    v2 = vlist[m];
	    distance = Length( v1->x, v1->y, v2->x, v2->y );

	    if( distance <= MAXWIDTH && SameAngle( v1, v2, vref1, vref2 ) ){
	      /** At this point, v2 meets all the criteria for being a vertex on a door
		  jamb.  Now check for a corresponding vertex. **/
	      wn = v2->wl;
	      do{
		if( GoodLayer( wn->w ) ){
		  vn = wn->w->vl;
		  while( vn->v != v2 ) vn = vn->next;

		  if( Length( v2->x, v2->y, vn->next->v->x, vn->next->v->y )
		      <= MAXWIDTH && IsBetween( v1, v2, vn->next->v ) ){
		    /** We have a match **/
		    found = true;
		    w2 = SplitWire( uf, wn->w, v2, vn->next->v );
		  }else if( Length( v2->x, v2->y, vn->next->v->x, vn->prev->v->y )
			    <= MAXWIDTH && IsBetween( v1, v2, vn->prev->v ) ){
		    /** Found a door jamb **/
		    found = true;
		    w2 = SplitWire( uf, wn->w, vn->prev->v, v2 );
		  }
		}

		wn = wn->next;
	      } while( wn != v2->wl && !found );
	    }

	    m += dir;
	  } /** end while **/

	  dir = -1;
	}
      }

      /** If it still hasn't been found, try to add a point. **/
      if( !found ){
	correction = Correct( uf, v1, vref1, vref2 );
	if( correction != NULL ){
	  ++create;
	  found = true;
	  w2 = SplitWire( uf, correction->w[0], correction->v[0], correction->v[1] );
	  free( correction );
	}
      }

      if( found ){
	++n;
	assert( GoodLayer( w2 ) );
	while( strcmp( w2->layer->name, "A_DOOR_JAMB" ) )
	  w2->layer = w2->layer->next;
      }

    }       /** end for < 0 <= j <= 1 > **/
  }

  /** Clean up the doors (merge the wires to make a single door) **/
  for( j = 0; doorVerts[j] != NULL; ++j ){
    doorVerts[j]->w[2] = MergeWires( doorVerts[j]->w[0], doorVerts[j]->w[1] );
  }

  for( j = 0; j < uf->hWire; ++j ){
    if( uf->wireList[j] != NULL && !strcmp( uf->wireList[j]->name, "extra" ) ){
      free( uf->wireList[j] );
      uf->wireList[j] = NULL;
    }
  }

  /** Print out statistics **/
  if( VERBOSE ){
    fprintf( stderr, "  o Found %d door jambs out of an expected %d.\n", n, 2 * i );
    if( n < 2 * i ) fprintf( stderr, "    The discrepency may mean nothing.\n" );
    fprintf( stderr, "    Out of those, %d were created.\n", create );
  }

  return;
}

/** FixFaces:
 *  Search through all of the wires and figure out which ones need to be made faces
 *  by identifying the first and last vertex and comparing them.
 */
void FixFaces( ugfile *uf )
{
  wire *w;
  vnode *vn1, *vn2;
  int counter;

  w = uf->firstWire;
  counter = 0;

  do{

    assert( w != NULL );
    if( !w->isFace ){

      vn1 = w->vl;
      vn2 = w->vl->prev;

      if( vn1->v == vn2->v ){
	/** We've identified that it's not already a face, but that its' first and
	    last points are the same. **/

	vn2->prev->next = vn1;
	vn1->prev = vn2->prev;
	w->isFace = true;
	free( vn2 );
	++counter;

      }else if( vn1->v->x == vn2->v->x && vn1->v->y == vn2->v->y ){
	/** It's not a face, and the two points aren't the same point, but they are
	    on top of one another. **/

	if( vn1->v->wl == vn1->v->wl->next ){
	  /** I'll get rid of vn1 because it's a 'loner' vertex, and we can get rid
	      of it all together when we write the file back out. **/

	  vn2 = vn1;

	}

	RemoveVertWire( vn2->v, w );
	w->isFace = true;
	++counter;

      }
    }

    w = w->next;
  } while( w != uf->firstWire );

  if( VERBOSE ) fprintf( stderr, "  o Changed %d wires into faces.\n", counter );

  return;
}

/** FixPlan:
 *  Add vertices where lines intersect.  This makes it easier for me to seperate
 *  out wires that belong in the junk layer, and easier for BMG to follow wall
 *  contours.
 */
void FixPlan( ugfile *uf, edge ***e, int *le, int *me )
{
  edge **vertical, **horizontal, **sloped;
  int lv, mv, lh, mh, ls, ms;     /** Counters for the arrays of edges **/
  int i, j, k, counter, iSect, badVal;

  counter = 0;
  iSect = 0;
  badVal = 0;

  /** Seperate the edges into v, h, and s because vertical lines aren't going to
      intersect vertical lines, etc. **/

  vertical   = EdgeSubset( e[0], VERTICAL, &lv, &mv );
  horizontal = EdgeSubset( e[0], HORIZONTAL, &lh, &mh );
  sloped     = EdgeSubset( e[0], SLOPED, &ls, &ms );

  for( i = 0; i < lv; ++i ){
    if( GoodLayer( vertical[i]->w ) ){

      /** Compare all vertical edges with horizontal edges for intersection points **/
      for( j = 0; j < lh; ++j ){
	if( GoodLayer( horizontal[j]->w ) ){
	  k = Intersect( uf, &vertical, &horizontal, e, i, j, &lv, &mv, &lh, &mh, le, me );

	  switch( k ){
	  case 2:
	    ++iSect;
	  case 1:
	    ++counter;
	    break;
	  case -1:
	    fprintf( stderr, "Error: On Line %d.  Both vertices in one edge were the same.\n", __LINE__ );
	  case -2:
	    ++badVal;
	  default:
	    break;
	  }

	}
      }             /** end for <horizontal counter> **/

      /** Compare all vertical edges with sloped edges **/
      for( j = 0; j < ls; ++j ){
	if( GoodLayer( sloped[j]->w ) ){
	  k = Intersect( uf, &vertical, &sloped, e, i, j, &lv, &mv, &ls, &ms, le, me );

	  switch( k ){
	  case 2:
	    ++iSect;
	  case 1:
	    ++counter;
	    break;
	  case -1:
	    fprintf( stderr, "Error: On Line %d.  Both vertices in one edge were the same.\n", __LINE__ );
	  case -2:
	    ++badVal;
	  default:
	    break;
	  }

	}
      }             /** end for <sloped counter> **/
    }
  }                 /** end for <vertical counter> **/

  for( i = 0; i < lh; ++i ){
    if( GoodLayer( horizontal[i]->w ) ){
      for( j = 0; j < ls; ++j ){
	if( GoodLayer( sloped[j]->w ) ){
	  k = Intersect( uf, &horizontal, &sloped, e, i, j, &lh, &mh, &ls, &ms, le, me );

	  switch( k ){
	  case 2:
	    ++iSect;
	  case 1:
	    ++counter;
	    break;
	  case -1:
	    fprintf( stderr, "Error: On Line %d.  Both vertices in one edge were the same.\n", __LINE__ );
	  case -2:
	    ++badVal;
	  default:
	    break;
	  }

	}
      }             /** end for <sloped counter> **/
    }
  }                 /** end for <horizontal counter> **/

  if( VERBOSE ) fprintf( stderr, "  o Added %d points, %d intersections.  %d bad values.\n",
			 counter, iSect, badVal );

  return;
}


/** RemoveEdges:
 *  Some of the extraneous lines appear to be a two dimensional wall.  This
 *  searches for those edges that don't have another, parallel edge, and
 *  filters them into the A_STOOPID layer.
 */
void RemoveEdges( ugfile *uf, edge **e )
{
  edge **vertical, **horizontal, **sloped;
  int vh, hh, sh;
  vert *v1, *v2, *v3, *v4;
  float minLength, maxError;
  int i, j, adder;

  minLength = .3;
  maxError  = .01;

  vertical   = EdgeSubset( e, VERTICAL, &vh, NULL );
  horizontal = EdgeSubset( e, HORIZONTAL, &hh, NULL );
  sloped     = EdgeSubset( e, SLOPED, &sh, NULL );

  /** Check each of the vertical edges to search for parallel lines **/
  for( i = 0; vertical[i] != NULL; ++i ){
    if( !vertical[i]->hasMatch && GoodLayer( vertical[i]->w ) && !vertical[i]->w->isFace ){

      /** Get the edge's two vertices **/
      v1 = vertical[i]->v[0];
      v2 = vertical[i]->v[1];

      if( fabs( v1->y - v2->y ) <= MAXWIDTH ){
	/** It's where an actual wall sticks out into a room.  We'll pretend
	    that it's got a match so that it doesn't get modified, later. **/
	vertical[i]->hasMatch = true;
      }else{

	if( v1->y > v2->y ){
	  v1 = v2;
	  v2 = vertical[i]->v[0];
	}

	j = i + 1;
	if( vertical[j] != NULL ){
	  while( fabs( vertical[j]->eq[B] - vertical[i]->eq[B] ) < MAXWIDTH ){
	    if( vertical[j]->eq[B] != vertical[i]->eq[B] &&
		vertical[j]->w->layer == vertical[i]->w->layer ){
	      /** They aren't on the same line, but they are parallel.  Now, just check
		  to see if they're close to eachother **/
	      v3 = vertical[j]->v[0];
	      v4 = vertical[j]->v[1];

	      if( ( v1->y < v3->y && v3->y < v2->y ) ||
		  ( v1->y < v4->y && v4->y < v2->y ) ||
		  ( v3->y < v1->y && v1->y < v4->y ) ||
		  ( v3->y < v2->y && v2->y < v4->y ) ||
		  ( v1->y == v3->y && v2->y == v4->y ) ||
		  ( v1->y == v4->y && v2->y == v3->y ) ){
		/** Just one test left: The corresponding wall has to be greater than or
		    equal to minLength * the length of this wall. **/
		if( fabs( v3->y - v4->y ) >= minLength * fabs( v1->y - v2->y ) )
		  vertical[i]->hasMatch = true;

	      }
	    }
	    ++j;
	    if( vertical[j] == NULL ) break;
	  } /** end while **/
	}

	j = i - 1;
	if( j >= 0 ){
	  while( fabs( vertical[j]->eq[B] - vertical[i]->eq[B] ) < MAXWIDTH && j > 0 ){
	    if( vertical[j]->eq[B] != vertical[i]->eq[B] &&
		vertical[j]->w->layer == vertical[i]->w->layer ){
	      /** They aren't on the same line, but they are parallel.  Now, just check
		  to see if they're close to each other **/
	      v3 = vertical[j]->v[0];
	      v4 = vertical[j]->v[1];
	      if( v3->y > v4->y ){
		v3 = v4;
		v4 = vertical[j]->v[0];
	      }

	      if( ( v1->y < v3->y && v3->y < v2->y ) ||
		  ( v1->y < v4->y && v4->y < v2->y ) ||
		  ( v3->y < v1->y && v1->y < v4->y ) ||
		  ( v3->y < v2->y && v2->y < v4->y ) ||
		  ( v1->y == v3->y && v2->y == v4->y ) ||
		  ( v1->y == v4->y && v2->y == v3->y ) ){
		/** Just one test left: The corresponding wall has to be greater than or
		    equal to minLength * the length of this wall. **/
		if( fabs( v3->y - v4->y ) >= minLength * fabs( v1->y - v2->y ) )
		  vertical[i]->hasMatch = true;
	      }
	    }
	    --j;
	  } /** end while **/
	}
      }
    }
  }       /** end for **/

  /** Check each of the horizontal edges to search for parallel lines **/
  for( i = 0; horizontal[i] != NULL; ++i ){
    if( !horizontal[i]->hasMatch && GoodLayer( horizontal[i]->w ) && !horizontal[i]->w->isFace ){

      /** Get the edge's two vertices **/
      v1 = horizontal[i]->v[0];
      v2 = horizontal[i]->v[1];

      if( fabs( v1->x - v2->x ) <= MAXWIDTH ){
	/** It's where an actual wall sticks out into a room.  We'll pretend
	    that it's got a match so that it doesn't get modified, later. **/
	horizontal[i]->hasMatch = true;
      }else{

	if( v1->x > v2->x ){
	  v1 = v2;
	  v2 = horizontal[i]->v[0];
	}

	j = i + 1;
	if( horizontal[j] != NULL ){
	  while( fabs( horizontal[j]->eq[B] - horizontal[i]->eq[B] ) < MAXWIDTH ){
	    if( horizontal[j]->eq[B] != horizontal[i]->eq[B] &&
		horizontal[j]->w->layer == horizontal[i]->w->layer ){
	      /** They aren't on the same line, but they are parallel.  Now, just check
		  to see if they're close to eachother **/
	      v3 = horizontal[j]->v[0];
	      v4 = horizontal[j]->v[1];
	      if( v3->x > v4->x ){
		v3 = v4;
		v4 = horizontal[j]->v[0];
	      }

	      if( ( v1->x < v3->x && v3->x < v2->x ) ||
		  ( v1->x < v4->x && v4->x < v2->x ) ||
		  ( v3->x < v1->x && v1->x < v4->x ) ||
		  ( v3->x < v2->x && v2->x < v4->x ) ||
		  ( v1->x == v3->x && v2->x == v4->x ) ||
		  ( v1->x == v4->x && v2->x == v3->x ) ){
		/** Just one test left: The corresponding wall has to be greater than or
		    equal to minLength * the length of this wall. **/
		if( fabs( v3->x - v4->x ) >= minLength * fabs( v1->x - v2->x ) )
		  horizontal[i]->hasMatch = true;
	      }
	    }
	    ++j;
	    if( horizontal[j] == NULL ) break;
	  } /** end while **/
	}

	j = i - 1;
	if( j >= 0 ){
	  while( fabs( horizontal[j]->eq[B] - horizontal[i]->eq[B] ) < MAXWIDTH && j > 0 ){
	    if( horizontal[j]->eq[B] != horizontal[i]->eq[B] &&
		horizontal[j]->w->layer == horizontal[i]->w->layer ){
	      /** They aren't on the same line, but they are parallel.  Now, just check
		  to see if they're close to each other **/
	      v3 = horizontal[j]->v[0];
	      v4 = horizontal[j]->v[1];
	      if( v3->x > v4->x ){
		v3 = v4;
		v4 = horizontal[j]->v[0];
	      }

	      if( ( v1->x < v3->x && v3->x < v2->x ) ||
		  ( v1->x < v4->x && v4->x < v2->x ) ||
		  ( v3->x < v1->x && v1->x < v4->x ) ||
		  ( v3->x < v2->x && v2->x < v4->x ) ||
		  ( v1->x == v3->x && v2->x == v4->x ) ||
		  ( v1->x == v4->x && v2->x == v3->x ) ){
		/** Just one test left: The corresponding wall has to be greater than or
		    equal to minLength * the length of this wall. **/
		if( fabs( v3->x - v4->x ) >= minLength * fabs( v1->x - v2->x ) )
		  horizontal[i]->hasMatch = true;
	      }
	    }
	    --j;
	  } /** end while **/
	}
      }
    }
  }       /** end for **/

  /** Check each of the sloped edges to search for parallel lines **/
  for( i = 0; sloped[i] != NULL; ++i ){
    if( !sloped[i]->hasMatch && GoodLayer( sloped[i]->w ) && !sloped[i]->w->isFace ){

      /** Get the edge's two vertices **/
      v1 = sloped[i]->v[0];
      v2 = sloped[i]->v[1];

      if( Length( v1->x, v1->y, v2->x, v2->y ) <= MAXWIDTH ){
	/** It's where an actual wall sticks out into a room.  We'll pretend
	    that it's got a match so that it doesn't get modified, later. **/

	sloped[i]->hasMatch = true;

      }else{

	if( v1->y > v2->y ){
	  v1 = v2;
	  v2 = sloped[i]->v[0];
	}

	adder = 1;

	for( adder = 1; adder != 0; ++adder ){
	  if( adder == 2 ) adder = -1;
	  j = i + adder;
	  if( j >= 0 && sloped[j] != NULL ){
	    while( fabs( sloped[j]->eq[M] - sloped[i]->eq[M] ) <= maxError ){

	      if( sloped[j]->eq[B] != sloped[i]->eq[B] &&
		  fabs( sloped[j]->eq[B] - sloped[j]->eq[B] ) <= MAXWIDTH &&
		  sloped[j]->w->layer == sloped[i]->w->layer ){
		/** They aren't on the same line, but they are parallel.  Now, just check
		    to see if they're close to eachother **/
		v3 = sloped[j]->v[0];
		v4 = sloped[j]->v[1];

		if( ( v1->y <= v3->y && v3->y <= v2->y ) ||
		    ( v1->y <= v4->y && v4->y <= v2->y ) ||
		    ( v3->y <= v1->y && v1->y <= v4->y ) ||
		    ( v3->y <= v2->y && v2->y <= v4->y ) ||
		    ( v1->y == v3->y && v2->y == v4->y ) ||
		    ( v1->y == v4->y && v2->y == v3->y ) ){
		  /** Just one test left: The corresponding wall has to be greater than or
		      equal to minLength * the length of this wall. **/
		  if( fabs( v3->y - v4->y ) >= minLength * fabs( v1->y - v2->y ) )
		    sloped[i]->hasMatch = true;
		}
	      }
	      j += adder;
	      if( j < 0 || sloped[j] == NULL ) break;
	    } /** end while **/
	  }
	}

      }
    }
  }       /** end for **/

  /** I wait to do the modifications, themselves, until the end so that the splitting
      of wires doesn't cause problems. **/

  Unmark( uf, vertical );
  Unmark( uf, horizontal );
  Unmark( uf, sloped );

  return;
}

/** Unmark:
 *  Search through the list of edges for those without matches, and move them into
 *  the layer, A_STOOPID.
 */
void Unmark( ugfile *uf, edge **e )
{
  int i;

  for( i = 0; e[i] != NULL; ++i ){
    if( !e[i]->hasMatch ){

      if( !( VertHasWire( e[i]->v[0], e[i]->w ) && VertHasWire( e[i]->v[1], e[i]->w ) ) ){
	e[i]->w = CommonWire( e[i]->v[0], e[i]->v[1] );
	assert( e[i]->w != NULL );
      }

      if( GoodLayer( e[i]->w ) && !e[i]->w->isFace ){
	e[i]->w = SplitWire( uf, e[i]->w, e[i]->v[0], e[i]->v[1] );

	while( strcmp( e[i]->w->layer->name, JUNKLAYER ) )
	  e[i]->w->layer = e[i]->w->layer->next;
      }
    }
  }

  return;
}

/** RemoveDisjoint:
 *  Get rid of wires [with only two vertices] for which one or both ends only
 *  connect to wires that are already in the junk layer.
 */
void RemoveDisjoint( ugfile *uf )
{
  bool valid;
  wire *w1;
  vert *v1, *v2;
  wnode *wn;
  lnode *junk;

  junk = uf->layerList;
  while( strcmp( junk->name, JUNKLAYER ) )
    junk = junk->next;

  w1 = uf->firstWire;
  do{
    /** Loop iterates through every wire in the file **/

    valid = false;

    if( GoodLayer( w1 ) && !w1->isFace ){
      /** There are only two points in this wire, and it's worth checking to see
          if it should be moved to A_STOOPID **/

      assert( w1->vl != NULL );

      v1 = w1->vl->v;
      v2 = w1->vl->prev->v;

      assert( v1 != v2 );

      wn = v1->wl;
      do{

	if( wn->w->layer == junk ){
	  valid = true;
	}else if( wn->w == w1 ){
	  /** Do nothing **/
	}else{
	  /** It attached to a wire that was not of the layer: JUNKLAYER **/
	  valid = false;
	  break;
	}

	wn = wn->next;
      } while( wn != v1->wl );

      if( !valid ){
	/** If one end was okay, check the other one **/
	wn = v2->wl;
	do{

	  if( wn->w->layer == junk ){
	    valid = true;
	  }else if( wn->w == w1 ){
	    /** Do nothing **/
	  }else{
	    /** It attached to a wire that was not of the layer: JUNKLAYER **/
	    valid = false;
	    break;
	  }

	  wn = wn->next;
	} while( wn != v2->wl );
      }

      if( valid || ( !w1->isFace && ( v1->wl == v1->wl->next || v2->wl == v2->wl->next ) ) ){
	/** If one of the ends was connected to a junk wire or one of the
	    ends was sticking out into nothingness, mark this as junk. **/
	while( strcmp( w1->layer->name, JUNKLAYER ) )
	  w1->layer = w1->layer->next;
      }
    }

    w1 = w1->next;
  } while( w1 != uf->firstWire );

  return;
}

/** GoodLayer:
 *  Useful for shortening those lengthy if-statements.  GoodLayer checks to see
 *  if the wire that is passed is in a layer that should be modified.
 */
bool GoodLayer( wire *w )
{
  if( !strcmp( w->layer->name, "A_WALL" ) ||
      !strcmp( w->layer->name, "A_WALL_CORE" ) )
    return true;

  return false;
}

/** AddLayer:
 *  Add a layer called A_STOOPID for lines that don't belong in the
 *  drawing if it doesn't already exist.  Return true if it was added
 *  and false if it already existed.  Also, add layers for elevators
 *  and stairs if they don't already exist.
 */
bool
AddLayer( ugfile *uf )
{
  lnode *currentLayer;
  ugstr *layerStr;
  bool junk, stair, elevator;

  junk     = false;
  stair    = false;
  elevator = false;

  /** Check the existing layers for A_STOOPID **/
  currentLayer = uf->layerList;
  assert( currentLayer != NULL );

  do{
    if( !strcmp( currentLayer->name, JUNKLAYER ) ){
      currentLayer->r = 1.0;
      currentLayer->g = 0.1;
      currentLayer->b = 0.1;
      junk = true;
    }else if( !strcmp( currentLayer->name, "A_STAIRCASE" ) ){
      currentLayer->r = 1.0;
      currentLayer->g = 1.0;
      currentLayer->b = 1.0;
      stair = true;
    }else if( !strcmp( currentLayer->name, "A_ELEVATOR" ) ){
      elevator = true;
    }
    currentLayer = currentLayer->next;
  } while( currentLayer != uf->layerList );

  /** If STOOPID didn't already exist, add it now **/
  for( layerStr = uf->end; layerStr->next != NULL; layerStr = layerStr->next );

  if( !junk ){
    currentLayer = (lnode*) malloc( sizeof( lnode ) );
    currentLayer->name = (char*) calloc( strlen( JUNKLAYER ), sizeof( char ) );
    strcpy( currentLayer->name, JUNKLAYER );
    currentLayer->r = 1.0;
    currentLayer->g = 0.1;
    currentLayer->b = 0.1;

    /** Insert it into the list **/
    currentLayer->next = uf->layerList;
    currentLayer->prev = uf->layerList->prev;
    uf->layerList->prev->next = currentLayer;
    uf->layerList->prev = currentLayer;

    /** Add a bit of color info at the end **/
    layerStr->next = (ugstr*) malloc( sizeof( ugstr ) );
    layerStr = layerStr->next;
    layerStr->line = (char*) calloc( strlen( "{ color 60 }\n" ), sizeof( char ) );
    strcpy( layerStr->line, "{ color 60 }\n" );
    layerStr->next = NULL;
  }

  /** Do the same for elevators **/
  if( !elevator ){
    currentLayer = (lnode*) malloc( sizeof( lnode ) );
    currentLayer->name = (char*) calloc( strlen( "A_ELEVATOR" ), sizeof( char ) );
    strcpy( currentLayer->name, "A_ELEVATOR" );
    currentLayer->r = 0.8;
    currentLayer->g = 0.1;
    currentLayer->b = 0.8;

    /** Insert Elevator into the list **/
    currentLayer->next = uf->layerList;
    currentLayer->prev = uf->layerList->prev;
    uf->layerList->prev->next = currentLayer;
    uf->layerList->prev = currentLayer;

    /** Add some color data **/
    layerStr->next = (ugstr*) malloc( sizeof( ugstr ) );
    layerStr = layerStr->next;
    layerStr->line = (char*) calloc( strlen( "{ color 61 }\n" ), sizeof( char ) );
    strcpy( layerStr->line, "{ color 61 }\n" );
    layerStr->next = NULL;
  }

  /** And finally for staircases. **/
  if( !stair ){
    currentLayer = (lnode*) malloc( sizeof( lnode ) );
    currentLayer->name = (char*) calloc( strlen( "A_STAIRCASE" ), sizeof( char ) );
    strcpy( currentLayer->name, "A_STAIRCASE" );
    currentLayer->r = 1.0;
    currentLayer->g = 1.0;
    currentLayer->b = 1.0;

    /** Insert it into the list **/
    currentLayer->next = uf->layerList;
    currentLayer->prev = uf->layerList->prev;
    uf->layerList->prev->next = currentLayer;
    uf->layerList->prev = currentLayer;

    /** Add a bit of color info at the end **/
    layerStr->next = (ugstr*) malloc( sizeof( ugstr ) );
    layerStr = layerStr->next;
    layerStr->line = (char*) calloc( strlen( "{ color 62 }\n" ), sizeof( char ) );
    strcpy( layerStr->line, "{ color 62 }\n" );
    layerStr->next = NULL;
  }

  if( junk ) return false;
  return true;
}

/** Correct:
 *  Called while searching for door jambs, if a particular door jamb isn't found,
 *  it is sometimes because a vertex is missing.  Awfully annoying, but it's too
 *  much to be fixed by DOF.  So try to find where a likely point should be
 *  located, what wire it should probably belong to, and put it there.  If there
 *  is no nearby wire, there is probably no door jamb to be found (sometimes the
 *  case).  Return a dwPts object containing the wire and the two vertices on
 *  which to seperate it if it exists.
 */
dwPts *Correct( ugfile *uf, vert *v, vert *ref1, vert *ref2 )
{
  wire *w;
  vert *v1, *v2, *testVert;
  vnode *vn;
  float eq[2], c;
  int direction;
  dwPts *match;

  v1       = uf->firstVert;
  match    = (dwPts*) malloc( sizeof( dwPts ) );
  testVert = (vert*)  malloc( sizeof( vert ) );

  /** Go through each vertex and test its usability as a door jamb. **/
  do{

    if( SameAngle( v, v1, ref1, ref2 ) && Length( v1->x, v1->y, v->x, v->y ) <= MAXWIDTH
	&& v1->wl == v1->wl->next && GoodLayer( v1->wl->w ) && v1 != v ){

      match->w[0] = v1->wl->w;

      /** Figure out what angle v-v1 is at and parameterize it, if necessary **/
      if( v->x == v1->x ){

	direction = VERTICAL;
	testVert->x = v->x;

      }else if( v->y == v1->y ){

	direction = HORIZONTAL;
	testVert->y = v->y;

      }else{

	direction = SLOPED;
	eq[M] = ( v->y - v1->y ) / ( v->x - v1->x );
	eq[B] = v->y - eq[M] * v->x;

      }

      /** Check the previous vertex, first, then the next one. **/
      w  = v1->wl->w;
      vn = w->vl;
      while( vn->v != v1 ) vn = vn->next;
      vn = vn->prev;
      v2 = vn->v;

      if( w->isFace || vn != w->vl ){

	if( direction == VERTICAL ) testVert->y = v2->y;
	else if( direction == HORIZONTAL ) testVert->x = v2->x;
	else{

	  /** y = ax + b, y = -1/a x + c ==>    **/
	  /** x = (a*(c-b))/(a^2+1), y = ax + b **/
	  c = v2->y + (1/eq[M]) * v2->x;
	  testVert->x = (eq[M]*(c-eq[B])) / (sq(eq[M])+1);
	  testVert->y = eq[M] * testVert->x + eq[B];

	}

	if( !( testVert->x == v1->x && testVert->y == v1->y ) ){
	  /** Got a valid point.  Now create it, and add it to the wire **/

	  match->v[1] = v1;
	  if( testVert->x == v->x && testVert->y == v->y ) match->v[0] = v;
	  else match->v[0] = AddVertex( uf, testVert->x, testVert->y, 0 );

	  AddVertWire( match->v[0], match->w[0], v2, v1 );

	  assert( IsBetween( v, match->v[0], match->v[1] ) );

	  free( testVert );
	  return match;

	}
      }

      /** If the previous vertex, v2, didn't work, try the next one.  Run all of
	  the same tests. **/

      vn = vn->next->next;
      if( w->isFace || vn != w->vl ){

	if( direction == VERTICAL ) testVert->y = v2->y;
	else if( direction == HORIZONTAL ) testVert->x = v2->x;
	else{

	  /** y = ax + b, y = -1/a x + c ==>    **/
	  /** x = (a*(c-b))/(a^2+1), y = ax + b **/
	  c = v2->y + (1/eq[M]) * v2->x;
	  testVert->x = (eq[M]*(c-eq[B])) / (sq(eq[M])+1);
	  testVert->y = eq[M] * testVert->x + eq[B];

	}

	if( !( testVert->x == v1->x && testVert->y == v1->y ) ){
	  /** Got a valid point.  Now create it, and add it to the wire **/

	  match->v[0] = v1;
	  if( testVert->x == v->x && testVert->y == v->y ) match->v[1] = v;
	  else{
	    match->v[1] = AddVertex( uf, testVert->x, testVert->y, 0 );
	    AddVertWire( match->v[1], match->w[0], v2, v1 );
	  }

	  free( testVert );
	  return match;

	}
      }
    }

    v1 = v1->next;
  } while( v1 != uf->firstVert );

  free( testVert );
  free( match );

  return NULL;
}

/** EdgeSubset:
 *  Get all of the edges in **e that are in the direction, dir.  It returns an array
 *  of these edges, and modifies n and m where n is the highest non-NULL value in the
 *  array and m is the size of the array.
 */
edge **EdgeSubset( edge **e, int dir, int *n, int *m )
{
  edge **el;
  int numEdges, i, j, k;

  /** Set up the edge list el. **/
  numEdges = 512;
  el = (edge**) calloc( numEdges, sizeof( edge* ) );
  i = 0;
  for( k = 0; k < numEdges; ++k ) el[k] = NULL;

  /** Go through all of the edges in **e **/
  for( j = 0; e[j] != NULL; ++j ){
    if( e[j]->direction == dir ){
      if( i >= numEdges - 1 ){
	/** Resize el if needed **/
	numEdges *= 2;
	el = (edge**) realloc( el, numEdges * sizeof( edge* ) );
	for( k = i; k < numEdges; ++k ) el[k] = NULL;
      }

      el[i] = e[j];

      ++i;
    }
  }

  EdgeSort( el, 0, i - 1 );

  if( n != NULL ) *n = i;
  if( m != NULL ) *m = numEdges;

  return el;
}

/** EdgeSort:
 *  Quicksort on edges.  EdgeSort( e, 0, numE )
 */
void EdgeSort( edge **e, int p, int r )
{
  int div;

  if( p < r ){
    div = EdgePart( e, p, r );
    EdgeSort( e, p, div );
    EdgeSort( e, div + 1, r );
  }

  return;
}

/** EdgePart:
 *  The partitioning function for quicksort on edges.
 */
int EdgePart( edge **e, int p, int r )
{
  edge *x, *swap;
  int i, j;

  x = e[(p+r)/2];
  i = p - 1;
  j = r + 1;

  while( (1) ){
    do{
      --j;
    } while( e[j]->sortValue > x->sortValue );
    do{
      ++i;
    } while( e[i]->sortValue < x->sortValue );

    if( i < j ){
      swap = e[i];
      e[i] = e[j];
      e[j] = swap;
    }else return j;
  }
}

/** VertSort:
 *  Quicksort vertices.  When called: VertSort( v, 0, v_length - 1 );
 */
void VertSort( vert **v, int p, int r )
{
  int div;

  if( p < r ){
    div = VertPart( v, p, r );
    VertSort( v, p, div );
    VertSort( v, div + 1, r );
  }

  return;
}

/** VertPart:
 *  Partitioning function used by VertSort.
 */
int VertPart( vert **v, int p, int r )
{
  vert *x, *swap;
  int i, j;
  x = v[(p+r)/2];
  i = p - 1;
  j = r + 1;

  while( (1) ){
    do{
      --j;
    } while( v[j]->sortValue > x->sortValue );
    do{
      ++i;
    } while( v[i]->sortValue < x->sortValue );

    if( i < j ){
      swap = v[i];
      v[i] = v[j];
      v[j] = swap;
    }else return j;
  }
}

/** TestPoint:
 *  See if the point ( x, y ) exists on the edge.  This is run with the presupposition
 *  that the point is on the same line.
 */
bool TestPoint( float x, float y, edge *e )
{
  vert *v1, *v2;

  v1 = e->v[0];
  v2 = e->v[1];

  if( e->direction == VERTICAL ){
    if( ( v1->y <= y && y <= v2->y ) || ( v2->y <= y && y <= v1->y ) ) return true;
  }else{
    if( ( v1->x <= x && x <= v2->x ) || ( v2->x <= x && x <= v1->x ) ) return true;
  }

  return false;
}

/** Intersect:
 *  Check to see whether the edges, segment1 and segment2 intersect.  If they do, and
 *  they don't already share a vertex, split them on the point and add the point to the
 *  wires.
 */
int Intersect( ugfile *uf, edge ***list1, edge ***list2, edge ***e, int i,
	       int j, int *l1, int *m1, int *l2, int *m2, int *le, int *me )
{
  vert *v1, *v2, *v3, *v4, *v5;
  float x, y;
  edge *segment1, *segment2, *newEdge;
  int returnValue;
  bool xSet = false, ySet = false, add1, add2;

  segment1 = list1[0][i];
  segment2 = list2[0][j];

  v1 = segment1->v[0];
  v2 = segment1->v[1];
  v3 = segment2->v[0];
  v4 = segment2->v[1];

  if( v1 == v2 || v3 == v4 ) return -1;

  /** If they intersect and share a point, don't bother adding a new point. **/
  if( v1 != v3 && v1 != v4 && v2 != v3 && v2 != v4 ){

    if( ( !segment1->w->isFace && v2 == segment1->w->vl->v && v1 == segment1->w->vl->prev->v ) ||
	( !segment2->w->isFace && v4 == segment2->w->vl->v && v3 == segment2->w->vl->prev->v ) )
      return -2;

    /** Set the x position if either segment is vertical **/
    if( segment1->direction == VERTICAL ){
      x = v1->x;
      xSet = true;
    }else if( segment2->direction == VERTICAL ){
      x = v3->x;
      xSet = true;
    }

    /** Set the y position if either segment is horizontal **/
    if( segment1->direction == HORIZONTAL ){
      y = v1->y;
      ySet = true;
    }else if( segment2->direction == HORIZONTAL ){
      y = v3->y;
      ySet = true;
    }

    /** Set either remaining x or y **/
    if( !xSet && !ySet ){
      /** Both segments are sloped:
	  y = ax + b, y = cx + d => x = (b-d)/(c-a) **/
      x = ( segment1->eq[B] - segment2->eq[B] )/( segment1->eq[M] - segment2->eq[M] );
      y = segment1->eq[M] * x + segment1->eq[B];
    }else if( !xSet ){
      /** x = (y-b)/m **/
      if( segment1->direction == SLOPED ) x = ( y - segment1->eq[B] ) / segment1->eq[M];
      else x = ( y - segment2->eq[B] ) / segment2->eq[M];
    }else if( !ySet ){
      /** y = mx + b **/
      if( segment1->direction == SLOPED ) y = segment1->eq[M] * x + segment1->eq[B];
      else y = segment2->eq[M] * x + segment2->eq[B];
    }

    if( !TestPoint( x, y, segment1 ) || !TestPoint( x, y, segment2 ) ) return 0;

    /** Now I have the x and y values of the new point.  See if the point already
	exists on either of the lines.  If it does, only add it to the other wire.
	If it doesn't, add it to both wires.  Make new segments where appropriate. **/

    add1 = add2 = false;

    if( x == v1->x && y == v1->y ){
      /** The point is v1 **/
      v5 = v1;
      segment2->v[1] = v5;
      add2 = true;
      returnValue = 2;
    }else if( x == v2->x && y == v2->y ){
      /** The point is v2 **/
      v5 = v2;
      segment2->v[1] = v5;
      add2 = true;
      returnValue = 2;
    }else if( x == v3->x && y == v3->y ){
      /** The point is v3 **/
      v5 = v3;
      segment1->v[1] = v5;
      add1 = true;
      returnValue = 2;
    }else if( x == v4->x && y == v4->y ){
      /** The point is v4 **/
      v5 = v4;
      segment1->v[1] = v5;
      add1 = true;
      returnValue = 2;
    }else{
      /** The point does not exist, yet. **/
      v5 = AddVertex( uf, x, y, 0 );
      segment1->v[1] = v5;
      segment2->v[1] = v5;
      add1 = true;
      add2 = true;
      returnValue = 1;
    }

    if( segment1->v[1] == v5 && add1 ){
      /** If segment1 was shortened, add another edge to list1 and put the
	  vertex in the wire being pointed to **/
      newEdge = (edge*) malloc( sizeof( edge ) );
      newEdge->direction = segment1->direction;
      newEdge->hasMatch  = false;
      newEdge->v[0]  = v5;
      newEdge->v[1]  = v2;
      newEdge->eq[M] = segment1->eq[M];
      newEdge->eq[B] = segment1->eq[B];
      newEdge->w     = segment1->w;

      if( newEdge->direction == VERTICAL ) newEdge->sortValue = v5->x;
      else if( newEdge->direction == HORIZONTAL ) newEdge->sortValue = v5->y;
      else newEdge->sortValue = segment1->sortValue;

      AddVertWire( v5, newEdge->w, v1, v2 );
      AddEdge( list1, newEdge, l1, m1 );
      AddEdge( e, newEdge, le, me );
    }

    if( segment2->v[1] == v5 && add2 ){
      /** If segment2 was shortened, add another edge to list2 and put the
	  vertex in the wire being pointed to **/
      newEdge = (edge*) malloc( sizeof( edge ) );
      newEdge->direction = segment2->direction;
      newEdge->hasMatch  = false;
      newEdge->v[0]  = v5;
      newEdge->v[1]  = v4;
      newEdge->eq[M] = segment2->eq[M];
      newEdge->eq[B] = segment2->eq[B];
      newEdge->w     = segment2->w;

      if( newEdge->direction == VERTICAL ) newEdge->sortValue = v5->x;
      else if( newEdge->direction == HORIZONTAL ) newEdge->sortValue = v5->y;
      else newEdge->sortValue = segment2->sortValue;

      AddVertWire( v5, newEdge->w, v3, v4 );
      AddEdge( list2, newEdge, l2, m2 );
      AddEdge( e, newEdge, le, me );
    }

    return returnValue;
  }

  return 0;
}

/** AddEdge:
 *  Add *insert to the end of the list of edges, ***e.
 */
void AddEdge( edge ***e, edge *insert, int *l1, int *m1 )
{
  int i;

  if( (*l1)+2 >= (*m1) ){
    /** Resize ***e, if necessary **/
    (*m1) *= 2;
    e[0] = (edge**) realloc( e[0], (*m1) * sizeof( edge* ) );
    for( i = (*l1) + 1; i < (*m1); ++i ) e[0][i] = NULL;
  }

  e[0][*l1] = insert;
  (*l1)++;

  return;
}

/** IsBetween:
 *  Tests whether *between lies on the same line as *v1 and *v2 and is
 *  between the two.  If it is, return true, else return false.
 */
bool IsBetween( vert *between, vert *v1, vert *v2 )
{
  int dist;

  if( SameAngle( between, v1, v1, v2 ) ){
    dist = Length( v1->x, v1->y, v2->x, v2->y );
    if( Length( between->x, between->y, v1->x, v1->y ) <= dist &&
	Length( between->x, between->y, v2->x, v2->y ) <= dist )
      return true;
  }

  return false;
}

/** Length:
 *  Find the length of a line segment given by x1, x2, y1, y2.
 */
float Length( float x1, float y1, float x2, float y2 )
{
  return sqrt( sq( y1 - y2 ) + sq( x1 - x2 ) );
}

/** SameAngle:
 *  Test the angles between v1-v2 and v3-v4.  If they are the same,
 *  return true; else, return false.
 */
bool SameAngle( vert *v1, vert *v2, vert *v3, vert *v4 )
{
  float maxError = 0.01;

  /** Test for horizontal, vertical and sloped: **/
  if( fabs( fabs( v1->y - v2->y ) - fabs( v3->y - v4->y ) )
      <= maxError ) return true;                             /** horizontal **/
  else if( fabs( fabs( v1->x - v2->x ) - fabs( v3->x - v4->x ) )
	   <= maxError ) return true;                        /** vertical **/
  else if( v1->x != v2->x && v3->x != v4->x ){
    if( fabs( ( v1->y - v2->y ) / ( v1->x - v2->x ) - ( v3->y - v4->y ) / ( v3->x - v4->x ) )
	<= maxError ) return true;                           /** sloped **/
  }

  return false;
}

/** WellPlaced:
 *  If the two wires passed are parallel, the same length, and their ends match,
 *  return the distance between them.  Else, return 0.
 */
float WellPlaced( wire *w1, wire *w2 )
{
  vert *v1, *v2, *v3, *v4;
  float slope;

  v1 = w1->vl->v;
  v2 = w1->vl->prev->v;
  v3 = w2->vl->v;
  v4 = w2->vl->prev->v;

  /** For each possibility of angle of the two wires (vertical, horizontal, sloped)
      test whether they meet the criteria and are, hence, possibilities as windows. **/

  if( v1->x == v2->x && v3->x == v4->x ){

    /** Vertical **/
    if( v1->y == v3->y && v2->y == v4->y )      return fabs( v1->x - v3->x );
    else if( v1->y == v4->y && v2->y == v3->y ) return fabs( v1->x - v4->x );

  }else if( v1->y == v2->y && v3->y == v4->y ){

    /** Horizontal **/
    if( v1->x == v3->x && v2->x == v4->x )      return fabs( v1->y - v3->y );
    else if( v1->x == v4->x && v2->x == v3->x ) return fabs( v1->y - v4->y );

  }else if( v1->x != v2->x && v3->x != v4->x ){

    /** Sloped **/
    slope = ( v1->y - v2->y ) / ( v1->x - v2->x );
    if( slope == ( v3->y - v4->y ) / ( v3->x - v4->x ) ){

      assert( slope != 0 );
      slope = -1/slope;
      if( ( v1->y - v3->y ) / ( v1->x - v3->x ) == slope &&
	  ( v2->y - v4->y ) / ( v2->x - v4->x ) == slope ){

	return Length( v1->x, v1->y, v3->x, v3->y );

      }else if( ( v1->y - v4->y ) / ( v1->x - v4->x ) == slope &&
		( v2->y - v3->y ) / ( v2->x - v3->x ) == slope ){

	return Length( v1->x, v1->y, v4->x, v4->y );

      }
    }
  }

  return 0.0;
}

/** Angle:
 *  Find the angle between the two lines specified by v1-v2, v2-v3.
 *  Return the angle in degrees.
 */
float Angle( vert *v1, vert *v2, vert *v3 )
{
  float a1, a2;

  /** Get the angle from horizontal of the line specified by v1-v2 **/
  if( v1->x == v2->x ){

    if( v1->y > v2->y ) a1 = 90;
    else a1 = -90;

  }else{

    a1 = (180/PI) * atan( ( v1->y - v2->y )/( v1->x - v2->x ) );

  }

  /** Get the angle from the horizontal of the line specified by v2-v3 **/
  if( v2->x == v3->x ){

    if( v3->y > v2->y ) a2 = 90;
    else a2 = -90;

  }else{

    a2 = (180/PI) * atan( ( v2->y - v3->y )/( v2->x - v3->x ) );

  }

  return fabs( a1 - a2 );
}

/** sq: square x **/
float sq( float x ){ return x * x; }

/** Emphasize:
 *  Used for debugging purposes.  Emphasize a certain layer to make it stand out when
 *  debugging a part of the program that deals with that layer.
 */
void Emphasize( ugfile *uf, char *lname )
{
  lnode *ln;

  ln = uf->layerList;
  while( strcmp( ln->name, lname ) ) ln = ln->next;
  ln->r = 1.0;
  ln->g = 0.5;
  ln->b = 0.5;

  return;
}

/** Usage:
 *  Print the help dialogue to the stderr and exit.
 */
void
Usage( )
{
  fprintf( stderr, "dejunk [-options] inputFile.ug [outputFile.ug]\n" );
  fprintf( stderr, "Possible options are:\n" );
  fprintf( stderr, "  -v Verbose:  Turn verbose mode on.  Gives a periodic update of how\n" );
  fprintf( stderr, "     the program is doing and what results it is achieving.\n" );
  fprintf( stderr, "  -h Help: Bring up this help dialogue and exit.\n" );
  fprintf( stderr, "  -c Clean: Turn clean mode off.  Program will not attempt to modify\n" );
  fprintf( stderr, "     the number of vertices in the file by finding intersections.\n" );
  fprintf( stderr, "  -d Doors: Turn off the search for doors, windows, stairs, door\n" );
  fprintf( stderr, "     jambs, and window jambs.\n" );
  fprintf( stderr, "  -m Merge: Don't merge vertices that are on top of one another.\n" );
  fprintf( stderr, "  -j Junk: Turn dejunk mode off.  Program will not search for extra\n" );
  fprintf( stderr, "     lines or line segments and mark them in a junk layer.\n" );

  exit( -1 );
}

