
/**********************************/
/** Author:       Allen Miu      **/
/** Last Updated: 11/8/2001      **/
/** Email:        aklmiu@mit.edu **/
/**********************************/

/******************/
/** HEADER FILES **/
/******************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h> 
#include "sug.h"

#include "../cdt/cdt.h"

#define TOLERANCE 0.001    /* used to test for equality to zero for floating point operations */


/**************************/
/** TYPEDEFS AND STRUCTS **/
/**************************/
typedef struct filter_record
{
    edge** e;
    int numEdges;
    int maxEdges;
} filter_record;


/***************************/
/** CONSTANTS AND GLOBALS **/
/***************************/

/* const float PI = 3.141592653589793238462643383279502884; */
/* const char JUNKLAYER[] = "A_STOOPID"; */

/** Flags **/
bool VERBOSE = false;

/***************************/
/** FUNCTION DECLARATIONS **/
/***************************/

bool Arguments( int argc, char **argv, char **inFile, char **outFile );
void Usage( );

void FindExtremes( ugfile *uf, double *maxX, double *maxY, double *minX, double *minY );
lnode *FindLayer( ugfile *uf, char* layername );
bool AddLayer( ugfile *uf );
bool ObstacleLayer( wire *w );
void AddEdge( ugfile *uf, lnode *layer, double x, double y, double xx, double yy );
void AddTriangle( ugfile *uf, lnode *layer, double x, double y, double xx, double yy, 
		  double xxx, double yyy );
void Emphasize( ugfile *uf, char *lname );
edge*** FilterEdges(edge **e, int *numEdges, int *maxEdges);

void FillCoordinates( ugfile *uf, void *mesh, double x, double y );
void AddConstrainedEdges( ugfile *uf, void *mesh );
void ExtractValidGraph( ugfile *uf, void *mesh, double x, double y );
void DrawMesh( ugfile *uf, void *mesh );
int WriteMesh( const char *fileName, void *mesh, double *bound );
int WriteMeshAll( const char *fileName, void *mesh, double *bound );

/**********/
/** MAIN **/
/**********/

/** main:
 *  Dejunk is run on a unigraphics formatted MIT file.  It's designed
 *  to seperate out all of the extraneous lines and wires from the
 *  floorplan and put them in a seperate layer.
 */
int main( int argc, char **argv )
{
  ugfile *uf;
  char *inFile, *outFile;
  /* double maxX, maxY, minX, minY; */
#define MINX 0
#define MINY 1
#define MAXX 4
#define MAXY 5
  double bound[8];

  /** Check the arguments **/
  if( !Arguments( argc, argv, (char**) &inFile, (char**) &outFile ) ) return -1;

  /** Load the file into the data structure **/
  uf = (ugfile*) malloc( sizeof( ugfile ) );
  //Load( inFile, uf );
  //if( VERBOSE ) fprintf( stderr, "Loaded file %s.\n", inFile );

  /** Begin seperating out wires.  See function definitions for
      specifics on each function. **/
  //xAddLayer( uf );
  printf("done AddLayer.\n");

  bound[MAXX] = 452.0;
  bound[MINY] = 0;
  bound[MINX] = 0;
  bound[MAXY] = 452.0;

  //FindExtremes( uf, &bound[MAXX], &bound[MAXY], &bound[MINX], &bound[MINY] );

  printf("done FindExtremes %f %f %f %f\n",
	 bound[MAXX], bound[MAXY], bound[MINX], bound[MINY] );

  bound[2] = bound[MAXX];
  bound[3] = bound[MINY];
  bound[6] = bound[MINX];
  bound[7] = bound[MAXY];

  void *mesh = create_mesh( 4, bound );
  printf("done create_mesh\n");

  AddConstrainedEdges( uf, mesh );
  printf("done AddConstrainedEdges\n");

  FillCoordinates( uf, mesh, 350, 350 );
  printf("done FillCoordinates\n");

  //  DrawMesh( uf, mesh );
  //  printf("done DrawMesh\n");

  //  ExtractValidGraph( uf, mesh, 500, 500 );
  //  printf("done ExtractValidGraph\n");

  char *mesh_outFile = "mesh.txt";
  if( WriteMesh( mesh_outFile, mesh, bound ) ) 
  /*  if( WriteMeshAll( mesh_outFile, mesh, bound ) ) */
      printf("error writing mesh to %s\n", mesh_outFile);
  printf("done WriteMesh\n");

  /** Write the ugfile into a file (or to the stdout) **/
  //  if( outFile == NULL ){

  //    outFile = (char*) calloc( strlen( "stdout" )+1, sizeof( char ) );
  //    strcpy( outFile, "stdout" );

  //  }

  //  Write( outFile, uf );
  //  if( VERBOSE ) fprintf( stderr, "Done writing to %s.\n", outFile );
  //  Unload( uf );
  //  if( VERBOSE ) fprintf( stderr, "Done.\n" );

  return 0;
}

/**************************/
/** FUNCTION DEFINITIONS **/
/**************************/

/** Arguments:
 *  Take the arguments, set flags, and the in/out files.
 */
bool Arguments( int argc, char **argv, char **inFile, char **outFile )
{
  int i = 1;
  char *p;

  inFile[0]     = NULL;
  outFile[0]    = NULL;

  if( argc < 2 ) Usage( );

  while( i < argc ){
    p = argv[i];

    if( *p == '-' ){
      /** Set flags **/

      ++p;
      if( *p == '\0' ){
	fprintf( stderr, "Error: Bad argument '-', use -h for help.\n" );
      }

      while( *p != '\0' ){
	switch( *p ){
	case 'h':
	  Usage( );
	  break;
	case 'v':
	  VERBOSE = true;
	  break;
	default:
	  fprintf( stderr, "Error: Bad argument '%c', use -h for help.\n", *p );
	  return false;
	  break;
	}
	++p;
      }

    }else{
      /** Set the in and out files **/

      inFile[0] = p;
      ++i;
      if( i < argc ) outFile[0] = argv[i];

      ++i;
      if( i < argc ){
	fprintf( stderr, "Error: Too many arguments.  Use -h for help.\n" );
	return false;
      }
    }

    ++i;
  }

  return true;
}

/** FindExtremes
 * Finds the rectangular bounding box dimensions
 */
void FindExtremes( ugfile *uf, double *maxX, double *maxY,
		   double *minX, double *minY )
{
  vert *v;
  
  v = uf->firstVert;
  *maxX = *minX = v->x;
  *maxY = *minY = v->y;

  do{
    if( v->x > *maxX )      *maxX = v->x;
    else if( v->x < *minX ) *minX = v->x;

    if( v->y > *maxY )      *maxY = v->y;
    else if( v->y < *minY ) *minY = v->y;

    v = v->next;

  } while( v != uf->firstVert );
}


/** FilterEdges
 * Return a subset of edges whose wires belong to an "obstacle layer"
 * (i.e. statisfy the condition defined by ObstacleLayer(wire *w)
 * 
 * Outputs the number of valid edges in the array and 
 * the size of the array.
 *
 * Destroys e
 *
 */
void FilterEdges(edge **e, int numEdges, filter_record *obs, filter_record *doors )
{
    int i;
    filter_record *r;

    /* initialize records */
    obs->numEdges = 0;
    obs->maxEdges = 64;
    obs->e = (edge**) calloc( obs->maxEdges, sizeof( edge* ) );

    doors->numEdges = 0;
    doors->maxEdges = 64;
    doors->e = (edge**) calloc( obs->maxEdges, sizeof( edge* ) );


    for(i = 0; i < numEdges; i++) {
	if(!strcmp( e[i]->w->layer->name, "A_DOOR_JAMB" ))
	    r = doors;
	else if(ObstacleLayer(e[i]->w))
	    r = obs;
	else
	    r = 0;

	if(r) {
	    if( r->numEdges >= r->maxEdges ){
		r->maxEdges *= 2;
		r->e = (edge**) realloc( r->e, r->maxEdges * sizeof( edge* ) );
	    }
	    r->e[r->numEdges] = e[i];
	    r->numEdges++;
	}
	else {
	    free(e[i]);
	}
    }
    free(e);
}

lnode *FindLayer( ugfile *uf, char* layername )
{
  lnode *result;

  assert(uf != NULL);

  result = uf->layerList;
  while( strcmp( result->name, layername ) ) {
    result = result->next;
    if(result == uf->layerList)
      assert(false); /* return NULL; */
  }
  
  return result;
}

double lengthSqr( double *pts ) 
{
    double dx = pts[0] - pts[2];
    double dy = pts[1] - pts[3];
    return dx*dx + dy*dy;
}

void get_center_coordinate( void *edge, double *xavg, double *yavg ) 
{
    double pts[4];
    *xavg = 0;
    *yavg = 0;

    void *u = edge;

    do {  // scan around to get the triangle center
	void *uu = rot( u );
	get_edge_endpoints( uu, pts );
	*xavg += pts[0];
	*yavg += pts[1];
	u = onext( u );
    } while (u != edge);
    
    *xavg /= 3;
    *yavg /= 3;
}

/* sample algorithm for dfs traversal; adds edge to ugfile for all unconstrained edges in mesh */
void extractvgraph( bool terminate, ugfile *uf, lnode *layer, void *edge, double x, double y )
{
    void *u, *uu, *v;
    double xavg, yavg;
    double pts[4];

    u = edge; // get the "dual" graph, left face
    mark_edge_org( u, BLACK );

    get_center_coordinate( u, &xavg, &yavg );

    //    printf("Adding Edge %f %f %f %f\n", x, y, xavg, yavg); 
    AddEdge( uf, layer, x, y, xavg, yavg);

    if(!terminate) {
	do {  // dfs
	    uu = rot( u );
	    v = sym( u );
	    get_edge_endpoints( uu, pts );
	    // printf("length: %f\n", lengthSqr(pts));
	    if(!is_constrained( uu ) && lengthSqr( pts ) > 16) {
		// recurse only if the face is not marked;
		// always draw an edge between the visible faces to create 
		// a graph, instead of tree
		extractvgraph( is_marked_org(v), uf, layer, v, xavg, yavg );
	    }

	    u = onext( u );
	} while (u != edge);
    }
}

#define VALID    0
#define INVALID  1
#define DOORJAMB 2

void AddConstrainedEdges( ugfile *uf, void *mesh )
{
    edge **edges;
    int numEdges, maxEdges;
    filter_record obstacles, doorjambs;
    int i;
    
    
    /* find all barrier edges */
    //    edges = GetEdges( uf, &numEdges, &maxEdges );
    //    FilterEdges(edges, numEdges, &obstacles, &doorjambs);


    /*
    for(i = 0; i < obstacles.numEdges; i++) {
	insert_edge_type(mesh,
			 INVALID,
			 obstacles.e[i]->v[0]->x,
			 obstacles.e[i]->v[0]->y,
			 obstacles.e[i]->v[1]->x,
			 obstacles.e[i]->v[1]->y);
	free(obstacles.e[i]);
    }
    free(obstacles.e);

    for(i = 0; i < doorjambs.numEdges; i++) {
	insert_edge_type(mesh,
			 DOORJAMB,
			 doorjambs.e[i]->v[0]->x,
			 doorjambs.e[i]->v[0]->y,
			 doorjambs.e[i]->v[1]->x,
			 doorjambs.e[i]->v[1]->y);
	free(doorjambs.e[i]);
    }
    free(doorjambs.e);
    */

    /*
    insert_edge_type(mesh, INVALID, 45,246,45,407);
    insert_edge_type(mesh, INVALID, 45,407,407,407);
    insert_edge_type(mesh, INVALID, 407,407,407,45);
    insert_edge_type(mesh, INVALID, 407,45,256,45);
    insert_edge_type(mesh, INVALID, 256,45,256,246);
    insert_edge_type(mesh, INVALID, 256,246,45,246);
    */

    insert_edge_type(mesh, INVALID, 40,241,40,412);
    insert_edge_type(mesh, INVALID, 40,412,412,412);
    insert_edge_type(mesh, INVALID, 412,412,412,40);
    insert_edge_type(mesh, INVALID, 412,40,251,40);
    insert_edge_type(mesh, INVALID, 251,40,251,241);
    insert_edge_type(mesh, INVALID, 251,241,40,241);

}

bool has_doorjamb( void *edge )
{
    void *u = edge;

    do {  // scan around to get the triangle center
	void *uu = rot( u );
	int type = get_edge_type( uu );
	if(type == DOORJAMB)	
	    return 1;
	u = onext( u );
    } while (u != edge);
    
    return 0;
}

void bfs( ugfile *uf, lnode *layer, void *mesh, void *edge, void(*func)( ugfile *u, lnode *l, void *e ))
{
#define Q_LEN 10000
    double pts[4];
    void *queue[Q_LEN];
    void *head;
    int queue_head = 0;
    int num_enqueued = 0;

    /* reset */
    unmark_mesh( mesh );

    /* enqueue */
    queue[0] = edge;
    mark_edge_org(edge, GREY);
    num_enqueued++;

    while(num_enqueued > 0) {

	head = queue[queue_head];

	// process
	func(uf, layer, head);

	void *u = head;
	do {  
	    void *uu = rot(u);
	    void *v = sym(u);
	    get_edge_endpoints( uu, pts );

	    /*
	    if(!is_marked_org( v ) && !is_constrained( uu ) && lengthSqr( pts ) > 16 &&
	       !has_doorjamb( v )) {
	    */
	    if(!is_marked_org( v ) && !is_constrained( uu ) && lengthSqr( pts ) > 16) {
		// enqueue
		queue[(queue_head+num_enqueued) % Q_LEN] = v;
		mark_edge_org(v, GREY);
		num_enqueued++;
		assert(num_enqueued <= Q_LEN);
	    }
	    u = onext( u );
	} while (u != head);

	/* dequeue */
	mark_edge_org( head, BLACK );
	queue_head++;
	queue_head %= Q_LEN;
	num_enqueued--;
    }
}

void draw_path( ugfile *uf, lnode *layer, void *edge ) 
{
    double pts[4];

    void *u = edge;
    do {
	void *uu = rot(u);
	void *v = sym(u);
	get_edge_endpoints( uu, pts );
	if(!is_marked_org( v ) && !is_constrained( uu ) && lengthSqr( pts ) > 16) {
	    get_edge_endpoints( u, pts );
	    printf("Adding Edge: %f %f %f %f\n",
		   pts[0], pts[1], pts[2], pts[3]);
	    AddEdge( uf, layer, pts[0], pts[1], pts[2], pts[3]);
	}
	u = onext( u );
    } while (u != edge);

}

void draw_tri( ugfile *uf, lnode *layer, void *edge ) 
{
    double tri[6];
    double pts[4];
    int c = 0;

    void *u = edge;
    do {
	void *uu = rot(u);
	void *v = sym(u);
	get_edge_endpoints( uu, pts );
	tri[c]   = pts[0];
	tri[c+1] = pts[1];
	assert(c < 5);
	c += 2;
	u = onext( u );
    } while (u != edge);

    AddTriangle( uf, layer, tri[0], tri[1], tri[2], tri[3], tri[4], tri[5]);
}


void ExtractValidGraph( ugfile *uf, void *mesh, double x, double y )
{
    void *edge = locate(mesh, x, y);

    double pts[4];
    get_edge_endpoints( edge, pts );
    printf("seed edge: %f %f %f %f\n", pts[0], pts[1], pts[2], pts[3]);
    lnode *layer = FindLayer( uf, "A_CDT_DUAL" );

    //extractvgraph( 0, uf, layer, rot( edge ), x, y );
    //    bfs( uf, layer, mesh, rot(edge), draw_tri );
        bfs( uf, layer, mesh, rot(edge), draw_path );
}

void fill_coord( ugfile *uf, lnode *layer, void *edge ) 
{
    double thisx, thisy;
    
    get_center_coordinate( edge, &thisx, &thisy );
    void *pt = create_point( thisx, thisy );
    void *u = edge;
    do {
	set_org_endpoint( u, pt );
	u = onext( u );
    } while (u != edge);
}

void FillCoordinates( ugfile *uf, void *mesh, double x, double y )
{
    void *edge = locate(mesh, x, y);

    double pts[4];
    get_edge_endpoints( edge, pts );
    printf("seed edge: %f %f %f %f\n", pts[0], pts[1], pts[2], pts[3]);
    //lnode *layer = FindLayer( uf, "A_CDT_DUAL" );
    lnode *layer = 0;

    bfs( uf, layer, mesh, rot(edge), fill_coord );
}

void AddTriangle( ugfile *uf, lnode *layer, double x, double y, double xx, double yy, 
		  double xxx, double yyy ) 
{
    wire *newWire = AddWire( uf, NULL );
    newWire->layer = layer;
    vert *t, *u, *v;
    u = AddVertex( uf, x, y, 0 );
    v = AddVertex( uf, xx, yy, 0 );
    t = AddVertex( uf, xxx, yyy, 0 );

    AddVertWire( u, newWire, NULL, NULL );
    AddVertWire( v, newWire, NULL, NULL );
    AddVertWire( t, newWire, NULL, NULL );
    AddVertWire( u, newWire, NULL, NULL );
}


void AddEdge( ugfile *uf, lnode *layer, double x, double y, double xx, double yy ) 
{
    wire *newWire = AddWire( uf, NULL );
    newWire->layer = layer;
    vert *u, *v;
    u = AddVertex( uf, x, y, 0 );
    v = AddVertex( uf, xx, yy, 0 );
    AddVertWire( u, newWire, NULL, NULL );
    AddVertWire( v, newWire, NULL, NULL );

}

void DrawEdge( void *u, void *e, int constrained ) 
{
    double pts[4];
    lnode *layer;
    ugfile *uf = (ugfile *)u;

    get_edge_endpoints( e, pts );

    int type = get_edge_type( e );
    
    layer = NULL;
    if(type == DOORJAMB)
	layer = FindLayer( uf, "A_DOORJAMB" );
    else if(!constrained)
	layer = FindLayer( uf, "A_CDT" ); // inefficient

    if(layer != NULL)
	AddEdge(uf, layer, pts[0], pts[1], pts[2], pts[3]);
}

void DrawMesh( ugfile *uf, void *mesh )
{
    apply_edges( mesh, uf, DrawEdge ); 
}

void write_quadedge(ugfile *uf, lnode *layer, void *edge )
{
    FILE *fp = (FILE*) uf;

    void *u = edge;
    do {
	void *v = sym(u);
	/*
	int seqnum = get_seqnum(v);
	if(seqnum == 1611)
	    write_edge( stdout, v );
	*/
	if(get_mark_org( v ) != BLACK)
	    write_edge( fp, v );
	u = onext( u );
    } while (u != edge);
}

/* reason for using bfs instead of apply_edges is to
   output only the stuff we need and prune away the
   others
*/
int WriteMesh( const char *fileName, void *mesh, double *bound )
{
    void *edge = locate(mesh, 350, 350);
    double pts[4];
    get_edge_endpoints( edge, pts );
    printf("seed edge: %f %f %f %f\n", pts[0], pts[1], pts[2], pts[3]);

    FILE *fp;
    if( !strcmp( fileName, "stdout" ) )
	fp = stdout;
    else if( !strcmp( fileName, "stderr" ) )
	fp = stderr;
    else {
	fp = fopen( fileName, "w" );
	if( !fp ) 
	    return 1;
    }

    fprintf( fp, "%f %f %f %f\n",
	     bound[MINX], bound[MINY] , bound[MAXX], bound[MAXY]);

    bfs( (ugfile*)fp, (lnode *) 0, mesh, rot(edge), write_quadedge );
    return fclose( fp );
}

void write_quadedge_all(void *apply_arg, void *edge, int constrained )
{
    FILE *fp = (FILE*) apply_arg;

    write_edge( fp, edge );
}

/* Same as WriteMesh except it outputs ALL edges */
int WriteMeshAll( const char *fileName, void *mesh, double *bound )
{
    void *edge = locate(mesh, 350, 350);
    double pts[4];
    get_edge_endpoints( edge, pts );
    printf("seed edge: %f %f %f %f\n", pts[0], pts[1], pts[2], pts[3]);

    FILE *fp;
    if( !strcmp( fileName, "stdout" ) )
	fp = stdout;
    else if( !strcmp( fileName, "stderr" ) )
	fp = stderr;
    else {
	fp = fopen( fileName, "w" );
	if( !fp ) 
	    return 1;
    }

    fprintf( fp, "%f %f %f %f\n",
	     bound[MINX], bound[MINY] , bound[MAXX], bound[MAXY]);

    apply_edges( mesh, fp, write_quadedge_all );

    return fclose( fp );
}


bool ObstacleLayer( wire *w )
{
  if( !strcmp( w->layer->name, "A_WALL" ) ||
      !strcmp( w->layer->name, "A_WALL_CORE" ) ||
      !strcmp( w->layer->name, "A_WINDOW" ) ||
      !strcmp( w->layer->name, "A_WINDOW_JAMB" ) )

      /* !strcmp( w->layer->name, "A_CLOSED_DOOR" ) || */
      /* !strcmp( w->layer->name, "A_STOOPID" ) || */
      /* !strcmp( w->layer->name, "A_STAIRCASE" */
      
      return true;
  
  return false;
}

void addlnode( ugfile *uf, char *lname, int color_num, float r, float g, float b )
{
    ugstr *layerStr;

    lnode *currentLayer;
    char buf[32];

    for( layerStr = uf->end; layerStr->next != NULL; layerStr = layerStr->next );
    
    currentLayer = (lnode*) malloc( sizeof( lnode ) );
    currentLayer->name = (char*) calloc( strlen( lname ), sizeof( char ) );
    strcpy( currentLayer->name, lname );
    currentLayer->r = r;
    currentLayer->g = g;
    currentLayer->b = b;
    
    /** Insert into the list **/
    currentLayer->next = uf->layerList;
    currentLayer->prev = uf->layerList->prev;
    uf->layerList->prev->next = currentLayer;
    uf->layerList->prev = currentLayer;

    sprintf(buf, "{ color %d }\n", color_num );
    
    /** Add a bit of color info at the end **/
    layerStr->next = (ugstr*) malloc( sizeof( ugstr ) );
    layerStr = layerStr->next;
    layerStr->line = (char*) calloc( strlen( buf ), sizeof( char ) );
    strcpy( layerStr->line, buf );
    layerStr->next = NULL;

}

/** AddLayer:
 *  Add a layer called A_PATH_PORTALS for lines with an endpoint at
 *  a doorway, and a layer called A_PATH_INTERIOR for interior 
 *  lines with neither endpoint at a doorway.
 */
bool
AddLayer( ugfile *uf )
{
  bool interior, dual, door, constrained ;

  dual = interior = door = constrained = false;

  /** Check the existing layers for A_STOOPID **/
  lnode *currentLayer = uf->layerList;
  assert( currentLayer != NULL );

  do{
      if( !strcmp( currentLayer->name, "A_CDT" ) )
	  interior = true;
      if( !strcmp( currentLayer->name, "A_CDT_DUAL" ) )
	  dual = true;
      if( !strcmp( currentLayer->name, "A_DOORJAMB" ) )
	  door = true;
      if( !strcmp( currentLayer->name, "A_CDT_CONSTRAINED" ) )
	  door = true;
      currentLayer = currentLayer->next;
  } while( currentLayer != uf->layerList );



  if( !interior )
      addlnode( uf, "A_CDT", 69, 0.75, 0, 0.75 );

  if( !dual )
      addlnode( uf, "A_CDT_DUAL", 70, 0, 1, 0 );

  if( !door )
      addlnode( uf, "A_DOORJAMB", 71, 1, 1, 0 );

  if( !constrained )
      addlnode( uf, "A_CDT_CONSTRAINED", 72, 1, 1, 1 );

  return true;
}



/** Emphasize:
 *  Used for debugging purposes.  Emphasize a certain layer to make it stand out when
 *  debugging a part of the program that deals with that layer.
 */
void Emphasize( ugfile *uf, char *lname )
{
  lnode *ln;

  ln = uf->layerList;
  while( strcmp( ln->name, lname ) ) ln = ln->next;
  ln->r = 1.0;
  ln->g = 0.5;
  ln->b = 0.5;

  return;
}

/** Usage:
 *  Print the help dialogue to the stderr and exit.
 */
void
Usage( )
{
  fprintf( stderr, "vis [-options] inputFile.ug [outputFile.ug]\n" );
  fprintf( stderr, "Possible options are:\n" );
  fprintf( stderr, "  -v Verbose:  Turn verbose mode on.  Gives a periodic update of how\n" );
  fprintf( stderr, "     the program is doing and what results it is achieving.\n" );
  fprintf( stderr, "  -h Help: Bring up this help dialogue and exit.\n" );

  exit( -1 );
}

