/**********************************/
/** Author:       Will Leiserson **/
/** Last Updated: 8/9/2001       **/
/** Email:        leisew@rpi.edu **/
/**********************************/

/******************************************************************/
/** sug.h : Simple UG header file.  Tired of that old, spaggetti **/
/**         code that passes for a header file to deal with UG   **/
/**         files?  Try this.                                    **/
/******************************************************************/

/**************************/
/** DEFINES AND TYPEDEFS **/
/**************************/


#ifndef bool
#define bool int

#ifndef true
#define true 1
#define false 0
#endif

#endif

typedef struct ugfile   ugfile;
typedef struct wire     wire;
typedef struct wnode    wnode;
typedef struct vert     vert;
typedef struct vnode    vnode;
typedef struct ugstr    ugstr;
typedef struct lnode    lnode;
typedef struct edge     edge;
typedef struct edgeList edgeList;

/***************************/
/** FUNCTION DECLARATIONS **/
/** see sug.c for details **/
/***************************/

bool  Load( const char *fileName, ugfile *uf );
bool  Write( const char *fileName, ugfile *uf );
bool  Unload( ugfile *uf );
void  Standardize( ugfile *uf );
vert  *AddVertex( ugfile *uf, float x, float y, float z );
wire  *AddWire( ugfile *uf, wire *w );
void  RemoveWire( ugfile *uf, wire *w );
bool  AddVertWire( vert *insert, wire *w, vert *v1, vert *v2 );
bool  RemoveVertWire( vert *remove, wire *w );
edge  **GetEdges( ugfile *uf, int *n, int *m );
wire  *SplitWire( ugfile *uf, wire *w, vert *v1, vert *v2 );
wire  *MergeWires( wire *w1, wire *w2 );
wnode *VertHasWire( vert *v, wire *w );
vnode *WireHasVert( wire *w, vert *v );
wire  *CommonWire( vert *v1, vert *v2 );
vert  *CommonVert( wire *w1, wire *w2 );
void  ReverseWire( wire *w );
float Area( wire *w );

/*************/
/** STRUCTS **/
/*************/

/** ugfile:
 *  The base of the structure.  This holds all of the data.
 */
struct ugfile
{
  char *name;           /** filename **/
  ugstr *init, *end;    /** extra data I don't know what to do with **/
  /** lists of wires and vertices, as well as the first and last entries **/
  wire **wireList, *firstWire;
  vert **vertList, *firstVert;
  int hWire, hVert;     /** indices to the final entries in the lists of
			    vertices and wires **/
  int maxWire, maxVert; /** Highest entry in wireList and vertList **/
  lnode *layerList;     /** list of layernames **/
};

/** wire:
 *  Could be a wire or a face.  Generally expressed in a .ug
 *  file as: w w130 ( vv3 vv7 vv22 ) A_LAYER_NAME ;
 */
struct wire
{
  char *name;           /** wire name **/
  bool isFace;          /** whether the wire is actually a face **/
  vnode *vl;            /** next and previous wires in the ugfile's list **/
  wire *next, *prev;
  lnode *layer;         /** points to an entry in uf->layerList **/
};

/** wnode:
 *  Used for individual vertices in creating a list of wires
 *  that use them.
 */
struct wnode
{
  wire *w;              /** pointer to a wire in uf->wireList **/
  wnode *next, *prev;
};

/** vert:
 *  Vertex.  Just a point.  In a .ug file, it would look like:
 *  v vv727 22.6 493.1 0 ;
 */
struct vert
{
  char *name;           /** of the form: vv%d **/
  float x, y, z;        /** position in space **/
  wnode *wl;            /** linked list of wires that use this vert **/
  bool isUsed;          /** if it is used at all **/
  int sortValue;        /** x + y + z **/
  int location;
  vert *next, *prev;
};

/** vnode:
 *  vnodes are used by wires to keep a linked list of the
 *  vertices they use.
 */
struct vnode
{
  vert *v;              /** pointer to a vert in uf->vertList **/
  vnode *next, *prev;
};

/** str:
 *  Extra data kept near the beginning of the file.  Don't
 *  know what it is, but figured I probably shouldn't get rid
 *  of it.
 */
struct ugstr
{
  char  *line;
  ugstr *next;
};

/** lnode:
 *  Layer data.  ugfile has a pointer to one of these.  Used
 *  as a linked list of all of the layers.
 */
struct lnode
{
  char *name;           /** The layer name itself **/
  float r, g, b;        /** The rgb values of the color this layer is
			    represented as, graphically **/
  lnode *next, *prev;
};

/** edge:
 *  Edge is used seperate from the rest of the data structure.
 *  It contains two vertices (connected in a wire or face), and
 *  an equation of the form y = mx + b.
 */
struct edge
{
#ifndef M
#define M 0
#endif
#ifndef B
#define B 1
#endif
#ifndef HORIZONTAL
#define HORIZONTAL 0
#endif
#ifndef VERTICAL
#define VERTICAL   1
#endif
#ifndef SLOPED
#define SLOPED     2
#endif
  vert *v[2];        /** Pointers to the two vertices in question **/
  wire *w;           /** Pointer to the appropriate wire (vertices have a list of wires) **/
  float eq[2];       /** m = eq[0], b = eq[1] **/
  float sortValue;   /** Used to sort the list of edges **/
  int direction;     /** Contains the value, HORIZONTAL, VERTICAL, or SLOPED **/
  bool hasMatch;     /** Used to figure out if there is a parallel line (not a 2D wall) **/

  edge *next, *prev; /** Allows for a linked list of edges **/
};

