package navigator;

import java.awt.*; // for Polygon...
import java.awt.geom.*; // for Point2D

public class Edge
{
    private static final int CLEAR = 0;
    private static final int GREY  = 1;
    private static final int BLACK = 2;
    private static int PLACEHOLDER = BLACK+1; // largest color plus one

    private static int clear = CLEAR;
    private static int grey  = GREY;
    private static int black = BLACK;


    private QuadEdge qEdge;
    private int  marked;
    private int     edgeID;
    private NavPoint2D data;

    /* target's information */
    Edge     next;
    String  quadID;
    int     quadEdgeID;

    /* optimization - cache transform coordinates for rendering */
    private Point2D ptrans;
    
    /* used for shortest path algorithm */
    TreeNode treeNode;

    public Edge(QuadEdge q, int i, String id, int n, float x, float y) 
    { 
	qEdge      = q;
	marked     = CLEAR;
	edgeID     = i;

	next       = null;
	quadID     = id;
	quadEdgeID = n;
	data       = new NavPoint2D((int)x, (int)y);
    }

    public String toString()
    {
	return qEdge.id+" "+type()+" "+edgeID+" "+data+" :::: "+
	    quadID+" "+quadEdgeID+ 
	    ((next != null) ? " || "+
	     next.qEdge().id+" "+next.type()+" "+next.edgeID+" "+
	     next.data+" "+next.quadID+" "+next.quadEdgeID : "");
    }

    // Return the dual of the current edge, directed from its right to
    // its left.
    public Edge rot() 
    { 
	return (edgeID < 3) ? qEdge.edges[edgeID+1] : qEdge.edges[0];
    }

    // Return the dual of the current edge, directed from its left to
    // its right.
    public Edge invRot() 
    {    
        return (edgeID > 0) ? qEdge.edges[edgeID - 1] : qEdge.edges[3];
    }

    // Return the edge from the destination to the origin of the current
    // edge.
    public Edge sym()
    {
        return (edgeID < 2) ? qEdge.edges[edgeID+2] : qEdge.edges[edgeID-2];
    }

    // Return the next ccw edge around (from) the origin of the current
    // edge.
    public Edge oNext() { return next; }

    // Return the next cw edge around (from) the origin of the current
    // edge.
    public Edge oPrev() 
    { 
	if(rot().oNext() == null)
	    return null;
	else 
	    return rot().oNext().rot(); 
    }

    // Return the next ccw edge around (into) the destination of the
    // current edge.
    public Edge dNext() { return sym().oNext().sym(); }

    // Return the next cw edge around (into) the destination of the
    // current edge.
    public Edge dprev() { return invRot().oNext().invRot(); }

    // Return the ccw edge around the left face following the current
    // edge.
    public Edge lNext() { return invRot().oNext().rot(); }

    // Return the ccw edge around the left face before the current edge.
    public Edge lPrev() { return oNext().sym(); }

    // Return the edge around the right face ccw following the current
    // edge.
    public Edge rNext() { return rot().oNext().invRot(); }

    // Return the edge around the right face ccw before the current
    // edge.
    public Edge rPrev() { return sym().oNext(); }

    public NavPoint2D org() { return data; }
    public NavPoint2D dest() { return sym().org(); }

    public int type() { return qEdge.type; }
    public QuadEdge qEdge() { return qEdge; }
    public void setMark(int color) { marked = color; }
    public boolean isMarked(int color) 
    { 
	if(color == getClear())
	    return marked < color;
	return marked == color; 
    }
    public boolean isInvalid() 
    { 
	//!quadID.equals(next.qEdge().id);
	// data.isInvalid()
	Util.assert(!data.isInvalid(), "Edge::isInvalid()");
	return qEdge.type == QuadEdge.INVALID;
    }
    
    public boolean isDoorJamb() { return qEdge.type == QuadEdge.DOORJAMB; }
    public boolean isValid() { return qEdge.type == QuadEdge.VALID; }
    
    /*
    public void unMarkAll() 
    { 
	Edge t = this;
	do {
	    t.setMark(CLEAR);
	    t.ptrans = null;
	    t = t.rot();
	} while (t != this);      
    }
    */

    public static int getClear() { return clear; }
    public static int getGrey() { return grey; }
    public static int getBlack() { return black; }

    public static void unMarkAll() 
    { 
	clear += PLACEHOLDER;
	grey  += PLACEHOLDER;
	black += PLACEHOLDER;
    }

    public boolean isFaceOriented()
    {
	return edgeID == 1 || edgeID == 3;
    }

    static int count = 0;

    public void markOrg(int color)
    {
	Edge t = this;
	do {
	    t.setMark(color);
	    //System.out.println(t);
	    t = t.oNext();

	    //System.out.println(this+" :::: "+t);
	    //Util.assert(t != null, "Edge::markOrg() "+this);

	    /*
	      if(count == 5)
	      System.exit(0);
	      if(qEdge.id.equals("1581"))
	      count++;
	    */
	} while (t != this);
    }

    public boolean isInFace(int x, int y)
    {
	Util.assert(isFaceOriented(), "Edge::isInFace()");
	
	//System.out.println(treeNode+" face "+x+" "+y);
	// I'm whimping out at testing to see if a point lies in a face!
	return treeNode.face.contains(x, y);
    }

    public Point2D getPtrans() { return ptrans; }
    public void setPtrans(Point2D pt)
    {
	Edge t = this;
	/* rewind as far backward as possible */
	while(t.oPrev() != null) {
	    t = t.oPrev();
	    if(t == this)
		break; // cycle
	}
	do {
	    ptrans = pt;
	    t = t.oNext();
	    if(t == null) 
		return;  // best effort; incomplete graph!!!
	} while ( t != this );
    }


    public static boolean hasLeftFace(Edge e)
    // Returns TRUE if there is a triangular face to the left of e.
    {
	return (NavPoint2D.isEq(e.lPrev().org(), e.lNext().dest()) &&
		Util.leftOf(e.lPrev().org(), e));
    }
	
    public static boolean hasRightFace(Edge e)
    // Returns TRUE if there is a triangular face to the right of e.
    {
	return hasLeftFace(e.sym());
    }

}
