package navigator;

import java.io.*;
import java.util.*;
import java.awt.geom.*;

public class Mesh
    implements GraphProcessor
{
    // possibly need one pointer to each graph partition
    public Edge currentEdge; 
    private Hashtable edgeTable;
    // subset of edge table that contain edges with incomplete references
    private Hashtable danglingEdges; 
    public Rectangle2D boundingBox;

    public Mesh(BufferedReader in)
    {
	edgeTable = new Hashtable();
	danglingEdges = new Hashtable();

	try {
	    String dim = in.readLine();
	    if(dim == null)
		throw new IOException();

	    System.out.println(dim);
	    StringTokenizer s = new StringTokenizer(dim);
	    float minX = Float.parseFloat(s.nextToken());
	    float minY = Float.parseFloat(s.nextToken());
	    float maxX = Float.parseFloat(s.nextToken());
	    float maxY = Float.parseFloat(s.nextToken());

	    boundingBox = new Rectangle2D.Float(minX, minY, maxX-minX,
						maxY-minY);

	    Util.assert(minX >= 0 && minY >= 0 &&
			maxX >= 0 && maxY >= 0,
			"NavigatorApp() invalid dimensions: "+
			minX+" "+minY+" "+maxX+" "+maxY);

	    String line = in.readLine();
	    Util.assert(line != null, "Mesh::Mesh()");
	    
	    currentEdge = parseAnEdge(line);
	    while((line = in.readLine()) != null)
		parseAnEdge(line);	    
	}
	catch(IOException e) {
	    e.printStackTrace();
	    Util.assert(false, "Mesh::Mesh() read error "+in);
	}
    }

    public QuadEdge getQuadEdge(int id)
    {
	return QuadEdge.getQuadEdge( edgeTable, id );
    }

    private Edge parseAnEdge(String line)
    {
	//System.out.println(line);
	StringTokenizer str = 
	    new StringTokenizer(line.substring(1, line.length()-1), ",");

	String id  = str.nextToken();
	int type   = Integer.parseInt(str.nextToken());

	String[] quadID     = new String[4];
	int[]    quadEdgeID = new int[4];
	float[]  x          = new float[4];
	float[]  y          = new float[4];
	float[]  z          = new float[4];
	
	for(int i = 0; i < 4; i++) {
	    quadID[i]  = str.nextToken();
	    quadEdgeID[i] = Integer.parseInt(str.nextToken());
	    x[i]       = Float.parseFloat(str.nextToken());
	    y[i]       = Float.parseFloat(str.nextToken());
	}

	QuadEdge result = new QuadEdge(id, type, quadID, quadEdgeID, x, y);
	result.insertTable(edgeTable, danglingEdges);
	return result.edges[0];
    }

    public void unMarkAll()
    {
	//faster implementation see Edge.isMarked, Edge.unMarkAll
	Edge.unMarkAll();

	/*
	Enumeration enum = edgeTable.elements();
	while(enum.hasMoreElements()) {
	    QuadEdge qe = (QuadEdge)enum.nextElement();
	    qe.edges[0].unMarkAll();
	}
	*/
    }

    public void resetTreeNodes()
    {
	Enumeration enum = edgeTable.elements();
	while(enum.hasMoreElements()) {
	    QuadEdge qe = (QuadEdge)enum.nextElement();
	    for(int i = 0; i < 4; i++ ) {
		Edge t = qe.edges[i];
		if(t.treeNode != null)
		    t.treeNode.reset();
	    }
	}
    }

    public boolean shouldSearch(Edge src, Edge dst)
    {
	return true;
    }

    // assumes triangulated subdivision
    public boolean procGraphEdge(Edge e)
    {
	Util.assert(e.isFaceOriented(), "Edge::isInFace()");
	TreeNode newNode = new TreeNode();
	Edge[] edges = new Edge[3];
	for(int i = 0; i < 3; i++) {
	    e.treeNode = newNode;
	    if(e.isDoorJamb())
		newNode.isDoor = true;
	    edges[i] = e;
	    e = e.oNext();
	}
	newNode.init(edges);
	return false;
    }

    public void packMesh()
    {
	bfs( currentEdge.rot(), this );
    }

    public void bfs( Edge edge, GraphProcessor gp)
    {
	ServiceTime.printElapsedTime("bfs");
	ArrayList queue = new ArrayList();
	ServiceTime.printElapsedTime("bfs new arraylist");
	Util.assert(edge.isFaceOriented(), "Mesh::bfs()");

	// save the locality of edge search, see locate
	//currentEdge = edge.rot();

	/* reset */
	unMarkAll();
	ServiceTime.printElapsedTime("bfs unmark");
	queue.add(edge);
	edge.markOrg(Edge.getGrey());

	while(queue.size() > 0) {
	    Edge head = (Edge) queue.remove(0);
	    if(gp.procGraphEdge(head))
		return;

	    Edge u = head;
	    do {  
		//System.out.println("whoa: "+u);
		Edge v = u.sym();
		//System.out.println("sym: "+v);
		if(v.next != null &&
		   gp.shouldSearch( u, v ) && v.isMarked(Edge.getClear())) {
		   //v.isMarked(Edge.CLEAR)) {
		    // enqueue
		    queue.add(v);
		    v.markOrg(Edge.getGrey());
		}
		u = u.oNext();
	    } while (u != head);
	    
	    head.markOrg(Edge.getBlack());
	}

    }

    // result is contained in edge.treeNode
    public Path shortestPath( Edge edge, int limit, int hopLimit )
    {
	return shortestPath( edge, null, limit, hopLimit );
    }

    public Path shortestPath( Edge edge, Edge dest )
    {
	return shortestPath( edge, dest, Integer.MAX_VALUE, 0 );
    }

    // root of the shortest path tree is contained in edge.treeNode
    // if(dest != null), then path (clone of tree nodes ) is returned
    // otherwise, returns null

    boolean astar = false;

    public Path shortestPath( Edge edge, Edge dest, int limit, 
			      int hopLimit )
    {
	PriorityQueue queue = new PriorityQueue();
	Util.assert(edge.isFaceOriented(), "Mesh::bfs()");

	/* reset */
	unMarkAll();
	ServiceTime.printElapsedTime("shortest unmark");

	resetTreeNodes();
	ServiceTime.printElapsedTime("shrotest reset treenode");


	Edge q = edge;
	do {  
	    Edge v = q.sym();
	    if(v.treeNode != null)
		Util.assert(v.treeNode.curDist > 10, "fac");
	    q = q.oNext();
	} while (q != edge);

	if(astar && dest != null)
	    edge.treeNode.curDist = edge.treeNode.p.distanceInt(dest.treeNode.p);
	else
	    edge.treeNode.curDist = 0;

	edge.treeNode.parent = null;

	queue.add(edge);
	edge.markOrg(Edge.getGrey());

	//	System.out.println("curEdge: "+edge);

	while(queue.size() > 0) {
	    Edge head = queue.removeMin();

	    //////////////// hop and limit
	    if(head.treeNode.curDist >= limit && 
	       head.treeNode.curHop >= hopLimit) {
		return null; // finish
	    }

	    if(dest != null && (dest.treeNode.equals(head.treeNode))) {
		// dest found, extract destination to root path
		Path v = new Path(); // first element becomes the leaf
		TreeNode t = dest.treeNode;
		do {
		    TreeNode tt = (TreeNode) t.clone();
		    if(dest != null && astar)
			tt.curDist -= tt.p.distanceInt(dest.treeNode.p);
		    if(v.size() > 0) {
			TreeNode prev = (TreeNode) v.get(v.size() - 1);
			prev.parent = tt;
		    }
		    v.add(tt); 
		    // prevent subsequent calls to
		    t = t.parent;
		} while (t != null);
		return v;
	    }

	    head.markOrg(Edge.getBlack());
	    if(head.treeNode.parent != null)
		head.treeNode.parent.children.add(head.treeNode);

	    Edge u = head;
	    do {  
		//		if(dest != null)
		    //		    System.out.println("whoa: "+u);
		Edge v = u.sym();
		//		if(dest != null)	
		    //		    System.out.println("sym: "+v);
		if(v.next != null && !v.isMarked(Edge.getBlack())) {
		    // RELAX
		    int newDist = 0;
		    
		    if(astar && dest != null) {
			newDist = 
			    u.treeNode.curDist + 
			    u.treeNode.p.distanceInt(v.treeNode.p) +
			    v.treeNode.p.distanceInt(dest.treeNode.p) -
			    u.treeNode.p.distanceInt(dest.treeNode.p);
		    }
		    else {
			int cur = u.treeNode.curDist;
			TreeNode vtn = v.treeNode;
			NavPoint2D pp = vtn.p;
			int add = u.treeNode.p.distanceInt(pp);
			newDist = cur+add;
			/*
			newDist = 		       
			    u.treeNode.curDist + 
			    u.treeNode.p.distanceInt(v.treeNode.p);
			*/
		    }

		    /*
		    System.out.println("dist: "+newDist+" "+
				       u.treeNode.curDist + " "+
				       v.treeNode.curDist);
		    */
		    if(v.treeNode.curDist > newDist) {
			
			/*
			System.out.println("updating dist for "+v+" newdist: "+
					   newDist);
			*/
			v.treeNode.curHop = u.treeNode.curHop + 1;
			v.treeNode.curDist = newDist;
			v.treeNode.parent = u.treeNode;
		    }

		    if(v.isMarked(Edge.getClear()) && v.isValid()) {
			queue.add(v);
			v.markOrg(Edge.getGrey());
		    }
		}
		u = u.oNext();
	    } while (u != head);
	}
	return null;
    }


    public void applyEdges( MeshProcessor gp )
    {
	Enumeration enum = edgeTable.elements();
	while(enum.hasMoreElements()) {
	    QuadEdge qe = (QuadEdge)enum.nextElement();
	    gp.procMeshEdge(qe.edges[0]);
	}
    }

    public boolean wideEnough(Edge u, Edge v) 
    {
	NavPoint2D p = u.org();
	NavPoint2D q = v.org();
	
	return p.distanceSqInt(q) > 16;
    }

    /*
      Nice algorithm from the original cdt package but won't work if
      there is an incomplete mesh or a mesh with holes.  Use the bfs
      approach instead.

    // The goals of this routine are as follows:
    // - if x coincides with one or more vertices in the mesh, return an edge
    //   whose origin is one such vertex;
    // - otherwise, return the edge nearest to x oriented ccw around a face
    //   containing x (up to numerical round-off); 
    // In the event of failure, we pass the search to BruteForceLocate.
    public Edge locate(Point2D x)
    {
	Edge e = currentEdge;
	int iterations = 0;

	while (true) {
	    iterations++;
	    
	    Util.assert(iterations <= edgeTable.size());
	    //return BruteForceLocate(x);

	    // first, examine the endpoints for coincidence:
	    if(NavPoint2D.isEq(x, e.org()))
		return e;
	    if(NavPoint2D.isEq(x, e.dest()))
		return e.sym();
		
	    // x isn't one of the endpoints.
	    // orient e s.t. x is not to its right:
	    if (Util.rightOf(x, e))
		e = e.sym();

	    double a = Line2D.ptSegDistSq(e.org().getX(),
					  e.org().getY(),
					  e.dest().getX(),
					  e.dest().getY(),
					  x.getX(), x.getY());
	    
		
	    // x is not to the right of e.
	    // now we need to test if x is in the left face of e:
	    if (Edge.hasLeftFace(e)) {
		
		// does x coincide with the third vertex?
		if (NavPoint2D.isEq(x, e->lPrev()->org()))
		    return e.lPrev();
		
		if (Util.leftOf(x, e->oNext())) {
		    e = e.oNext();
		    continue;
		}
		if (Util.leftOf(x, e.dPrev())) {
		    e = e.dPrev();
		    continue;
		}
		
		// x is inside the left face of e!
		// return the edge nearest to x:
		double b = Line2D.ptSegDistSq(e.dest().getX(),
					      e.dest().getY(),
					      e.lPrev().org().getX(),
					      e.lPrev().org().getY(),
					      x.getX(), x.getY());
		double c = Line2D.ptSegDistSq(e.lPrev().org().getX(),
					      e.lPrev().org().getY(),
					      e.org().getX(),
					      e.org().getY(),
						  x.getX(), x.getY());
		switch(Util.imin(Math.abs(a), Math.abs(b), Math.abs(c))) {
		case 0:
		    return e;
		case 1:
		    return e.lNext();
		case 2:
		    return e.lPrev();
		}
	    }
		
	    // there is no left face!
	    if (a <= NavPoint2D.EPS) // on edge
		return e.sym();

	    // ok, let's try to get closer using the edges of the right face
	    if (Util.rightOf(x, e.oPrev())) {
		e = e.oPrev().sym();
		continue;
	    }
	    if (Util.rightOf(x, e.dNext())) {
		e = e.dNext().sym();
		continue;
	    }

	    System.out.println("Locate: I'm stuck and I can't move on");
	    return BruteForceLocate(x);
	}	
    }
    */
}


