/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package pathfinder;

import javax.swing.border.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import cricketdaemon.clientlib.data.*;

/**
 * GUI for the PathFinderApp.
 *
 * @author Allen Miu
 */
public class GUI extends JPanel
{
    protected JFrame f;
    protected StatusBar status;

    class StatusBar extends JPanel {
	
	public JLabel x, y, orient;
	
	public StatusBar () {
	    super();
	    
	    setLayout(new FlowLayout());
	    _addInfoLabel(" Device Coords", null);
	    x = _addInfoLabel(" x:", "0");
	    y = _addInfoLabel(" y:", "0");    
	    
	    /* to make window fit in ipaq 
	       if( s != null )
	       orient = _addInfoLabel(" angle:", "0.0");    
	    */
	}
	
	private JLabel _addInfoLabel(String title, String value) 
	{
	    JLabel t = new JLabel(title, JLabel.LEFT);
	    add(t);
	    
	    if( value != null ) {
		JLabel v = new JLabel(value, JLabel.RIGHT);
		add(v);
		return v;
	    }
	    return null;
	}
	
    }

    Image im;
    Graphics2D imgc;
    JList pathList, destList;

    //final static BasicStroke stroke = new BasicStroke(2.0f);
    final static BasicStroke wideStroke = new BasicStroke(8.0f);

    private double curArrow_cx, curArrow_cy, nextArrow_cx, nextArrow_cy;

    private static String FONTNAME = "Arial";
    private static int FONTSTYLE = Font.BOLD;
    private static int FONTSIZE = 22;

    private FontMetrics fontMetrics;

    private PathFinderApp pathfinder;
    public DestinationGUI destGUI;

    private double scale = 1.0;

    private void setFontDefault(Graphics g)
    {
	Font f = new Font(FONTNAME, Font.PLAIN, 12);
	g.setFont(f);
	fontMetrics = g.getFontMetrics();
    }

    private void setFonts(Graphics g)
    {
	Font f = new Font(FONTNAME, FONTSTYLE, FONTSIZE);
	g.setFont(f);
	fontMetrics = g.getFontMetrics();
    }

    public GUI(String title, PathFinderApp p) 
    {
	pathfinder = p;
	preferredSize = new Dimension(220, 220);
	curArrow_cx = nextArrow_cx = preferredSize.width/2;
	curArrow_cy = preferredSize.height/4;
	nextArrow_cy = 3*preferredSize.width/4;

	setBorder(createBorder());

	f = new CustomFrame(title);
	Container container = f.getContentPane();
	container.setLayout(new BoxLayout(container, BoxLayout.Y_AXIS));

	JToolBar toolbar = new JToolBar();
	JButton destButt = new JButton("Destinations");
	destButt.addMouseListener(new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
		    Util.dbgPrintln("Destinations clicked");
		    launchDestinationsGUI();
		}
	    });
	toolbar.setFloatable(false);
	toolbar.add(destButt);
	container.add(toolbar);


	container.add(this);

	pathList = new JList();
	pathList.setPrototypeCellValue("0123456789");

	destList = new JList();
	destList.setPrototypeCellValue("0123456789");
	destList.addMouseListener(DestinationGUI.selectDestListener);
	
	JScrollPane scroll = new JScrollPane(pathList);
	scroll.setBorder(createBorder());
	container.add(scroll);

	status = new StatusBar();
	
	status.setPreferredSize(new Dimension(preferredSize.width, 
					      preferredSize.height/10));
	status.setBorder(BorderFactory.createLoweredBevelBorder());
	container.add(status);

	f.setBackground(Color.white);
	f.validate();
	f.pack();
	f.setVisible(true);


	im = createImage(preferredSize.width, preferredSize.height);
	imgc = (Graphics2D) im.getGraphics();
        imgc.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
			      RenderingHints.VALUE_ANTIALIAS_ON);
	imgc.setStroke( wideStroke );

	//	imgc.setXORMode(Color.red);
	//	imgc.setColor(Color.blue);

	setFonts(imgc);

	Util.dbgPrintln("view created and exposed");
    }

    /* public API for program */
    public void setX(double s) {
	status.x.setText(Util.textOf(s, 3));
    }

    public void setY(double s) {
	status.y.setText(Util.textOf(s, 3));
    }

    public void setZ(double s) {
	//status.z.setText(s);
    }

    public void setQueryPathStatus(boolean querying) 
    {
	return;
    }

    public double dist(double x, double y)
    {
	return Math.sqrt(x*x + y*y);
    }
    public void update(Position cur, Vector path,
		       int targetIndex, int nextIndex)
    {
	updatePath(path, targetIndex, nextIndex);

	imgc.clearRect(0, 0, preferredSize.width, preferredSize.height);
	
	Waypoint target = (Waypoint) path.get(targetIndex);
	double curArrow_dx = target.pos.x - cur.x + 0.00000000001;
	double curArrow_dy = target.pos.y - cur.y;
	double curArrow_angle = Math.atan(curArrow_dy/curArrow_dx);
	if(curArrow_dx < 0)
	    curArrow_angle += Math.PI;
	double curArrow_l  = scale*dist(curArrow_dx, curArrow_dy);
	drawArrow(curArrow_cx, curArrow_cy, curArrow_l/2,
		   curArrow_angle);

	if(nextIndex >= 0) {
	    Waypoint next = (Waypoint) path.get(nextIndex);
	    double nextArrow_dx = next.pos.x - cur.x + 0.00000000001;
	    double nextArrow_dy = next.pos.y - cur.y;
	    double nextArrow_angle = Math.atan(nextArrow_dy/nextArrow_dx);
	    double nextArrow_l  = scale*dist(nextArrow_dx, nextArrow_dy);
	    drawArrow(nextArrow_cx, nextArrow_cy, nextArrow_l/2,
			     nextArrow_angle);
	}

	repaint();
    }

    private void updatePath(Vector path, int target, int next)
    {
	int selection[] = new int[2];
	selection[0] = target;
	selection[1] = next;

	pathList.setListData(path);
	pathList.setSelectedIndices(selection);
    }

    public void setDestinations(Vector dests)
    {
	destList.setListData(dests);
    }
    
    private void drawArrow(double x, double y, double r, double a)
    {
	double ix = x;
	double iy = y;
	double x2 = r*Math.cos(a);
	double y2 = r*Math.sin(a);

	int x3 = (int)(ix+(1.1)*x2);
	int y3 = (int)(iy+(1.1)*y2);

	System.out.println("angle: "+a);

	imgc.setColor(Color.black);
	imgc.draw(new Line2D.Double(ix, iy, ix+x2, iy+y2));
	imgc.setColor(Color.red);
	imgc.fillOval(x3-5, y3-5, 10, 10);
	imgc.setColor(Color.black);
    }

    /* Java AWT stuff */
    public void paintComponent(Graphics g) 
    {
        super.paintComponent(g);
	if( im != null )
	    g.drawImage(im,0,0,null);
    }

    Dimension preferredSize;
    public Dimension getPreferredSize() 
    {
	return preferredSize;
    }

    private Border createBorder() 
    {
	Border raisedbevel = BorderFactory.createRaisedBevelBorder();
	Border loweredbevel = BorderFactory.createLoweredBevelBorder();
	
	return
	    BorderFactory.createCompoundBorder(raisedbevel, loweredbevel);
    }

    private void launchDestinationsGUI() 
    {
	if(destGUI != null)
	    return;
	destGUI = new DestinationGUI(pathfinder, this, destList);
	destGUI.setStatusBusy(true);
	pathfinder.queryDestinations();
	destGUI.setStatusBusy(false);
    }
}



