package streamredirector;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

/**
 * This is class is used to monitor the run-time values of the data members of the
 * specified object. It reads the user's input on the console and executes the appropriate
 * command (for more details on the allowed commands, read the printUsage() method).
 *
 * @author Jorge Rafael Nogueras
 */

public class ClassMonitor implements Runnable
{
  /** the object whose data members will be monitored */
  protected Object object;
  
  /** the thread in which we will be running, reading the user's input */
  protected Thread thread;
  
  /** the list of the variables contained in the object (keyed by the name of each
      field and whose value is the Field object that describes it */
  protected Hashtable variablesList;
  
  /**
   * Initializes the class to examine the data members of the specified object.
   *
   * @param the_object the object whose data members will be examined
   */

  public ClassMonitor (Object the_object)
  {
    object = the_object;
    
    variablesList = new Hashtable();
    
    // we get a list of all the fields (data members) declared in the specified object
    Field [] classFields = the_object.getClass().getDeclaredFields();
    
    // we add all the Field values to the hashtable, keying each one with its name
    for (int i = 0; i < classFields.length; i++)
    {
      Field thisField = classFields [i];
      variablesList.put (thisField.getName(), thisField);
    }
    
    // we start the thread that reads the user's input
    thread = new Thread (this, "ClassMonitor reader #" + hashCode());
    thread.start();
  }
  
  /**
   * This method continously reads the user's input on the console and processes each
   * command typed.
   */

  public void run()
  {
    BufferedReader br = new BufferedReader (new InputStreamReader (System.in));
    System.out.println ("Examining class " + object.getClass().getName() + "...");

    while (true) 
    {
      try
      {
        System.out.print ("Command [? for help] ==> ");
        String str = br.readLine();
        
        if (str == null || str.equals (""))
          continue;
    
        StringTokenizer st = new StringTokenizer (str);

        if (!st.hasMoreElements())
          continue;
    
        // commands are case insensitive, so we compare them always in lowercase
        String cmd = st.nextToken().toLowerCase();
    
        if (cmd.equals ("q") || cmd.equals ("x") ||
            cmd.equals ("quit") || cmd.equals ("exit") ||
            cmd.equals ("shutdown") || cmd.equals ("close"))
        {
          // exits the JVM normally
          System.out.println ("Closing down...");
          System.exit (0);
        }
        else if (cmd.equals ("pa") || cmd.equals ("printall"))
        {
          // prints all the data members, their types and their current values
          Enumeration list = variablesList.keys();
          
          while (list.hasMoreElements())
          {
            String variableName = (String) list.nextElement();
            printVariable (variableName);
          }
        }
        else if (cmd.equals ("p") || cmd.equals ("print"))
        {
          // prints the type and value of the specified data member
          try
          {
            String variableName = st.nextToken();
            printVariable (variableName);
          }
          catch (Exception e)
          {
            System.err.println ("==> Correct usage is: \"print <variable-name>\"");
          }
        }
        else if (cmd.equals ("gc"))
        {
          // "suggests" to the JVM to run the garbage collector and displays how much memory was regained
          long memory = Runtime.getRuntime().freeMemory();
          System.out.println ("==> Running garbage collector...");
          System.gc();
          System.out.println ("==> Memory regained: " + (Runtime.getRuntime().freeMemory() - memory) + " bytes");
        }
        else if (cmd.equals ("free") || cmd.equals ("mem") || cmd.equals ("memory"))
        {
          // displays how much free memory there is in the JVM
          System.out.println ("==> Free memory: " + Runtime.getRuntime().freeMemory() + " bytes");
        }
        else if (cmd.equals ("?") || cmd.equals ("h") || cmd.equals ("help") ||
                 cmd.equals ("usage"))
        {
          // prints a help screen displaying all the available commands and their usage
          printUsage();
        }
        else
        {
          // if the command is unrecognized, print the usage screen
          System.err.println ("==> Command \"" + cmd + "\" unknown.");
          printUsage();
        }

      }
      catch (Exception e)
      {
        e.printStackTrace();
        continue;
      }
    }
  }
  
  /**
   * Prints the specified data member and its current value.
   *
   * @param variableName the data member whose current value will be printed
   */

  protected void printVariable (String variableName)
  {
    try
    {
      if (variableName == null)
      {
        System.err.println ("==> There is no variable named: " + variableName);
      }
      else
      {
        // find the Field object that describes the specified data member
        Field thisField = (Field) variablesList.get (variableName);
        
        if (thisField == null)
        {
          System.err.println ("==> There is no variable named: " + variableName);
        }
        else
        {
          // get the current value of the specified data member in the specified object
          Object variableValue = thisField.get (object);
          Class variableType = thisField.getType();
          
          // print out the data member's name, its type and its current value
          System.out.println ("==>\t" + variableName + " ["  + variableType.getName() + "] => " + variableValue);
        }
      }
    }
    catch (Exception e)
    {
      System.err.println ("==> Error getting information on variable named: " + variableName);
      e.printStackTrace();
    }
  }
  
  /**
   * Prints a help screen displaying all the available commands and their usage.
   */

  protected void printUsage()
  {
    System.err.println ("==> Command usage:\n\t" +
      "pa | printall\t\t\t==> prints all the variables along with their types and their values\n\t" +
      "p  | print <variable-name>\t==> print the type and value of variable <variable-name>\n\t" +
      "gc\t\t\t\t==> runs the garbage collector\n\t" +
      "free | mem | memory\t\t==> shows the available (free) memory in the JVM\n\t" +
      "x | q | exit | quit\n\t\t| shutdown | close\t==> to exit the program and shut down the JVM\n\t" +
      "? | h | help | usage\t\t==> print this usage message\n\t");
  }
}