package streamredirector;

/**
 * This interface should be implemented by any class that will listen to commands received
 * by a ListeningTransport; right now it is implemented by the SinkProxy.
 *
 * @see streamredirector.ListeningTransport, streamredirector.SinkProxy
 * @author Jorge Rafael Nogueras
 */

public interface CommandListener
{
	/**
	 * This method will get called when the establishConnection command is received with
	 * two parameters.
	 *
   * @param server embodies the connection with the server
   * @param sink where the stream will be redirected by default
   * @exception Exception if there are errors creating the piped output streams or connection errors
   * @return a long value representing the cookie that uniquely identifies the connection
	 */
	 
	public Long establishConnection (ServerConnection server, SinkConnection sink) throws Exception;
	
	/**
	 * This method will get called when the establishConnection command is received with
	 * only one parameter (assuming the default sink).
	 *
   * @param server embodies the connection with the server
   * @exception Exception if there are errors creating the piped output streams or connection errors
   * @return a long value representing the cookie that uniquely identifies the connection
	 */

	public Long establishConnection (ServerConnection server) throws Exception;

	/**
	 * This method will get called when the sendCommand command is received.
	 *
	 * @param cookie identifies uniquely the connection where the command must be sent
   * @param command the command that the server will be sent
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a return code indicating success or failure
	 */

	public ReturnCode sendCommand (Long cookie, Command command) throws Exception;

  /**
   * This method will get called when the renewLease command is received.
   *
   * @param cookie identifies uniquely the connection whose lease should be renewed
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a return code indicating success or failure
   */

  public ReturnCode renewLease (Long cookie) throws Exception;

  /**
   * This method will get called when the endConnection command is received.
   *
   * @param cookie identifies uniquely the connection that must be terminated
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a ServerCommand that resumes the connection from the point where it was ended
   */

  public ServerCommand endConnection (Long cookie) throws Exception;
}