package streamredirector;

/**
 * This is a convenience class that simplifies reneweing connection leases.
 * By default, if the proxy does not receive a renewLease() command every
 * so often, it will forcefully terminate the connection between the server and
 * the sink (this is done in order to respond to controllers who die or lose their
 * their connection to the sink proxy). It has a thread that every so often (specified
 * by the user) renews the connection lease for a specific connection.
 *
 * @author Jorge Rafael Nogueras
 */

public class ConnectionLeaseRenewer implements Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** the default delay between lease renewals (in milliseconds) */
  protected final static long DEFAULT_DELAY = 10 * 1000;
  
	/** the reference to transport where commands are written */
	protected Transport transport;
	
	/** cookie that identifies the connection that is being periodically renewed */
	protected Long cookie;

	/** amount of time (in milliseconds) between connection lease renewals */
	protected long delay;

	/** the timer thread that will close the connection after some time */
	protected Thread timer;

	/** this variable, if true, will cause the thread to stop running */
	protected boolean interrupted;
	
  /**
   * Initializes the connection lease renewer so that it renews the connection identified
   * by the specified cookie every DEFAULT_DELAY milliseconds; the command to renew the lease
   * will be written to the specified transport.
	 *
	 * @param the_transport the transport where the command to renew the lease will be written to 
   * @param the_cookie identifies the connection to be renewed
   */

  public ConnectionLeaseRenewer (Transport the_transport, Long the_cookie)
  {
  	this (the_transport, the_cookie, DEFAULT_DELAY);
  }
  
  /**
   * Initializes the connection lease renewer so that it renews the connection identified
   * by the specified cookie every "delay" milliseconds; the command to renew the lease
   * will be written to the specified transport.
	 *
	 * @param the_transport the transport where the command to renew the lease will be written to 
   * @param the_cookie identifies the connection to be renewed
   * @param the_delay amount of time (in milliseconds) between connection lease renewals
   */

  public ConnectionLeaseRenewer (Transport the_transport, Long the_cookie, long the_delay)
	{
		transport = the_transport;
		cookie = the_cookie;
		delay = the_delay;

		interrupted = false;

		timer = new Thread (this, "Lease Renewer #" + hashCode());
		timer.start();
    
    if (DEBUG)
      System.out.println ("ConnectionLeaseRenewer::<init> -> Starting lease renewer #" + hashCode());
	}
	
	/**
	 * This method makes the timer thread stop running; it should be called when the connection
	 * state object is removed from the hash table so the timer stops running.
	 */
	 
	public void stop()
	{
    if (DEBUG)
      System.out.println ("ConnectionLeaseRenewer::stop -> Stopping timer...");
    
		interrupted = true;
		timer.interrupt();
	}
	
	/**
	 * The timer will be running this method until the interrupt() method is called or until an
	 * unexpected exception (something other than an InterruptedException) gets raised.
	 */
	 
	public void run()
	{
		while (!interrupted)
		{
			try
			{
				// sleep for the specified amount of time
//System.out.println ("ConnectionLeaseRenewer::run -> About to sleep for " + delay + " seconds...");
				Thread.sleep (delay);

//System.out.println ("ConnectionLeaseRenewer::run -> Lease renewed!...");

				transport.renewLease (cookie);
			}
			catch (InterruptedException ie)
			{
				// do nothing, simply go to sleep again
//        ie.printStackTrace();
			}
			catch (Exception e)
			{
				// an unexpected exception was raised (probably in the call to sendCommand()); this could
				// signal that there is a connection error, in which case future attempts to renew the 
				// connection lease will be futile; we thus make the thread stop running				
//				e.printStackTrace();
				
				return;
			}
		}
    
    if (DEBUG)
      System.out.println ("ConnectionLeaseRenewer::run -> shutting down ConnectionLeaseRenewer timer #" + hashCode());
	}

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

	public String toString()
	{
		return "[" + this.getClass().getName() + ": transport = " + transport + ", cookie = " + cookie + ", delay = " + delay + "]";
	}
}