package streamredirector;

/**
 * This class keeps all the state necessary to identify a connection between a stream
 * server and a sink. It gets created whenever a connection is established. It is also
 * in charge of starting a timer thread that should be renewed every so often (by way
 * of the method "renewLease()"). If the timer finishes its countdown a few times without
 * getting renewed, it will assume that the controller has died or has been disconnected
 * and it will manually shut down the connection between the server and the sink.
 *
 * @see streamredirector.SinkProxy
 * @author Jorge Rafael Nogueras
 */

public class ConnectionState implements Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** the maximum number of consecutive timeouts there can be before the connection is closed */
  protected final static int MAX_TIMEOUTS = 2;
  
	/** the cookie that uniquely identifies this connection */
	public Long cookie;
	
	/** the reference to the server */
	public ServerConnection server;

	/** the reference to the sink */
	public SinkConnection sink;

	/** the reference to the sink proxy that listens for commands: this is kept so that
      the connection can be closed by calling endConnection() on the SinkProxy if the
      connection state's lease timer expires */
	protected CommandListener proxy;
	
	/** the timer thread that will close the connection after some time */
	protected Thread timer;

	/** this variable, if true, will cause the thread to stop running */
	protected boolean interrupted;

	/** keeps count of how many successive timeouts there have been */
	protected int timeouts;
	
  /**
   * Initializes the connection state with the appropriate parameters
   *
   * @param the_cookie the cookie that uniquely identifies this connection
   * @param the_server the reference to the server
   * @param the_sink the reference to the sink
   * @param the_proxy the reference to the sink proxy that listens for commands
   */

   public ConnectionState (Long the_cookie, ServerConnection the_server,
													SinkConnection the_sink, CommandListener the_proxy)
	{
		cookie = the_cookie;
		server = the_server;
		sink = the_sink;
		proxy = the_proxy;
		
		timeouts = 0;

		interrupted = false;

		timer = new Thread (this, "Connection State Timer #" + hashCode());
		timer.start();
	}
	
	/**
	 * This method resets the number of timeouts to zero and interrupts the sleeping timer
	 * but without stopping the thread itself; the run() method will still run and thus the
	 * timer will go to sleep once again. This method should be called by the controller in
	 * regular intervals so the connection between the server and the sink is not prematurely
	 * closed.
	 */
	 
	public void renewLease()
	{
//System.out.println ("ConnectionState::renewLease -> Cookie (" + cookie + "): Lease renewed! Interrupting timer...");

		timeouts = 0;

		timer.interrupt();
	}
	
	/**
	 * This method makes the timer thread stop running; it should be called when the connection
	 * state object is removed from the hash table so the timer stops running.
	 */
	 
	public void stop()
	{
//System.out.println ("ConnectionState::interrupt -> Cookie (" + cookie + "): Stopping timer...");
		interrupted = true;
		timer.interrupt();
	}
	
	/**
	 * The timer will be running this method until the interrupt() method is called or until an
	 * unexpected exception (something other than an InterruptedException) gets raised.
	 */
	 
	public void run()
	{
		while (!interrupted)
		{
			try
			{
				// sleep for the default number of milliseconds
        if (DEBUG)
          System.out.println ("ConnectionState::run -> Cookie (" + cookie + "): About to sleep for 30 seconds...");

				Thread.sleep (ConnectionLeaseRenewer.DEFAULT_DELAY);
				
				// if the timer was not interrupted, increase the number of timeouts
				timeouts++;

        if (DEBUG)
          System.out.println ("ConnectionState::run -> Cookie (" + cookie + "): No lease renewed! Incrementing timeouts to: " + timeouts);
				
				// when the maximum number of consecutive timeouts is reached, force the call of disconnection in the proxy
				if (timeouts == MAX_TIMEOUTS)
				{
          if (DEBUG)
            System.out.println ("ConnectionState::run -> Cookie (" + cookie + "): Timeouts exceeded! Sending end command!");
	
          proxy.endConnection (cookie);
				}
			}
			catch (InterruptedException ie)
			{
				// do nothing, simply go to sleep again
			}
			catch (Exception e)
			{
				// an unexpected exception was raised (probably in the call to sendCommand()); this could
				// signal that the connection state object was removed but not interrupted. To provide for this
				// case, we will make the timer stop running by setting "interrupted" to true.
				e.printStackTrace();
				
				return;
			}
		}
    
    if (DEBUG)
      System.out.println ("ConnectionState::run -> shutting down ConnectionState timer #" + hashCode());
	}

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

	public String toString()
	{
		return "[" + this.getClass().getName() + ": cookie = " + cookie + ", server = " + server + ", sink = " + sink + "]";
	}
}