package streamredirector;

import java.util.*;

/**
 * This class abstracts the logic for connection establishment, moving a connection from one sink
 * to another and ending a connection. It also has helper methods to discover sinks and the like.
 * A Controller is an INS Application that gets its location information from the Cricket system.
 * It can use the INS network to find sink proxies that have announced their existence to the INS
 * network and it can ask for specific proxies that are in the Controller's current location.
 *
 * @see streamredirector.NetworkResourceManager, streamredirector.PhysicalLocationManager
 * @author Jorge Rafael Nogueras
 */

public class Controller extends LocationUpdateRegistrar implements LocationUpdateListener
{
  /** this variable sets whether or not debug println's are printed */
  public static final boolean DEBUG = true;
  
  /** keeps track of the connection state for the different server-client proxy connections */
  protected Hashtable connectionTable;
  
  /** this is the object that interfaces with the network resource discovery system */
  protected NetworkResourceManager networkResourceManager;

  /** this is the object that interfaces with the physical location discovery system */
  protected PhysicalLocationManager physicalLocationManager;

  /**
   * Initializes a Controller with the objects that interface with the network resource discovery
   * manager and the one that interfaces with the physical location discovery system.
   *
   * @param listener the object that is interested in learning when the location changes
   * @param network_resource_manager the object that interfaces with the network resource discovery system
   * @param physical_location_manager the object that interfaces with the physical location discovery system
   */

  public Controller (LocationUpdateListener listener, NetworkResourceManager network_resource_manager,
                     PhysicalLocationManager physical_location_manager)
	{
    connectionTable = new Hashtable();
    locationUpdateListeners = new Hashtable();

    addLocationUpdateListener (listener);
//System.out.println ("Controller::<init> " + this + " -> After adding listener, Controller's listeners are: " + locationUpdateListeners);    

    networkResourceManager = network_resource_manager;
    
    physicalLocationManager = physical_location_manager;
    physicalLocationManager.configure (this);
	}

  public void updateLocation (LocationInformation info)
  {
    // we inform the network resource manager that the location has changed
    networkResourceManager.updateLocation (info);
//System.out.println ("Controller::updateLocation " + this + " -> Notifying Controller's listeners: " + locationUpdateListeners);    
    // notify all listeners that the location has changed
    fireLocationUpdatedEvent (info);
  }

  /**
   * This method asks the network resource discovery system for all sink proxies that have announced themselves to the
   * network. It can either block indefinitely until one active sink proxy has been found (in which case it should be
   * called from within another thread) or it can return immediately regardless of how many active proxies are found.
   * Only active sinks (those whose listening transports are actually listening) will be returned.
   *
   * @param block if set to BLOCK, we will keep searching until at least one active sink is found; if set
   *              to NO_BLOCK, we will return immediately (the vector returned will be empty if no proxies are found)
   * @return a vector of SinkProxyDescriptor which is the list of the names for the active sink proxies that
   *              were found, or an empty (size == 0) vector if none are found
   */

  public Vector getAllSinks (boolean block)
  {
    return networkResourceManager.getAllSinks (block);
  }

  /**
   * This method asks the INS network for a sink proxy that exists in the current location
   * of the controller. Note that this method will block until at least one sink proxy is found
   * on the network: it will return "null" if none of the sink proxies found is at the same
   * location as the controller.
   *
   * @param block if set to BLOCK, we will keep searching until at least one local sink is found; if set
   *              to NO_BLOCK, we will return immediately (the vector returned will be empty if no proxies are found)
   * @return a vector of SinkProxyDescriptor which are the names of the sink proxies that were found 
   *         in the same location as the controller, or an empty (size == 0) vector if none are found
   */

  public Vector getLocalSinks (boolean block)
  {
    return networkResourceManager.getLocalSinks (block);
  }
  
  /**
   * Gets all the sinks local to the controller (i.e. those that are in the same location) and
   * arbitrarily returns the first one. This should be used when, given a choice of several local sinks
   * they are all equally good. If no local sinks are found, "null" will be returned.
   *
   * @param block if set to BLOCK, we will keep searching until at least one local sink is found; if set
   *              to NO_BLOCK, we will return immediately (the vector returned will be empty if no proxies are found)
   * @return the SinkProxyDescriptor of an active sink that is local to the controller
   */

  public SinkProxyDescriptor getAnyLocalSink (boolean block)
  {
    return networkResourceManager.getAnyLocalSink (block);
  }

  /**
   * This method will get called when the user wants to establish a connection
   * between a server and a sink. It specifies the server where the stream will be coming
   * from and the sink where data should be written
   *
   * @param serverInfo embodies the server's information
   * @param sinkInfo embodies the sink proxy's information
   * @exception Exception if there are errors sending the server and sink connection objects or connecting to the sink proxy
   * @return a long value representing the cookie that uniquely identifies the connection
   */

  public Long establishConnection (ServerInfo serverInfo, SinkInfo sinkInfo) throws Exception
  {
    if (DEBUG)
      System.out.println ("Controller::establishConnection -> Server info: " + serverInfo + ", sink info: " + sinkInfo);

    // get the pertinent ServerConnection object
    ServerConnection server = serverInfo.getServerConnection();

    // get the pertinent SinkConnection and Transport objects
    SinkConnection sink = sinkInfo.getSinkConnection();
    Transport transport = sinkInfo.getTransport();

    Long cookie = null;

    // if no sink was specified, this means the sink proxy should use its default sink; thus we
    // send the "establishConnection()" command with only one parameter, which is the ServerConnection object
    if (sink == null)
    {
      cookie = transport.establishConnection (server);
    }
    else
    {
      cookie = transport.establishConnection (server, sink);
    }
    
    // a lease renewer object is created to periodically renew the lease on this connection
    ConnectionLeaseRenewer renewer = new ConnectionLeaseRenewer (transport, cookie);
    
    // a controller connection state object is created and stored in the connection table under
    // the connection's (unique) cookie
    connectionTable.put (cookie, new ControllerConnectionState (cookie, serverInfo, sinkInfo,
                                                                server, sink, transport, renewer));
    
    return cookie;
  }

  /**
   * This method will get called when the user wants to send a command to the sink proxy after
   * the connection has been established.
   *
   * @param cookie the value that uniquely identifies the connection
   * @param command the command that will be sent to the sink
   * @exception Exception if there are communication errors
   * @return a return code indicating success or an error condition
   */

  public ReturnCode sendCommand (Long cookie, Command command) throws Exception
  {
    ControllerConnectionState state = (ControllerConnectionState) connectionTable.get (cookie);
    
    // if we can not find the state corresponding to that cookie, it means that the connection
    // does not exist or that it was already closed; in this case, we simply return an error code
    if (state == null)
      return ReturnCode.COMMAND_ERROR;
    else
      return state.transport.sendCommand (cookie, command);
  }

  /**
   * This method will get called when the user wants to end the connection.
   *
   * @param cookie identifies uniquely the connection that will be terminated
   * @exception Exception if there are communication errors
   * @return a ServerCommand that resumes the connection from the point where it was ended
   */

  public ServerCommand endConnection (Long cookie) throws Exception
  {
    if (DEBUG)
      System.out.println ("Controller::endConnection -> Ending connection with cookie: " + cookie);

    // we get the stored state for the connection that must be ended; note that said state
    // is also removed from the hashtable
    ControllerConnectionState state = (ControllerConnectionState) connectionTable.remove (cookie);

    // if we can not find the state corresponding to that cookie, it means that the connection
    // does not exist or that it was already closed; in this case, we simply return null
    if (state == null)
    {
      return null;
    }
    else
    {
      try
      {
        // when a connection is ended, its lease renewer is stopped permanently
        state.renewer.stop();
        
        return state.transport.endConnection (cookie);
      }
      catch (Exception e)
      {
        // if there are problems ending a previous connection, just return null to indicate that
        // the connection can not be resumed 
        return null;
      }
    }
  }

  /**
   * This method will get called when the user wants to move the connection from one
   * sink to another.
   *
   * @param cookie identifies uniquely the connection that will be terminated
   * @param sinkInfo embodies the information of the new sink proxy where the connection will be moved
   * @param returnCode (out parameter) where the return code will be stored (must be instantiated before the call)
   * @exception Exception if there are communication errors
   * @return a long value representing the cookie that uniquely identifies the newly-established connection
   */

  public Long moveConnection (Long cookie, SinkInfo sinkInfo, ReturnCode returnCode) throws Exception
  {
    // we get the stored state for the connection that must be moved
    ControllerConnectionState state = (ControllerConnectionState) connectionTable.get (cookie);
    
    // if we can not find the state corresponding to that cookie, it means that the connection
    // does not exist or that it was already closed; in this case, we simply set the ReturnCode
    // to an error and we return an "error-value" cookie, -1L.
    if (state == null)
    {
      if (returnCode != null)
      {
        returnCode.setReturnCode (ReturnCode.COMMAND_ERROR);
      }
      
      return new Long (-1);
    }

    // we ask that the current connection be ended, and, if the transfer was not completed, we should
    // get back a ServerCommand that can be used to "move" the connection, that is, re-establish it
    // on a different sink and have it continue where it left off
    ServerCommand resumeCommand = endConnection (cookie);
    
    // if the transfer was completed, the resume command will be null
    if (resumeCommand != null)
    {
      // we ask that the resume command be configured: depending on the actual server being used,
      // it is possible that the resume command must be configured somehow; to provide for this,
      // we call the configureResumeCommand() method on the ServerInfo class and use the configured
      // command
      resumeCommand = state.serverInfo.configureResumeCommand (resumeCommand);
      
      if (DEBUG)
        System.out.println ("Controller::moveConnection -> ended command and got: " + resumeCommand);

      // we get the appropriate sink connection from the SinkInfo object passed as a parameter
      state.sink = sinkInfo.getSinkConnection();
      state.transport = sinkInfo.getTransport();
      
// <HERE> I will remove this as soon as we have two physically separate sinks
//System.out.println ("Controller::moveConnection -> Artificially waiting 5 seconds so the player empties its buffer");
//try { Thread.sleep (5000); } catch (Exception e) { }
      
      if (state.sink == null)
      {
        cookie = state.transport.establishConnection (state.server);
      }
      else
      {
        cookie = state.transport.establishConnection (state.server, state.sink);
      }
      
      // update cookie's value in the state object to the cookie of the new connection established
      state.cookie = cookie;
      
      // we create a new lease renewer with the new cookie value but reusing the old transport
      state.renewer = new ConnectionLeaseRenewer (state.transport, cookie);
//System.out.println ("Controller::moveConnection -> re-storing connection state: " + state);

      // since some of the connection state has changed, we will store it again, this time
      // using the new connection's cookie as the key
      connectionTable.put (cookie, state);
//System.out.println ("Controller::moveConnection -> After re-storing, connection state is: " + state);
      
      // we send the resume command to the new sink proxy where the connection has been moved and
      // we set the return code to whatever that command returns
      returnCode.setReturnCode (state.transport.sendCommand (cookie, resumeCommand));
      
      return cookie;
    }
    else
    {
      if (returnCode != null)
      {
        // if the connection could not be moved because the original one had progressed to completion,
        // we signal that condition by setting the appropriate return code
        returnCode.setReturnCode (ReturnCode.COMMAND_COULD_NOT_BE_RESUMED);
      }

      // the cookie returned is -1, which signals that it is not a valid cookie and the result of unexpected behavior
      return new Long (-1);
    }
  }
}