package streamredirector;

import java.io.*;

import streamredirector.player.*;

/**
 * This is class is a sink connection that writes all the input it receives to
 * a specified filename. It has a listening thread that gets data from an input stream
 * and writes it out to the file. Note that only the fileName data member is not transient
 * since it is the only one that should be serialized and sent along with the object;
 * the rest of the variables will get initialized only after the "establishConnection()"
 * method is called.
 *
 * @see streamredirector.SinkConnection
 * @author Jorge Rafael Nogueras
 */

public class FileSinkConnection implements SinkConnection, Runnable
{
	/** the name of the file in the sink where the data from the server will be written */
	protected String fileName;
	
	/** where data coming going to the file are written */
	protected transient BufferedOutputStream bos;
	
	/** where data coming from the server are read */
  protected transient BufferedInputStream server;
  
  /** a long number that uniquely identifies this connection */
  protected transient Long cookie;
  
	/** this read is in charge of copying data read from the server out to the file */
  protected transient Thread thread;
  
  /** this variable, if true, will cause the thread to stop running */
	protected transient boolean interrupted;
  
	/**
	 * The constructor only takes a file name representing where the data should be written.
	 *
	 * @param the_fileName where data coming going to the file are written
	 */
	 
	public FileSinkConnection (String the_fileName)
	{
		fileName = the_fileName;
	}
	
	/**
	 * This method is called by the sink proxy when a controller has requested it that a connection
	 * be established; it creates an file output stream where data read from the server will be written.
	 * It holds off on starting the thread until the server's input stream has been initialized.
	 *
   * @param the_cookie a long number that uniquely identifies this connection
	 * @exception Exception if there is an error creating the FileOutputStream
	 */
	 
	public void establishSinkConnection (Long the_cookie) throws Exception
	{
    cookie = the_cookie;

// <HERE> adding support for web server's directory
System.out.println ("Creating file in: " + WebServerManager.getResourceDirectory().getAbsolutePath());

    File cookieDirectory = new File (WebServerManager.getResourceDirectory(), cookie.toString());

System.out.println ("Creating cookie directory: " + cookieDirectory);
    cookieDirectory.mkdir();
    cookieDirectory.deleteOnExit();
    
    File resourceFile = new File (cookieDirectory, fileName);

System.out.println ("Creating resource file: " + resourceFile);
    resourceFile.createNewFile();
    resourceFile.deleteOnExit();
    
    bos = new BufferedOutputStream (new FileOutputStream (resourceFile), 5 * 1024);

		thread = new Thread (this, "FileSinkConnection reader/writer #" + hashCode());
		interrupted = false;
	}
	
	/**
	 * Initializes the input stream where data coming from the server should be read and starts
	 * the thread that is in charge of reading data from the server and writing it to the file.
	 *
	 * @param in where data coming from the server are read
	 * @exception Exception if there is an error creating the BufferedInputStream
	 */
	 
	public void setInputStream (InputStream in) throws Exception
	{
		server = new BufferedInputStream (in, 5 * 1024);
	
		thread.start();
	}

	/**
	 * This method is called by the proxy when a controller has requested it to close the connection; it
	 * interrupts and stops the running thread and closes all appropriate streams.
	 *
	 * @exception Exception if there is an error closing the streams
	 */

	public void closeSinkConnection() throws Exception
	{
		interrupt();

		server.close();
		bos.close();
	}
	
  /**
   * This method is called when a controller has requested that a command be sent to the sink.
   * Since no commands may be sent to a FileSinkProxy, its implementation does nothing.
   *
   * @param command the command to be sent to the sink
   * @return a return code indicating success
   */
   
  public ReturnCode sendCommand (SinkCommand command)
  {
    return ReturnCode.COMMAND_OK;
  }

  /**
	 * This method makes the thread stop running, but not before having finished writing any data it may
	 * have read from the server to the file.
	 */
	 
	public void interrupt()
	{
		interrupted = true;
	}
	
	/**
	 * The thread will be running this method until the interrupt() method is called. It reads the data
	 * from the server and writes it out to the file.
	 */
	 
	public void run()
	{
		while (!interrupted)
		{
			try
			{
				byte [] buffer = new byte [5 * 1024];

        boolean started = false;
				int bytesRead;
        
        int time = 0;
//int total = 0;
				while (!interrupted && (bytesRead = server.read (buffer)) != -1)
				{
					bos.write (buffer, 0, bytesRead);
//total += bytesRead;

//System.out.println ("FileSinkConnection::run -> wrote to file: " + bytesRead + "/" + total + " bytes");
//System.out.println ("FileSinkConnection::run -> available in stream: " + server.available());

//System.out.println ("FileSinkConnection::run -> read/wrote to file: " + new String (buffer, 0, bytesRead));

					bos.flush();
  time++;
          if (!started && time != 10)
            try
            {
              started = true;
//try { Thread.sleep (10 * 1000); } catch (Exception e) { }
              Player p = PlayerManager.getPlayer ("audio/mpeg");
System.out.println ("FileSinkConnection::run -> Starting player #" + p + " and playing file: " + fileName);

// <HERE> made command be a URL
//              p.play (fileName);
              p.play ("http://localhost:" + WebServerManager.getLocalPort() + "/" + cookie + "/" + fileName);
            }
            catch (Exception e)
            {
System.out.println ("FileSinkConnection::run -> Unable to start player!");
              e.printStackTrace();
            }
        }

System.out.println ("FileSinkConnection::run -> finished writing file in directory " + cookie + ", closing");
				bos.close();
				return;
				
			}
			catch (InterruptedIOException iioe)
			{
			}
			catch (Exception e)
			{
//				e.printStackTrace();
			}
		}

System.out.println ("HTTPServerConnection::run -> shutting down FileSinkConnection reader/writer #" + hashCode());
  }
	
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

	public String toString()
	{
		return "[" + this.getClass().getName() + ": fileName = " + fileName + "]";
	}
}