package streamredirector;

import java.io.*;
import java.net.*;
import java.util.*;
import java.rmi.*;
import java.security.*;

import ins.api.*;
import ins.inr.*;
import ins.namespace.*;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;

import streamredirector.player.*;

/*
TO-DO:

- Test with all the new changes
- Figure out why pressing Move quickly makes it stop working
- Add separate Play/Stop buttons on ControllerGUI.

- check sink's available MIME types before connecting
- protect against bad locations in Cricket
- add Cricket's constant to our file
- decide what to do when a bad location is encountered: pass it up
  to updateLocation() so the sound can be stopped while a new sink is
  found?

- make sure that InvalidResourceException is actually thrown if resource is invalid

- make web proxy's port be settable from command line
*/

/**
 * This is the main executable class that runs at the sink. It instantiates a SinkProxy to handle the
 * data transfer between the server and the file that is the ultimate sink of the data.
 * It configures the proxy to use TCP as the listening transport and a file sink connection
 * object that will be where data will be written by default if the controller does not
 * specify a specific sink upon connection establishment.
 *
 * @see streamredirector.TCPListeningTransport, streamredirector.SinkConnection, streamredirector.FileSinkConnection
 * @author Jorge Rafael Nogueras
 */

public class FileSinkProxy extends Application implements Callback
{
  /** this variable sets whether or not debug println's are printed */
  public static final boolean DEBUG = false;
  
  /** the default filename of the player list */
  protected static final String defaultPlayerListFileName = "players.dat";
  
  /** the default virtual space name */
  protected static final String defaultVSpaceName = "none";
  
  /** the default DSR name */
  protected static final String defaultDSRName = "localhost";
  
	/** the generic proxy that will handle the server to sink connection */
	protected static SinkProxy proxy;

  /** the name specifier that describes this sink proxy */
  protected NameSpecifier sourceNS;
  
  /** the attribute-value entity that describes the current location */
  protected AVelement proxyLocation;

  /** the name of the virtual space name where the sink proxy is (taken from controllerLocation) */
  protected String vspaceName;
  
  protected boolean announcerStarted;

  /**
   * Initializes a FileSinkProxy with the specified virtual space name and using the local host
   * as the default DSR.
   *
   * @param the_vspaceName the virtual space name
   * @param port the port where we will listen for incoming connections
   * @exception Exception if there are errors connecting to the DSR
   */

  public FileSinkProxy (String the_vspaceName, int port) throws Exception
  {
    this ("localhost", the_vspaceName, port);
  }

  /**
   * Initializes a FileSinkProxy with the specified virtual space name and using the specified DSR.
   *
   * @param dsrName the DSR to be used
   * @param the_vspaceName the virtual space name
   * @param port the port where we will listen for incoming connections
   * @exception Exception if there are errors connecting to the DSR
   */

  public FileSinkProxy (String dsrName, String the_vspaceName, int port) throws Exception
  {
    this (dsrName, new NameSpecifier ("[vspace=" + the_vspaceName + "]" +
            "[service=sink_proxy][transport_type=TCP]" +
            "[hostname="  + InetAddress.getLocalHost().getHostAddress() + "][port=" + port + "]"));
  }

  /**
   * Initializes a FileSinkProxy with the specified name specifier and using the specified DSR.
   *
   * @param dsrName the DSR to be used
   * @param the_sourceNS the name specifier to use in the announcements
   * @exception Exception if there are errors connecting to the DSR
   */

  public FileSinkProxy (String dsrName, NameSpecifier the_sourceNS) throws Exception
  {
		super (dsrName);

		sourceNS = the_sourceNS;
    vspaceName = the_sourceNS.getVspace();
    
    // we add the MIME types contained in the player list file to the name specifer
    addMIMETypes (PlayerManager.getAllMIMETypes());
  }

  /**
   * This method satisfies Cricket's LocationAware interface: this method is called whenever
   * the controller's location changes.
   */

   public void callback (CricketData data, BitSet mask)
   {
    String location = data.getCurrentSpace();

// <HERE> temporary while Cricket's stuff is fixed    
    //int spaceIndex = location.indexOf (" ");
    //if (spaceIndex != -1)
    //  location = location.substring (0, spaceIndex);
    location = "[location=mit[building=ne43[floor=1[room=" + location + "]]]]";

    System.out.println ("FileSinkProxy::update -> Cricket says location is: " + location);
    
// <HERE> protect against bad Cricket packets
    if (new String (location.getBytes()).equals (new String (new byte[] {1, 1, 1, 0})))
    {
      System.out.println ("FileSinkProxy::update -> Bad location: "  + location);
      return;
    }
    
    if (location != null)
    {
//try { Thread.sleep (3000); } catch (Exception e) {}
      // the location of the proxy is updated to reflect the current location
      proxyLocation = new AVelement (location);

      synchronized (sourceNS)
      {
        // we remove the previous location announcement (if there was one)
        if (announcerStarted)
        {
System.out.println ("FileSinkProxy::update -> Removing announcement for: " + sourceNS);
          // we remove the previous announcement with this source name specifier
          removeAnnouncement (sourceNS);
        }

        AVelement locationElement = sourceNS.getAVelement (new Attribute ("location"));
//System.out.println ("FileSinkProxy::update -> New LocationElement: " + locationElement);
        
        if (locationElement != null)
        {
          sourceNS.removeAVelement (locationElement);
        }
          
        sourceNS.addAVelement (new AVelement (location));

        // we will remove the previous location announcement and add the new location
        // only if the announcer had already been started
        if (announcerStarted)
        {
          // we add the announcement containing the current location with a metric of 1
          addAnnouncement (sourceNS, 1);
        }
        else//if (!announcerStarted)
        {
          // if the announcer has not yet been started, we start it with the source name specifier
          startAnnouncer (new NameSpecifier[] {sourceNS}, new int[]{1}, 10000);//100000);
          announcerStarted = true;

          System.out.println ("FileSinkProxy::update -> Announcing: " + sourceNS);
        }
      }

    }

//System.out.println ("FileSinkProxy::update -> New location is " + sourceNS);
  }

  /**
   * Adds the MIME types specified to the intentional name being announced.
   *
   * @param mimeTypes the MIME types that will get added to the intentional name
   */

  protected void addMIMETypes (String [] mimeTypes)
  {
    StringBuffer sb = new StringBuffer ("[MIME_types=");
    sb.append (mimeTypes.length);
    
    for (int i = 0; i < mimeTypes.length; i++)
    {
      sb.append ("[");
      sb.append (i);
      sb.append ("=");
      sb.append (mimeTypes [i]);
      sb.append ("]");
    }

    sb.append ("]");

    // we create an attribute-value element with the constructed string
    AVelement addition = new AVelement (sb.toString());

//    synchronized (sourceNS)
//    {
      // we add the MIME types to the source name specifier
    sourceNS.addAVelement (addition);
      
      // we will only start the announcer once both the location has been set and the
      // MIME type information has been added to the name specifier
//      if (!announcerStarted && proxyLocation != null)
//      {
//  System.out.println ("FileSinkProxy::addMIMETypes -> Starting announcer with sourceNS: " + sourceNS);
//  
//        // we start the announcer with the source name specifier
//        startAnnouncer (new NameSpecifier[] {sourceNS}, new int[]{1}, 10000);//100000);
//        announcerStarted = true;
//      }

//    mimeTypesWereAdded = true;
  }
  
  protected static void printUsage()
  {
    System.err.println ("Usage:\n\n" +
        "streamredirector.FileSinkProxy -p <player list filename> -v <vspace name> -d <DSR name> [-l <location>]\n\n" +
        "\twhere:\n" +
        "\t\t<player list filename> is the filename containing the MIME types and respective players\n" +
        "\t\t\t(default value: \"" + defaultPlayerListFileName + "\")\n" +
        "\t\t<vspace name> is name of the INS virtual space where all the nodes are\n" +
        "\t\t\t(default value: \"" + defaultVSpaceName + "\")\n" +
        "\t\t<DSR name> is name of the host where the INS' Domain Space Resolver is running\n" +
        "\t\t\t(default value: \"" + defaultDSRName + "\")\n" +
        "\t\t<location> is the static location name specifier (in case the sink proxy's location never changes)\n" +
        "\t\t\t(it's in INS' name specifer format: \"[location=value]\")\n");
  }
  
  public static void main (String args []) throws Exception
	{
    String arg_PlayerListFileName = null;
    String arg_vspaceName = null;
    String arg_DSRName = null;
    String arg_StaticLocation = null;
    
    for (int i = 0; i < args.length; i++)
    {
      try
      {
        if (args [i].equals ("-p"))
        {
          if (DEBUG)
            System.out.println ("FileSinkProxy::main -> Read player file command line parameter: " + args [i+1]);

          arg_PlayerListFileName = args [++i];
        }
        else if (args [i].equals ("-v"))
        {
          if (DEBUG)
            System.out.println ("FileSinkProxy::main -> Read vspace command line parameter: " + args [i+1]);

          arg_vspaceName = args [++i];
        }
        else if (args [i].equals ("-d"))
        {
          if (DEBUG)
            System.out.println ("FileSinkProxy::main -> Read DSR command line parameter: " + args [i+1]);

          arg_DSRName = args [++i];
        }
        else if (args [i].equals ("-l"))
        {
          if (DEBUG)
            System.out.println ("FileSinkProxy::main -> Read location command line parameter: " + args [i+1]);

          arg_StaticLocation = args [++i];
        }
      }
      catch (Exception e)
      {
        e.printStackTrace();
        printUsage();
      }
    }
    
    // we will only install this security manager if none was set in the command line
    if (System.getSecurityManager() == null)
    {
      if (DEBUG)
        System.out.println ("FileSinkProxy::main -> Installing RMI security manager");

      // we set a basic security manager and instruct it to allow connections to remote sockets
      // (from where it can load the bytecodes of class files not in the current classpath)
      System.setSecurityManager (new RMISecurityManager()
      {
        public void checkPermission (Permission p) { }
      });
    }

    // the first command-line argument must be the filename of the player list file

    // initialize the PlayerManager, which is in charge of loading the information
    // related to the available players and the MIME types they can play
    try
    {
      // if the user did not specify a filename, a default is assumed
      if (arg_PlayerListFileName == null)
      {
        PlayerManager.initialize (defaultPlayerListFileName);
      }
      else
      {
        PlayerManager.initialize (arg_PlayerListFileName);
      }
    }
    catch (Exception e)
    {
      e.printStackTrace();
      System.err.println (e);
      printUsage();
      System.exit (0);
    }
    
    // <HERE> Kalpak's modification for non-native players to work
    // (had been commented out when we began using the JMF-based
    // players)

    // initialize the local web proxy: the media player will request
    // files through this web proxy
    try
    {
      WebProxyManager.initialize();
    }
    catch (Exception e)
    {
      e.printStackTrace();
      System.err.println (e);
      System.exit (0);
    }

    // if no virtual space name is specified, the default value will be used
    String vspace = (arg_vspaceName != null)? arg_vspaceName: defaultVSpaceName;
    
    // if no DSR is specified, the default will be used
    String dsrName = (arg_DSRName != null)? arg_DSRName: defaultDSRName;

    // we instantiate a SinkProxy object (which is the all-purpose proxy that manages connections
    // and piping data between a ServerConnection and a SinkConnection
    proxy = SinkProxy.getInstance();
    
    // we instantiate a listening transport to receive commands on an anonymous port, and we pass
    // the proxy as the CommandListener, the class that is interested in receiving the commands
    // read in by the listening transport
    TCPListeningTransport listeningTransport = new TCPListeningTransport (proxy);
    
    try
    {
      proxy.configure (listeningTransport, new NullSinkConnection());
    }
    catch (Exception e)
    {
      System.err.println ("FileSinkProxy::main -> Error creating ListeningTransport: " + e);
      e.printStackTrace();
      System.exit (0);
    }

    FileSinkProxy fileSinkProxy = null;
    
    try
    {
      if (args.length < 3)
      {
        printUsage();
        System.err.println ("FileSinkProxy::main -> Proceeding with default values for vspace = \"" + vspace + "\", DSR name = \"" + dsrName + "\"");
      }

      // we instantiate a file sink proxy with the specified DSR and virtual space names
      fileSinkProxy = new FileSinkProxy (dsrName, vspace, listeningTransport.getLocalPort());

      // if the sink proxy will always be announcing the same static location, we do not start the thread
      // that listens for Cricket's location updates
      if (arg_StaticLocation != null)
      {
        fileSinkProxy.proxyLocation = new AVelement (arg_StaticLocation);
        fileSinkProxy.sourceNS.addAVelement (fileSinkProxy.proxyLocation);

        // we start the announcer with the source name specifier
        fileSinkProxy.startAnnouncer (new NameSpecifier[] {fileSinkProxy.sourceNS}, new int[]{1}, 10000);//100000);
        fileSinkProxy.announcerStarted = true;
      
        System.out.println ("FileSinkProxy::main -> Announcing: " + fileSinkProxy.sourceNS);
      }
      else
      {
        // we will now start the code that receives Cricket's notification of location changes
        ServerBroker broker = new ServerBroker();
        
        // we specify that we are inly interested in space (location) information
        BitSet mask = new BitSet();
        mask.set(Broker.SPACE);
        
        // we register ourselves and start the thread that communicates with Cricket
        broker.register (fileSinkProxy, mask);
        broker.start();
      }
    }
    catch (Exception e)
    {
      System.err.println (e);
      printUsage();
      System.exit (0);
    }
    
    // start a process that allows the user to monitor this class's data members at run-time
    new ClassMonitor (fileSinkProxy);
  }
}
