package streamredirector;

/**
 * This sub-class of TCPSinkInfo knows how to create a TCPTransport object to
 * connect to the sink proxy and it can also return a customized FileSinkConnection object.
 * To do so, this class may be constructed with a file name that should be the ultimate
 * sink in the sink proxy. If this class is constructed without said filename, the
 * SinkConnection returned will be "null," which signifies that the default sink in the
 * sink proxy should be used.
 *
 * @see streamredirector.TCPTransport, streamredirector.FileSinkConnection
 * @author Jorge Rafael Nogueras
 */

public class FileTCPSinkInfo extends TCPSinkInfo
{
  /** the name of the sink filename (optional, may be null) */
  protected String sinkFileName;

  /**
   * Initializes the object with the information necessary to connect to the
   * specified sink.
   *
   * @param the_hostName the hostname of the sink proxy
   * @param the_port prefix the port where the sink proxy is listening
   */

  public FileTCPSinkInfo (String the_hostName, int the_port)
  {
    this (the_hostName, the_port, null);
  }

  /**
   * Initializes the object with the information necessary to connect to the
   * specified sink.
   *
   * @param the_hostName the hostname of the sink proxy
   * @param the_port prefix the port where the sink proxy is listening
   * @param the_sinkFileName the name of the sink filename (may be null)
   */

  public FileTCPSinkInfo (String the_hostName, int the_port, String the_sinkFileName)
  {
    super (the_hostName, the_port);
    sinkFileName = the_sinkFileName;
  }

  /**
   * This method will create a FileSinkConnection object that specifies what the ultimate
   * file sink is. If a sink file name has not been specified (it is null), then the default
   * sink should be used: in that case, we return a null SinkConnection, and the connection
   * will be established using the sink proxy's default sink,
   *
   * @return a FileSinkConnection that indicates the ultimate file sink in the proxy; it
   *          may be null to indicate that the default sink should be used
   * @exception Exception if there is an error creating the sink connection object
   */

  public SinkConnection getSinkConnection() throws Exception
  {
    if (sinkFileName == null)
    {
      return null;
    }
    else
    {
      return new FileSinkConnection (sinkFileName);
    }
  }
  
  /**
   * Gets the name of the sink filename.
   *
   * @return the name of the sink filename (may be null)
   */

  public String getSinkFileName()
  {
    return sinkFileName;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": host = " + hostName + ":" + port +
           ", sink file name = " + (sinkFileName==null?"<NONE: default sink>": sinkFileName) + "]";
  }
}