package streamredirector;

/**
 * This is a concrete implementation of a ServerCommand that describes an HTTP command. Since
 * HTTP is a simple protocol, creating an HTTP command is as simple as passing it as a string and
 * translating said command into a "wire format" is as simple as getting the bytes for the command.
 * This class is used as a base for creating more specific HTTP commands so that knowledge of the actual
 * protocol is embodied in the command object.
 *
 * @see streamredirector.Command, streamredirector.ServerCommand, streamredirector.HTTPGetCommand, streamredirector.HTTPResumeCommand
 * @author Jorge Rafael Nogueras
 */

public class HTTPCommand extends ServerCommand
{
	/** the HTTP command as a string */
	protected String command;
  
  /** the MIME type of the resource */
  protected String mimeType;
  
  /** the object that parses the HTTP command */
  protected HTTPRequestParser parser;
	
  /**
   * Creates an HTTP command, initializing the command that describes it.
   *
   * @param the_command the HTTP command as a string
   */

  public HTTPCommand (String the_command, String the_mimeType)
	{
		command = the_command;
    mimeType = the_mimeType;
    
    try
    {
      parser = new HTTPRequestParser (the_command);

//System.out.println ("Parsed: method = " + parser.getMethod() + ", resource = " + parser.getResource() + ", headers = " + parser.getHeaders());
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
	}
	
  /**
   * Returns the MIME type of the resource.
   *
   * @return the MIME type of the resource
   */

  public String getMimeType()
  {
    return mimeType;
  }
  
  /**
   * Returns the object used to parse the HTTP command.
   *
   * @return the object used to parse the HTTP command
   */

  public HTTPRequestParser getRequestParser()
  {
    return parser;
  }
  
	/**
	 * Returns a wire representation of the command that is suitable for writing to the server.
	 * In this case, since the command is just a string, we return the bytes that make up the string.
   *
   * @return the wire representation of the command (in this case, a byte array of its characters)
	 */
	
	public Object translateCommand()
	{
		return command.getBytes();
	}
	
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

	public String toString()
	{
		return "[" + this.getClass().getName() + ": command = " + command + "]";
	}
}