package streamredirector;

import java.io.*;
import java.util.*;

/**
 * This class parses an HTTP request to identify its method, the resource being requested, the
 * HTTP version being used, and the request headers included.
 *
 * @see streamredirector.WebServerInstance
 * @author Jorge Rafael Nogueras
 */

public class HTTPRequestParser implements Serializable
{
  /** the HTTP method specified in this request */
  protected String method;

  /** the resource specified in this request */
  protected String resource;
  
  /** the HTTP version specified in this request */
  protected String version;
  
  /** a list of the headers specified in the request, keyed by the header name and
      whose value is the header's value */
  protected Properties headers;
  
  /**
   * Initializes this object with a string that represents the full request being parsed.
   *
   * @param request textual representation of the request we will parse
   * @exception Exception if there are errors reading the request or parsing it
   */

  public HTTPRequestParser (String fullRequest) throws Exception
  {
    this (new BufferedReader (new StringReader (fullRequest)));
  }

  /**
   * Initializes this object with a buffered reader from where the different elements of the
   * request will be parsed.
   *
   * @param reader the buffered reader from where the request will be read
   * @exception Exception if there are errors reading the request or parsing it
   */

  public HTTPRequestParser (BufferedReader reader) throws Exception
  {
    // read the first line of the full request, which should be the request line of the form:
    // <method> <resource> <HTTP-version>\r\n
    String requestLine = reader.readLine();
    
//System.out.println ("HTTPRequestParser::<init> -> Request is: " + request);

    // now we parse the request line appropriately
    StringTokenizer tok = new StringTokenizer (requestLine);
    
    method = tok.nextToken();
    resource = tok.nextToken();
    version = tok.nextToken();

    headers = new Properties();
    
    String header = null;
    
    // we keep reading the headers, line by line, until a terminating sequence of '\r\n' indicates
    // that the request header is over (this will be signalled by reading an empty string)
    while (!(header = reader.readLine()).equals (""))
    {
//System.out.println ("HTTPRequestParser::<init> -> read header: " + header);
      int separatorPosition = header.indexOf (":");
      
      if (separatorPosition != -1)
      {
        // we read each header's name and value and store in the headers properties object
        String headerName = header.substring (0, separatorPosition);
        String headerValue = header.substring (separatorPosition + 1);
        headers.setProperty (headerName, headerValue);
      }
    }
    
//System.out.println ("HTTPRequestParser::<init> -> Finished reading header: " + this);
//System.out.println ("HTTPRequestParser::<init> -> All headers: " + headers);
  }

  /**
   * Gets the method specified in the request
   *
   * @return the method specified in the request
   */

  public String getMethod()
  {
    return method;
  }

  /**
   * Gets the resource specified in the request
   *
   * @return the resource specified in the request
   */

  public String getResource()
  {
    return resource;
  }

  /**
   * Gets the HTTP version specified in the request
   *
   * @return the HTTP version specified in the request
   */

  public String getVersion()
  {
    return version;
  }

  /**
   * Gets the headers specified in the request
   *
   * @return the headers specified in the request
   */

  public Properties getHeaders()
  {
    return headers;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": method = " + method + ", resource = " + resource +
           ", version = " + version + ", headers = " + headers + "]";
  }
}