package streamredirector;

import java.util.*;

/**
 * This is an HTTP command used for resuming a download. This command is returned when an
 * "endConnection" command is sent; it is used to continue downloading the resource at the point where
 * it left off. When the command is created, a place holder for the filename is left, and this
 * file name may be "filled-in" at any point. This is done like so because when the proxy ends a
 * connection, it has no knowledge of the name of the resource that was being downloaded, because
 * this is a protocol-dependent concept. It is thus the controller's job to fill in the name of the
 * resource once it receives this resume command.
 *
 * @see streamredirector.Command, streamredirector.ServerCommand, streamredirector.HTTPCommand, streamredirector.HTTPGetCommand
 * @author Jorge Rafael Nogueras
 */

public class HTTPResumeCommand extends HTTPCommand
{
  /** the file name of the resource being retrieved */
  protected String fileName;
  
  /** the byte offset where the transfer should be resumed */
  protected long lastOffset;
  
  /**
   * Creates an HTTP resume command, initializing the command that describes it.
   * The byte offset where the transfer should be resumed is passed as an argument.
   * Note that a the special sequence "%f" is a place holder for the actual file name
   * of the resource that will be resumed (this file name has to be set using the setFileName()
   * method)
   *
   * @param last_offset the byte offset where the transfer should be resumed 
   */

	public HTTPResumeCommand (long last_offset, String the_mimeType)
	{
		super ("GET %f HTTP/1.0\r\nRange: bytes=" + last_offset + "-\r\n\r\n", the_mimeType);
    lastOffset = last_offset;
	}

  /**
   * Creates an HTTP resume command, initializing the command that describes it using
   * the file name and the offset specified.
   *
   * @param the_fileName the name of the resource being downloaded
   * @param the_lastOffset the byte offset where the transfer should be resumed 
   */

  public HTTPResumeCommand (String the_fileName, long the_lastOffset, String the_mimeType)
  {
    super ("GET " + the_fileName + " HTTP/1.0\r\nRange: bytes=" + the_lastOffset + "-\r\n\r\n", the_mimeType);
    fileName = the_fileName;
    lastOffset = the_lastOffset;
  }

	/**
	 * Fills in the command with the appropriate file name. It parses the command, looking for
	 * the placeholder ("%f") and substituting it with the specified file name, so the command
	 * is now a valid HTTP command.
	 *
	 * @param file_name the file name that will replace the special place holder sequence
	 */

	public void setFileName (String file_name)
	{
    fileName = file_name;

		StringTokenizer st = new StringTokenizer (command, "%f");

		// there must always be two tokens: the one before the "%f" and the one after; this exception
		// should never be raised since we created the command ourselves in the constructor and we
		// made sure the special place holder sequence exists
		if (st.countTokens() != 2)
			throw new NoSuchElementException ("HTTP resume command must contain one and only one instance of \"%f\" " +
								"as the placeholder for the filename whose download is being resumed");
			
		command = st.nextToken() + file_name + st.nextToken();
//System.out.println ("HTTPResumeCommand::setFileName -> parsed command is: " + command);
	}

  /**
   * Gets the offset from where the transfer will be resumed.
   *
   * @return the offset from where the transfer will be resumed
   */
  
  public long getLastOffset()
  {
    return lastOffset;
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": GET " + fileName + " (resuming from byte offset " + lastOffset + ")]";
  }
}