package streamredirector;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This is a concrete implementation of a server connection: it embodies the connection
 * to an HTTP (web) server. It is instantiated with a URL that represents the address of
 * the server and the port where it is running (it assumes port 80 by default).
 * this object is instantiated by the Controller with the appropriate server address and
 * it is sent through the network to the sink proxy, which will use it to communicate with
 * the server. Note that only the serverAdress data member is not transient since it is the
 * only one that should be serialized and sent along with the object; the rest of the variables
 * will get initialized only after the "establishConnection()" method is called.
 * 
 * Commands are written to the HTTP server and the corresponding data are read from it, and then
 * written out to the sink.
 *
 * @see streamredirector.ServerConnection
 * @author Jorge Rafael Nogueras
 */

public class HTTPServerConnection implements ServerConnection, Runnable
{
  /** the size (in bytes) of the buffer where data will be read into and
      also the size of the buffer in the buffered stream classes */
  protected static final int BUFFER_SIZE = 10 * 1024;

	/** the host address (and port number, if necessary) of the HTTP server */
	protected URL serverAddress;
	
  /** the HTTP request that was sent to the server */
  protected transient HTTPCommand request;
  
	/** where commands going to the HTTP server are written */
	protected transient BufferedOutputStream bos;

	/** where data coming from the HTTP server are read */
	protected transient BufferedInputStream bis;
	
	/** where data read from the HTTP server are written so the sink can read them */
	protected transient BufferedOutputStream sink;

  /** a long number that uniquely identifies this connection */
  protected transient Long cookie;
  
	/** this read is in charge of copying data read from the server out to the sink */
	protected transient Thread thread;
	
	/** this variable, if true, will cause the thread to stop running */
	protected transient boolean interrupted;
	
	/** this variable determines whether the resource being read from the server exists or not */
	protected transient boolean validResource;
	
	/** this variable determines whether the resource has been completely or only partially read */
	protected transient boolean resourceCompletelyRead;
	
	/** holds the offset of the last byte read from the server and written to the sink*/
	protected transient long lastOffset;
	
	/**
	 * The constructor only takes a URL representing the address of the HTTP server.
	 *
	 * @param address the host address (and port number, if necessary) of the HTTP server
	 */
	 
	public HTTPServerConnection (URL address)
	{
		serverAddress = address;
	}
	
	/**
	 * This method is called by the proxy when a controller has requested it that a connection be
	 * established; it opens up a socket connection to the server, gets the appropriate input and
	 * output streams but holds off on starting the thread until the sink's output stream has been
	 * initialized.
	 *
   * @param the_cookie a long number that uniquely identifies this connection
	 * @exception Exception if there is an error creating a connection to the server
	 */
	 
	public void establishServerConnection (Long the_cookie) throws Exception
	{
    cookie = the_cookie;

//System.out.println ("host: " + serverAddress.getHost() + ", port = " + serverAddress.getPort());
		int port = serverAddress.getPort();

		Socket socket = new Socket (serverAddress.getHost(), (port == -1)? 80: port);

		bos = new BufferedOutputStream (socket.getOutputStream(), BUFFER_SIZE);
		bis = new BufferedInputStream (socket.getInputStream(), BUFFER_SIZE);

		interrupted = false;
		validResource = false;
		resourceCompletelyRead = false;
		lastOffset = 0;
		
		thread = new Thread (this, "HTTPServerConnection reader/writer #" + hashCode());
	}
	
	/**
	 * Initializes the output stream where data read from the server should be written and starts
	 * the thread that is in charge of reading data from the server and writing it to the sink.
	 *
	 * @param out where data read from the HTTP server are written so the sink can read them
	 * @exception Exception if there is an error creating the BufferedOutputStream
	 */
	 
	public void setOutputStream (OutputStream out) throws Exception
	{
		sink = new BufferedOutputStream (out, BUFFER_SIZE);
		thread.start();
	}
	
	/**
	 * This method is called by the proxy when a controller has requested it to close the connection; it
	 * interrupts and stops the running thread and closes all appropriate streams.
	 *
	 * @exception Exception if there is an error closing the streams
	 */

	public void closeServerConnection() throws Exception
	{
		interrupt();
		
		bis.close();
		bos.close();
		sink.close();
	}
	
	/**
	 * This method is called when a controller has requested that a command be sent to the HTTP
	 * server. It gets the command from the ServerCommand object and writes it out to the server,
	 * and then it waits for the thread to read and parse the return header. If it determines that
	 * the command proceeded without problems (i.e. the response code was "200"), the thread will notify
	 * our waiting thread so it can proceed. If the resource is somehow invalid, an appropriate
	 * exception is thrown.
	 *
	 * @param command the command to be sent to the HTTP server
	 * @exception InvalidResourceException if the resource does not exist or cannot be opened in the server
   * @return a return code indicating success or an error condition
	 */
	 
	public ReturnCode sendCommand (ServerCommand command) throws Exception
	{
    request = (HTTPCommand) command;
    
		byte [] requestBytes = (byte []) command.translateCommand();

//System.out.println ("HTTPServerConnection::run -> writing command: " + command);
		bos.write (requestBytes, 0, requestBytes.length);
		bos.flush();
		
		// wait until the reading thread parses the HTTP response header and sets the variable
		// "validResource" to its appropriate valie
		synchronized (thread)
		{
			thread.notifyAll();
//System.out.println ("HTTPServerConnection::sendCommand -> Waiting for header to be read");

			thread.wait();
			
//System.out.println ("HTTPServerConnection::sendCommand -> CONTINUED: Resource valid: "  + validResource);
			if (validResource)
			{
				return ReturnCode.COMMAND_OK;
			}
			else
			{
        // since the resource is invalid, nothing will be read from it and the thread can be stopped
				stop();
				throw new InvalidResourceException ("Server could not locate or open the specified resource");
			}
		}
	}

	/**
	 * This method constructs a command that enables the controller to resume the stream connection
	 * from the current byte offset. It is usually called after a connection has been ended if the
	 * controller later wants to resume it at the point it was interrupted (this will be done if the
	 * stream should be redirected to a different sink). The HTTP command returned is structured so
	 * that by inserting the name of the filename that should be resumed it can be issued to the
	 * server and it can begin sending the data from the appropriate byte offset.
	 *
	 * @exception Exception if there is an error creating the resume command 
   * @return a ServerCommand that resumes the connection or "null" if the resource has been completely read
	 */

	public ServerCommand getResumeCommand() throws Exception
	{
		if (resourceCompletelyRead)
      return null;
		else
			return new HTTPResumeCommand (lastOffset, request.getMimeType());
	}

	/**
	 * This method makes the thread stop running, but not before having finished writing any data it may
	 * have read from the server to the sink; it should be called when the connection is closed but there
	 * is no error.
	 */
	 
	public void interrupt()
	{
		// wake up the thread if it was waiting, so it notices that we want the thread to stop running
		synchronized (thread)
		{
			thread.notifyAll();
		}
		
		interrupted = true;
	}
	
	/**
	 * This method makes the thread stop immediately; it should be called if there is an unrecoverable
	 * error, such as if the server can not open the requested resource: in this case, nothing should be
	 * written to the sink at all and the connection should be closed immediately.
	 */
	 
	public void stop()
	{
		interrupted = true;
		thread.interrupt();
	}

	/**
	 * The thread will be running this method until the interrupt() method is called. It reads the data
	 * requested from the HTTP server and writes it out to the sink's output stream. Note that at first
	 * it must remove the HTTP response header, and only writes the resource content itself to the sink.
	 */
	 
	public void run()
	{
		while (!interrupted)
		{
			// we will wait at the top of the loop until a command is received
			synchronized (thread)
			{
				try
				{
					thread.wait();
				}
				catch (InterruptedException ie)
				{
					// if the thread is forcefully interrupted, this means that the thread should stop running,
					// so we return from the run() method, effectively stopping the thread
					return;
				}
			}

			// if we have woken up the thread only to tell it it should stop executing, the following
			// statement will be executed and the thread will stop
			if (interrupted)
			{
				return;
			}
				
			try
			{
				// here we will read and consume the input until we have detected 2 consecutive '\r\n' combinations,
				// which in HTTP signal the division between the HTTP header and the content body
				boolean found = false;
				int ch;
				StringBuffer buf = new StringBuffer();
				
				while (!interrupted && (ch = bis.read()) != -1)
				{
					buf.append ((char) ch);

//System.out.print ((char) ch);
					if (((char) ch) == '\r')
					{
						if (((char) (ch = bis.read())) == '\n')
						{
							if (found)
								break;
							else
								found = true;
						}
					}
					else
						found = false;
				}

//System.out.println ("Header is: "  + buf.toString());

				StringTokenizer tok = new StringTokenizer (buf.toString(), " ");
				
				// the first token should be protocol identification, i.e. "HTTP/1.0"
				tok.nextToken();
				
				// the following token should be the numeric HTTP return code
				String code = tok.nextToken();
				
//System.out.println ("HTTP code is: "  + code);

				synchronized (thread)
				{
					// return code 2xx means the file was found and everything is OK; otherwise,
					// it means some error or unexpected condition has occurred
					if (code.startsWith ("2"))
						validResource = true;
					else
						validResource = false;

					// we notify the main thread, which is waiting to see if the resource was valid or not
					thread.notifyAll();
					
					// if the resource is invalid, we will return to the beginning of the loop and thus prevent the
					// thread from reading or writing any more data; the thread will then wait until another
					// command is received (this should not happen with normal HTTP since there is only one request
					// and response per connection, but in case persistent connection are used, it does not hurt
					// to have the capabilities of doing so).
					if (!validResource)
						continue;
				}
				
				byte [] buffer = new byte [BUFFER_SIZE];

				int bytesRead = 0;

				while (!interrupted && (bytesRead = bis.read (buffer)) != -1)
				{
					// increment the last offset read from the stream by how many bytes were read
					lastOffset += bytesRead;
					
//System.out.println ("!!HTTPServerConnection::run -> wrote to stream: " + bytesRead + "/" + total + " bytes");
//System.out.println ("HTTPServerConnection::run -> read/wrote line: " + new String (buffer, 0, bytesRead));

					sink.write (buffer, 0, bytesRead);
					sink.flush();
				}

				// since we have finished reading the resource, we set this variable to true
				if (bytesRead == -1)
				{
System.out.println ("HTTPServerConnection::run -> finished reading file with cookie " + cookie + "!!");
					resourceCompletelyRead = true;
// <HERE> TEMP
          sink.close();
				}
else
System.out.println ("HTTPServerConnection::run -> file transfer was interrupted!");

			}
			catch (InterruptedIOException iioe)
			{
				iioe.printStackTrace();
				// if the thread is forcefully interrupted, this means that the thread should stop running,
				// so we return from the run() method, effectively stopping the thread
				return;
			}
			catch (Exception e)
			{
				synchronized (thread)
				{
					// if an exception is caught
System.out.println ("HTTPServerConnection::run -> caught exception for cookie " + cookie + ": resource invalid, notifying and waiting...");
					validResource = false;
          e.printStackTrace();
					
					try
					{
						thread.notifyAll();
					}
					catch (Exception ie) { }
				}

//				e.printStackTrace();
			}
		}
    
System.out.println ("HTTPServerConnection::run -> shutting down HTTPServerConnection reader/writer #" + hashCode());
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

	public String toString()
	{
		return "[" + this.getClass().getName() + ": URL = " + serverAddress + "]";
	}
}