package streamredirector;

import java.util.*;
import java.lang.reflect.*;

import ins.api.*;
import ins.inr.*;
import ins.namespace.*;

/**
 *
 * @see ins.api.Application, ins.namespace.NameSpecifier, streamredirector.Controller
 * @author Jorge Rafael Nogueras
 */

public class INSNetworkResourceManager extends Application implements NetworkResourceManager, Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;//Controller.DEBUG;

  /** the interval between refreshing the sink proxies list */
  protected static final long SLEEP_TIME = 60 * 1000;
  
  /** the name of the package where this class is stored (to be used for reflection purposes) */
  protected static final String packageName = "streamredirector"; //INSNetworkResourceManager.class.getPackage().getName();

  /** the name specifier that identifies all sink proxies (it need only be initialized once) */
  protected final NameSpecifier sinkProxyNameSpecifier;
  
//  /** the name specifier that describes this Controller */
//  protected NameSpecifier sourceNS;
  
  /** the name of the virtual space name where the controller is */
  protected String vspaceName;
  
  /** the attribute-value entity that describes the current location */
  protected AVelement currentLocation;

  /** the array of name specifiers that keeps the list of all the proxies discovered most recently */
  protected NameSpecifier[] discoveredProxies = new NameSpecifier[] {};
  
  /** the timer thread that will refresh the sink proxies found on the network every so often */
  protected Thread timer;

  /** this variable, if true, will cause the thread to stop running */
  protected boolean interrupted;

  /**
   * Initializes a INSNetworkResourceManager using the default DSR running on the local machine and
   * the specified INS virtual space name.
   */

  public INSNetworkResourceManager (String the_vspaceName) throws Exception
  {
    this ("localhost", the_vspaceName);
  }

  /**
   * Initializes a INSNetworkResourceManager using the DSR and vspace specified.
   *
   * @param dsrName the host name of the machine where the DSR is running
   */

  public INSNetworkResourceManager (String dsrName, String the_vspaceName) throws Exception
  {
    super (dsrName);

    NameSpecifier sourceNS = new NameSpecifier ("[vspace=" + the_vspaceName + "][service=controller]");
    vspaceName =  the_vspaceName; 
    
    // we initialize the name specifier that identifies all sink proxies in INS: note that the virtual
    // space name is taken directly from the controller's vspace, which means that the sink proxies must be
    // in the same virtual space as the controller
    sinkProxyNameSpecifier = new NameSpecifier ("[vspace=" + vspaceName + "][service=sink_proxy]");
    
//    // we announce the controller's presence in the INS network
    startAnnouncer (new NameSpecifier[] {sourceNS}, new int[]{1}, 10000);

// <HERE> this location is set here for testing without Cricket purposes; should be removed
currentLocation = new AVelement ("[location=mit[building=ne43[floor=1[room=DEFAULT]]]]");

    interrupted = false;

    timer = new Thread (this, "Sink Proxy Refresh Timer #" + hashCode());
    timer.start();
  }
  
  /**
   * This method interrupts the timer thread and resets it.
   */
   
  public void interrupt()
  {
    if (DEBUG)
      System.out.println ("INSNetworkResourceManager::interrupt -> Refresh timer reset!");

    timer.interrupt();
  }
  
  /**
   * The timer will be running this method until the interrupt() method is called or until an
   * unexpected exception (something other than an InterruptedException) gets raised.
   */
   
  public void run()
  {
    while (!interrupted)
    {
      try
      {
        Thread.sleep (SLEEP_TIME);
        
        if (DEBUG)
          System.out.println ("INSNetworkResourceManager::run -> Refresh timer expired: refreshing sink proxies...");
        
        synchronized (discoveredProxies)
        {
          // we query the INS network for any sink proxies and refresh the list
          refreshSinkProxies();
        }
        
      }
      catch (Exception e)
      {
      }
    }
  }
  
  /*
   * This method gets called when the location information has changed. The current location
   * information will be updated.
   *
   * @param info the object that contains the location information (new and old)
   */

  public void updateLocation (LocationInformation info)
  {
    // the location of the controller is updated to reflect the new location
    currentLocation = new AVelement (info.getNewLocation());
  }

  /**
   * Verifies that the sink proxy described by the name specified is active, that it, that
   * it has a listening transport actually listening to commands. This precaution is taken because
   * sink proxies may announce their existence to the INS network and the fail, and their name specifier
   * will still be found for some time. We do not want the application to have some certainty that the
   * sink proxies it is given are actually running, so we "ping" each sink proxy to make sure they are
   * still running.
   *
   * @param proxy the SinkProxyDescriptor of the proxy whose liveness we want to test
   * @return true if the proxy responds to our "ping" message, false otherwise
   */

  public static boolean proxyIsActive (INSTCPSinkProxyDescriptor proxy)
  {
    try
    {
      // first we get the transport type from the sink proxy's name specifier; note that every sink proxy
      // must include a "transport_type" value that describes the transport it uses. Note that this name must
      // correspond to what is used for creating the SinkInfo subclass; i.e., if the proxy transport specified
      // is "TCP," there should be a "TCPSinkInfo" subclass that knows how to create an appropriate transport
      // for that type of sink.
      String transportType = proxy.getTransportType();
      
      // gets the class information for the SinkInfo subclass that knows how to create an appropriate transport
      // to communicate with the sink proxy: we will assume that the name of said subclass will be the package
      // name ("streamredirector") followed by the transport name an followed by "SinkInfo" (i.e., if the proxy
      // transport is "TCP," we will look for the "streamredirector.TCPSinkInfo" subclass
      Class sinkInfoClass = Class.forName (packageName + "." + transportType + "SinkInfo");

      // all SinkInfo subclasses must have a constructor that takes a SinkProxyDescriptor object that should be
      // used to instantiate the SinkInfo subclass (using the fields from said name specifier)
      Constructor sinkInfoConstructor = sinkInfoClass.getConstructor (new Class[] {SinkProxyDescriptor.class});
      
      // we instantiate the SinkInfo subclass passing it the SinkProxyDescriptor we were given
      SinkInfo sinkInfo = (SinkInfo) sinkInfoConstructor.newInstance (new Object[] {proxy});
   
      // we ask the SinkInfo subclass for an appropriate Transport instance that may be used to connect to the
      // sink proxy to verify that it is listening for commands; if there is an error creating this transport instance,
      // we will assume the transport is unreachable and return "false"
      Transport transport = sinkInfo.getTransport();
      
      // we call the "ping" method, which ascertains whether or not there is a listening transport on the other
      // side actually listening for commands; it will return "true" if there is one, in which case we will also
      // return "true": if there is a connection error at this point, we will return "false"
      return transport.ping();
    }
    catch (Exception e)
    {
//      e.printStackTrace();
      // any errors that may have occurred above would signal that there is something wrong either with the class
      // that was specified as transport type, creating the transport object or connecting to the counterpart
      // listening transport; in any of these cases, we will simply return that the proxy is not active
      return false;
    }
  }
  
  /**
   * This method queries INS for the list of all the sink proxies that have announced their existence on the
   * network. It is synchronized to make sure we are not updating the list of discovered proxies at the same time
   * we are using it.
   */

  protected void refreshSinkProxies()
  {
    synchronized (discoveredProxies)
    {
      // we query the INS network for any sink proxies
      discoveredProxies = discoverNames (sinkProxyNameSpecifier, true);
    }

    if (DEBUG)
    {
      System.out.println ("INSNetworkResourceManager::refreshSinkProxies -> Refreshed proxies list");

      for (int i = 0; i < discoveredProxies.length; i++)
      {
        System.out.println ("INSNetworkResourceManager::refreshSinkProxies -> Sink found: " + discoveredProxies [i]);
      }
    }
  }
  
  /**
   * This method asks the INS network for any sink proxies that have announced their existence to the
   * network. It can ask that the list for sink proxies discovered on the network be refreshed, or it can use the
   * last list cached. Only active sinks (those whose listening transports are actually listening) will be returned.
   *
   * @param refresh if set to REFRESH, we will refresh the list of sink proxies on the network; if it set to
   *              NO_REFRESH, we will use the last-cached list of sink proxies
   * @return a vector of INSTCPSinkProxyDescriptor which is the list of the names for the active sink proxies that
   *        were found, or an empty (size == 0) vector if none are found
   */
  
  public Vector getAllSinks (boolean refresh)
  {
    // if refresh is set to true, we will ask INS to refresh the list of sink proxies found on the network;
    // otherwise, we will use the latest list of discovered proxies without asking INS for it
    if (refresh)
    {
      // we query the INS network for any sink proxies and refresh the list
      refreshSinkProxies();
      
      // we interrupt the refresher thread so it resets its timer
      interrupt();
    }
    
    // we will create a vector in which we will add just the sinks that are active
    Vector activeProxiesList = null;
    
    // we must synchronize to make sure we are not updating the list of discovered proxies at the same time
    // we are using it
    synchronized (discoveredProxies)
    {
      activeProxiesList = new Vector (discoveredProxies.length);
      
      for (int i = 0; i < discoveredProxies.length; i++)
      {
        if (DEBUG)
          System.out.println ("INSNetworkResourceManager::getAllSinks -> Sink found: " + discoveredProxies [i]);
  
        INSTCPSinkProxyDescriptor proxy = new INSTCPSinkProxyDescriptor (discoveredProxies [i]);
        
        // if the proxy is active, we add it to the active proxies list
        if (proxyIsActive (proxy))
        {
          activeProxiesList.addElement (proxy);
        }
      }
    }
    
    if (DEBUG)
      System.out.println ("INSNetworkResourceManager::getAllSinks -> All active sinks found: " + activeProxiesList);

    return activeProxiesList;
  }
  
  /**
   * This method asks the INS network for sink proxies that exists in the current location
   * of the controller. It will return "null" if none of the sink proxies found is at the same location.
   *
   * @param refresh if set to REFRESH, we will refresh the list of sink proxies on the network; if it set to
   *              NO_REFRESH, we will use the last-cached list of sink proxies
   * @return a vector of SinkProxyDescriptor which are the names of the sink proxies that were found 
   *         in the same location as the controller, or an empty (size == 0) vector if none are found
   */

  public Vector getLocalSinks (boolean refresh)
  {
    Vector allProxies = getAllSinks (refresh);
    
    if (DEBUG)
    {
      System.out.println ("INSNetworkResourceManager::getLocalSinks -> List of all proxies before looking for locals is: " + allProxies);
      System.out.println ("INSNetworkResourceManager::getLocalSinks -> Current location: " + currentLocation);
    }

    for (Iterator list = allProxies.iterator(); list.hasNext();)
    {
      INSTCPSinkProxyDescriptor proxy = (INSTCPSinkProxyDescriptor) list.next();

      AVelement proxyLocation = proxy.getAVelement ("location");
      
      if (DEBUG)
        System.out.println ("INSNetworkResourceManager::getLocalSinks -> Proxy location: " + proxyLocation);
  
      // if this proxy is not in the same location as the controller, remove it from the list
      if (!currentLocation.equals (proxyLocation))
      {
        if (DEBUG)
          System.out.println ("INSNetworkResourceManager::getLocalSinks -> Removed proxy from list: " + proxy);

        list.remove();
      }
    }
    
    if (DEBUG)
      System.out.println ("INSNetworkResourceManager::getLocalSinks -> List of local proxies is: " + allProxies);

    return allProxies;
  }
  
  /**
   * Gets all the sinks local to the controller (i.e. those that are in the same location) and
   * arbitrarily returns the first one. This should be used when, given a choice of several local sinks
   * they are all equally good. If no local sinks are found, "null" will be returned.
   *
   * @param refresh if set to REFRESH, we will refresh the list of sink proxies on the network; if it set to
   *              NO_REFRESH, we will use the last-cached list of sink proxies
   * @return the SinkProxyDescriptor of an active sink that is local to the controller
   */

  public SinkProxyDescriptor getAnyLocalSink (boolean refresh)
  {
    try
    {
      return (SinkProxyDescriptor) getLocalSinks (refresh).firstElement();
    }
    catch (Exception e)
    {
      return null;
    }
  }
}