package streamredirector;

import java.util.*;

import ins.namespace.*;

/**
 * This class is initialized with a NameSpecifier that describes a sink proxy and has
 * methods to get at the different attributes contained therein. It encapsulates the
 * knowledge of the structure of the name specifier belonging to a sink proxy.
 *
 * @see ins.namespace.NameSpecifier
 * @author Jorge Rafael Nogueras
 */

public class INSTCPSinkProxyDescriptor implements TCPSinkProxyDescriptor
{
  /** the name specifier that describes the sink proxy */
  protected NameSpecifier nameSpecifier;
  
  /**
   * Initializes the object with the name specifier of the sink proxy.
   *
   * @param the_nameSpecifier the name specifier that describes the sink proxy
   */

  public INSTCPSinkProxyDescriptor (String the_nameSpecifier)
  {
    this (new NameSpecifier (the_nameSpecifier));
  }

  /**
   * Initializes the object with the name specifier of the sink proxy.
   *
   * @param the_nameSpecifier the name specifier that describes the sink proxy
   */

  public INSTCPSinkProxyDescriptor (NameSpecifier the_nameSpecifier)
  {
    nameSpecifier = the_nameSpecifier;
  }

  /**
   * Get the value of the specified attribute in the name specifier as a string.
   *
   * @param attribute the attribute whose value we want
   * @return the value of the attribute as a string
   * @exception Exception if an illegal attribute is given
   */

  public String getValue (String attribute) throws Exception
  {
    return getAVelement (attribute).getValue().toString();
  }
  
  /**
   * Gets the name specifier of the sink proxy.
   *
   * @return the name specifier of the sink proxy
   */

  public NameSpecifier getNameSpecifier()
  {
    return nameSpecifier;
  }

  public AVelement getAVelement (String attribute)
  {
    return nameSpecifier.getAVelement (new Attribute (attribute));
  }
  
  /**
   * Gets the type of transport used by the sink proxy.
   *
   * @return the type of transport used by the sink proxy or "null" if the attribute does not exist
   */

  public String getTransportType()
  {
    try
    {
      return getValue ("transport_type");
    }
    catch (Exception e)
    {
      e.printStackTrace();
      return null;
    }
  }

  /**
   * Gets the hostname of the sink proxy.
   *
   * @return the hostname of the sink proxy or "null" if the attribute does not exist
   */

  public String getHostName()
  {
    try
    {
      return getValue ("hostname");
    }
    catch (Exception e)
    {
      return null;
    }
  }

  /**
   * Gets the port where the sink proxy is listening.
   *
   * @return the port where the sink proxy is listening  or "null" if the attribute does not exist
   */

  public int getPort()
  {
    try
    {
      return Integer.parseInt (getValue ("port"));
    }
    catch (Exception e)
    {
      return -1;
    }
  }
  
  /**
   * Gets the location of the sink proxy.
   *
   * @return the location of the sink proxy or "null" if the attribute does not exist
   */

  public String getLocation()
  {
    AVelement av = nameSpecifier.getAVelement (new Attribute ("location"));
    
    if (av == null)
    {
      return null;
    }
    
    String value = av.getValue().toString();
    
    StringBuffer location = new StringBuffer (value);

    Enumeration e = av.getAVelements();
    
    while (e.hasMoreElements())
    {
      String s = ((AVelement) e.nextElement()).toString();
      
      location.append (s);
    }
    
//System.out.println ("Full location: " + location);

    return location.toString();
  }
  
  /**
   * Returns an array of string containing all the MIME types that the sink proxy supports.
   *
   * @return an array of string containing all the MIME types that the sink proxy supports
   * @exception Exception if there are missing attributes or values in the name specifier
   */

  public String[] getAllMIMETypes() throws Exception
  {
    AVelement allMIMETypes = getAVelement ("MIME_types");
    int n = Integer.parseInt (allMIMETypes.getValue().toString());

    String[] mimeTypes = new String [n];

    for (int i = 0; i < n; i++)
    {
      mimeTypes [i] = allMIMETypes.getAVelement (new Attribute (Integer.toString (i))).getValue().toString();
    }
    
    return mimeTypes;
  }
  
  /**
   * Looks in the list of MIME types this sink proxy supports and returns true if it can find the
   * specified MIME type or false if it can not.
   *
   * @param mimeType the MIME type we are looking for
   * @return true if it can find the specified MIME type or false if it can not
   * @exception Exception if there are missing attributes or values in the name specifier
   */

  public boolean findMIMEType (String mimeType) throws Exception
  {
    String[] allMIMETypes = getAllMIMETypes();
    
    for (int i = 0; i < allMIMETypes.length; i++)
    {
      if (mimeType.equals (allMIMETypes [i]))
        return true;
    }
    
    return false;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": name specifier = " + nameSpecifier + "]";
  }
}