package streamredirector;

import java.io.*;
import java.net.*;
import java.util.*;

import streamredirector.player.*;

/**
 * This is a concrete implementation of a server connection: it embodies knowledge of the connection
 * between a media player and the HTTP (web) server from where the resource is being downloaded.
 *
 * The way this works is as follows: the media player running on this sink is instructed to download the
 * resource using the HTTP protocol, but it must first be configured to go through a web proxy.
 * The web proxy it goes through is actually our very own version of a web proxy, which listens on a
 * well-known port and performs all the normal functions of a web proxy. This web proxy must have
 * already been started by the executable that instantiated the sink proxy. The web proxy's job is first,
 * to keep state indicating for a specific connection what was the last offset of data sent to the media
 * player, and second, to request the resource from said offset if there is already state in the
 * web proxy saying where it left off.
 *
 * When this server connection receives a command, it augments the URL of the requested resource by
 * prepending the connection's cookie before the actual resource path. This augmented path is what the
 * media player is told to download through our web proxy. The augmented URL looks like this:
 *
 *   http://<HTTP-server-host>:<port>/<cookie>/<actual-resource-path>
 *
 * The web proxy will receive the HTTP command, strip away the cookie's value and use it for storing state
 * pertaining to the connection (such as the last offset transferred) and contacts the actual web * server with the actual resource URL, forwarding the data to the media player as it receives it.
 *
 * If the command it receives is an HTTPResumeCommand, it means that the connection had started before
 * (most probably on another sink), was interrupted and migrated to this sink. This means that the
 * connection must actually be resumed at the point indicated by the offset in the HTTPResumeCommand.
 * However, since this is a new connection, the web proxy will not have any state associated with the
 * transfer, and will therefore not know to resume the transfer where it left off. For this reason, we
 * will create state in the web proxy so that when the media player requests the resource using the
 * connection's cookie, it will know from which offset to request the resource.
 *
 * @see streamredirector.WebProxyManager, streamredirector.WebProxyConnectionState
 * @author Jorge Rafael Nogueras
 */

public class NativePlayerServerConnection implements ServerConnection
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** the host address (and port number, if necessary) of the HTTP server */
  protected URL serverAddress;
  
  /** this output stream will be passed to us but we will do nothing with it */
  protected transient OutputStream os;

  /** the HTTP request that was sent to the server */
  protected transient HTTPCommand request;
  
  /** a long number that uniquely identifies this connection */
  protected transient Long cookie;
  
  protected transient Player player;
  
  /**
   * The constructor only takes a URL representing the address of the HTTP server.
   *
   * @param address the host address (and port number, if necessary) of the HTTP server
   */
   
  public NativePlayerServerConnection (URL address)
  {
    serverAddress = address;
  }
  
  /**
   * This method is called by the proxy when a controller has requested it that a connection be
   * established; it opens up a socket connection to the server, gets the appropriate input and
   * output streams but holds off on starting the thread until the sink's output stream has been
   * initialized.
   *
   * @param the_cookie a long number that uniquely identifies this connection
   */
   
  public void establishServerConnection (Long the_cookie)
  {
    cookie = the_cookie;
  }
  
  /**
   * Sets the output stream: however, we will do nothing with it, except to keep a reference to
   * it so we can close it when the server connection is closed.
   *
   * @param out the output stream we will be storing
   */
   
  public void setOutputStream (OutputStream out)
  {
    os = out;
  }
  
  /**
   * This method is called by the proxy when a controller has requested it to close the connection; it
   * interrupts the web proxy's transfer if it is still going on and closes the output stream.
   *
   * @exception Exception if there is an error closing the stream
   */

  public void closeServerConnection() throws Exception
  {
    if (os != null)
    {
      player.stop();
      os.close();
    }
  }
  
  /**
   * This method is called when a controller has requested that a command be sent to the HTTP
   * server. It directs a Player to get the resource from the HTTP server, augmenting the pathname
   * with the cookie. This means that the full URL will be of the form:
   *
   *   http://<HTTP-server-host>:<port>/<cookie>/<actual-resource-path>
   *
   * If the command is a resume command, this means that we should make sure that the transfer
   * starts at the offset specified in the HTTPResumeCommand object. To accomplish this, we create
   * an entry in the web proxy's connection table associating the connection's cookie with the offset
   * where the transfer must start at.
   *
   * @param command the command to be sent to the HTTP server
   * @exception Exception if there are errors manipulating URLs or with the player
   * @return a return code indicating success or an error condition
   */
   
  public ReturnCode sendCommand (ServerCommand command) throws Exception
  {
    // we must cast it to an HTTPCommand to get its request parser
    request = (HTTPCommand) command;

    if (DEBUG)
      System.out.println ("NativePlayerServerConnection::sendCommand -> Received command: " + request);

    // this is the URL we will tell the player to access: it is of the form:
    // http://<HTTP-server>:<port>/<cookie>/<actual-resource-path>
    URL serverURL = new URL ("http", serverAddress.getHost(), serverAddress.getPort(), request.getRequestParser().getResource());

    if (DEBUG)
      System.out.println ("NativePlayerServerConnection::sendCommand -> Server address is: " + serverAddress);

    long offset = 0L;
    
    // if the command received is meant to resume a transfer, we must create state in the
    // web proxy so that when the media player contacts the proxy with the full path (which
    // includes the cookie of this connection plus the path to the actual resource) there is
    // offset information and the web proxy can request the resource from the given offset
    if (command instanceof HTTPResumeCommand)
    {
      // this is the offset we will store in the connection state is the one sent in the resume command:
      // note that we must cast the command to a HTTPResumeCommand so we can access the offset
      offset = ((HTTPResumeCommand) request).getLastOffset();
      
      if (DEBUG)
        System.out.println ("NativePlayerServerConnection::sendCommand -> Received resume command from offset: " + offset);
    }
    
    String mimeType = request.getMimeType();
    
//System.out.println ("NativePlayerServerConnection::sendCommand -> MIME type of resource: " + mimeType);
    if (mimeType != null)
    {
      player = PlayerManager.getPlayer (request.getMimeType());
    }
    else
    {
      player = PlayerManager.getPlayer ("audio/mpeg");
    }
//    player = new NativePlayer();
    
    if (DEBUG)
    {
      System.out.println ("NativePlayerServerConnection::sendCommand -> Starting player #" + player + " and playing file: " + request.getRequestParser().getResource());
      System.out.println ("NativePlayerServerConnection::sendCommand -> URL is: " + serverURL);
    }
    
    player.play (serverURL.toString(), offset);

    return ReturnCode.COMMAND_OK;
  }

  /**
   * This method constructs a command that enables the controller to resume the stream connection
   * from the current byte offset. It is usually called after a connection has been ended if the
   * controller later wants to resume it at the point it was interrupted (this will be done if the
   * stream should be redirected to a different sink). The HTTP command returned is structured so
   * that by inserting the name of the filename that should be resumed it can be issued to the
   * server and it can begin sending the data from the appropriate byte offset.
   *
   * @exception Exception if there is an error creating the resume command 
   * @return a ServerCommand that resumes the connection or "null" if the resource has been completely read
   */

  public ServerCommand getResumeCommand() throws Exception
  {
    // if no request was ever sent, we simply return a "null" to indicate that the command
    // can not be resumed
    if (request == null)
    {
      return null;
    }
    else
    {
      long offset = player.getLastByteOffset();
      
      if (DEBUG)
        System.out.println ("NativePlayerServerConnection::getResumeCommand -> offset is: " + offset);

      return new HTTPResumeCommand (request.getRequestParser().getResource(), offset, request.getMimeType());
    }
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": URL = " + serverAddress + "]";
  }
}