package streamredirector;

import java.net.*;

/**
 * This concrete implementation of ServerInfo has the knowledge of how to create
 * an JavaPlayerServerConnection. Thus, when a ServerInfo object must be created in order
 * to connect to an HTTP server through a web proxy, the concrete object that must be used
 * is this one.
 *
 * @see streamredirector.JavaPlayerServerConnection
 * @author Jorge Rafael Nogueras
 */

public class NativePlayerServerInfo extends ServerInfo
{
  /**
   * Initializes the object with the information necessary to connect to the
   * specified web server and download the specified resource.
   *
   * @param the_hostName the hostname of the web server
   * @param the_port prefix the port where the web server is running
   * @param the_path the full path of the filename on the server
   * @param the_selectedResource the resource that was selected for download
   */

  public NativePlayerServerInfo (String the_hostName, int the_port, String the_path, ResourceEntry the_selectedResource)
  {
    super (the_hostName, the_port, the_path, the_selectedResource);
  }

  /**
   * Initializes the object with the corresponding server's information
   *
   * @param url describes the server information as read from the database
   * @param resourceEntry describes the resource information as read from the database
   */

  public NativePlayerServerInfo (URL url, ResourceEntry resourceEntry)
  {
    super (url, resourceEntry);
  }
  
  /**
   * This method will create an JavaPlayerServerConnection object that embodies the connection
   * to the web server.
   *
   * @return the JavaPlayerServerConnection object that embodies the connection to the web server
   * @exception Exception if there is an error creating the server connection object
   */

  public ServerConnection getServerConnection() throws Exception
  {
    return new NativePlayerServerConnection (new URL ("http://" + hostName + ":" + port + "/"));
  }

  /**
   * This method will configure the HTTP resume command used to re-establish a connection when
   * the connection must migrate to another sink. In our case, nothing needs to be done because
   * the HTTP command already has all the necessary information to be resumed.
   *
   * @return a configured HTTPResumeCommand object that can be used to re-establish a connection
   */

  public ServerCommand configureResumeCommand (ServerCommand resumeCommand)
  {
    return resumeCommand;
  }
}
