package streamredirector;

import java.io.*;

/**
 * This is class is a null sink connection: it should be used when the specific implementation
 * does not require the use of a sink connection object. The NullSinkConnection simply fulfills the
 * contract of all sink connections and stores the variables that are passed to it in case they
 * are needed, but does nothing else aside from that.
 *
 * @see streamredirector.SinkConnection
 * @author Jorge Rafael Nogueras
 */

public class NullSinkConnection implements SinkConnection
{
  /** where data coming from the server are read */
  protected transient InputStream is;
  
  /** a long number that uniquely identifies this connection */
  protected transient Long cookie;
  
  /**
   * Initializes the object.
   */
   
  public NullSinkConnection() { }
  
  /**
   * This method is called by the sink proxy when a controller has requested it that a connection
   * be established.
   *
   * @param the_cookie a long number that uniquely identifies this connection
   * @exception Exception if there is an error establishing the connection (should never be thrown)
   */
   
  public void establishSinkConnection (Long the_cookie) throws Exception
  {
    cookie = the_cookie;
  }
  
  /**
   * Initializes the input stream where data coming from the server should be read.
   *
   * @param in where data coming from the server are read
   * @exception Exception if there is an error with the input stream (should never be thrown)
   */
   
  public void setInputStream (InputStream in) throws Exception
  {
    is = in;
  }

  /**
   * This method is called by the proxy when a controller has requested it to close the connection.
   *
   * @exception Exception if there is an error closing the connection
   */

  public void closeSinkConnection() throws Exception
  {
    if (is != null)
    {
      is.close();
    }
  }
  
  /**
   * This method is called when a controller has requested that a command be sent to the sink.
   * Since no commands may be sent to a NullSinkProxy, its implementation does nothing.
   *
   * @param command the command to be sent to the sink
   * @return a return code indicating success
   */
   
  public ReturnCode sendCommand (SinkCommand command)
  {
    return ReturnCode.COMMAND_OK;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": cookie = " + cookie + "]";
  }
}