package streamredirector;

import java.util.*;
import java.lang.reflect.*;

/**
 * This class describes a resource entry read from the
 * database. "Resource" is a general term that refers to any file that
 * may be gotten from the server. A resource is mainly described by
 * its MIME type (which says the type of media file it is), and it may
 * also have additional (optional) data like a description of the
 * resource, or the size of the resource in bytes.  The act of filling
 * out these values will use reflection so that it is easy to add more
 * optional values to describe a resource: to do this, a public String
 * variable must be added to this class, and the corresponding
 * information must be placed in the resource description in the
 * server database.  Note that only the MIME type is required to
 * describe the resource entry: the rest of the information must be
 * added to the resource entry in the order in which the variables are
 * declared. For instance, a line in the server database is as
 * follows:
 *
 * <resource-URL> = <MIME-type> [| <description>] [| <size>] [| ...]
 *
 * In the basic implementation there are two other information items
 * associated with each resource entry, a description and the resource
 * size. Note that first comes the description, then the separator
 * character ("|") and then the size information; this matches the
 * order in which the public String variables "description" and "size"
 * are declared in this class.  If another information item (like say,
 * bitrate) is added, a public String variable must be added to the
 * class (after the last variable), and each line in the server
 * database must have the new information item.
 *
 * @see streamredirector.ServerDatabaseManager
 * @author Jorge Rafael Nogueras */

public class ResourceEntry
{
  /** the character that separates the field information read from the database */
  protected static final String separator = "|";

  /** array of the fields in the class; can be static since it does
      not change and is common to all instances */
  protected static final Field [] fields = ResourceEntry.class.getFields();
  
  /** array of the fields in the class; can be static since it does
      not change and is common to all instances */
  protected static Vector fieldNamesVector = new Vector (fields.length);
  
  /** the MIME type corresponding to the resource */
  public String MIME_Type;

  /** the Description of the resource (optional) */
  public String Description;

  /** the Size of the resource (optional) */
  public String Size;

  static
  {
    // add the field names to the field names vector
    for (int i = 0; i < fields.length; i++)
    {
      fieldNamesVector.addElement (fields [i].getName());
    }
  }
  
  /**
   * Initializes the entry that describes a server resource with the
   * corresponding information.
   *
   * @param unparsedInformation contains the information for the resource entry */

  public ResourceEntry (String unparsedInformation)
  {
    parseInformation (unparsedInformation);
  }

  /**
   * Initializes the entry that describes a server resource with the corresponding information.
   *
   * @param the_MIME_Type the MIME type corresponding to the resource
   * @param the_Description the Description of the resource
   * @param the_Size the Size of the resource
   */

  public ResourceEntry (String the_MIME_Type, String the_Description, String the_Size)
  {
    MIME_Type = the_MIME_Type;
    Description = the_Description;
    Size = the_Size;
  }
  
  /**
   * Parses the information used to initialize this resource entry. It
   * tokenizes the unparsed line read from the server database, where
   * each token is marked by the separator character.  It sets the
   * values read from each token to the public variables found in this
   * class: these must correspond, in order, to the information read
   * from the server database.
   *
   * @param unparsedInformation the line read from the server database
   * containing the unparsed information */

  protected void parseInformation (String unparsedInformation)
  {
    StringTokenizer tok = new StringTokenizer (unparsedInformation, separator);
    
    // we will iterate through all the public fields in this class
    for (int i = 0; i < fields.length; i++)
    {
      // any errors or exceptions caught here will be silently
      // discarded and we will proceed to the next field for
      // examination
      try
      {
        // we assign the next public field the value from the token
        // read, trimming whitespace from the beginning and end of
        // each item
        fields [i].set (this, tok.nextToken().trim());
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
      catch (Error e1)
      {
        e1.printStackTrace();
      }
    }
  }
  
  /**
   * Encodes the public fields from this class into a string that
   * follows the format used in the server database file (it basically
   * adds the separator character in between the values of the
   * variables and returns it as one string).
   *
   * @return a string encoding the values from this resource entry */

  public String encodeResourceEntry()
  {
    StringBuffer buf = null;
    
    try
    {
      // we start out the unparsed string with the first field: we can
      // be sure we have at lease one field because the MIME type of
      // the resource is a required field and thus, it must always
      // exist in the resource entry
      buf = new StringBuffer (fields [0].get (this).toString());
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
    
    // we will iterate through all the public fields in this class
    for (int i = 1; i < fields.length; i++)
    {
      // any errors or exceptions caught here will be silently
      // discarded and we will proceed to the next field for
      // examination
      try
      {
        // before each new field we insert the separator character
        buf.append (separator);
        
        // and then we append the value of the current field being processed
        buf.append (fields [i].get (this).toString());
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
      catch (Error e1)
      {
        e1.printStackTrace();
      }
    }
    
    return buf.toString();
  }
  
  /**
   * Returns the names of all the public fields in this class in a Vector.
   *
   * @return the names of all the public fields in this class in a Vector
   */

  public static Vector getFieldNames()
  {
    return fieldNamesVector;
  }

  /**
   * Returns the values of all the public fields in this class in a Vector.
   *
   * @return the values of all the public fields in this class in a Vector
   */

    public Vector getValues() {
	Vector v = new Vector();
	for (int i = 0; i < fields.length; i++) {
	    try {
		v.addElement(fields[i].get(this));
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
	return v;
    }

    public Object getValue(int i) {
	Object o = null;
	try {
	    o = fields[i].get(this);
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return o;
    }
    
  /**
   * Gets the MIME type of the resource
   *
   * @return the MIME type corresponding to the resource
   */

  public String getMimeType()
  {
    return MIME_Type;
  }
  
  /**
   * Gets the Description of the resource
   *
   * @return the Description of the resource
   */

  public String getDescription()
  {
    return Description;
  }

  /**
   * Gets the Size of the resource
   *
   * @return the Size of the resource
   */

  public String getSize()
  {
    return Size;
  }

  /**
   * Prints a useful Description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    StringBuffer desc =  new StringBuffer ("[" + this.getClass().getName() + ": ");

    try
    {
      desc.append (fields [0].getName());
      desc.append (" = ");
      desc.append (fields [0].get (this));
      
      // prints the values for the variables there are in the class
      for (int i = 1; i < fields.length; i++)
      {
        desc.append (", ");
        desc.append (fields [i].getName());
        desc.append (" = ");
        desc.append (fields [i].get (this));
      }
  
      desc.append ("]");
      return desc.toString();
    }
    catch (Exception e)
    {
      return null;
    }
  }
}
