package streamredirector;

import java.io.*;

/**
 * This is a repository of useful constants used as return codes. It can
 * also be used to return object values by subclassing it and including references
 * to said object values (which must themselves be serializable, since this class is
 * sent across the network).
 *
 * @author Jorge Rafael Nogueras
 */

public class ReturnCode implements Serializable
{
  /** internal value used to denote an uninitalized return code object */
  protected final static int COMMAND_UNINITIALIZED_VALUE = -1;

  /** internal value used to denote an error return code object  */
  protected final static int COMMAND_ERROR_VALUE = 0;
  
  /** internal value used to denote an OK return code object  */
  protected final static int COMMAND_OK_VALUE = 1;

  /** internal value used to denote a "could not be resumed" return code object  */
  protected final static int COMMAND_COULD_NOT_BE_RESUMED_VALUE = 2;

  /** used to denote an uninitalized return code object */
  public final static ReturnCode COMMAND_UNINITIALIZED = new ReturnCode (COMMAND_UNINITIALIZED_VALUE);

  /** returned to denote an error condition */
	public final static ReturnCode COMMAND_ERROR = new ReturnCode (COMMAND_ERROR_VALUE);
	
	/** returned when there are no errors to report (or exceptions to throw) */
	public final static ReturnCode COMMAND_OK = new ReturnCode (COMMAND_OK_VALUE);

  /** returned when a connection cannot be resumed because the original transfer was completed */
  public final static ReturnCode COMMAND_COULD_NOT_BE_RESUMED = new ReturnCode (COMMAND_COULD_NOT_BE_RESUMED_VALUE);
  
  /** the return code stored in this object */
  protected int returnCode;

  /**
   * Initializes the object with the default value of "uninitialized".
   *
   * @param the_returnCode the return code stored in this object
   */

  protected ReturnCode()
  {
    // assume the value of COMMAND_UNINITIALIZED by default
    this (COMMAND_UNINITIALIZED_VALUE);
  }
  
  /**
   * Initializes the object with the appropriate return code.
   *
   * @param the_returnCode the return code stored in this object
   */

  protected ReturnCode (int the_returnCode)
  {
    returnCode = the_returnCode;
  }
  
  /**
   * Sets this return code object's value equal to the specified return code's value.
   *
   * @param otherReturnCode the return code whose value will be copied into this one
   */

  public void setReturnCode (ReturnCode otherReturnCode)
  {
    returnCode = otherReturnCode.getReturnCode();
  }

  /**
   * Gets the return code.
   *
   * @return the return code stored in this object
   */

  protected int getReturnCode()
  {
    return returnCode;
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    StringBuffer buf = new StringBuffer ("[");
    buf.append (this.getClass().getName());
    buf.append (": return code = ");
    
    switch (returnCode)
    {
      case COMMAND_UNINITIALIZED_VALUE: buf.append ("<uninitialized value>"); break;
      case COMMAND_ERROR_VALUE: buf.append ("<Error>"); break;
      case COMMAND_OK_VALUE: buf.append ("<OK>"); break;
      case COMMAND_COULD_NOT_BE_RESUMED_VALUE: buf.append ("<Could not be resumed>"); break;
      
      default: buf.append ("<uninitialized value>");
    }
    
    buf.append ("]");

    return buf.toString();
  }
}