package streamredirector;

import java.io.*;

/**
 * This interface should be implemented by any object that will embody the connection from
 * the proxy to the server. Its methods allow the proxy to connect to the server, tell it where
 * it should write out its data, close the connection, send it a command or query for a command
 * that would allow the connection to be resumed from its current point.
 *
 * @author Jorge Rafael Nogueras
 */

public interface ServerConnection extends Serializable
{
	/**
	 * This method is called by the proxy when a controller has requested it that a connection be
	 * established.
	 *
   * @param the_cookie a long number that uniquely identifies this connection
	 * @exception Exception if there is an error creating a connection to the server
	 */

	public void establishServerConnection (Long the_cookie) throws Exception;
	
	/**
	 * Initializes the output stream where data read from the server should be written.
	 *
	 * @param out where data read from the server are written so the sink can read them
	 * @exception Exception if there is an error setting the new output stream
	 */
	 
	public void setOutputStream (OutputStream out) throws Exception;
	
	/**
	 * This method is called by the proxy when a controller has requested it to close the connection.
	 *
	 * @exception Exception if there is an error closing the connection
	 */

	public void closeServerConnection() throws Exception;
	
	/**
	 * This method is called when a controller has requested that a command be sent to the
	 * server. It gets the command from the ServerCommand object and writes it out to the server,
	 * If the resource named in the command is somehow invalid, an appropriate exception is thrown.
	 *
	 * @param command the command to be sent to the server
	 * @exception InvalidResourceException if the resource does not exist or cannot be opened in the server
   * @return a return code indicating success or an error condition
	 */
	 
	public ReturnCode sendCommand (ServerCommand command) throws Exception;

	/**
	 * This method constructs a command that enables the controller to resume the stream connection
	 * from where it left off. It is usually called after a connection has been ended if the
	 * controller later wants to resume it at the point it was interrupted (this will be done if the
	 * stream should be redirected to a different sink).
	 *
	 * @exception Exception if there is an error creating the resume command
   * @return a ServerCommand that resumes the connection from the point where it was ended, or "null" if the resource has been completely read
	 */

	public ServerCommand getResumeCommand() throws Exception;
}