package streamredirector;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * This class accesses the server database existing in the controller.
 *
 * All its methods are static, and before using any of them, the initialize()
 * method must be called. If a method is called before the class is initialized,
 * an UninitializedManagerException will be thrown.
 *
 * The "server database" is a plain-text properties file (that is, it observes
 * Java's properties file syntax) that contains information about the resources
 * users may ask to download. In other words, it is a list of "favorites." Its
 * format is explained in the header of the file itself, but a brief description follows:
 *
 * Every line is of the format: (items between square brackets are optional)
 *
 * <resource-URL> = <MIME-type> [| <description>] [| <size>] [| ...]
 *
 * where <resource-URL> follows the format of any URL:
 *
 * <protocol>://<host>[:<port>]/[path-to-resource]
 *
 * Note that optional information must be separated by "|" symbols, which may not occur
 * in the information text itself; furthermore, the format is flexible in that other information items
 * may be included by simply including an appropriate String variable in the ResourceEntry class
 * (more information may be found there).
 *
 * This class reads the specified database file and parses it, construction a Properties
 * object that contains the information read from the file. There is also a hashtable that
 * contains a list of the resource URLs and the ResourceEntry objects that contain each
 * resource's information.
 *
 * @see streamredirector.ResourceEntry
 * @author Jorge Rafael Nogueras
 */

public class ServerDatabaseManager
{
  /** the default filename for the server database */
  protected static final String defaultFileName = "serverDB.dat";
  
  /** the header that will get written to the database file when it is saved */
  protected static final String header =
    " The server database is a plain-text properties file (that is, it observes\n" +
    "# Java's properties file syntax) that contains information about the resources\n" +
    "# users may ask to download. Its format is as follows:\n" +
    "#\n" +
    "# Every line is of the format: (items between square brackets are optional)\n" +
    "#\n" +
    "# <resource-URL> = <MIME-type> [| <description>] [| <size>] [| ...]\n" +
    "#\n" +
    "# where <resource-URL> follows the format of any URL:\n" +
    "#\n" +
    "# <protocol>://<host>[:<port>]/[path-to-resource]\n" +
    "#\n" +
    "# An example of a line would be:\n" +
    "#\n" +
    "# http://www.myhost.com/~mp3/mysong.mp3 = audio/mpeg | A song I recorded | 3.4 MB\n" +
    "#\n" +
    "# Note that optional information must be separated by \"|\" symbols, which may not occur\n" +
    "# in the information text itself; furthermore, the format is flexible in that other information items\n" +
    "# may be included by simply including an appropriate String variable in the ResourceEntry class\n" +
    "# (more information may be found there).\n";

  /** this flag is true after the class has been properly initialized */
  protected static boolean initialized = false;

  /** the name of the text file that has the server database information */
  protected static String fileName;
  
  /** the object that maintains the information read from the database file */
  protected static Properties serverInformation;
  
  /** the table that indexes the ResourceEntry objects created (keyed by the URL of the
      resource and whose value is the ResourceEntry object that describes that entry) */
  protected static HashMap resourceEntriesList;
  
  /**
   * Initializes the ServerDatabaseManager so it can be used statically from then on,
   * using the default server database filename.
   * This initialization method must be called before any other static method is used.
   *
   * @exception Exception if there are problems reading the filename or creating the server list
   */

  public static void initialize() throws Exception
  {
    initialize (defaultFileName);
  }
  
  /**
   * Initializes the ServerDatabaseManager so it can be used statically from then on.
   * This initialization method must be called before any other static method is used.
   *
   * @param file_name the name of the text file that has the server database information
   * @exception Exception if there are problems reading the filename or creating the server list
   */

  public static void initialize (String file_name) throws Exception
  {
    if (!initialized)
    {
      initialized = true;
    }
    
    fileName = file_name;
    resourceEntriesList = new HashMap();
  
    // load the database information into the Properties object
    serverInformation = new ResourceProperties();
//    serverInformation.load (ClassLoader.getSystemResourceAsStream (file_name));

    // loads the server database from the current directory
    serverInformation.load (new FileInputStream (file_name));

System.out.println ("Read properties: " + serverInformation);

    // fill the table with the resource entry information read from the database
    createResourceEntriesList();
  }
  
  /**
   * Adds the specified resource entry to the table, keyed by the URL to the resource.
   *
   * @param theURL the URL identifying the resource
   * @param entry the resource entry object being added
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static void addEntry (URL theURL, ResourceEntry entry) throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call addEntry() without initializing the ServerDatabaseManager");
    }

    resourceEntriesList.put (theURL, entry);
  }
  
  /**
   * Removes the specified resource entry from the table.
   *
   * @param theURL the URL that uniquely identifies the resource entry object being removed
   * @return the same resource entry object that was just removed, or null if it did not exist
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static ResourceEntry removeEntry (URL theURL) throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call removeEntry() without initializing the ServerDatabaseManager");
    }

    return (ResourceEntry) resourceEntriesList.remove (theURL);
  }
  
  /**
   * Retrieves the specified resource entry from the table.
   *
   * @param theURL the URL that uniquely identifies the resource entry object being retrieved
   * @return the resource entry object that describes the resource specified by the URL
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static ResourceEntry getEntry (URL theURL) throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getEntry() without initializing the ServerDatabaseManager");
    }

    return (ResourceEntry) resourceEntriesList.get (theURL);
  }
  
  /**
   * Replaces a resource entry in the table with the specified entry.
   *
   * @param theURL the URL that uniquely identifies the resource entry object being replaced
   * @param entry the resource entry object that will replace the old one
   * @return the resource entry object that was replaced by the specified entry
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static ResourceEntry replaceEntry (URL theURL, ResourceEntry entry) throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call replaceEntry() without initializing the ServerDatabaseManager");
    }

    ResourceEntry oldEntry = removeEntry (theURL);
    addEntry (theURL, entry);
    
    return oldEntry;
  }
  
  /**
   * Returns an object that can be used to iterate through the URLs that identify the
   * resources read from the database.
   *
   * @return the iterator object used to go through the URLs read from the database
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static Iterator getURLs() throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getURLs() without initializing the ServerDatabaseManager");
    }

    return resourceEntriesList.keySet().iterator();
  }
  
  /**
   * Returns an object that can be used to iterate through the resource entry objects that
   * describe the resources read from the database.
   *
   * @return the iterator object used to go through the resource entry objects
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */

  public static Iterator getResourceEntries() throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getResourceEntries() without initializing the ServerDatabaseManager");
    }

    return resourceEntriesList.values().iterator();
  }
  
  /**
  * <HERE>
   * This method is for debugging purposes: it prints the database's contents.
   */
   
  public static void printDatabaseInfo() throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call printDatabaseInfo() without initializing the ServerDatabaseManager");
    }

    serverInformation.list (System.out);
  }
  
  /**
   * Reads in the database file and creates ResourceEntry objects that contain the information specified
   * therein. It puts these ResourceEntry objects in the resource list table, keyed by the URL that
   * identifies the resource.
   *
   * @exception UninitializedManagerException if the ServerDatabaseManager has not been initialized
   */
   
  protected static void createResourceEntriesList() throws UninitializedManagerException
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call createResourceEntriesList() without initializing the ServerDatabaseManager");
    }

    Enumeration keys = serverInformation.keys();
    Enumeration values = serverInformation.elements();
    
    while (keys.hasMoreElements())
    {
      try
      {
        // go through the textual data read from the server database file; add entries to the
        // table containing all the resource entry objects (using the resource's URL as the
        // key and the resource entry as the value)
        addEntry (new URL ((String) keys.nextElement()),
                  new ResourceEntry ((String) values.nextElement()));
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
  }
  
  /**
   * Writes the information contained in the list of resource entries into the server database,
   * overwriting what was there before.
   *
   * @exception Exception if there is an error writing the database file
   */

 public static void writeServerDatabase() throws Exception
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call writeServerDatabase() without initializing the ServerDatabaseManager");
    }

    Iterator URLs = getURLs();
    Iterator entries = getResourceEntries();

    // we must first empty out the current property object
    serverInformation.clear();

    // we iterate through all the items read from the database
    while (URLs.hasNext())
    {
      try
      {
        // we add the information as a line in the server database, encoding the URL and the
        // resource entry as a string
        serverInformation.setProperty (((URL) URLs.next()).toString(),
                                       ((ResourceEntry) entries.next()).encodeResourceEntry());
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }

    // finally, we write the new server information object to the file it was read from
    serverInformation.store (new FileOutputStream (fileName), header);
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": server information = " + serverInformation +
           ", database filename = " + fileName + ", list of servers: " + resourceEntriesList + "]";
  }
}