package streamredirector;

import java.net.URL;
import java.net.URLEncoder;

/**
 * This abstract class embodies the information necessary to be able to connect to a
 * server. It should be able to return an appropriate ServerConnection object that is
 * appropriate to the concrete sub-class. This is the object that is passed to the
 * Controller so it can connect to the server.
 *
 * @see streamredirector.Controller, streamredirector.ServerConnection
 * @author Jorge Rafael Nogueras
 */

public abstract class ServerInfo
{

    /** the URL of the selected server resource **/
    protected URL serverURL;

    /** the hostname of the server */
    protected String hostName;
  
    /** the port where the server is running */
    protected int port;

    /** the full path of the filename on the server **/
    protected String path;
  
    /** the description of the resource that was selected for download */
    protected ResourceEntry selectedResource;
  
  /**
   * Initializes the object with the information necessary to connect to the
   * specified server and download the specified resource.
   *
   * @param the_hostName the hostname of the server
   * @param the_port prefix the port where the server is running
   * @param the_path the full path of the filename on the server
   * @param the_selectedResource the description of the resource that was selected for download
   */

  public ServerInfo (String the_hostName, int the_port, String the_path, ResourceEntry the_selectedResource)
  {
    hostName = the_hostName;
    port = the_port;
    path = the_path;
    selectedResource = the_selectedResource;
  }
    
  /**
   * Initializes the object with the corresponding server and resource information.
   *
   * @param url describes the server information as read from the database
   * @param resourceEntry describes the resource information as read from the database
   */

  public ServerInfo (URL url, ResourceEntry resourceEntry)
  {
      serverURL = url;
      hostName = serverURL.getHost();
      port = serverURL.getPort();
      path = serverURL.getPath();
      selectedResource = resourceEntry;
  }

  /**
   * This abstract method will create a ServerConnection object that depends on the type of
   * server we are dealing with. This is because specific sub-classes of ServerConnection may
   * be constructed in different manners, and we want our interface to the controller to be
   * as abstract and flexible as possible.
   *
   * @return ServerConnection object that is tailored to the specific server in question
   * @exception Exception if there is an error creating the server connection object
   */

  abstract public ServerConnection getServerConnection() throws Exception;
  
  /**
   * This abstract method will configure the resume command used to re-establish a connection when
   * the connection must migrate to another sink.
   *
   * @return a configured ServerCommand object that can be used to re-establish a connection
   * @exception Exception if there is an error configuring the resume command
   */

  abstract public ServerCommand configureResumeCommand (ServerCommand resumeCommand) throws Exception;
  
  /**
   * Gets the hostname of the server.
   *
   * @return the hostname of the server
   */

  public String getHostName()
  {
    return hostName;
  }

  /**
   * Gets the port where the server is running.
   *
   * @return the port where the server is running
   */

  public int getPort()
  {
    return port;
  }

  public String getPath()
  {
    return path;
  }

  
  /**
   * Gets the resource that was selected for download.
   *
   * @return the resource that was selected for download
   */

  public ResourceEntry getSelectedResource()
  {
    return selectedResource;
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
      return "[" + this.getClass().getName() + 
	  " URL: " + serverURL +
	  ", selected resource info: " + selectedResource + "]";
  }
}
