package streamredirector;

import java.io.*;

/**
 * This is interface should be implemented by any object that will embody the connection from
 * the proxy to the sink. Its methods allow the proxy to connect to the sink, tell it from where
 * it should read its data, and close the connection.
 *
 * @author Jorge Rafael Nogueras
 */

public interface SinkConnection extends Serializable
{
	/**
	 * This method is called by the sink proxy when a controller has requested it that a connection
	 * be established; it establishes a connection to the sink.
	 *
   * @param the_cookie a long number that uniquely identifies this connection
	 * @exception Exception if there is an error establishing the connection to the sink
	 */
	 
	public void establishSinkConnection (Long the_cookie) throws Exception;
	
	/**
	 * Initializes the input stream where data coming from the server should be read.
	 *
	 * @param in where data coming from the server are read
	 * @exception Exception if there is an error setting the new input stream
	 */
	 
	public void setInputStream (InputStream in) throws Exception;
	
	/**
	 * This method is called by the proxy when a controller has requested it to close the connection.
	 *
	 * @exception Exception if there is an error closing the connection
	 */

	public void closeSinkConnection() throws Exception;
  
  /**
   * This method is called when a controller has requested that a command be sent to the sink.
   * Since some sinks need not process any commands, it is acceptable for a class that implements
   * this interface to have an emtpy implementation of it.
   *
   * @param command the command to be sent to the sink
   * @exception Exception if there are errors sending the command to the sink
   * @return a return code indicating success or an error condition
   */
   
  public ReturnCode sendCommand (SinkCommand command) throws Exception;
}