package streamredirector;

/**
 * This abstract class embodies the information necessary to be able to connect to a
 * sink proxy. It should be able to return an appropriate SinkConnection object and
 * a Transport that are appropriate to the concrete sub-class. This is the object that
 * is passed to the Controller so it can connect to the sink proxy.
 *
 * @see streamredirector.Controller, streamredirector.SinkConnection, streamredirector.Transport
 * @author Jorge Rafael Nogueras
 */

public abstract class SinkInfo
{
  /**
   * Initializes the object.
   */

  public SinkInfo() { }
  
  /**
   * Initializes the object, using the sink proxy's descriptor to initialize it.
   *
   * @param descriptor describes the sink proxy
   */

  public SinkInfo (SinkProxyDescriptor descriptor) { }
  
  /**
   * This abstract method will create a SinkConnection object that depends on the type of
   * sink proxy we are dealing with. This is because specific sub-classes of SinkConnection may
   * be constructed in different manners, and we want our interface to the controller to be
   * as abstract and flexible as possible.
   *
   * @return SinkConnection object that is tailored to the specific sink proxy in question
   * @exception Exception if there is an error creating the sink connection object
   */

  abstract public SinkConnection getSinkConnection() throws Exception;
  
  /**
   * This abstract method will create a Transport object that depends on the type of
   * sink proxy we are dealing with. This is because different sink proxies may use different
   * Transports and these transports may be constructed in different manners, and we want our
   * interface to the controller to be as abstract and flexible as possible.
   *
   * @return Transport object that is tailored to the specific sink proxy in question
   * @exception Exception if there is an error creating the transport object
   */

  abstract public Transport getTransport() throws Exception;
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "]";
  }
}