package streamredirector;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class embodies all the knowledge regarding the communication with a Controller. It is
 * fully configurable: one must pass it an appropriate listening transport (whose job is to
 * communicate with the Controller using an agreed-upon protocol) and also a default sink
 * (which is where the stream will be directed if the Controller decides to establish a connection
 * to the sink proxy without specifying a sink; the default sink may be null if it is known
 * beforehand that the Controller always will always specify a sink on the sink proxy).
 *
 * The SinkProxy takes care of all the details regarding establishment of connections, sending
 * commands to the server, and re-direction of streams.
 *
 * There will only be one instance of a SinkProxy that will be shared amongst all classes; thus,
 * the only way to get to that instance is through the static getInstance() method, which will
 * either instantiate a SinkProxy if none had been created before, or return the existing instance.
 *
 * @see streamredirector.CommandListener
 * @author Jorge Rafael Nogueras
 */

public class SinkProxy implements CommandListener
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** the one and only instance of this class that will ever be returned */
  protected static SinkProxy sinkProxy;
  
	/** the object that communicates with the controller and listens for connections */
	protected ListeningTransport listeningTransport;
	
	/** where the stream will be redirected by default */
	protected SinkConnection defaultSink;
	
	/** this specialized hash table makes sure that when connection state objects are removed,
	   	their interrupt() method is called to ensure the timer threads stop running;  it is
      keyed by a Long cookie and its value is the ConnectionState that represents that connection */
	protected ConnectionStateHashtable connectionTable;
	
	/** generates random numbers from where the unique cookies are derived */
	protected Random random;
	
  /**
   * Note that all this does is initialize the connection table and the random-number generator;
   * the listening transport and the default sink must be specified with configure(). This is
   * because a transport usually requires a reference to the sink proxy object itself when it
   * is created, so initialization of a SinkProxy can be a two-step process: first, the SinkProxy
   * instance is gotten and then configure() is called, passing in the transport and the default sink.
   */

  protected SinkProxy()
	{
		connectionTable = new ConnectionStateHashtable();
		random = new Random (System.currentTimeMillis());
	}

  /**
   * Creates a fully-initialized SinkProxy: this constructor can be used to create a SinkProxy
   * in one step, if the listening transport does not require a reference to the sink proxy
   * in order to be created. The default sink may be null if it is known beforehand that the
 	 * Controller will always specify a sink to use on the sink proxy.
   *
   * @param listening_transport the object that communicates with the controller and listens for connections
   * @param default_sink where the stream will be redirected by default
   */

  protected SinkProxy (ListeningTransport listening_transport, SinkConnection default_sink)
	{
		this();
		
		configure (listening_transport, default_sink);
	}
	
  /**
   * Retrieves the one and only SinkProxy instance; if it is null (i.e., never before created),
   * it will instantiate one and then return it. If the instance was already created, it will
   * simply return it.
   *
   * @return the only SinkProxy instance for this class
   */

  public static SinkProxy getInstance()
  {
    if (sinkProxy == null)
    {
      sinkProxy = new SinkProxy();
    }
    
    return sinkProxy;
  }
  
  /**
   * Retrieves the one and only SinkProxy instance; if it is null (i.e., never before created),
   * it will instantiate one and then return it. If the instance was already created, it will
   * simply return it.
   *
   * @param listening_transport the object that communicates with the controller and listens for connections
   * @param default_sink where the stream will be redirected by default
   * @return the only SinkProxy instance for this class
   */

  public static SinkProxy getInstance (ListeningTransport listening_transport, SinkConnection default_sink)
  {
    if (sinkProxy == null)
    {
      sinkProxy = new SinkProxy (listening_transport, default_sink);
    }
    
    return sinkProxy;
  }
  
  /**
   * Configures a SinkProxy by setting its listening transport and default sink: the default sink may
   * be null if it is known beforehand that the Controller will always specify a sink to use.
   *
   * @param listening_transport the object that communicates with the controller and listens for connections
   * @param default_sink where the stream will be redirected by default
   */

  public void configure (ListeningTransport listening_transport, SinkConnection default_sink)
	{
		listeningTransport = listening_transport;
		defaultSink = default_sink;
		
    // inform the listening transport that everything has been set up and it should start listening
    // for commands
		listeningTransport.startListening();
	}

  /**
   * This method gets called by the ListeningTransport when the "establishConnection" command
   * is received from the Controller. It specifies the server where the stream will be coming
   * from (the sink used will be this proxy's default sink, which should not be null in this
   * case). It calls the "establishConnection()" method with the specified server and the
   * default sink.
   *
   * @param server embodies the connection with the server
   * @exception Exception if there are errors creating the piped output streams or establishing the connection
   * @return a long value representing the cookie that uniquely identifies the connection
   */

  public Long establishConnection (ServerConnection server) throws Exception
	{
		return establishConnection (server, defaultSink);
	}
	
  /**
   * This method gets called by the ListeningTransport when the "establishConnection" command
   * is received from the Controller. It specifies the server where the stream will be coming
   * from and the sink where the data should be going to.
   * It sets up some state to keep track of the connection and generates a "cookie,"
   * which is a random number that uniquely identifies this connection.
   * It tries to establish a connection with both the server and the specified sink and ties
   * their input and output streams so that whatever is read from the server is written out
   * to the sink. It returns the cookie for the established connection.
   *
   * @param server embodies the connection with the server
   * @param sink where the stream data will be redirected
   * @exception Exception if there are errors creating the piped output streams or establishing the connection
   * @return a long value representing the cookie that uniquely identifies the connection
   */

  public Long establishConnection (ServerConnection server, SinkConnection sink) throws Exception
	{
    // generate a new, unique cookie value that will describe this connection
    Long cookie = getUniqueNumber();
    
    if (DEBUG)
      System.out.println ("SinkProxy::establishConnection -> Establishing connection with cookie: " + cookie);
    
		try
		{
//System.out.println ("SinkProxy::establishConnection -> connecting server...");
			server.establishServerConnection (cookie);
		}
		catch (Exception e)
		{
			throw new CannotEstablishServerConnectionException (e.getMessage());
		}

		try
		{
//System.out.println ("SinkProxy::establishConnection -> connecting sink...");
			sink.establishSinkConnection (cookie);
		}
		catch (Exception e)
		{
			throw new CannotEstablishSinkConnectionException (e.getMessage());
		}
		
    // these piped input and output streams "bind" the server's output stream to the
    // sink's input stream, so that anything that is written by the server is
    // automatically read by the sink
		PipedOutputStream pos = new PipedOutputStream();
		PipedInputStream pis = new PipedInputStream (pos);

		try
		{
			server.setOutputStream (pos);
			sink.setInputStream (pis);
		}
		catch (Exception e)
		{
			throw new CannotConnectServerToSinkException (e.getMessage());
		}

    // create state in the connection table that is associated with the connection's cookie:
    // each state object contains references to the ServerConnection, SinkConnection and the
    // SinkProxy itself
		connectionTable.put (cookie, new ConnectionState (cookie, server, sink, this));

//System.out.println ("SinkProxy::establishConnection -> state hashtable = " + connectionTable);
//System.out.println ("SinkProxy::establishConnection -> returning cookie: " + cookie);

		return cookie;
	}
	
  /**
   * This method gets called by the ListeningTransport when the "sendCommand" command
   * is received from the Controller. It specifies the cookie for the connection in question:
   * the corresponding ServerConnection object is gotten from the connection table
   * and the specified command is sent to the server by calling its "sendCommand()" method
   * if it is a ServerCommand (that is, protocol-specific commands meant for the server). 
   *
   * @param cookie identifies uniquely the connection where the command must be sent
   * @param command the command that the server will be sent
   * @exception Exception if there are errors getting connection state or if there are communication errors
   * @return a return code indicating success or an error condition
   */
   
	public synchronized ReturnCode sendCommand (Long cookie, Command command) throws Exception
	{
//System.out.println ("SinkProxy::sendCommand -> received command");

		ConnectionState state = (ConnectionState) connectionTable.get (cookie);

		if (state == null)
		{
			throw new NoSuchElementException ("Could not find connection corresponding to cookie " + cookie);
		}

//System.out.println ("SinkProxy::sendCommand -> found connection with state: " + state);

		ServerConnection server = state.server;
		SinkConnection sink = state.sink;
		
		if (server == null || sink == null)
		{
			throw new NoSuchElementException ("Connection state was improperly created and contains nothing: " + state);
		}
			
		if (command instanceof ServerCommand)
		{
//System.out.println ("SinkProxy::sendCommand -> executing server command: " + command);

			return server.sendCommand ((ServerCommand) command);
		}
    else if (command instanceof SinkCommand)
    {
//System.out.println ("SinkProxy::sendCommand -> executing sink command: " + command);

      return sink.sendCommand ((SinkCommand) command);
    }

    return ReturnCode.COMMAND_ERROR;
	}
	
  /**
   * This method gets called by the ListeningTransport when the "renewLease" command
   * is received from the Controller. It specifies the cookie for the connection whose lease
   * must be renewed. It simply fetches the appropriate state information and renews the lease
   * for that connection.
   *
   * @param cookie identifies uniquely the connection whose lease should be renewed
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a return code indicating success or an error condition
   */

  public ReturnCode renewLease (Long cookie) throws Exception
  {
//System.out.println ("SinkProxy::renewLease -> executing refresh lease connection command");

    ConnectionState state = (ConnectionState) connectionTable.get (cookie);

    if (state == null)
    {
      throw new NoSuchElementException ("Could not find connection corresponding to cookie " + cookie);
    }

//System.out.println ("SinkProxy::renewLease -> found connection with state: " + state);

    // we renew the connection lease on this connection
    state.renewLease();
    
    return ReturnCode.COMMAND_OK;
  }

  /**
   * This method will get called when the endConnection command is received.
   *
   * @param cookie identifies uniquely the connection that must be terminated
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a ServerCommand that resumes the connection from the point where it was ended
   */

  public ServerCommand endConnection (Long cookie) throws Exception
  {
    if (DEBUG)
      System.out.println ("SinkProxy::endConnection -> executing end connection command");

    ConnectionState state = (ConnectionState) connectionTable.get (cookie);

    if (state == null)
    {
      throw new NoSuchElementException ("Could not find connection corresponding to cookie " + cookie);
    }

//System.out.println ("SinkProxy::endConnection -> found connection with state: " + state);

    ServerConnection server = state.server;
    SinkConnection sink = state.sink;
    
    if (server == null || sink == null)
    {
      throw new NoSuchElementException ("Connection state was improperly created and contains nothing: " + state);
    }
    
    // close the connection to the server and the sink
    server.closeServerConnection();
    sink.closeSinkConnection();
    
    // get the resume command from the server before deleting the connection's state
    ServerCommand resumeCommand = server.getResumeCommand();
    
    // delete the connection's state from the connection table
    connectionTable.remove (cookie);

    return resumeCommand;
  }
  
  /**
   * Queries the connection table for a ConnectionState object corresponding to
   * the specified cookie.
   *
   * @param cookie the cookie whose connection is being requested
   * @returns the ConnectionState object corresponding to the specified cookie
   */

  public ConnectionState getConnectionState (Long cookie)
  {
    return (ConnectionState) connectionTable.get (cookie);
  }
  
	/**
	 * Returns a long random number that is unique to the connection table. Before returning the next
	 * long random number, it makes sure that it does not already exist in the table, and keeps selecting
	 * a new random number until it finds one that does note exist in the table. We never return "-1" as a
   * cookie because we reserve that value to represent an error.
   *
   * @return a unique random long number
	 */

	protected Long getUniqueNumber()
	{
		// get a positive long random number (it should be positive because the cookie will
    // be used as a directory name in the web server and directory names can not have
    // "-" in them
    long cookie = Math.abs (random.nextLong());

		// if there was a connection stored with that number already, pick the next random
		// number, stopping when an unused number is found; also, never return "-1" as a cookie, since
    // we reserve that value to represent an error
		while ((cookie != -1L) && (connectionTable.get (new Long (cookie)) != null))
			cookie = random.nextLong();

		return new Long (cookie);
	}
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": listening transport = " + listeningTransport +
           ", default sink = " + defaultSink + ", connection table = " + connectionTable + "]";
  }
}