package streamredirector;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This is the counterpart to a TCPTransport: its job is to listen for TCP connections by binding to a
 * a specific TCP port and spawning off TCPListeningTransportInstances that in turn read the messages
 * written to the connection (this is done to account for different controllers connecting at the same time).
 * It keeps a list of all the TCPListeningTransportInstance that have been spawned-off in case it is
 * asked to stop listening, in which case it will shut down all the instances, too.
 *
 * @see streamredirector.Transport, streamredirector.TCPTransport, streamredirector.TCPListeningTransport, streamredirector.TCPListeningTransportInstance, streamredirector.CommandListener
 * @author Jorge Rafael Nogueras
 */

public class TCPListeningTransport implements ListeningTransport, Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** object that is interested in receiving the messages read from the connection to the Transport */
	protected CommandListener listener;
	
	/** socket where we are listening for incoming connections */
	protected ServerSocket serverSocket;
	
  /** the local port used to listen to incoming connections */
  protected int localPort;
  
	/** thread that continues listening for incoming connections */
	protected Thread thread;
	
	/** this variable, if true, will cause the thread to stop running */
	protected boolean interrupted;

	/** keeps a list of all TCPListeningTransportInstance that have been spawned-off */
	protected Hashtable listeningInstancesList;
	
  /**
   * Initializes the listening transport and passes it a reference to the CommandListener that is
   * interested in receiving notifications of messages as they are read and the port where this
   * listening transport should bind.
   * The listening thread is not yet created.
   * Since the "port" argument is absent, we will bind to a random port.
   *
   * @param the_listener to the object that is interested in receiving notifications of messages as they are read 
   * @exception Exception if there are errors creating the server socket
   */

  public TCPListeningTransport (CommandListener the_listener) throws Exception
  {
    this (the_listener, 0);
  }
  
	/**
	 * Initializes the listening transport and passes it a reference to the CommandListener that is
	 * interested in receiving notifications of messages as they are read and the port where this
	 * listening transport should bind.
	 * The listening thread is not yet created.
   * If the "port" argument is 0, we will bind to a random port.
   *
   * @param the_listener to the object that is interested in receiving notifications of messages as they are read 
   * @param port the port where this listening transport should bind and listen
   * @exception Exception if there are errors creating the server socket
   */

	public TCPListeningTransport (CommandListener the_listener, int port) throws Exception
	{
		listener = the_listener;
		
		interrupted = true;
		serverSocket = new ServerSocket (port);
		
    localPort = serverSocket.getLocalPort();
		listeningInstancesList = new Hashtable();
	}
	
	/**
	 * Notifies the listening transport that it should start listening for incoming connections (it
	 * creates and starts a listening thread only if there was not one already running).
	 */

	public void startListening()
	{
		if (interrupted)
		{
			interrupted = false;
			thread = new Thread (this, "TCPListeningTransport listener #" + hashCode());
			thread.start();
		}
	}
	
	/**
	 * Notifies the listening transport that it should stop listening for incoming connections (it
	 * interrupts the running thread and makes it stop running). It also goes through the list of
	 * TCPListeningTransportInstance it has spawned-off and requests them to stop listening, also.
	 */

	public void stopListening()
	{
		Enumeration e = listeningInstancesList.elements();
		
		while (e.hasMoreElements())
		{
      if (DEBUG)
        System.out.println ("TCPListeningTransport::stopListening -> shutting down listening transport instance");

      ((TCPListeningTransportInstance) e.nextElement()).stopListening();
		}

    if (DEBUG)
      System.out.println ("TCPListeningTransport::stopListening -> !!Interrupting!");

    interrupt();
	}

  /**
   * Removes the specified listener from the list of listener instances.
   *
   * @param instance the TCPListeningTransportInstance to be removed
   * @return the instance that was removed
   */
   
  public TCPListeningTransportInstance removeListener (TCPListeningTransportInstance instance)
  {
    return (TCPListeningTransportInstance) listeningInstancesList.remove (instance);
  }
  
  /**
   * Queries for the local port used to listen to incoming connections.
   *
   * @return the local port used to listen to incoming connections
   */
   
  public int getLocalPort()
  {
    return localPort;
  }
  
	/**
	 * This method makes the listening thread stop running.
	 */
	 
	protected void interrupt()
	{
		interrupted = true;
		thread.interrupt();
	}
	
	/**
	 * This is the code that will be run by the listening thread: it blocks while waiting for a new
	 * connection and spawns off a new TCPListeningTransportInstance that will read the messages it
	 * receives.
	 */

	public void run()
	{
		// <HERE> For single sink proxy architecture, we only spawn off 1 TransportInstance
		// which will in turn await for commands and send them up to the CommandListener
		while (!interrupted)
		{
			try
			{
				Socket newSocket = serverSocket.accept();
        TCPListeningTransportInstance instance = new TCPListeningTransportInstance (listener, newSocket, this);
        listeningInstancesList.put (instance, instance);
			}
			catch (Exception e)
			{
				e.printStackTrace();
				// if there was an error accepting a new socket connection, this is probably an unrecoverable error
				// in that case, we will exit the loop and let the thread exit normally to shut down this listening transport
				System.err.println ("TCPListeningTransport::run -> Error reading from TCP Listening transport: " + e);
				interrupted = true;
				continue;
			}
		}

    if (DEBUG)
      System.out.println ("TCPListeningTransport::run -> shutting down TCPListeningTransport listener #" + hashCode());
	}
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": server socket = " + serverSocket +
        ", list of listening instances = " + listeningInstancesList + "]";
  }
}