package streamredirector;

import java.io.*;
import java.net.*;
import java.rmi.*;

/**
 * The job of this class is to read the commands written to the socket it is listening to 
 * and relaying them to a CommandListener. A CommandListener is an object that is
 * interested in getting notified when messages are received (concretely, the SinkProxy is a
 * CommandListener).
 *
 * It uses serialization to receive the parameters from the network; however, since we
 * do not know beforehand which ServerConnection, SinkConnection and Command subclasses
 * we will receive, we must load the bytecodes for these specific subclasses at runtime. To do this,
 * whenever we receive one of these types of objects over the network, we will be receiving in reality
 * a MarshalledObject which encapsulates the desired object. To get at the object it is wrapping, we call
 * its get() method and cast it to the appropriate type. If we do not have the bytecodes for the specific
 * subclass received, the MarshalledObject will load it from an appropriate codebase (which in our case
 * is a URL that points to a jarfile that contains all the classes that resides on the controller and that
 * is being served by its local web server).
 *
 * In order for this automatic remote class loading to work, the main executable for the sink proxy must
 * have installed a SecurityManager that allows connections to remote sockets.
 *
 * @see streamredirector.Transport, streamredirector.TCPTransport, streamredirector.TCPListeningTransport, streamredirector.TCPListeningTransportInstance, streamredirector.CommandListener, java.rmi.MarshalledObject
 * @author Jorge Rafael Nogueras
 */

public class TCPListeningTransportInstance implements ListeningTransport, Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

	/** object that is interested in receiving the messages read from the connection to the Transport */
	protected CommandListener listener;
	
  /** the "parent" listening transport that spawns all the listening instances */
  protected TCPListeningTransport parent;
  
	/** where return values are written out to the connection */
	protected ObjectOutputStream oos;
	
	/** where messages (method descriptors and parameters) are read from */
	protected ObjectInputStream ois;

	/** thread that continues listening for incoming messages */
	protected Thread thread;

	/** this variable, if true, will cause the thread to stop running */
	protected boolean interrupted;
	
	/**
	 * Initializes the listening transport instance and passes it a reference to the CommandListener that is
	 * interested in being passed notifications of messages as they are received and the socket representing the
	 * connection that was previously established and from where messages should be read.
   *
   * @param the_listener to the object that is interested in receiving notifications of messages as they are read 
   * @param the_socket the connection from where messages should be read and return values should be written
   * @param the_parent the "parent" listening transport that spawns all the listening instances
   * @exception Exception if there are errors getting the input and output streams for the socket
   */

	public TCPListeningTransportInstance (CommandListener the_listener, Socket the_socket, TCPListeningTransport the_parent) throws Exception
	{
		listener = the_listener;
		parent = the_parent;
    
		oos = new ObjectOutputStream (the_socket.getOutputStream());
		ois = new ObjectInputStream (the_socket.getInputStream());

		interrupted = true;

		startListening();
	}

	/**
	 * Notifies the listening transport instance that it should start listening for incoming messages (it
	 * creates and starts a listening thread only if there was not one already running).
	 */

	public void startListening()
	{
		if (interrupted)
		{
			interrupted = false;
			thread = new Thread (this, "TCPListeningTransportInstance listener #" + hashCode());
			thread.start();
		}
	}
	
	/**
	 * Notifies the listening transport instance that it should stop listening for incoming messages (it
	 * interrupts the running thread and makes it stop running).
	 */

	public void stopListening()
	{
		interrupt();
	}

	/**
	 * This method makes the listening thread stop running if it has not been previously interrupted.
	 */
	 
	protected void interrupt()
	{
		if (!interrupted)
		{
			interrupted = true;
  		thread.interrupt();
		}
	}
	
	/**
	 * This is the code that will be run by the listening thread: it blocks while waiting for a new
	 * messages to arrive and keeps listening until it is interrupted.
	 */

	public void run()
	{
    if (DEBUG)
      System.out.println ("TCPListeningTransportInstance::run -> started transport instance #" + hashCode());

		while (!interrupted)
		{
			try
			{
				listenForCommands();
			}
			catch (Exception e)
			{
				// if there was an error writing the output, this means that the connection was probably closed
				// due to an error; in that case, we will exit the loop and let the thread exit normally to
				// shut down this listening transport instance 
        if (DEBUG)
				  System.err.println ("TCPListeningTransportInstance::run -> Error reading command from controller: connection was possibly closed");
//				e.printStackTrace();
				interrupted = true;
				continue;
			}
		}

    // remove this listening transport instance from the parent's list of currently-running instances
    parent.removeListener (this);
    
    if (DEBUG)
      System.out.println ("TCPListeningTransportInstance::run -> shutting down TCPListeningTransport listener #" + hashCode());
	}

	/**
	 * This method is called to read a message; first it expects a MethodDescriptor that specifies the
	 * name of the method that will follow and the number and types of its parameters. After having read
	 * the parameters, the appropriate method is called in the CommandListener object and the appropriate
	 * return object is written back. If at any point an exception is thrown by the executing command,
	 * this exception is written as the return object so it can be re-thrown at the controller.
	 *
	 * @exception Exception if there are errors writing out the return object
	 */

	protected void listenForCommands() throws Exception
	{
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> listening...");

		// this contains the object that will be returned to the controller: if an exception is caught,
		// it will be sent as the return object so the controller is aware of the error
		Object returnObject = null;
		
		MethodDescriptor method = null;
		
		try
		{
			// the method descriptor is read; if there is some error reading it, the exception will be
			// caught and written to the output stream
			method = (MethodDescriptor) ois.readObject();
		}
		catch (Exception e)
		{
      if (DEBUG)
        System.out.println ("TCPListeningTransportInstance::listenForCommands -> error reading method descriptor...");
//e.printStackTrace();
			writeException (e);
			return;
		}
		
		String methodName = method.getName();
		
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> read method name: " + methodName);

    if (methodName.equals ("ping"))
    {
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> Received ping message");
      // write the return object back to the connection
      oos.writeObject (new Boolean (true));
      oos.flush();
    }
		else if (methodName.equals ("establishConnection"))
		{
			try
      {
        // when establishing a connection, we will first read the ServerConnection object where the
        // sink proxy should connect
        MarshalledObject marshalledServer = (MarshalledObject) ois.readObject();
        ServerConnection server = (ServerConnection) marshalledServer.get();
  			
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> server: " + server);
			
				Class [] parameters = method.getParameters();
			
				// if there is only one parameter for this method, this means that only the ServerConnection
				// object has been sent; the sink will be whatever the default sink is for the proxy
				if (parameters.length == 1)
	  		{
		  		returnObject = listener.establishConnection (server);
	  		}
	  		else if (parameters.length == 2)
	  		{
          // if the SinkConnection is explicitly sent, we read it next
          MarshalledObject marshalledSink = (MarshalledObject) ois.readObject();
          SinkConnection sink = (SinkConnection) marshalledSink.get();

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> sink: " + sink);
	
          // finally, we call the command listener's establishConnection() method with the server and
          // sink connection objects read
					returnObject = listener.establishConnection (server, sink);
	  		}
			}
			catch (Exception e)
			{
        e.printStackTrace();
				writeException (e);
				return;
			}
  		
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> writing cookie: " + returnObject);

			// write the return object back to the connection
			oos.writeObject (returnObject);
			oos.flush();
		}
		else if (methodName.equals ("sendCommand"))
		{
			try
			{
        // when a command is being sent to the sink proxy, first we read the cookie for the connection
        // and then the command itself
        Long cookie = (Long) ois.readObject();
        MarshalledObject marshalledCommand = (MarshalledObject) ois.readObject();
        Command command = (Command) marshalledCommand.get();

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> cookie: " + cookie);
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> command: " + command);
			
        // finally, we call the command listener's sendCommand() method with the cookie and command read
				returnObject = listener.sendCommand (cookie, command);
			}
			catch (Exception e)
			{
        if (DEBUG)
          System.out.println ("TCPListeningTransportInstance::listenForCommands -> caught exception: " + e);
e.printStackTrace();
        writeException (e);
				return;
			}

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> return code: " + returnObject);

			// write the return object back to the connection
			oos.writeObject (returnObject);
			oos.flush();
		}
    else if (methodName.equals ("renewLease"))
    {
      try
      {
        // when a lease is being renewed, we must first read the cookie for the connection
        Long cookie = (Long) ois.readObject();

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> cookie: " + cookie);
      
        // finally, we call the command listener's renewLease() method with the cookie read
        returnObject = listener.renewLease (cookie);
      }
      catch (Exception e)
      {
        if (DEBUG)
          System.out.println ("TCPListeningTransportInstance::listenForCommands -> caught exception: " + e);
        writeException (e);
        return;
      }

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> return code: " + returnObject);

      // write the return object back to the connection
      oos.writeObject (returnObject);
      oos.flush();
    }
    else if (methodName.equals ("endConnection"))
    {
      try
      {
        // when a connection is being ended, we must first read the cookie for the connection
        Long cookie = (Long) ois.readObject();
        
//        if (DEBUG)
        if (true)
          System.out.println ("TCPListeningTransportInstance::listenForCommands -> ending connection with cookie: " + cookie);
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> cookie: " + cookie);
      
        // finally, we call the command listener's endConnection() method with the cookie read
        returnObject = listener.endConnection (cookie);
      }
      catch (Exception e)
      {
//System.out.println ("TCPListeningTransportInstance::listenForCommands -> caught exception: " + e);
e.printStackTrace();
        writeException (e);
        return;
      }

//System.out.println ("TCPListeningTransportInstance::listenForCommands -> return code: " + returnObject);

      // write the return object back to the connection
      oos.writeObject (returnObject);
      oos.flush();
    }
		else
		{
			System.err.println ("TCPListeningTransportInstance::listenForCommands -> Invalid method name: " + methodName);
			writeException (new InvalidMethodNameException ("Received invalid method name: " + methodName));
			return;
		}
	}
	
	/**
	 * Writes the specified exception out to the output stream.
	 *
	 * @param exception the exception object that will get written out to the output stream
	 * @exception Exception if there are errors writing to the output stream
	 */

	protected void writeException (Exception exception) throws Exception
	{
 		oos.writeObject (exception);
		oos.flush();
	}

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + " #" + hashCode() + "]";
  }
}