package streamredirector;

/**
 * This concrete implementation of SinkInfo knows how to create a TCPTransport object to
 * connect to the sink proxy. It can be used when a TCPTransport object is desired to connect
 * to the sink proxy and where the default sink should be used in the proxy (when asked for
 * a SinkConnection object, it will always return "null," which indicates that the default
 * sink in the proxy should be used). If we want to be able to specify an ultimate sink in
 * the proxy other than the default, an appropriate subclass (like FileTCPSinkInfo, for instance)
 * should be used instead.
 *
 * @see streamredirector.TCPTransport, streamredirector.FileTCPSinkInfo
 * @author Jorge Rafael Nogueras
 */

public class TCPSinkInfo extends SinkInfo
{
  /** the hostname of the sink proxy */
  protected String hostName;
  
  /** the port where the sink proxy is listening */
  protected int port;
  
  public TCPSinkInfo (SinkProxyDescriptor descriptor)
  {
    this (((TCPSinkProxyDescriptor) descriptor).getHostName(), ((TCPSinkProxyDescriptor) descriptor).getPort());
  }

  /**
   * Initializes the object with the information necessary to connect to the
   * specified sink using TCP.
   *
   * @param the_hostName the hostname of the sink proxy
   * @param the_port prefix the port where the sink proxy is listening
   */

  public TCPSinkInfo (String the_hostName, int the_port)
  {
    hostName = the_hostName;
    port = the_port;
  }

  /**
   * When a TCPSinkInfo object is used, it means that the client proxy will always use its
   * default sink; this is why we will always return "null" when asked for a SinkConnection
   * object. If we want to be able to specify an ultimate sink in the proxy other than the
   * default, we must create a subclass and override this method to return the desired subclass
   * of SinkConnection.
   *
   * @return always returns "null" to indicate that the default sink should be used in the proxy
   * @exception Exception if there is an error creating the sink connection object (should never be thrown)
   */

  public SinkConnection getSinkConnection() throws Exception
  {
    return null;
  }
  
  /**
   * This method will create a TCPTransport object that connects to the sink proxy
   * using a TCP connection.
   *
   * @return TCPTransport object that will be used to connect to the sink proxy
   * @exception Exception if there is an error creating the transport object
   */

  public Transport getTransport() throws Exception
  {
    return new TCPTransport (hostName, port);
  }
  
  /**
   * Gets the hostname of the sink proxy.
   *
   * @return the hostname of the sink proxy
   */

  public String getHostName()
  {
    return hostName;
  }

  /**
   * Gets the port where the sink proxy is listening.
   *
   * @return the port where the sink proxy is listening
   */

  public int getPort()
  {
    return port;
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": host = " + hostName + ":" + port + "]";
  }
}