package streamredirector;

import java.net.*;
import java.io.*;
import java.rmi.*;

/**
 * This concrete implementation of a Transport communicates with its corresponding ListeningTransport
 * using TCP sockets. It uses serialization to send the parameters over the network; however, since the
 * remote sink proxies do not know beforehand which ServerConnection, SinkConnection and Command subclasses
 * they will receive, it is necessary for them to load the bytecodes for these specific subclasses at
 * runtime. To do this, whenever we write one of these types of objects over the network, we use the
 * MarshalledObject class. It encapsulates the object we want to serialize, and it "annotates" each object
 * with the codebase that can be used by the remote JVM to load its bytecodes.
 *
 * In our case, the controller executable must start a web server and set the "java.rmi.server.codebase"
 * property to the URL where the jarfile that contains all the classes can be found. This way, when the
 * remote TCPTransportListener deserializes the MarshalledObjects and gets their encapsulated objects,
 * it will know from where to load their bytecodes.
 *
 * @see streamredirector.Transport, streamredirector.ListeningTransport, streamredirector.TCPListeningTransport, java.rmi.MarshalledObject
 * @author Jorge Rafael Nogueras
 */

public class TCPTransport implements Transport
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** this method descriptor describes the "ping" method diagnostic method */
  protected static final MethodDescriptor pingMethodDescriptor = new MethodDescriptor ("ping",
                                                        new Class[] {});

  /** this method descriptor describes the "establishConnection" method with only 1 parameter */
	protected static final MethodDescriptor establishConnectionMethodDescriptor1 = new MethodDescriptor ("establishConnection",
																												new Class[] {ServerConnection.class});

	/** this method descriptor describes the "establishConnection" method with 2 parameters */
	protected static final MethodDescriptor establishConnectionMethodDescriptor2 = new MethodDescriptor ("establishConnection",
																												new Class[] {ServerConnection.class, SinkConnection.class});

  /** this method descriptor describes the "sendCommand" method */
	protected static final MethodDescriptor sendCommandMethodDescriptor = new MethodDescriptor ("sendCommand",
																												new Class[] {Long.class, Command.class});

  /** this method descriptor describes the "renewLease" method */
  protected static final MethodDescriptor renewLeaseMethodDescriptor = new MethodDescriptor ("renewLease",
                                                        new Class[] {Long.class});

  /** this method descriptor describes the "endConnection" method */
  protected static final MethodDescriptor endConnectionMethodDescriptor = new MethodDescriptor ("endConnection",
                                                        new Class[] {Long.class});

	/** the output stream where the method and parameter information will be written */
	protected ObjectOutputStream oos;

	/** the input stream from where return object will be read */
	protected ObjectInputStream ois;

	/** the description of this object (used in the toString() method) */
	protected String description;

	/**
   * Creates a TCPTransport that connects to the specified address and port. It also sets up
   * the corresponding input and output streams.
   *
   * @param remoteAddress the host where the corresponding TCPListeningTransport is running
   * @param remotePort the port where the corresponding TCPListeningTransport is listening
   * @exception Exception if there are errors establishing a connection or creating the input and output streams
   */

	public TCPTransport (String remoteAddress, int remotePort) throws Exception
	{
		Socket socket = new Socket (remoteAddress, remotePort);
		
		oos = new ObjectOutputStream (socket.getOutputStream());
		ois = new ObjectInputStream (socket.getInputStream());
		
		description = "[" + this.getClass().getName() + ": connected to: " + remoteAddress + ", port = " + remotePort + "]";
	}
	
  /**
   * This method will get called to verify that the listening transport counterpart is
   * actually there to receive command requests. It sends a ping request to the listening
   * transport and waits for the response; once it gets it, it closes the connection and
   * returns true; if it gets a false value or an exception is thrown, it returns false.
   *
   * @exception Exception if there are errors sending the ping request
   * @return true if the listening counterpart is active, false otherwise
   */

  public boolean ping() throws Exception
  {
    try
    {
      // we write the method descriptor object
      oos.writeObject (pingMethodDescriptor);
        
      oos.flush();
      
      // we wait until the connection is established and read the results of the operation
      Object returnObject = ois.readObject();
    
//System.out.println ("TCPTransport::ping -> return object is: " + returnObject);

      if (returnObject instanceof Exception)
      {
        // when the return object is of type Exception, it means something went wrong on the other
        // side of the connection: we will thus throw the exception so it is caught by the controller:
        // note that we rethrow the exception as it was read, but with the stack trace filled in,
        // since by default stack traces are not serialized along with the Exception object
//System.out.println ("TCPTransport::ping -> throwing Exception: " + returnObject);
        return false;
      }
      else
      {
//System.out.println ("TCPTransport::ping -> boolean return value was: " + returnObject);
        return ((Boolean) returnObject).booleanValue();
      }
    }
    catch (Exception e)
    {
      return false;
    }
    finally
    {
      oos.close();
      ois.close();
    }
  }
  
  /**
   * This method will get called when the controller wants to establish a connection
   * between a server and a sink. Since only the server object has been specified, it will
   * call the establishConnection() method passing a null SinkConnection to signal that it
   * should not be written to the sink proxy.
   *
   * @param server embodies the connection with the server
   * @exception Exception if there are errors writing or reading objects
   * @return a long value representing the cookie that uniquely identifies the connection
   */

	public Long establishConnection (ServerConnection server) throws Exception
	{
    return establishConnection (server, null);
	}

  /**
   * This method will get called when the controller wants to establish a connection.
   * between a server and a sink. It sends the appropriate method descriptor followed by the
   * ServerConnection and SinkConnection objects specified. It then reads the return object
   * and if it is an exception, it throws that exception, or simply returns it as the long
   * cookie uniquely identfying the connection just established.
   * If the SinkConnection object is null, it will assume that it should not be written
   * to the sink proxy.
   *
   * @param server embodies the connection with the server
   * @param sink embodies the connection with the sink
   * @exception Exception if there are errors writing or reading objects
   * @return a long value representing the cookie that uniquely identifies the connection
   */

	public Long establishConnection (ServerConnection server, SinkConnection sink) throws Exception
	{
    // if the SinkConnection object is null, it will assume that it should not be written
    // to the sink proxy so we will use the method descriptor with only one parameter
    if (sink == null)
    {
      if (DEBUG)
        System.out.println ("TCPTransport::establishConnection -> writing method descriptor (only server connection)");

      // we write the method descriptor followed by the ServerConnection object
      oos.writeObject (establishConnectionMethodDescriptor1);

//System.out.println ("Before writing server connection, codebase is: " + System.getProperty ("java.rmi.server.codebase"));
      // we write the server object wrapped in a MarshalledObject
      oos.writeObject (new MarshalledObject (server));
//      oos.writeObject (server);
    }
    else
    {
      if (DEBUG)
        System.out.println ("TCPTransport::establishConnection -> writing method descriptor and server/sink");

      // we write the method descriptor followed by the ServerConnection and SinkConnection object
      oos.writeObject (establishConnectionMethodDescriptor2);
      
//System.out.println ("Before writing server connection, codebase is: " + System.getProperty ("java.rmi.server.codebase"));

      // we write the server and sink objects wrapped in MarshalledObjects
      oos.writeObject (new MarshalledObject (server));
      oos.writeObject (new MarshalledObject (sink));
//      oos.writeObject (server);
//      oos.writeObject (sink);
    }
    
		oos.flush();
		
//System.out.println ("TCPTransport::establishConnection -> reading cookie");

		// we wait until the connection is established and read the results of the operation
		Object returnObject = ois.readObject();
		
//System.out.println ("TCPTransport::establishConnection -> return object is: " + returnObject);

    if (returnObject instanceof Exception)
		{
			// when the return object is of type Exception, it means something went wrong on the other
			// side of the connection: we will thus throw the exception so it is caught by the controller:
			// note that we rethrow the exception as it was read, but with the stack trace filled in,
			// since by default stack traces are not serialized along with the Exception object
      if (DEBUG)
        System.out.println ("TCPTransport::establishConnection -> throwing Exception: " + returnObject);
        
			throw (Exception) ((Exception) returnObject).fillInStackTrace();
		}
		else
		{
      if (DEBUG)
        System.out.println ("TCPTransport::establishConnection -> cookie was: " + returnObject);

        return (Long) returnObject;
		}
	}

  /**
   * This method will get called when the controller wants to send a command to the sink proxy.
   * It sends the appropriate method descriptor followed by the cookie that uniquely identifies
   * the connection where the command should operate and the actual command object.
   * It then reads the return object and if it is an exception, it throws that exception, or
   * simply returns it as the object as it was sent (the class of the object may differ, since
   * it can be an Integer error code or a protocol-dependent command that may be used to resume
   * the connection if it was closed).
   * The method is synchronized because there may be different threads trying to send commands
   * simultaneously (for instance, the controller and the ConnectionLeaseRenewer).
   *
   * @param cookie the cookie that uniquely identifies the connection where the command should operate 
   * @param command the command that should operate on the specified connection
   * @exception Exception if there are errors writing or reading objects
   * @return a return code indicating success or an error condition
   */

	public synchronized ReturnCode sendCommand (Long cookie, Command command) throws Exception
	{
//System.out.println ("TCPTransport::sendCommand -> writing method descriptor: " + sendCommandMethodDescriptor);
    if (DEBUG)
      System.out.println ("TCPTransport::sendCommand -> writing command: " + command);

		// we write the method descriptor followed by the cookie and the actual command
  	oos.writeObject (sendCommandMethodDescriptor);

		oos.writeObject (cookie);
		
   // we write the command object wrapped in a MarshalledObject
    oos.writeObject (new MarshalledObject (command));
		oos.flush();
		
//System.out.println ("TCPTransport::sendCommand -> reading return object");

		// we wait until the command is executed and read the results of the operation
		Object returnObject = ois.readObject();

		if (returnObject instanceof Exception)
		{
			// when the return object is of type Exception, it means something went wrong on the other
			// side of the connection: we will thus throw the exception so it is caught by the controller:
			// note that we rethrow the exception as it was read, but with the stack trace filled in,
			// since by default stack traces are not serialized along with the Exception object
      if (DEBUG)
        System.out.println ("TCPTransport::sendCommand -> throwing Exception: " + returnObject);

      throw (Exception) ((Exception) returnObject).fillInStackTrace();
		}
		else
		{
      if (DEBUG)
        System.out.println ("TCPTransport::sendCommand -> return code: " + returnObject);
        
			return (ReturnCode) returnObject;
		}
	}
	
  /**
   * This method will get called when the controller wants a connection lease to be renewed.
   * It will send to the corresponding ListeningTransport the cookie that uniquely identifies
   * the connection whose lease must be renewed.
   *
   * @param cookie identifies uniquely the connection whose lease should be renewed
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a return code indicating success or an error condition
   */

  public ReturnCode renewLease (Long cookie) throws Exception
  {
//System.out.println ("TCPTransport::renewLease -> writing method descriptor: " + renewLeaseMethodDescriptor);

    // we write the method descriptor followed by the cookie
    oos.writeObject (renewLeaseMethodDescriptor);

    oos.writeObject (cookie);
    oos.flush();
    
//System.out.println ("TCPTransport::renewLease -> reading return object");

    // we wait until the command is executed and read the results of the operation
    Object returnObject = ois.readObject();

    if (returnObject instanceof Exception)
    {
      // when the return object is of type Exception, it means something went wrong on the other
      // side of the connection: we will thus throw the exception so it is caught by the controller:
      // note that we rethrow the exception as it was read, but with the stack trace filled in,
      // since by default stack traces are not serialized along with the Exception object
      if (DEBUG)
        System.out.println ("TCPTransport::renewLease -> throwing Exception: " + returnObject);

      throw (Exception) ((Exception) returnObject).fillInStackTrace();
    }
    else
    {
//System.out.println ("TCPTransport::renewLease -> return code: " + returnObject);
      return (ReturnCode) returnObject;
    }
  }

  /**
   * This method will get called when the endConnection command is received.
   *
   * @param cookie identifies uniquely the connection that must be terminated
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a ServerCommand that resumes the connection from the point where it was ended
   */

  public ServerCommand endConnection (Long cookie) throws Exception
  {
    if (DEBUG)
      System.out.println ("TCPTransport::endConnection -> writing method descriptor: " + endConnectionMethodDescriptor);

    // we write the method descriptor followed by the cookie
    oos.writeObject (endConnectionMethodDescriptor);

    oos.writeObject (cookie);
    oos.flush();
    
//System.out.println ("TCPTransport::endConnection -> reading return object");

    // we wait until the command is executed and read the results of the operation
    Object returnObject = ois.readObject();

    if (returnObject instanceof Exception)
    {
      // when the return object is of type Exception, it means something went wrong on the other
      // side of the connection: we will thus throw the exception so it is caught by the controller:
      // note that we rethrow the exception as it was read, but with the stack trace filled in,
      // since by default stack traces are not serialized along with the Exception object
      if (DEBUG)
        System.out.println ("TCPTransport::endConnection -> throwing Exception: " + returnObject);

      throw (Exception) ((Exception) returnObject).fillInStackTrace();
    }
    else
    {
      if (DEBUG)
        System.out.println ("TCPTransport::endConnection -> return code: " + returnObject);

      return (ServerCommand) returnObject;
    }
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
	{
		return description;
	}
}