package streamredirector;

/**
 * This interface describes a class whose responsibility it is to communicate to the sink proxy and
 * relay a message from the part of the controller (namely, either that it wishes that a connection be
 * established between a specific server and a sink, or that it wishes a specific command to be sent
 * to the sink proxy). The communication between the transport and its counterpart in the proxy (the
 * ListeningTransport), is specific between them: this allows several transport methods to be used
 * between a controller and a sink proxy in a matter that is transparent to the general operation of
 * both the controller and the sink proxy (examples of transport could be TCP sockets, UDP datagrams,
 * RMI, INS late-binding, etc).
 *
 * @see streamredirector.ListeningTransport, streamredirector.TCPTransport
 * @author Jorge Rafael Nogueras
 */

public interface Transport
{
  /**
   * This method will get called to verify that the listening transport counterpart is
   * actually there to receive command requests. Exactly how the listening transport proves that
   * it is there listening is up to the specific implementation.
   *
   * @exception Exception if there are errors sending the ping request
   * @return true if the listening counterpart is active, false otherwise
   */

  public boolean ping() throws Exception;
  
  /**
   * This method will get called when the controller wants to establish a connection.
   * between a server and a sink. It specifies the server where the stream will be coming
   * from: this object is relayed to the corresponding ListeningTransport.
   *
   * @param server embodies the connection with the server
   * @exception Exception if there are errors sending the server connection object or connecting to the listening transport
   * @return a long value representing the cookie that uniquely identifies the connection
   */

	public Long establishConnection (ServerConnection server) throws Exception;

  /**
   * This method will get called when the controller wants to establish a connection
   * between a server and a sink. It specifies the server where the stream will be coming
   * from and the sink where data should be written: these objects are relayed to the corresponding ListeningTransport.
   *
   * @param server embodies the connection with the server
   * @param sink embodies the connection with the sink
   * @exception Exception if there are errors sending the server and sink connection objects or connecting to the listening transport
   * @return a long value representing the cookie that uniquely identifies the connection
   */

	public Long establishConnection (ServerConnection server, SinkConnection sink) throws Exception;
	
  /**
   * This method will get called when the controller wants to send a command. It will send to
   * the corresponding ListeningTransport the cookie that uniquely identifies the connection and
   * the command that should be performed on said connection.
   *
   * @param cookie identifies uniquely the connection where the command will be sent
   * @param command the command that the proxy will be sent
   * @exception Exception if there are errors sending the cookie or the command object or connecting to the listening transport
   * @return a return code indicating success or an error condition
   */
   
	public ReturnCode sendCommand (Long cookie, Command command) throws Exception;

  /**
   * This method will get called when the controller wants a connection lease to be renewed.
   * It will send to the corresponding ListeningTransport the cookie that uniquely identifies
   * the connection whose lease must be renewed.
   *
   * @param cookie identifies uniquely the connection whose lease should be renewed
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a return code indicating success or an error condition
   */

  public ReturnCode renewLease (Long cookie) throws Exception;

  /**
   * This method will get called when the endConnection command is received.
   *
   * @param cookie identifies uniquely the connection that must be terminated
   * @exception Exception if there are errors getting connection state or communication errors
   * @return a ServerCommand that resumes the connection from the point where it was ended
   */

  public ServerCommand endConnection (Long cookie) throws Exception;
}