package streamredirector;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class is spawned by the WebProxyManager to process a connection to a client.
 * It works as a web proxy between the client and the web server: it reads the request
 * written by the client, establishes a connection to the web server asking it for the
 * specified resource and transfers the data from the web server to the client.
 *
 * It keeps track of the state of each connection (which includes the cookie for the connection
 * as well as the offset of the last byte written) and stores it in the (static) connection
 * table in the WebProxyManager class.
 *
 * This web proxy accepts GET requests of the following format:
 *
 *    http://<web-server-host>:<port>/<cookie>/<actual-path>
 *
 * where <web-server-host>:<port> is the address and port of the web server where the resource
 * resides, <cookie> is the numeric cookie that uniquely identifies each connection, and
 * <actual-path> is the path of the desired resource in the web server.
 *
 * This web proxy will strip away the cookie part of the URL and establish a connection to the
 * web server, passing through all the request headers intact.
 *
 * @see streamredirector.WebProxyManager
 * @author Jorge Rafael Nogueras
 */

public class WebProxyInstance implements Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** the size (in bytes) of the buffer where data will be read into */
  protected static final int BUFFER_SIZE = 30 * 1024;
  
  /** where requests coming read from the client are read */
  protected BufferedReader br;
  
  /** where data going to the client are written */
  protected BufferedOutputStream bos;

  /** thread that continues listening for incoming requests */
  protected Thread thread;

  /** this variable, if true, will cause the thread to stop running */
  protected boolean interrupted;
  
  /**
   * Initializes a web proxy instance that will read and process the request from the client
   * connected to the specified socket.
   *
   * @param socket connection from where HTTP requests should be read and data should be written
   * @exception Exception if there are errors getting the input and output streams for the socket
   */

  public WebProxyInstance (Socket socket) throws Exception
  {
    br = new BufferedReader (new InputStreamReader (socket.getInputStream()));
    bos = new BufferedOutputStream (socket.getOutputStream());

    interrupted = false;
    thread = new Thread (this, "WebProxyInstance #" + hashCode());
    thread.start();
  }

  /**
   * This method makes the listening thread stop running if it has not been previously interrupted.
   */
   
  public void interrupt()
  {
    if (!interrupted)
    {
      interrupted = true;
      thread.interrupt();
    }
  }

  /**
   * This is the code that will be run by the listening thread: it blocks while waiting for a new
   * request to arrive and keeps listening until it is interrupted.
   */

  public void run()
  {
    if (DEBUG)
      System.out.println ("WebProxyInstance::run -> Spawned off WebProxyInstance #" + hashCode());

    while (!interrupted)
    {
      try
      {
        // process the request written by the client
        processRequest();
      }
      catch (Exception e)
      {
        // if there was an error writing the output, this means that the connection was probably closed
        // due to an error; in that case, we will exit the loop and let the thread exit normally to
        // shut down
        System.err.println ("WebProxyInstance::run -> Error reading command from client: connection was possibly closed");
//        e.printStackTrace();
        interrupted = true;
        continue;
      }
      finally
      {
        try
        {
          // finally, close the connections to the web server and the client
          br.close();
          bos.close();
        }
        catch (Exception e)
        {
          e.printStackTrace();
        }
      }
    }

    if (DEBUG)
      System.out.println ("WebProxyInstance::run -> Shutting down WebProxyInstance #" + hashCode());
  }

  /**
   * Processes the client's request.
   *
   * @exception Exception if there are errors reading from the web server or writing to the client
   */

  protected void processRequest() throws Exception
  {
//System.out.println ("WebProxyInstance::processRequest -> listening...");
    // parse the request header read from the BufferedReader
    HTTPRequestParser parser = new HTTPRequestParser (br);

    // the resource should be in the following form:
    // http://<host>:<port>/<cookie>/<actual-path>
    String resource = parser.getResource();

//System.out.println ("WebProxyInstance::processRequest -> Method: " + parser.getMethod());
System.out.println ("WebProxyInstance::processRequest -> Get Resource: " + parser.getResource());
//System.out.println ("WebProxyInstance::processRequest -> Properties: " + parser.getHeaders());

    if (DEBUG)
      System.out.println ("WebProxyInstance::processRequest -> Resource: " + parser.getResource());

    if (parser.getMethod().equals ("GET"))
    {
      // we create a URL object with the resource specification and get its
      // full path, which is of the form: /<cookie>/<actual-path>
      URL originalURL = new URL (parser.getResource());
      String fullPath = originalURL.getPath();

//System.out.println ("WebProxyInstance::processRequest -> originalURL = " + originalURL + ", fullPath = " + fullPath);
  
      Long cookie;
      String actualPath;
      
      // we look for the second "/" in the path (that is, we ignore the first one and look
      // for the next one, since the URL will always start with a "/")
      int indexSeparator = fullPath.indexOf ("/", 1);

      // if we cannot find a "/", then the resource name is invalid; we set the cookie to "-1" to
      // signal it as an invalid one, and we assign the full path to be the actual path
      if (indexSeparator < 0)
      {
        cookie = new Long (-1);
        actualPath = fullPath;
      }
      else
      {
        try
        {
          // the cookie will be everything up the second "/" (that is, not counting the initial "/":
          // this is why we start from offset 1 into the string)
          cookie = new Long (fullPath.substring (1, indexSeparator));
  
          // we assume that from the second "/" to the end of the string it is the actual resource path
          actualPath = fullPath.substring (indexSeparator);
        }
        catch (Exception e)
        {
          // if there is an error with the parsing (or perhaps because the first element of the
          // path was not a long cookie), we will use values that allow us to continue requesting the
          // resource from the web server: this allows this class to function as a regular web proxy
          cookie = new Long (-1);
          actualPath = fullPath;
        }
      }
//System.out.println ("WebProxyInstance::processRequest -> cookie = " + cookie + ", actualPath = " + actualPath + ", full path = " + fullPath);

      // we find out if this is a previous interrupted connection being resumed, in which case
      // we will find its state and resume the file from that offset
      WebProxyConnectionState state = WebProxyManager.getWebProxyConnectionState (fullPath);
//System.out.println ("WebProxyInstance::processRequest -> Proxy connection state: " + state);
      
      if (state == null)
      {
        // since this is a new connection, we create a new state object for it, specifying the cookie for this connection,
        // the full path (which includes the cookie and the actual resource path), the offset (which is zero since we want
        // to transfer the file from the beginning), a reference to this WebProxyInstance, and we set the status of the
        // transfer to "IN_PROGRESS"
        state = new WebProxyConnectionState (cookie, fullPath, 0, this, WebProxyConnectionState.IN_PROGRESS);
        
        // we store this state object in the WebProxyManager's list of established connections and we key it using the
        // full path, which is guaranteed unique because it includes the cookie and the name of the resource being downloaded
        WebProxyManager.setWebProxyConnectionState (fullPath, state);
System.out.println ("WebProxyInstance::processRequest -> Due to null state, created new state: " + state);
      }
      else
      {
//System.out.println ("WebProxyInstance::processRequest -> Found resuming state: " + state);
        // we have found a connection state object matching the full path: however, we must still
        // update the connection's state reference to its WebProxyInstance to point to this one
        state.webProxy = this;
        
        // we also reset the status of the connection to "IN_PROGRESS"
        state.status = WebProxyConnectionState.IN_PROGRESS;
      }

      // now we create the URL for the actual resource we will be downloading: we basically construct the URL from the fields
      // of the original URL, with the difference that the cookie has been stripped from its path so the resource path
      // can be correctly located in the web server
      URL resourceURL = new URL (originalURL.getProtocol(), originalURL.getHost(), originalURL.getPort(), actualPath);
      
//System.out.println ("WebProxyInstance::processRequest -> resourceURL = " + resourceURL);

      // we open a connection to this URL (although the request is not yet written out since we still need to pass the request
      // headers from the original request)
      URLConnection connection = resourceURL.openConnection();
      
      // we check if the "Range" property was set in the request header (this would happen, for instance, if
      // the media player is being fast-forwarded or rewound)
      String rangeProperty = parser.getHeaders().getProperty ("Range");
      
//System.out.println ("Range before: " + parser.getHeaders().getProperty ("Range"));
      // if no range property was sent by the media player, we set set the range request to be
      // whatever the current offset is (this is "0" if it is a new transfer)
      if (rangeProperty == null)
      {
        // we will request the resource starting from the offset specified in the connection state object: this means there is
        // automatic resuming of interrupted transfers
        parser.getHeaders().setProperty ("Range", "bytes=" + state.offset + "-");
      }
      else
      {
        // if the range property was already set, we do not change it, and we parse it to reset the current offset to
        // be the offset of the range requested; the format of a range request is as follows:
        // Range: bytes=<starting offset>-<end offset>
        // we will thus set the state's offset to be the starting offset (the number between the "=" and the "-")
        long offset = Long.parseLong (rangeProperty.substring (rangeProperty.indexOf ("=") + 1, rangeProperty.indexOf ("-")));
//System.out.println ("Range offset is: " + offset);
        state.offset = offset;
      }
      
      // now we get a list of the request headers contained in the original request and write them to the request
      // we will send the real web server
      Enumeration list = parser.getHeaders().propertyNames();
      
      while (list.hasMoreElements())
      {
        String requestProperty = (String) list.nextElement();
//System.out.println ("WebProxyInstance::processRequest -> setting request property: " + requestProperty + " to: " + parser.getHeaders().getProperty (requestProperty));
        connection.setRequestProperty (requestProperty, parser.getHeaders().getProperty (requestProperty));
      }
      
      try
      {
//        String header = null;
//System.out.println ("Range after: " + connection.getRequestProperty ("Range"));
//        int i = 0;
//        
//        do
//        {
//          header = connection.getHeaderField (i);
//          System.out.println ("Header " + i + ": " + connection.getHeaderFieldKey (i) + " = " + connection.getHeaderField (i));
//          i++;
//        }
//        while (header != null);
      }
      catch (Exception e)
      {
      }
      
      BufferedInputStream bis = null;
      
      try
      {
        // we finally establish the connection to the web server and write out the request and the request headers
        bis = new BufferedInputStream (connection.getInputStream());
      }
      catch (Exception e)
      {
        // an exception here means that an error has occurred in the establishment of the connection: more likely than
        // not, it means that the resource did not exist in the web server; we set the connection status to "NOT_FOUND"
        state.status = WebProxyConnectionState.NOT_FOUND;
        
        String serverName = connection.getHeaderField ("Server");
//System.out.println ("WebProxyInstance::processRequest -> server = " + serverName);

        // we will write out an appropriate error response header
        Properties headers = new Properties();
        headers.setProperty ("Server", (serverName == null)? "Stream Redirector Local Web Server/1.0": serverName);
        headers.setProperty ("Connection", "close");
        headers.setProperty ("Content-Type", "text/html; charset=iso-8859-1");

        // we send the response header saying that the resource was not found
        writeResponseHeader (404, "Not Found", headers);

        StringBuffer body = new StringBuffer ("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n");
        body.append ("<HTML><HEAD>\n");
        body.append ("<TITLE>404 Not Found</TITLE>\n");
        body.append ("</HEAD><BODY>\n");
        body.append ("<H1>Not Found</H1>\n");
        body.append ("The requested URL ");
        body.append (resource);
        body.append (" was not found on this server.<P>\n");
        body.append ("<HR>\n");
        body.append ("</BODY></HTML>\n");
        
        byte [] bodyBytes = body.toString().getBytes();
        
        bos.write (bodyBytes, 0, bodyBytes.length);

        // we re-throw the exception so it is caught by the main loop in the run() method and all the
        // connections are properly closed
        throw e;
      }
      
      // since the connection was successful, we read the response headers from the web server and write them out
      // to the client
      String serverName = connection.getHeaderField ("Server");
      
      Properties headers = new Properties();
      headers.setProperty ("Server", (serverName == null)? "Stream Redirector Local Web Server/1.0": serverName);
      headers.setProperty ("Connection", "close");
      headers.setProperty ("Content-Length", Integer.toString (connection.getContentLength()));
      headers.setProperty ("Content-Type", connection.getContentType());
  
      // we send the response header saying that the resource was found
      writeResponseHeader (200, "OK", headers);

      // now we write the actual resource as the content body of the response, reading from the BufferedInputStream
      // from where we are reading the data from the web server
      writeContentBody (bis, state);
    }
    else
    {
      System.err.println ("WebProxyInstance::processRequest -> Received unsupported method: " + parser.getMethod());
    }
  }
  
  /**
   * Writes a response header to the client on its output stream.
   *
   * @param code numeric HTTP response code
   * @param explanation the textual explanation of the response code
   * @param headers the response headers that will be included in the response
   * @exception Exception if there are errors writing to the client
   */
   
  protected void writeResponseHeader (int code, String explanation, Properties headers) throws Exception
  {
    // format the full HTTP response status line
    StringBuffer response = new StringBuffer ("HTTP/1.1 ");
    response.append (code);
    response.append (" ");
    response.append (explanation);
    response.append ("\r\n");

    // now add each header specified
    Enumeration headerNames = headers.propertyNames();
    
    while (headerNames.hasMoreElements())
    {
      String name = (String) headerNames.nextElement();
      String value = headers.getProperty (name);
      
      // each header is of the form: <header-name>:<space><header-value>\r\n
      response.append (name);
      response.append (": ");
      response.append (value);
      response.append ("\r\n");
    }

    // terminate the header part of the response
    response.append ("\r\n");

    byte [] responseBytes = response.toString().getBytes();

//System.out.println ("WebProxyInstance::writeResponseHeader -> Writing out response header: " + response);

    bos.write (responseBytes, 0, responseBytes.length);
    bos.flush();
  }
  
  /**
   * Writes the content body to the client on its output stream.
   *
   * @param bis the BufferedInputStream to the web server from where the data will be read
   * @param state the state of this connection whose "offset" value must be updated
   * @exception Exception if there are errors reading from the web server or writing to the client
   */

  protected void writeContentBody (BufferedInputStream bis, WebProxyConnectionState state) throws Exception
  {
    // this is the buffer where data will be read
    byte [] buffer = new byte [BUFFER_SIZE];
    int bytesRead = 0;
    
    // the total number of bytes that have been read will be initialized with the current offset
    long bytesTotal = state.offset;

    try
    {
      // we will read until we read the end of the resource or if the thread is interrupted
      while (!interrupted && (bytesRead = bis.read (buffer)) != -1)
      {
        bytesTotal += bytesRead;
        
        if (DEBUG)
          System.out.println ("WebProxyInstance::writeContentBody -> wrote out: " + bytesRead + " bytes, " + bytesTotal + " total");
//System.out.println ("WebProxyInstance::writeContentBody -> writing: " + new String (buffer));
        bos.write (buffer, 0, bytesRead);

        state.offset = bytesTotal;
      }
    }
    catch (Exception e)
    {
// <HERE> this catch clause is for debug purposes only

      if (DEBUG)
        System.out.println ("WebProxyInstance::writeContentBody -> caught exception while writing file: " + e);
//e.printStackTrace();
throw e;

    }
    finally
    {
      // if we have reached the end of the file, we mark its status as "FINISHED"
      if (bytesRead == -1)
      {
        state.status = WebProxyConnectionState.FINISHED;
      }
      else
      {
        // otherwise we simply mark the connection as "STOPPED"
        state.status = WebProxyConnectionState.STOPPED;
      }
      
      if (DEBUG)
        System.out.println ("WebProxyInstance::writeContentBody -> stopping connection when interrupted: " + interrupted + ", bytesRead: "  + bytesRead + " and bytesTotal: "  + bytesTotal);

      // finally, we flush the data and close the connection to the web server
      bos.flush();
      bis.close();
      
      synchronized (state.webProxy.thread)
      {
        thread.notifyAll();
      }

    }
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "#" + hashCode() + ": running = " + !interrupted + "]";
  }
}