package streamredirector;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * This class starts up a private web proxy: it will establish connections to the desired
 * web server and pass the data through to the client.
 *
 * It listens on the specified port and spawns off a new WebProxyInstance to handle each
 * connection request received.
 *
 * All its methods are static, and before using any of them, the initialize()
 * method must be called. If a method is called before the class is initialized,
 * an UninitializedManagerException will be thrown.
 *
 * @see streamredirector.WebProxyConnectionState, streamredirector.WebProxyInstance
 * @author Jorge Rafael Nogueras
 */

public class WebProxyManager
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** this flag is true after the class has been properly initialized */
  protected static boolean initialized = false;

  /** the port where the web proxy is running */
  protected static int localPort;

  /** the table that indexes the WebProxyConnectionState objects created (keyed by the cookie
      the connection they represent and whose value is the WebProxyConnectionState object
      that describes that connection) */
  protected static Hashtable connectionTable;
  
  /**
   * Initializes the WebProxyManager so it can be used statically from then on,
   * using the default resource directory.
   * This initialization method must be called before any other static method is used.
   *
   * @exception Exception if there are problems creating or accesing the resource directory
   */

  public static void initialize() throws Exception
  {
    initialize (0);
  }
  
  /**
   * Initializes the WebProxyManager so it can be used statically from then on.
   * This initialization method must be called before any other static method is used.
   *
   * @param the_directoryName the directory where the downloaded resources will be kept
   * @exception Exception if there are problems creating or accesing the resource directory
   */

  public static void initialize (int the_localPort) throws Exception
  {
    if (!initialized)
    {
      initialized = true;
    }
    
    connectionTable = new Hashtable();
    
    // bind the listening socket to a random port
// <HERE> uncomment this
//    final ServerSocket listeningSocket = new ServerSocket (the_localPort);
    final ServerSocket listeningSocket = new ServerSocket (8080);
    localPort = listeningSocket.getLocalPort();

    // create a thread that listens for incoming connections and spawns off a web proxy instance for each
    new Thread()
    {
      public void run()
      {
        while (true)
        {
          try
          {
            if (DEBUG)
              System.out.println ("WebProxyManager::run -> Waiting for HTTP proxy connection on port: " + localPort + "...");

            Socket socket = listeningSocket.accept();
            new WebProxyInstance (socket);

//            new WebProxyInstance (listeningSocket.accept());
//System.err.println ("WebProxyManager::run -> Accepted a connection!");
          }
          catch (Exception e)
          {
            e.printStackTrace();
            // if there was an error accepting a new socket connection, this is probably an unrecoverable error
            // in that case, we will exit the loop and let the thread exit normally to shut down the manager
            if (DEBUG)
              System.out.println ("WebProxyManager::run -> Error accepting a new web proxy connection: " + e);
            return;
          }
        }
      }

    }.start();
  }
  
  /**
   * Queries the web proxy manager for the IP of the host where it is running.
   *
   * @return the IP of the host where the web proxy is running as a string, or "127.0.0.1" if an error occurs
   */
   
  public static String getLocalHost()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getLocalHost() without initializing the WebProxyManager");
    }

    try
    {
      return InetAddress.getLocalHost().getHostAddress();
    }
    catch (Exception e)
    {
      return "127.0.0.1";
    }
  }

  /**
   * Queries the web proxy manager for the port where it is listening for incoming connections.
   *
   * @return the local port where the web proxy is listening for incoming connections
   */
   
  public static int getLocalPort()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getLocalPort() without initializing the WebProxyManager");
    }

    return localPort;
  }

  /**
   * Queries the web proxy manager for all the connections currently established.
   *
   * @return a list of the WebProxyConnectionState objects representing current connections
   */
   
  public static Enumeration getConnections()
  {
    return connectionTable.elements();
  }
  
  /**
   * Queries the connection table for the state of the connection matching the specified
   * identifier.
   * The unique identifier is formed by the full path of the resource being downloaded,
   * which is of the form:
   *   /<cookie>/<filename>
   * where "cookie" is the number that uniquely identifies the connection between the stream
   * server and the sink, and "filename" is the name of the resource being downloaded.
   * Note that it is synchronized to avoid race conditions when dealing with the connection
   * table from different threads.
   *
   * @param identifier the unique identifier used to find the desired connection state
   * @return the WebProxyConnectionState object that matches the specified identifier
   */

  public static synchronized WebProxyConnectionState getWebProxyConnectionState (String identifier)
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getWebProxyConnectionState() without initializing the WebProxyManager");
    }

    return (WebProxyConnectionState) connectionTable.get (identifier);
  }
  
  /**
   * Adds the specified connection state keyed by the specified identifier.
   * The unique identifier is formed by the full path of the resource being downloaded,
   * which is of the form:
   *   /<cookie>/<filename>
   * where "cookie" is the number that uniquely identifies the connection between the stream
   * server and the sink, and "filename" is the name of the resource being downloaded.
   * Note that it is synchronized to avoid race conditions when dealing with the connection
   * table from different threads.
   *
   * @param identifier the unique identifier used to identify the specified connection state
   * @param state the state of the HTTP connection
   */

  public static synchronized void setWebProxyConnectionState (String identifier, WebProxyConnectionState state)
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call setWebProxyConnectionState() without initializing the WebProxyManager");
    }

    connectionTable.put (identifier, state);
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": local port = " + localPort +
           ", connection table = " + connectionTable + "]";
  }
}