package streamredirector;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class is spawned off by the WebServerManager each time a new connection is received
 * in order to process the request.
 *
 * @see streamredirector.WebServerManager
 * @author Jorge Rafael Nogueras
 */

public class WebServerInstance implements Runnable
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** the size (in bytes) of the buffer where data will be read into */
  protected static final int BUFFER_SIZE = 10 * 1024;
  
  /** where requests coming read from the client are read */
  protected BufferedReader br;
  
  /** where data going to the client are written */
  protected BufferedOutputStream bos;

  /** thread that continues listening for incoming messages */
  protected Thread thread;

  /** this variable, if true, will cause the thread to stop running */
  protected boolean interrupted;
  
  /**
   * Initializes the web server instance with the socket from where the connection
   * was established.
   *
   * @param socket connection from where HTTP requests should be read and data should be written
   * @exception Exception if there are errors getting the input and output streams for the socket
   */

  public WebServerInstance (Socket socket) throws Exception
  {
    br = new BufferedReader (new InputStreamReader (socket.getInputStream()));
    bos = new BufferedOutputStream (socket.getOutputStream());

    interrupted = false;
    thread = new Thread (this, "WebServerInstance #" + hashCode());
    thread.start();
  }

  /**
   * This method makes the listening thread stop running if it has not been previously interrupted.
   */
   
  public void interrupt()
  {
    if (!interrupted)
    {
      interrupted = true;
      thread.interrupt();
    }
  }
  
  /**
   * This is the code that will be run by the listening thread: it blocks while waiting for a new
   * request to arrive and keeps listening until it is interrupted.
   */

  public void run()
  {
    if (DEBUG)
      System.out.println ("WebServerInstance::run -> Spawned off WebServerInstance #" + hashCode());

    while (!interrupted)
    {
      try
      {
        processRequest();
      }
      catch (Exception e)
      {
        // if there was an error writing the output, this means that the connection was probably closed
        // due to an error; in that case, we will exit the loop and let the thread exit normally to
        // shut down
        System.err.println ("WebServerInstance::run -> Error reading command from client: connection was possibly closed");
//        e.printStackTrace();
        interrupted = true;
        continue;
      }
    }

    if (DEBUG)
      System.out.println ("WebServerInstance::run -> Shutting down WebServerInstance #" + hashCode());
  }

  /**
   * Process the request the client has sent to the web server. We will read the request and its headers
   * and return the desired resource to the client.
   *
   * @exception Exception if there are errors writing out the resource to the client
   */

  protected void processRequest() throws Exception
  {
//System.out.println ("WebServerInstance::processRequest -> listening...");
    // we parse the request
    HTTPRequestParser parser = new HTTPRequestParser (br);

    // the resource should be in the following form: /<cookie>/<filename>
    // However, to allow this class to be used as a generic web server, if the resource does not
    // have a cookie value, we will still try to serve out the resource from our local directory
    String resource = parser.getResource();

//System.out.println ("WebServerInstance::processRequest -> Method: " + parser.getMethod());
//System.out.println ("WebServerInstance::processRequest -> Range: " + parser.getHeaders().getProperty ("Range"));

    if (parser.getMethod().equals ("GET"))
    {
      // we create a file object for the concatenation of the directory of the web server where
      // all resources must be placed and the name of the specific resource we have been requested
      File resourceFile = new File (WebServerManager.getResourceDirectory(), resource);
//System.out.println ("WebServerInstance::processRequest -> Resource: " + resource + ", file: " + resourceFile.getAbsolutePath());

      // if we are able to find the specified resource and it is indeed a file, we proceed to send it to the client
      if (resourceFile.isFile())// && WebServerManager.isAccessibleResource (resource))
      {
        Properties headers = new Properties();
        headers.setProperty ("Server", "Stream Redirector Local Web Server/1.0");
        headers.setProperty ("Connection", "close");
        headers.setProperty ("Content-Length", Long.toString (resourceFile.length()));
        headers.setProperty ("Content-Type", guessContentType (resourceFile.getName()));
    
        // we send the response header saying that the resource was found
        writeResponseHeader (200, "OK", headers);
    
        // we write the response header and the content body (the resource that was requested)
        writeContentBody (resourceFile);
      }
      else
      {
        // if the requested resource could not be found, an appropriate error response is sent to the client
        Properties headers = new Properties();
        headers.setProperty ("Server", "Stream Redirector Local Web Server/1.0");
        headers.setProperty ("Connection", "close");
        headers.setProperty ("Content-Type", "text/html; charset=iso-8859-1");

        // we send the response header saying that the resource was not found
        writeResponseHeader (404, "Not Found", headers);

        StringBuffer body = new StringBuffer ("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n");
        body.append ("<HTML><HEAD>\n");
        body.append ("<TITLE>404 Not Found</TITLE>\n");
        body.append ("</HEAD><BODY>\n");
        body.append ("<H1>Not Found</H1>\n");
        body.append ("The requested URL ");
        body.append (resource);
        body.append (" was not found on this server.<P>\n");
        body.append ("<HR>\n");
        body.append ("</BODY></HTML>\n");
        
        byte[] bodyBytes = body.toString().getBytes();
        
        bos.write (bodyBytes, 0, bodyBytes.length);
      }

      // finally, close the connection
      bos.close();
    }
  }
  
  /**
   * Writes a response header to the client on its output stream.
   *
   * @param code numeric HTTP response code
   * @param explanation the textual explanation of the response code
   * @param headers the response headers that will be included in the response
   * @exception Exception if there are errors writing to the client
   */
   
  protected void writeResponseHeader (int code, String explanation, Properties headers) throws Exception
  {
    // format the full HTTP response status line
    StringBuffer response = new StringBuffer ("HTTP/1.1 ");
    response.append (code);
    response.append (" ");
    response.append (explanation);
    response.append ("\r\n");

    // now add each header specified
    Enumeration headerNames = headers.propertyNames();
    
    while (headerNames.hasMoreElements())
    {
      String name = (String) headerNames.nextElement();
      String value = headers.getProperty (name);
      
      // each header is of the form: <header-name>:<space><header-value>\r\n
      response.append (name);
      response.append (": ");
      response.append (value);
      response.append ("\r\n");
    }

    // terminate the header part of the response
    response.append ("\r\n");

    byte[] responseBytes = response.toString().getBytes();

//System.out.println ("WebServerInstance::writeResponseHeader -> Writing out response header: " + response);

    bos.write (responseBytes, 0, responseBytes.length);
    bos.flush();
  }
  
  /**
   * Writes the content body to the client on its output stream.
   *
   * @param resourceFile the resource that will be sent to the client
   * @exception Exception if there are errors reading from the file or writing to the client
   */

  protected void writeContentBody (File resourceFile) throws Exception
  {
    // we open the file we will be reading from
    BufferedInputStream bis = new BufferedInputStream (new FileInputStream (resourceFile), BUFFER_SIZE);
    
//System.out.println ("WebServerInstance::writeContentBody -> guessed type: " + guessContentType (resourceFile.getName()));

    byte[] buffer = new byte [BUFFER_SIZE];
    int bytesRead = 0;
    
    long bytesTotal = 0;

    try
    {
      while (!interrupted && (bytesRead = bis.read (buffer)) != -1)
      {
        bytesTotal += bytesRead;
System.out.println ("WebServerInstance::writeContentBody -> wrote out: " + bytesRead + " bytes, " + bytesTotal + " total");
//System.out.println ("WebServerInstance::writeContentBody -> writing: " + new String (buffer));
        bos.write (buffer, 0, bytesRead);
      }
    }
    catch (Exception e)
    {
// <HERE> this catch clause is for debug purposes only
      if (DEBUG)
        System.out.println ("WebServerInstance::writeContentBody -> caught exception while writing file: " + e);
//e.printStackTrace();
throw e;
    }
    finally
    {
      if (DEBUG)
      {
        System.out.println ("WebServerInstance::writeContentBody -> wrote out bytes, " + bytesTotal + " total");
        System.out.println ("WebServerInstance::writeContentBody -> stopping connection when interrupted: " + interrupted + " and bytesRead: "  + bytesRead);
      }

      // we flush the output stream and close the file's input stream
      bos.flush();
      bis.close();
    }
  }
  
  /**
   * Tries to guess the content (MIME) type of the resource being sent from its extension.
   *
   * @param fileName the file whose content type will be guessed
   * @return the content (MIME) type of the file, or "text/plain" by default
   */
   
  protected String guessContentType (String fileName)
  {
    if (fileName == null)
    {
      return "text/plain";
    }

    // look at the extension of the file name
    int extensionSeparator = fileName.lastIndexOf (".");
    
    // if there was no "." in the filename, guess it is plain text
    if (extensionSeparator < 0)
    {
      return "text/plain";
    }
      
    String extension = fileName.substring (extensionSeparator + 1);

//System.out.println ("WebServerInstance::guessContentType -> examining extension: " + extension);

    if (extension.equals ("mp3") || extension.equals ("mp2"))
    {
      return "audio/mpeg";
    }
    if (extension.equals ("wav"))
    {
      return "audio/x-wav";
    }
    else if (extension.equals ("mpg") || extension.equals ("mpeg"))
    {
      return "video/mpeg";
    }
    else if (extension.equals ("zip"))
    {
      return "application/zip";
    }
    else if (extension.equals ("html") || extension.equals ("htm"))
    {
      return "text/html";
    }
    else if (extension.equals ("txt") || extension.equals ("gz"))
    {
      return "text/plain";
    }
    else
    {
      return "text/plain";
    }
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "#" + hashCode() + ": running = " + !interrupted + "]";
  }
}