package streamredirector;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * This class starts up a web server on a random port. When a connection is detected,
 * a WebServerInstance will be created to deal with that request.
 *
 * All its methods are static, and before using any of them, the initialize()
 * method must be called. If a method is called before the class is initialized,
 * an UninitializedManagerException will be thrown.
 *
 * @author Jorge Rafael Nogueras
 */

public class WebServerManager
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;
  
  /** the default directory where the downloaded resources will be kept */
  protected static final String defaultDirectory = ".";
  
  /** this flag is true after the class has been properly initialized */
  protected static boolean initialized = false;

  /** the port where the web server is running */
  protected static int localPort;

  /** the name of the directory where the downloaded resources will be kept */
  protected static String directoryName;

  /** the object that represents the directory where the downloaded resources will be kept */
  protected static File resourceDirectory;
  
  /** the list of resources that are accessible for download (keys are the string names
      of each pathname and the values are the File objects corresponding to said pathnames
      (where each path is taken to start from the resources directory) */
  protected static Hashtable accessibleResources;
  
  /**
   * Initializes the WebServerManager so it can be used statically from then on, using
   * the default resource directory.
   * This initialization method must be called before any other static method is used.
   *
   * @exception Exception if there are problems creating or accesing the resource directory
   */

  public static void initialize() throws Exception
  {
    initialize (defaultDirectory, new String [] {});
  }
  
  /**
   * Initializes the WebServerManager so it can be used statically from then on, specifying
   * the directory where the resources are kept.
   * This initialization method must be called before any other static method is used.
   *
   * @param the_directoryName the directory where the resources will be kept
   * @param the_accessibleResources the list of resources that are accessible for download
   * @exception Exception if there are problems creating or accesing the resource directory
   */

  public static void initialize (String the_directoryName, String[] the_accessibleResources) throws Exception
  {
    if (!initialized)
    {
      initialized = true;
    }
    
    directoryName = the_directoryName;
    resourceDirectory = new File (the_directoryName);
    
    // if the directory does not exist, create it
    if (!resourceDirectory.isDirectory())
    {
      resourceDirectory.mkdirs();
    }
    
    accessibleResources = new Hashtable();
    
    for (int i = 0; i < the_accessibleResources.length; i++)
    {
//System.err.println ("WebServerManager::<init> -> Adding resource: " + new File (resourceDirectory, the_accessibleResources [i]).getAbsolutePath());
      accessibleResources.put (the_accessibleResources [i], new File (resourceDirectory, the_accessibleResources [i]));
    }
    
//System.err.println ("WebServerManager::<init> -> Serving resources from directory: " + resourceDirectory.getAbsolutePath());
//System.err.println ("WebServerManager::<init> -> Serving the following resources: " + accessibleResources);

    // bind the listening socket to a random port
    final ServerSocket listeningSocket = new ServerSocket (0);
//    final ServerSocket listeningSocket = new ServerSocket (8080);

    localPort = listeningSocket.getLocalPort();

    // create a thread that listens for incoming connections and spawns off a web server instance for each
    new Thread()
    {
      public void run()
      {
        while (true)
        {
          try
          {
System.err.println ("WebServerManager::run -> Waiting for web connection on port: " + localPort + "...");
            new WebServerInstance (listeningSocket.accept());

//            new WebServerInstance (listeningSocket.accept());
//System.err.println ("WebServerManager::run -> Accepted a connection!");
          }
          catch (Exception e)
          {
            e.printStackTrace();
            // if there was an error accepting a new socket connection, this is probably an unrecoverable error
            // in that case, we will exit the loop and let the thread exit normally to shut down the manager
            System.err.println ("WebServerManager::run -> Error accepting a new web server connection: " + e);
            return;
          }
        }
      }

    }.start();
  }
  
  public static boolean isAccessibleResource (String resource)
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call isAccessibleResource() without initializing the WebServerManager");
    }

    File verifyingResourceFile = new File (resourceDirectory, resource);
//System.err.println ("WebServerManager::isAccessibleResource -> Verifying the resource: " + verifyingResourceFile.getAbsolutePath() + "...");

    Enumeration list = accessibleResources.elements();

    while (list.hasMoreElements())
    {
//      File f = (File) list.nextElement();
//System.err.println ("WebServerManager::isAccessibleResource -> Seeing if it equals: " + f);
      if (verifyingResourceFile.equals ((File) list.nextElement()))
      {
        return true;
      }
    }
    
    return false;
  }
  
  /**
   * Queries the web server manager for the IP of the host where it is running.
   *
   * @return the IP of the host where the web server is running as a string, or "127.0.0.1" if an error occurs
   */
   
  public static String getLocalHost()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getLocalHost() without initializing the WebServerManager");
    }

    try
    {
      return InetAddress.getLocalHost().getHostAddress();
    }
    catch (Exception e)
    {
      return "127.0.0.1";
    }
  }

  /**
   * Queries the web server manager for the port where it is listening for incoming connections.
   *
   * @return the local port where the web server is listening for incoming connections
   */
   
  public static int getLocalPort()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getLocalPort() without initializing the WebServerManager");
    }

    return localPort;
  }

  /**
   * Queries the web server manager for the directory where the downloaded resources will be kept.
   *
   * @return the directory where the downloaded resources will be kept
   */
   
  public static File getResourceDirectory()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getResourceDirectory() without initializing the WebServerManager");
    }

    return resourceDirectory;
  }

  /**
   * Queries the web server manager for the list of resources that are accessible for download.
   *
   * @return the list of resources that are accessible for download
   */
   
  public static Hashtable getAccessibleResources()
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getAccessibleResources() without initializing the WebServerManager");
    }

    return accessibleResources;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": resource directory = " + directoryName +
           ", accessible resources = " + accessibleResources + "]";
  }
  
  public static void main (String args[]) throws Exception
  {
    String serverDirectory = ".";
    String fileServed = "temp.txt";
    
    try
    {
      WebServerManager.initialize (serverDirectory, new String [] {fileServed});
    }
    catch (Exception e)
    {
      System.err.println (e);
      System.exit (0);
    }
          
    String webServerHost = WebServerManager.getLocalHost();
    int webServerPort = WebServerManager.getLocalPort();

    System.out.println ("WebServerManager::main -> Started web server on: " + webServerHost + " port " + webServerPort);
    System.out.println ("WebServerManager::main -> Web server directory is: " + new File (serverDirectory).getAbsolutePath());
    System.out.println ("WebServerManager::main -> Serving file: " + fileServed);
  }
}