/*
 * BookmarkPanel.java
 *
 * Bookmark Panel used by ControllerGUI to add/select
 * server resources (URLs) 
 */

package streamredirector.gui;

import streamredirector.*;

import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;


/**
 *
 * @author  Kalpak Kothari
 */
public class BookmarkPanel extends JFrame implements ActionListener {

    // Variables declaration
    private URL curServerEntry = null;
    private ResourceEntry curResourceEntry;
    private ControllerGUI controllerGUI;

    /** this variable sets whether or not debug println's are printed */
    protected static final boolean DEBUG = Controller.DEBUG;

    // GUI variables
    private JScrollPane treeScrollPane;
    private JTree bookmarkJTree;
    private DefaultMutableTreeNode top;
    private DefaultTreeModel treeModel;

    private JScrollPane detailScrollPane;
    private JTextArea detailTextArea;
    private JPanel commandPanel;
    private JButton applyButton;
    private JButton closeButton;
    private JButton editButton;


    // choose JTree line styles.  Possible values are
    //"Angled", "Horizontal", and "None" (the default).
    private boolean playWithLineStyle = false;
    private String lineStyle = "Angled";
    
    // variable used for double-click events.  if double-click occurs
    // and nodeIsLeaf==true, we simulate "Apply" button action.
    private boolean nodeIsLeaf = false;

    // End of variables declaration

    /** Creates new form BookmarkPanel */
    public BookmarkPanel() {
	initBookmarks();
        initComponents();
    }

    public BookmarkPanel(ControllerGUI connGUI) {
	controllerGUI = connGUI;
	initBookmarks();
        initComponents();
    }

    /** This method is called from within the constructor to
     * initialize the bookmark database.
     */
    private void initBookmarks() {
	try {
	    ServerDatabaseManager.initialize("serverDB.dat");
	}
	catch (Exception e) {
	    System.out.println("Controller::main -> Error creating database manager: " + e);
	}
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     */
    private void initComponents() {
        setTitle("ControllerGUI:: Bookmarks");
        getContentPane().setLayout(new GridBagLayout());

        treeScrollPane = new JScrollPane();	
        detailScrollPane = new JScrollPane();	
        detailTextArea = new JTextArea();
        commandPanel = new JPanel();
        applyButton = new JButton("Apply");
        closeButton = new JButton("Close");
        editButton = new JButton("Edit");

        GridBagConstraints gridBagConstraints5;


	/*** begin bookmarkJTree configuration */

	// Create the nodes
        top = new DefaultMutableTreeNode("URLs");
        createNodes(top);
	treeModel = new DefaultTreeModel(top);

	// Create a tree that allows one selection at a time
        bookmarkJTree = new JTree(treeModel);
	bookmarkJTree.setFont(new Font ("Dialog", 0, 10));
	bookmarkJTree.getSelectionModel().setSelectionMode
	    (TreeSelectionModel.SINGLE_TREE_SELECTION);
	
	bookmarkJTree.setCellRenderer(new MyRenderer());
	bookmarkJTree.setRootVisible(false);

	// Listen for when the selection changes
        bookmarkJTree.addTreeSelectionListener(new TreeSelectionListener() {
		public void valueChanged(TreeSelectionEvent e) {
		    DefaultMutableTreeNode node = (DefaultMutableTreeNode)
			bookmarkJTree.getLastSelectedPathComponent();
		    
		    if (node == null) {
			nodeIsLeaf = false;
			return;
		    }

		    Object nodeInfo = node.getUserObject();
		    if (node.isLeaf()) {
			//TreePath path = bookmarkJTree.getSelectionPath();
			//DefaultMutableTreeNode parent = (DefaultMutableTreeNode)path.getPathComponent(1);
			//curServerEntry = ((ServerNode)parent.getUserObject()).serverURL;
			//curResourceEntry = ((ResourceNode)nodeInfo).resourceEntry;
			curServerEntry = ((ServerNode)nodeInfo).serverURL;
			curResourceEntry = ((ServerNode)nodeInfo).resourceEntry;
			nodeIsLeaf = true;
			if(DEBUG)
			    System.out.println("BookmarkPanel::valueChanged() - node is Leaf\n" + curServerEntry + curResourceEntry);
			printBookmarkDetails();
			detailScrollPane.repaint();	// added because scrollbars don't refresh properly
			applyButton.setEnabled(true);
		    } else {
			nodeIsLeaf = false;
			if(DEBUG)
			    System.out.println("BookmarkPanel::valueChanged() - node not Leaf\n");
		    }
		}
	    });
	
	if (playWithLineStyle) {
            bookmarkJTree.putClientProperty("JTree.lineStyle", lineStyle);
        }


	MouseAdapter treeMouseListener = new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
		    if (e.getClickCount() == 2 && nodeIsLeaf == true) {
			applyButtonActionPerformed();
		    }
		}
	    };
	
	bookmarkJTree.addMouseListener(treeMouseListener);
	
	/* end bookmarkJTree configuration ***/

	// configure JScrollPane for bookmarkJTree
        treeScrollPane.getViewport().setLayout(new FlowLayout());
        treeScrollPane.setBorder(new TitledBorder(null, "Bookmarks", TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));
        treeScrollPane.setViewportBorder(new EtchedBorder());
        treeScrollPane.setToolTipText("Select/Edit Bookmarks");
        treeScrollPane.setAutoscrolls(true);
	treeScrollPane.setPreferredSize(new Dimension(220, 140));
        treeScrollPane.getViewport().add(bookmarkJTree);

	// configure detailScrollPane and detailTextArea	
        detailScrollPane.setPreferredSize(new Dimension(220, 110));
        detailScrollPane.setMinimumSize(new Dimension(220, 110));
        detailScrollPane.setBorder(new TitledBorder(null, "Details", TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));
        detailScrollPane.setViewportBorder(new EtchedBorder());
        detailScrollPane.setAutoscrolls(true);
	
        detailTextArea.setLineWrap(true);
	//	detailTextArea.setWrapStyleWord(true);
        detailTextArea.setEditable(false);
        detailTextArea.setFont(new Font ("Dialog", 0, 10));
        detailScrollPane.setViewportView(detailTextArea);

        // configure and add edit/apply/cancel buttons to commandPanel
        editButton.setFont(new Font ("Dialog", 0, 10));
	editButton.setMargin(new Insets(2,1,2,1));
        editButton.setToolTipText("Edit Bookmarks");
	editButton.setMnemonic(KeyEvent.VK_E);
	editButton.setActionCommand("editbutton");
	editButton.addActionListener(this);


        applyButton.setFont(new Font ("Dialog", 0, 10));
	applyButton.setMargin(new Insets(2,1,2,1));
	applyButton.setToolTipText("Apply selection");
	applyButton.setMnemonic(KeyEvent.VK_A);
	applyButton.setEnabled(false);
	applyButton.setActionCommand("applybutton");
	applyButton.addActionListener(this);

        closeButton.setFont(new Font ("Dialog", 0, 10));
	closeButton.setMargin(new Insets(2,1,2,1));
        closeButton.setToolTipText("Close Window");
	closeButton.setMnemonic(KeyEvent.VK_C);
	closeButton.setActionCommand("closebutton");
	closeButton.addActionListener(this);

	commandPanel.add(editButton);
	commandPanel.add(applyButton);
        commandPanel.add(closeButton);
        
	// add treeScrollPane
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 0;
        gridBagConstraints5.anchor = GridBagConstraints.NORTH;
        getContentPane().add(treeScrollPane, gridBagConstraints5);

	// add detailScrollPane
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 1;
        getContentPane().add(detailScrollPane, gridBagConstraints5);

	// add commandPanel
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 2;
        getContentPane().add(commandPanel, gridBagConstraints5);

	pack();
    }

    private void printBookmarkDetails() {
	detailTextArea.setText("URL: " + curServerEntry + "\n");
	detailTextArea.append("Size: " + 
			      ((curResourceEntry.getSize() == null) ? "Unknown" : curResourceEntry.getSize()) + "\n");
	detailTextArea.append("Description: " + 
			      ((curResourceEntry.getDescription() == null) ? "Unknown" : curResourceEntry.getDescription()) + "\n");
	detailTextArea.append("MIME Type: " + 
			      ((curResourceEntry.getMimeType() == null) ? "Unknown" : curResourceEntry.getMimeType()) + "\n");
    }

    public void actionPerformed(ActionEvent e) {
	// GUI event handler
	if(e.getActionCommand().equals("applybutton")) {
	    applyButtonActionPerformed();
        }
	else if(e.getActionCommand().equals("closebutton")) {
	    this.hide();
	}
	else if(e.getActionCommand().equals("editbutton")) {
	    editButtonActionPerformed();
	}
    }

    private void applyButtonActionPerformed() {
	// set the server URL in controllerGUI
	//    controllerGUI.setServerURL(url);
	Runnable doSetServerURLAndConnect = new Runnable() {
		public void run() {
		    controllerGUI.setServerURLAndConnect(curServerEntry);
		    System.out.println("BookmarkPanel::applyButtonActionPerformed -> Initiating connect: "
				       + Thread.currentThread());
		}
	    };

	try {
	    SwingUtilities.invokeLater(doSetServerURLAndConnect);
	} catch(Exception e) {
	    System.out.println ("BookmarkPanel::applyButtonActionPerformed -> Error in invokeLater(): " + e);
	    if (DEBUG)
		e.printStackTrace();
	}
	applyButton.setEnabled(false);
    }

    private void editButtonActionPerformed() {
	if(curServerEntry != null)
	    new BookmarkEditPanel(this, curServerEntry).show();
	else new BookmarkEditPanel(this).show();
    }

    public void refreshBookmarks() {
	// refresh bookmark list
	initBookmarks();
	top = new DefaultMutableTreeNode("URLs");
        createNodes(top);
	treeModel = new DefaultTreeModel(top);
	bookmarkJTree.setModel(treeModel);
    }

    /** method to create nodes for bookmarkJTree */
    private void createNodes(DefaultMutableTreeNode top) {
	DefaultMutableTreeNode host = null;
	DefaultMutableTreeNode resource = null;
	Hashtable allEntries = new Hashtable();

	Iterator e = ServerDatabaseManager.getURLs();
	Iterator f = ServerDatabaseManager.getResourceEntries();

	while(e.hasNext()) {
	    URL serverURL = (URL)e.next();
	    ResourceEntry resourceEntry = (ResourceEntry)f.next();
	    host = new DefaultMutableTreeNode(new ServerNode(serverURL, resourceEntry));
	    top.add(host);
	}
    }
    
    private class ServerNode {
	private String serverName;
	private URL serverURL;
	private ResourceEntry resourceEntry;
	
	public ServerNode(URL url, ResourceEntry resource) {
	    serverURL = url;
	    resourceEntry = resource;
	}
	
	public String toString() {
	    return resourceEntry.getDescription();
	}
    }


    class MyRenderer extends DefaultTreeCellRenderer {
	ImageIcon audioIcon;
	ImageIcon videoIcon;

	public MyRenderer() {
	    // images courtesy of OmniHTTP web server
	    audioIcon = new ImageIcon("audio.gif");
	    videoIcon = new ImageIcon("video.gif");
	}

	public Component getTreeCellRendererComponent(JTree tree,
						      Object value,
						      boolean sel,
						      boolean expanded,
						      boolean leaf,
						      int row,
						      boolean hasFocus) {

	    super.getTreeCellRendererComponent(tree, value, sel,
					       expanded, leaf, row,
					       hasFocus);

	    if(leaf) {
		switch (fileType(value)) {
		case 1:
		    setIcon(audioIcon);
		    break;
		case 2:
		    setIcon(videoIcon);
		    break;
		default:
		    break;
		}
	    }
	    return this;
	}

	protected int fileType(Object value) {
	    DefaultMutableTreeNode node =
		(DefaultMutableTreeNode)value;
	    ServerNode nodeInfo = (ServerNode)(node.getUserObject());
	    String filename = (nodeInfo.serverURL).getPath();
	    
	    if ((filename.indexOf(".mpg") >= 0) ||
		(filename.indexOf(".mpeg") >= 0) ||
		(filename.indexOf(".avi") >= 0)) {
		return 2;
	    }
	    else if((filename.indexOf(".mp3") >= 0) ||
		    (filename.indexOf(".wav") >= 0) ||
		    (filename.indexOf(".mp2") >= 0)) {
		return 1;
	    }
	    
	    return 0;
	}
    }
}


