package streamredirector.player;

//import java.util.*;

/**
 * This class is used to get a media player (although technically it is useful to run
 * any executable) to play a file. It uses the Runtime class to execute the external
 * program specified by the executable path and invokes it along with the specified
 * file name in the play() method.
 *
 * A player can only be created through the static getInstance() method, which will keep
 * track of all the players previously created and will return a previously-instantiated
 * player if one with the same executable path is requested. Thus, if the same player is used to
 * play different media types, there will be only one instance of said media player that
 * will be shared among the different types.
 *
 * It can be used as-is for generic players, but can be extended to have more player-
 * specific commands if so desired.
 *
 * This class gets automatically instantiated using reflection in the PlayerManager
 * and gets used from within FileSinkConnection to actually play the file.
 *
 * @see streamredirector.FileSinkConnection, streamredirector.PlayerManager
 * @author Jorge Rafael Nogueras
 */

public class ExecutablePlayer extends Player
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** the reference to the only instance of this class that will be returned */
  protected static ExecutablePlayer thisPlayer;
  
  /** reference to the Runtime class which allows running external programs */
  protected final static Runtime runtime = Runtime.getRuntime();
  
  /** the absolute path and filename to the executable */
  protected String executablePath;
  
  /**
   * Initializes a player with the specified executable path: it must be called
   * by subclasses to initialize the path. It is protected so a Player may only be
   * instantiated from within this class or any subclass using the getInstance() method.
   *
   * @param the_executablePath the executable path to the media player
   */

  protected ExecutablePlayer (String the_executablePath)
  {
    executablePath = the_executablePath;
  }
  
  /**
   * Gets an instance of the player with the specified path; in the case of this simple type of
   * player, the parameters are simply the path to the executable; however, subclasses
   * may overload this method and receive all the parameters in one string and then
   * parse the string to separate the executable path from the other parameters.
   * This method is the one that gets called when this class is instantiated using
   * reflection in the PlayerManager; therefore, subclasses must always overload this method.
   *
   * @param executablePath the path to the executable
   * @exception Exception provided in case any subclass should want to throw an Exception
   */

  public static Player getInstance (String executablePath) throws Exception
  {
    if (thisPlayer == null)
    {
      thisPlayer = new ExecutablePlayer (executablePath);
    }
    
    return thisPlayer;

//    Player thePlayer = (Player) playerList.get (executablePath);
//    
//    // if no player was found, create one and put it in the list
//    if (thePlayer == null)
//    {
//      thePlayer = new ExecutablePlayer (executablePath);
//      playerList.put (executablePath, thePlayer);
//    }
//
//    return thePlayer;
  }
  
  /**
   * Plays the file identified by the locator by executing the player, passing the locator as a
   * command-line parameter. We assume that the executable player can be started by simply executing
   * its binary executable and passing it as a command line the locator pointing to the media file.
   * The offset parameter is ignored by default: if a specific executable player can use this value
   * as a command-line parameter, a subclass to ExecutablePlayer must be created that knows specifically
   * how to specify these parameters.
   *
   * @param mediaLocator identifies the location of the media resource to play
   * @param offset the byte offset from where the media file should begin playing (ignored by default)
   * @exception Exception if there are errors executing the program
   */

  public void play (String mediaLocator, long offset) throws Exception
  {
    if (DEBUG)
      System.out.println ("ExecutablePlayer::play -> Player " + hashCode() + " executing: " + executablePath + " " + mediaLocator);

    execute (executablePath + " " + mediaLocator);
  }

  /**
   * Executes the specified command as an external program using the Runtime class.
   *
   * @param command the command to be exected as an external program
   * @exception Exception if there are errors executing the command
   * @return the Process class generated by the exec() command
   */

  protected Process execute (String command) throws Exception
  {
    return runtime.exec (command);
  }

  /**
   * Does nothing by default since the ExecutablePlayer cannot be stopped programmatically:
   * subclasses may overload this method if it is possible to stop the player.
   *
   * @exception Exception if there are errors stopping the player
   */

  public void stop() throws Exception
  {
    if (DEBUG)
      System.out.println ("ExecutablePlayer::stop -> Player " + hashCode() + " stopping file");
  }

  /**
   * Returns -1 since the ExecutablePlayer cannot determine the offset programmatically:
   * subclasses may overload this method if it is possible to get the offset.
   *
   * @exception Exception if there are errors retrieving the offset
   */

  public long getLastByteOffset() throws Exception
  {
    long lastByteOffset = -1L;
    
    if (DEBUG)
      System.out.println ("ExecutablePlayer::getLastByteOffset -> Player " + hashCode() + " returning offset: " + lastByteOffset);

    return lastByteOffset;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "#" + hashCode() + ": executable path = " + executablePath + "]";
  }
}