package streamredirector.player;

//import java.util.*;

/**
 * This class is used to execute an MP3 player, although technically it is useful to run
 * any executable that may require additional command-line parameters. It uses the Runtime
 * class to execute the external program specified by the executable path and invokes it
 * along with the specified file name in the play() method.
 *
 * An MP3 player can only be created through the static getInstance() method, which will keep
 * track of all the players previously created and will return a previously-instantiated
 * player if one with the same parameters is requested. Thus, if the same player is used to
 * play different media types, there will be only one instance of said media player that
 * will be shared among the different types.
 *
 * This class gets automatically instantiated using reflection in the PlayerManager
 * and gets used from within FileSinkConnection to actually play the file.
 *
 * @see streamredirector.Player, streamredirector.FileSinkConnection, streamredirector.PlayerManager
 * @author Jorge Rafael Nogueras
 */

public class MP3Player extends ExecutablePlayer
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** any command-line parameters passed along to the player */
  protected String parameters;
  
  /**
   * Initializes an MP3 player with the specified executable path and parameters. If the
   * parameters are null, we will substitute them with an empty string.
   *
   * @param the_executablePath the executable path to the MP3 player
   * @param the_parameters the (optional) parameters to initialize the player
   */

  protected MP3Player (String the_executablePath, String the_parameters)
  {
    super (the_executablePath);
    
    // the parameters should never be null: if they are, we assign them an empty string
    parameters = the_parameters==null? "": the_parameters;
  }
  
  /**
   * Gets an instance of the MP3 player; we will parse the unparsed line and assume everything before
   * the "%" is the path to the executable and everything afterwards are any parameters that the player
   * requires when it is executing. Since the parameters are optional, if no "%" separator character is
   * found, we will assume that the entire unparsed line is the executable path and that there are no
   * parameters.
   *
   * @param unparsedLine contains the executable path and any optional parameters the player needs
   * @exception Exception if there are errors parsing the unparsed line
   */

  public static Player getInstance (String unparsedLine) throws Exception
  {
    String executablePath = null;
    String parameters = null;
    
    int indexOfSpace = unparsedLine.lastIndexOf ("%");
    
    // if there is no "%", we assume no parameters
    if (indexOfSpace == -1)
    {
      executablePath = unparsedLine;
      parameters = "";
    }
    else
    {
      executablePath = unparsedLine.substring (0, indexOfSpace);
      parameters = unparsedLine.substring (indexOfSpace + 1);
    }

//    Player thePlayer = (Player) playerList.get (unparsedLine);
    
    if (thisPlayer == null)
    {
      thisPlayer = new MP3Player (executablePath, parameters);
//      playerList.put (unparsedLine, thePlayer);
    }

    return thisPlayer;
  }
  
  /**
   * Plays the specified file by executing the player and passing it the parameters specified.
   *
   * @param fileName the name of the file to play
   * @exception Exception if there are errors executing the program
   */

  public void play (String fileName) throws Exception
  {
    if (DEBUG)
      System.out.println ("MP3Player::play -> Player " + hashCode() + " executing: " + executablePath + " " + parameters + " " + fileName);
    
    execute (executablePath + " " + parameters + " " + fileName);
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "#" + hashCode() + ": executable path = " + executablePath +
            ", parameters = " + parameters + "]";
  }
}