package streamredirector.player;

//import streamredirector.*;
//import java.net.*;
//import java.io.*;

/**
 * This class extends the abstract Player class. Its function is to create
 * native players written in Java. It keeps a reference to itself, so only one
 * instance of itself will ever be returned by the getInstance() method.
 * It also keeps track of the latest JavaPlayer created: a new JavaPlayer is
 * created for each media file that is played. This mandates that only one
 * media file may be playing at the same time (this restriction may be relaxed
 * if necessary).
 * 
 * @see streamredirector.player.PlayerManager, streamredirector.player.JavaPlayer
 * @author Jorge Rafael Nogueras
 */

public class NativePlayer extends Player
{
  /** this variable sets whether or not debug println's are printed */
  public static final boolean DEBUG = false;

  /** the reference to the only instance of this class that will be returned */
  protected static Player thisPlayer;
  
  /** the reference to the only instance of the actual player that is used at any given time */
  protected static JavaPlayer javaPlayer;
  
  /**
   * Initializes this object.
   */

  protected NativePlayer()
  {
    if (DEBUG)
      System.out.println ("NativePlayer::<init> -> Creating new NativePlayer #" + hashCode());
  }
  
  /**
   * Gets an instance of the native player.
   *
   * @param parameters may be null since no parameters are necessary
   * @exception Exception provided in case any subclass should want to throw an Exception
   */

  public static Player getInstance (String parameters) throws Exception
  {
    if (DEBUG)
      System.out.println ("NativePlayer::getInstance -> Returning new NativePlayer");

    if (thisPlayer == null)
    {
      thisPlayer = new NativePlayer();
    }
    
    return thisPlayer;
  }
  
  /**
   * Plays the file identified by the locator by executing the player, passing the locator as a
   * command-line parameter, and instructing the player to start at the specified offset.
   *
   * @param mediaLocator identifies the location of the media resource to play
   * @param offset the byte offset from where the media file should begin playing
   * @exception Exception if there are errors executing the JavaPlayer
   */

  public void play (String mediaLocator, long offset) throws Exception
  {
    if (DEBUG)
      System.out.println ("NativePlayer::play -> NativePlayer #" + hashCode() + " playing: " + mediaLocator + " from offset: " + offset);
    
    // if a JavaPlayer had been previously created, it must be stopped before creating
    // the new one
    if (javaPlayer != null)
    {
      // <HERE> check this out..
      stop();
    }

    // create a JavaPlayer instance: it will immediately try to play the specified media file
    // from the specified offset
    javaPlayer = new JavaPlayer (mediaLocator, offset);
  }
  
  /**
   * This method stops the media file that is being played and deallocates the JavaPlayer,
   * releasing the hardware resources it had obtained and allowing for the creation of
   * new JavaPlayers.
   *
   * @exception Exception if there are errors stopping or deallocating the player
   */

  public void stop() throws Exception
  {
    if (DEBUG)
      System.out.println ("NativePlayer::stop -> NativePlayer  #" + hashCode() + " stopping file");
      
    javaPlayer.stop();
    javaPlayer.deallocate();
  }

  /**
   * Returns the offset of the last byte played by the JavaPlayer: if for some reason no
   * player had been created, it returns -1.
   *
   * @exception Exception if there are errors retrieving the offset
   */

  public long getLastByteOffset() throws Exception
  {
    if (javaPlayer == null)
    {
      return -1L;
    }
    
    long lastByteOffset = javaPlayer.getLastByteOffset();
    
    if (DEBUG)
      System.out.println ("NativePlayer::getLastByteOffset -> NativePlayer  #" + hashCode() + " returning offset: " + lastByteOffset);

    return lastByteOffset;
  }

  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + "#" + hashCode() + "]";
  }
  
/*
  public static void main (String args []) throws Exception
  {
//    String resource = "http://rafaeln.ne.mediaone.net/~secret/38Special-HoldOnLoosely.mp3";
//    String resource = "http://localhost/song.mp3";
//    String resource = "http://beacon.lcs.mit.edu/speech.mp3";
//    String resource = "http://beacon.lcs.mit.edu/38Special-HoldOnLoosely.mp3";
    String resource = "http://india.mit.edu:80/downloads/foo.mp3";

//    NativePlayer np = new NativePlayer();
//    np.play (resource, 3399945);
//    
//    try { Thread.sleep (5000); } catch (Exception e) { e.printStackTrace(); }
//    
//    np.stop();
//
//    long offset = np.getLastByteOffset(); 
//System.out.println ("NativePlayer::main -> Stopped, got last offset: " + offset);
//
//    np = new NativePlayer();
//    np.play (resource, offset);
//    
//    try { Thread.sleep (15000); } catch (Exception e) { e.printStackTrace(); }
//
//    offset = np.getLastByteOffset();
//System.out.println ("NativePlayer::main -> Stopped, got final offset: " + offset);

    URL resourceURL = new URL (resource);
    NativePlayerServerConnection conn = new NativePlayerServerConnection (new URL ("http://india.mit.edu:80/"));
    conn.establishServerConnection (new Long (0));
    conn.setOutputStream (new FileOutputStream ("caca"));
    conn.sendCommand (new HTTPGetCommand (resourceURL.getPath(), "audio/mpeg"));

    try { Thread.sleep (7000); } catch (Exception e) { e.printStackTrace(); }

    conn.closeServerConnection();
    
    HTTPResumeCommand resume = (HTTPResumeCommand) conn.getResumeCommand();
    conn.sendCommand (resume);
  }
*/
}