package streamredirector.player;

/**
 * This class extends the abstract Player class. Its function is to create
 * native video players written in Java. It keeps a reference to itself, so only one
 * instance of itself will ever be returned by the getInstance() method.
 * It also keeps track of the latest video player created: a new video player is
 * created for each media file that is played. This mandates that only one
 * media file may be playing at the same time (this restriction may be relaxed
 * if necessary).
 * 
 * @see streamredirector.player.PlayerManager, streamredirector.player.VideoPlayer
 * @author Nikos Michalakis
 */

public class NativeVideoPlayer extends Player
{
  /** the reference to the only instance of this class that will be returned */
  protected static Player thisPlayer;
  
  /** the reference to the only instance of the actual player that is used at any given time */
  protected static Cut videoPlayer;

  /**
   * Gets an instance of the native video player.
   *
   * @param parameters may be null since no parameters are necessary
   * @exception Exception provided in case any subclass should want to throw an Exception
   */

  public static Player getInstance (String parameters) throws Exception
  {
    if (thisPlayer == null)
    {
      thisPlayer = new NativeVideoPlayer();
    }
    
    return thisPlayer;
  }

  /**
   * Plays the file identified by the locator by executing the player, passing the locator as a
   * command-line parameter, and instructing the player to start at the specified offset.
   *
   * @param mediaLocator identifies the location of the media resource to play
   * @param offset the byte offset from where the media file should begin playing
   * @exception Exception if there are errors executing the JavaPlayer
   */

  public void play (String locator, long offset) throws Exception
  {
    videoPlayer = new Cut();
    videoPlayer.play (locator,offset);
  }

  /**
   * This method stops the media file that is being played and deallocates the player,
   * releasing the hardware resources it had obtained and allowing for the creation of
   * new players.
   *
   * @exception Exception if there are errors stopping or deallocating the player
   */

  public void stop() throws Exception
  {
    videoPlayer.stop();
  }

  /**
   * Returns the offset of the last byte played by the player: if for some reason no
   * player had been created, it returns -1.
   *
   * @exception Exception if there are errors retrieving the offset
   */

  public long getLastByteOffset() throws Exception
  {
    return videoPlayer.getLastByteOffset();
  }

  public static void main (String args[]) throws Exception
  {
    String inputURL="file:/d|/My%20Documents/INS_V1.0/streaming/pantsoff.avi";

    NativeVideoPlayer p = new NativeVideoPlayer();
    p.play (inputURL,10000000000L);

System.out.println ("About to sleep");
try { Thread.sleep (2000); } catch (Exception e) {}
System.out.println ("Awoken");

    p.stop();
    long o = p.getLastByteOffset();
    
try { Thread.sleep (2000); } catch (Exception e) {}

    p = new NativeVideoPlayer();
    p.play (inputURL, o);
  }
}