package streamredirector.player;

import java.util.*;

/**
 * This abstract class must be extended by any class that will create a player.
 * Its main job is to play a media file and stop the player if the user desires.
 * Also, it must be able to query the player for the last byte offset that was
 * played.
 * Subclasses of this class are created using reflection by the PlayerManager class;
 * they should always override the static method getInstance() so that it returns
 * the appropriate subclass (and for efficiency, they should only instantiate one
 * subclass object).
 *
 * @see streamredirector.PlayerManager
 * @author Jorge Rafael Nogueras
 */

public abstract class Player
{
  /**
   * Gets an instance of the player with the specified parameters (encoded in a single
   * string). This string may be null if nothing is needed to instantiate the Player.
   * For memory efficiency, all subclasses must make sure that only one instance of the subclass is
   * created; successive calls to getInstance() should return a reference to the one and
   * only instance.
   * This method is the one that gets called (in the subclass) when it is instantiated using
   * reflection in the PlayerManager.
   * This method should *ALWAYS* be overloaded by the subclass, since this basic implementation always
   * returns a null Player (unfortunately, Java does not allow for abstract static methods, so this was
   * the only way to "make sure" that subclasses have this static method).
   *
   * @param parameters the parameters necessary to instantiate the Player (may be null if nothing is needed)
   * @exception Exception provided in case any subclass should want to throw an Exception
   */

  public static Player getInstance (String parameters) throws Exception
  {
    return null;
  }
  
  /**
   * Plays the specified file, identified by its media locator. A media locator may be a URL,
   * a pathname on the local host, et cetera. By default, we will call the abstract version
   * of this method that also takes an offset, specifying a 0 offset (meaning the start
   * of the file).
   *
   * @param mediaLocator identifies how the player should connect to the desired media file
   * @exception Exception if there are errors playing the media file
   */

  public void play (String mediaLocator) throws Exception
  {
    play (mediaLocator, 0L);
  }

  /**
   * Plays the specified file, identified by its media locator. A media locator may be a URL,
   * a pathname on the local host, et cetera. The player should start playing the file, if
   * possible, from the specified offset.
   *
   * @param mediaLocator identifies how the player should connect to the desired media file
   * @param offset the byte offset from where the media file should begin playing
   * @exception Exception if there are errors playing the media file
   */

  public abstract void play (String mediaLocator, long offset) throws Exception;

  /**
   * Stops playing the media file; does nothing if no media file was being played or if the Player
   * cannot be stopped programmatically.
   *
   * @exception Exception if there are errors stopping the player
   */

  public abstract void stop() throws Exception;

  /**
   * Returns the byte offset of the last byte played by the Player, or -1 if the Player cannot
   * determine said offset programmatically.
   *
   * @exception Exception if there are errors retrieving the offset
   */

  public abstract long getLastByteOffset() throws Exception;
}