package streamredirector.player;

import java.util.*;
import java.lang.reflect.*;
import java.io.*;

import streamredirector.*;

/**
 * This class is in charge of reading a text file containing the list of MIME types and
 * the players (exectuable programs) used for each. It keeps a hashtable of all the MIME
 * types and the Player used for each. A Player may be used for more than one MIME type,
 * but each MIME type can be assigned only one Player.
 *
 * All its methods are static, and before using any of them, the initialize()
 * method must be called. If a method is called before the class is initialized,
 * an UninitializedManagerException will be thrown.
 *
 * @see streamredirector.Player
 * @author Jorge Rafael Nogueras
 */

public class PlayerManager
{
  /** this variable sets whether or not debug println's are printed */
  protected static final boolean DEBUG = false;

  /** the default filename for the player list file */
  protected static final String defaultFileName = "players.dat";
  
  /** this flag is true after the class has been properly initialized */
  protected static boolean initialized = false;

  /** the name of the text file that has the player list information */
  protected static String fileName;
  
  /** the object that maintains the information read from the player list file */
  protected static Properties playerInformation;
  
  /** the table that indexes the Player objects created (keyed by the MIME type and whose
      value is the Player object that should be used to play that MIME type) */
  protected static Hashtable playerList;
  
  /**
   * Initializes the PlayerManager so it can be used statically from then on,
   * using the default player list filename. This initialization method must be called before any
   * other static method is used.
   *
   * @exception Exception if there are problems reading the filename or creating the player list
   */

  public static void initialize() throws Exception
  {
    initialize (defaultFileName);
  }
  
  /**
   * Initializes the PlayerManager so it can be used statically from then on.
   * This initialization method must be called before any other static method is used.
   *
   * @param file_name the name of the text file that has the player list information
   * @exception Exception if there are problems reading the filename or creating the player list
   */

  public static void initialize (String file_name) throws Exception
  {
    if (!initialized)
    {
      initialized = true;
    }

    fileName = file_name;
    playerList = new Hashtable();
    
    // load the player information into the Properties object
    playerInformation = new Properties();
    playerInformation.load (ClassLoader.getSystemResourceAsStream (file_name));
    
    // fill the hashtable with the player information
    createPlayerList();
  }
  
  /**
   * Returns the Player in charge of playing for the specified MIME type, or null if none was
   * found.
   *
   * @param mimeType the MIME type whose player is being returned
   * @return the Player that is in charge of playing the specified MIME type
   */

  public static Player getPlayer (String mimeType)
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getPlayer() without initializing the PlayerManager");
    }

    return (Player) playerList.get (mimeType);
  }
  
  /**
   * Reads in the player list file and creates Player objects that should be used to play the
   * specified MIME types. It puts these Player objects in the player list hashtable, keyed by the MIME type
   * (as a String).
   */
   
  protected static void createPlayerList() throws Exception
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getPlayer() without initializing the PlayerManager");
    }

    // get the list of all the MIME types contained in the player list file
    Enumeration mimeTypeList = playerInformation.keys();

    // now we go through all the MIME types and instantiate the appropriate Player for each
    while (mimeTypeList.hasMoreElements())
    {
      String mimeType = (String) mimeTypeList.nextElement();
      
      // the player information is the value of the property keyed by each MIME type
      // (in terms of the player list file, everything after the "=" is the player information)
      String playerInfo = playerInformation.getProperty (mimeType);
      
      try
      {
        // there should be two pieces of information for each MIME type: the first one (until the first
        // space) is the fully-qualified classname of the Player that will be used to play the specified
        // MIME type, and the second one is a String which is the parameter with which the Player will be
        // instantiated
        int indexOfSeparator = playerInfo.trim().indexOf (" ");
        
        // we read the fully classified name of the player class (all the text until the first space): if
        // only the classname was specified with nothing after it, then we assign the whole line to the variable
        String playerClassName = (indexOfSeparator != -1)? playerInfo.substring (0, indexOfSeparator): playerInfo;
        
        // the parameters used to instantiate the player is the rest of the line; if no space was specified,
        // there are no parameters, so we assign it a null value
        String playerParameters = (indexOfSeparator != -1)? playerInfo.substring (indexOfSeparator + 1): null;

//System.out.println ("PlayerManager::createPlayerList -> Instantiating player: " + playerClassName + " with parameter: " + playerParameters);
/*
        // we get the appropriate Class object using the name of the class and ask it to return
        // the constructor with one String parameter
        Constructor playerClassConstructor = Class.forName (playerClassName).getConstructor (new Class [] {String.class});

//System.out.println ("PlayerManager::createPlayerList -> Constructor: " + playerClassConstructor);
        // now we instantiate a Player object by passing the parameters to its constructor
        Player playerObject = (Player) playerClassConstructor.newInstance (new String [] {playerParameters});
*/
        if (DEBUG)
          System.out.println ("PlayerManager::createPlayerList -> Instantiating player: " + playerClassName + " with parameter: " + playerParameters);

        Method getInstanceMethod = Class.forName (playerClassName).getMethod ("getInstance", new Class [] {String.class});

	//        if (DEBUG)
	//          System.out.println ("PlayerManager::createPlayerList -> Method: " + getInstanceMethod);

        Player playerObject = (Player) getInstanceMethod.invoke (null, new String [] {playerParameters});

	//System.out.println ("PlayerManager::createPlayerList -> Player: " + playerObject);

        // we place the instance of the player in the hashtable, keyed by its MIME type
        playerList.put (mimeType, playerObject);
      }
      catch (Exception e)
      {
        System.err.println ("Error while creating Player object for MIME type " + mimeType + ": " + e);
        e.printStackTrace();
      }
    }
    
    if (DEBUG)
      System.out.println ("PlayerManager::createPlayerList -> Player list: " + playerList);
  }
  
  /**
   * Returns an array of string containing all the MIME types in the player list file.
   *
   * @return an array of string containing all the MIME types in the player list file
   * @exception Exception if there are missing attributes or values in the sink proxy descriptor
   */

  public static String [] getAllMIMETypes() throws Exception
  {
    if (!initialized)
    {
      throw new UninitializedManagerException ("Tried to call getAllMIMETypes() without initializing the PlayerManager");
    }

    Enumeration list = playerList.keys();
    
    String [] mimeTypes = new String [playerList.size()];
    
    for (int i = 0; i < mimeTypes.length; i++)
    {
      mimeTypes [i] = (String) list.nextElement();
    }
    
    return mimeTypes;
  }
  
  /**
   * Prints a useful description of this object.
   *
   * @return a String with a useful description of this object
   */

  public String toString()
  {
    return "[" + this.getClass().getName() + ": player list filename = " + fileName + ", list of players: " + playerList + "]";
  }
}
