/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package viewfinder;

import cricketdaemon.clientlib.data.BeaconRecord;
import cricketdaemon.clientlib.data.Position;
import java.util.*;

/** 
 *  Type representing the coordinate of beacons, services, and/or the
 *  listener device, as determined by the CricketClient.  All
 *  mathematical operations ignore the Z component.  This is useful when
 *  the application is only interested in the (x,y)-position of the
 *  listener device relative to other services and beacons.  
 *
 * @author Allen Miu
 */
public class Coord 
{
    public double x = -1, y = -1;
    public String name; // name of device or client
    public String space;
    public int beaconID;

    public static final double fullCircleInRadians = 2*Math.PI;

    public void setName() { name = space+"-"+beaconID; }
    public void setName(String n) { name = n; }

    public Coord() {}

    /**
     * Constructs a Beacon Coord
     */
    public Coord(double x, double y, String space, int id)
    {
	this.x = x; this.y = y;
	this.space = space; this.beaconID = id;
	setName();
    }

    /**
     * Constructs a Device or a Service Coord.
     */
    public Coord(double x, double y, String name) 
    {
	this.x=x; this.y = y; this.name = name;
	space    = null;
	beaconID = -1;
    }

    // constructs a coordinate with an offset of (r, angle)
    private Coord(double x, double y, double r, double angle, String name) 
    {
	this(x, y, name);
	double x2 = r*Math.cos(angle);
	double y2 = r*Math.sin(angle);
	this.x += x2;
	this.y += y2;
    }

    private Coord(Coord c, double r, double angle)
    {
	this(c.x, c.y, r, angle, c.name);
    }

    public String toString() 
    {
	int ix = (int) x;
	int iy = (int) y;
	String result = name;
	if(ix >= 0 && iy >= 0)
	    result += " @( "+ix+" "+iy+" )";
	return result;
    }

    public boolean equals(Object o)
    {
	if(o == null)
	    return false;

	if(!(o instanceof Coord))
	    return false;

	Coord other = (Coord) o;
	return 
	    (x - other.x < 0.0000001) &&
	    (y - other.y < 0.0000001) &&
	    beaconID == other.beaconID &&	    
	    ((space == null && other.space == null) ||
	     space.equals(other.space)) &&
	    ((name ==  null && other.name == null) ||
	     name.equals(other.name));
    }

    // Input
    // Three points a, b, and c; origin at a.
    // "solA" and "solB" are the valid solutions for the "scalar" angle
    // formed between vectors ab and ac.
    //
    // Returns the correct solution for the angle from ab to ac,
    // positive counterclockwise.
    //  
    public static double uniqueAngle(Coord ab, Coord ac, 
				     double solA, double solB) 
    {
	// Do a positive and a negative angle transform on ab and see
	// which coordinate is closer to ac.  Note that ab and ac may
	// have different lengths so the transforms may not match the
	// coordinates of ac; we simply pick the one that is closer

	Coord solA_ab = angleTransform(ab, solA);
	Coord solB_ab = angleTransform(ab, solB);
	double solA_dist = dist(ac, solA_ab);
	double solB_dist = dist(ac, solB_ab);

	/*
	  System.out.println("uniqueAngle");
	  System.out.println("ab: "+ ab.x + " " + ab.y);
	  System.out.println("ac: "+ ac.x + " " + ac.y);
	  System.out.println("solA_ab: "+ solA_ab.x + " " + solA_ab.y);
	  System.out.println("solB_ab: "+ solB_ab.x + " " + solB_ab.y);
	  System.out.println("dist: "+ solA_dist + " " + solB_dist);
	*/

	// solA angle is closer to ac so the solA angle is the solution
	if(solA_dist < solB_dist)
	    return solA;
	else // vice versa
	    return solB;
    }

    /**
     * Computes the distance between the two coordinate <code>a,
     * b</code>, ignoring the Z component.  
     */
    public static double dist(Coord a, Coord b) 
    {
	double temp = (a.x - b.x)*(a.x - b.x) + (a.y - b.y)*(a.y - b.y);
	return Math.sqrt(temp);
    }

    public double dist(double xx, double yy, double zz)
    {
	double temp = (x - xx)*(x - xx) + (y - yy)*(y - yy) + zz*zz;
	return Math.sqrt(temp);
    }

    public static double dotProduct(Coord a, Coord b) 
    {
	return a.x*b.x + a.y*b.y;
    }

    public static Coord subtract(Coord a, Coord b) 
    {
	return new Coord(a.x - b.x, a.y - b.y, null);
    }

    public static Coord angleTransform(Coord a, double t)
    {
	// angle transform matrix
	// [ cos t  -sin t ]
	// [ sin t   cos t ]
	double x = a.x * Math.cos(t) - a.y * Math.sin(t);
	double y = a.x * Math.sin(t) + a.y * Math.cos(t);
	return new Coord(x, y, null);
    }

    public static double normalizeAngle(double angle)
    {
	angle %= fullCircleInRadians;
	if(angle < 0)
	    angle += fullCircleInRadians;
	return angle;
    }

    // convert absolute orientation to relative orientation
    public static double absToRelOrient(Coord beacon, 
					Coord device, 
					double abs)
    {
	return _convertOrient(beacon, device, abs, true);
    }

    public static double relToAbsOrient(Coord beacon, 
					 Coord device, 
					 double rel)
    {
	return _convertOrient(beacon, device, rel, false);
    }

    public static final Coord zeroZeroCoord = new Coord(0,0,null);
    // may need to adjust the parameter value (50, 50, null)
    // to suit the scale of the map.... FIXME
    public static final Coord globalRefCoord = new Coord(50,50,null);

    private static double _convertOrient(Coord beacon, 
					 Coord device, 
					 double angle, 
					 boolean absToRel)
    {
	double solA;
	double solB;

	Coord absPerpDeviceCoord;
	Coord referenceVector = subtract(beacon, device);
	if(referenceVector.x < 0.000001)
	    referenceVector.x += 0.000001;
	
	if(absToRel) {
	    absPerpDeviceCoord = 
		new Coord(0, 0, Coord.dist(beacon, device), angle, null);
	    solA = angle - Math.atan(referenceVector.y/referenceVector.x);
	}
	else {
	    //	    solA = angle + Math.atan((double) d.y/d.x);
	    // can't do the above because I don't know the exact
	    // vector of the line perpendicular to the Device
	    // So I might as well just do the transform.
	    absPerpDeviceCoord = angleTransform(referenceVector, angle);
	    // now calculate absolute angle using new coordinate
	    solA = Math.atan(absPerpDeviceCoord.y/absPerpDeviceCoord.x);
	    // set reference vector with respect to global ref
	    referenceVector = globalRefCoord;
	}

	// find unique angle
	solA = normalizeAngle(solA);
	solB = normalizeAngle(solA+Math.PI);
	double result = uniqueAngle(referenceVector, absPerpDeviceCoord, 
				    solA, solB);

	/*
	System.out.println("relOrient: "+result+" A "+solA+" B "+solB+
			   " distDD " + dist(zeroZeroCoord, absPerpDeviceCoord) +
			   " distDB " + dist(beacon, device)+
			   " d.x " + referenceVector.x);
	*/
	return result;
    }

    public static ArrayList toCoordArrayList(ArrayList beaconRecords)
    {
	ArrayList result = new ArrayList();
	for(int i = 0; i < beaconRecords.size(); i++) {
	    BeaconRecord br = (BeaconRecord) beaconRecords.get(i);
	    result.add( toCoord(br) );
	}
	return result;
    }

    public static Coord toCoord(BeaconRecord br)
    {
	double x = -1;
	double y = -1;
	if(br.pos != null) {  // FIXME: pos should never be null
	    x = br.pos.x;
	    y = br.pos.y;
	}
	int id = -1;
	if(br.id != null)
	    id = br.id.intValue();
	return new Coord(x, y, br.space, id);
    }

    public static Coord toCoord(Position br)
    {
	return new Coord(br.x, br.y, "device");
    }
}







