/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package viewfinder;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;
import cricketdaemon.clientlib.sim.*;
import cricketdaemon.server.*;
import java.util.*;

/**
 * Implements the ViewFinder application presented in the
 * Mobicom 2001 paper.  Not yet completed.
 *
 * @author Allen Miu
 */
class ViewFinderApp extends Thread implements Callback
{
    public static Random r;

    // screen size
    static final int maxXCoord  = 220;
    static final int maxYCoord  = 220;

    // update occurs only if orientation changes by this angle threshold    
    // static final double delta = Math.PI/16.0; 

    // from asking the server for the service coordinates
    ArrayList serviceCoords;
    ArrayList beaconCoords;

    // from polling
    Coord anchorBeaconCoord;
    Coord deviceCoord;
    double orientation; //in radians

    String currentSpace;
    ArrayList beaconsHeard;

    // user defined parameters
    double defaultViewHalfAngle; // in radians
    double viewingDistSquared;

    // timers
    long orientTimeOut = 0; 
    static final long TIMEOUT = 10000; // used to fire an update when no events come in 

    GUI gui;
    Broker cricket;

    public ViewFinderApp(double angle, double dist, Broker c)
    {
	orientation  = 5*Math.PI; // large value to force first-time update
	defaultViewHalfAngle = angle;
	viewingDistSquared = dist*dist;
	// set up connections, etc...
	cricket = c;
	gui = new GUI("ViewFinder", maxXCoord, maxYCoord, cricket);
    }

    // Queries INS location server for a list of device and their
    // coordinates
    private void getServiceCoords()
    {
	serviceCoords = cricket.requestServiceCoords();
    }

    /* should map BEACONID to BEACON COORD, etc. */
    synchronized public void callback(CricketData data, BitSet mask)
    {
	double oldorientation = orientation;
	Coord oldDeviceCoord  = deviceCoord; 

	//System.out.println(data);

	if(mask.get(Broker.SPACE)) {
	    /* SpaceModule */
	    BeaconRecord cs = data.getCurrentSpaceObject();
	    currentSpace = cs.uniqueName;
	}
	
	if(mask.get(Broker.DEVICECOORD)) {
	    /* Position Module */
	    System.out.println("devicecoord");
	    deviceCoord  = Coord.toCoord(data.getDevicePosition());
	    //notifyAll();
	    doWork();
	}

	if(mask.get(Broker.BEACONSHEARDLIST)) {
	    /* Stat Module */
	    beaconsHeard = Coord.toCoordArrayList(data.getBeaconsHeard());

	    // for simplicity, we assume that the set of beacons indicated
	    // on the "map" is the same as the set of beacons heard.
	    // This way, we don't have to worry about including beacon
	    // information in the "map."
	    gui.setBeacons( beaconsHeard );
	}

	if(mask.get(Broker.ANCHORBEACON)) {
	    /* Orientation Module */
	    anchorBeaconCoord  = Coord.toCoord(data.getAnchorBeacon());
	}

	if(mask.get(Broker.ORIENTATION)) {
	    orientation  = 2*Math.PI - data.getOrient();
	    if(orientation < 0) {
		orientation = 0;
	    }
	    if(cricket instanceof SimBroker)  // simulation mode
		orientTimeOut = Long.MAX_VALUE;
	    else
		orientTimeOut = System.currentTimeMillis() + TIMEOUT;
	    System.out.println("orient");
	    //notifyAll();
	    //doWork();
	}
	
	// fixme: need to calculate position difference too...
	//} while (Math.abs(orientation - oldorientation) < delta);
	//System.out.println("callback");
	//notifyAll();
    }

    private Vector doView(double viewHalfAngle)
    {
	if(deviceCoord == null || serviceCoords == null)
	    //anchorBeaconCoord == null 
	    return null;

	Vector result = new Vector();

	for(int i = 0; i < serviceCoords.size(); i++) {
	    Coord p = (Coord) serviceCoords.get(i);
	    if( inDistance(p) && inAngle(p, viewHalfAngle))
		result.addElement(p);
	}

	return result;
    }

    private boolean inDistance(Coord p) 
    {
	double x = deviceCoord.x;
	double y = deviceCoord.y;
	double dist = (p.x - x)*(p.x - x) + (p.y - y)*(p.y - y);
	return dist <= viewingDistSquared;
    }

    private boolean inAngle(Coord p, double viewHalfAngle)
    {
	//Util.assert(anchorBeaconCoord != null, "ViewFinderApp::inAngle()");
	double ratio = 0;
	Coord i = null;
	Coord j = Coord.subtract(p, deviceCoord);
	double disti = 0;
	double distj = Coord.dist(p, deviceCoord);

	if(anchorBeaconCoord != null) {
	    i = Coord.subtract(anchorBeaconCoord, deviceCoord);
	    disti = Coord.dist(anchorBeaconCoord, deviceCoord);
	}
	else {
	    // assume orientation is absolute angle wrt an anchor that
	    // is at the same y coordinate as the device
	    i = new Coord(10, 0, null);
	    disti = 10;
	}

	ratio  = (double) Coord.dotProduct(i, j);

	if(disti <= 0) {
	    Util.dbgPrintln("ViewFinderApp::inAngle() gives up! AnchorBeacon is directly over device");
	    return false;
	}
	
	if(distj <= 0) // device coincides with service location
	    return true;
	
	ratio /= disti*distj;

	if(ratio > 1) {
	    ratio = 1;
	}
	else if(ratio < -1) {
	    ratio = -1;
	}

	double serviceAngle = Math.acos(ratio);


	//System.out.println("Service "+p.name+" x "+p.x+" y "+
	//p.y+" angle "+serviceAngle+" ratio "+ratio);


	double solutionA = Math.abs(serviceAngle);
	double solutionB = Coord.fullCircleInRadians-solutionA;
	//check the sign of the angle
	serviceAngle = Coord.uniqueAngle(i, j, solutionA, solutionB);
					 
	double upBound    = orientation+viewHalfAngle;
	double lowBound   = orientation-viewHalfAngle;  

	//System.out.println("returned angle "+serviceAngle+" upBound "+upBound+" lowBound "+lowBound);
	
	// now see if the serviceAngle is within viewHalfAngle
	
	boolean inAngle = false;

	Util.assert(serviceAngle >= 0 && serviceAngle <= 
		    Coord.fullCircleInRadians,
		    "invalid ServiceAngle "+ serviceAngle +" ratio "+ratio );
	// the following stuff takes care of wrap-arounds
	if(upBound > Coord.fullCircleInRadians) {
	    upBound -= Coord.fullCircleInRadians;
	    //Util.assert(upBound >= 0, "invalid upBound 1 "+lowBound+" "+upBound);
	    //Util.assert(lowBound >= upBound, "invalid lowBound 2 "+lowBound+" "+upBound);
	    inAngle = serviceAngle <= upBound || serviceAngle >= lowBound;
	}
	else if(lowBound < 0) {
	    lowBound += Coord.fullCircleInRadians;
	    //Util.assert(upBound >= 0, "invalid upBound 3 "+lowBound+" "+upBound);
	    //Util.assert(lowBound >= upBound, "invalid lowBound 4"+lowBound+" "+upBound);
	    inAngle = serviceAngle <= upBound || serviceAngle >= lowBound;
	}
	else {
	    inAngle = serviceAngle <= upBound && serviceAngle >= lowBound;
	}
	 
	return inAngle;
    }

    private void updateGUI(Vector viewable, double viewHalfAngle) 
    {
	// convert orientation relative to beacon into absolute
	// orientation
	double absOrient = 0.0;
	boolean doViewCone = viewHalfAngle < Math.PI;

	if(deviceCoord != null) {
	    if(doViewCone) {
		if(anchorBeaconCoord == null) {
		    absOrient = Coord.relToAbsOrient(new Coord(deviceCoord.x+10, deviceCoord.y, null),
						     deviceCoord, 
						     orientation);
		    //System.out.println(orientation+" "+deviceCoord);
		}
		else {
		    absOrient = Coord.relToAbsOrient(anchorBeaconCoord, 
						     deviceCoord, 
						     orientation);
		    gui.highlightAnchorBeacon(anchorBeaconCoord.name);
		}
	    }

	    gui.drawDevice(deviceCoord.x, deviceCoord.y, absOrient,
			   viewHalfAngle, Math.sqrt(viewingDistSquared),
			   doViewCone);
	}
	if(viewable != null)
	    gui.highlightViewableServices(viewable);
	if(beaconsHeard != null)
	    gui.highlightBeaconsHeard(beaconsHeard);
	if(currentSpace != null)
	    gui.highlightSpaceBeacon(currentSpace);
	gui.repaint();
	
    }

    private void showResults(Vector v) 
    {
	if(v == null)
	    return;
	System.out.println("services in view: "+orientation);
	for(int i = 0; i < v.size(); i++) {
	    Coord p = (Coord) v.elementAt(i);
	    System.out.println("Service "+p.name+" at "+p.x+" "+p.y);
	}
    }

    private void initGUI() 
    {
	gui.setService( serviceCoords );
	if(cricket instanceof SimBroker) // simulation mode
	    gui.setBeacons( ((SimBroker) cricket).requestBeaconCoords() );
    }

    public void init2()
    {
	getServiceCoords();
	initGUI();
    }
    
    synchronized public void run() 
    {
	try {
	    while(true) {
		wait(TIMEOUT);
		doWork();
	    }
	}
	catch(InterruptedException e) {
	    e.printStackTrace();
	    Util.dbgPrintln("ViewFinderApp::run() interrupted");
	}
    }

    private void doWork()
    {
	double viewHalfAngle  = defaultViewHalfAngle;
	
	/*
	  if(orientTimeOut < System.currentTimeMillis()) {
	  //anchorBeaconCoord = new Coord(0,0,"ReferenceAnchor");
	  orientation       = 0;
	  viewHalfAngle     = Math.PI;
	  }
	*/
	
	Vector viewableServices = doView(viewHalfAngle);
	updateGUI(viewableServices, viewHalfAngle);
	System.out.println("############## updateGUI");
	//showResults(viewableServices);
    }
    
    public static void main(String[] args) 
    {
	r = new Random(0);
	Broker s = null;

	if (args.length > 0) {
	    if(args[0].equals("-s"))
	    s = new SimBroker(0, 0, maxXCoord, maxYCoord);
	    else
		s = new ServerBroker(args[0], ServerModule.PORT);
	}
	else {
	    System.err.println("Usage: java viewfinder.ViewFinderApp [-s] addr");
	    System.exit(1);
	}

	BitSet mask = new BitSet();
	mask.set(Broker.SPACE);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.ORIENTATION);
	mask.set(Broker.ANCHORBEACON);
	mask.set(Broker.BEACONSHEARDLIST);

	ViewFinderApp vf = 
	    new ViewFinderApp(Math.PI/8.0, 100.0, s);
	//	    new ViewFinderApp(Math.PI, 75.0, s);
	s.register(vf, mask);
	s.start();
	vf.init2();
	//vf.start();
	//try { vf.join(); } catch (InterruptedException ie) { }
    }
}
