/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.beaconstat;

import cricketdaemon.reader.Beacon;
import cricketdaemon.*;
import java.util.*;


/**
 * A module that keeps track of distance readings for one beacon
 * within a specified window period.
 *
 * @author Allen Miu
 */
public class BeaconDistStat implements Comparable
{
    public Beacon beacon;
    public String key;
    public String space;
    public int id;
    public ArrayList distances;

    /* stats */
    public Stat stat = null;
    public long lastUpdate;

    public double distanceEst = 0;

    public boolean equals(Object o)
    {
	if(!(o instanceof BeaconDistStat))
	    return false;
	BeaconDistStat other = (BeaconDistStat) o;
	return 
	    space.equals(other.space) &&
	    id == other.id;
    }

    /**
     * Note that compareTo == 0 does not imply equals(o) returns true.
     * Two different beacons can be the same distance away.
     */
    public int compareTo(Object o)
    {
	double otherDist = ((BeaconDistStat) o).distanceEst;
	if(distanceEst > otherDist)
	    return 1;
	else if(distanceEst == otherDist)
	    return 0;
	return -1;
    }

    public class BeaconDist
    {
	public Integer dist;
	public Integer duration;
	public Long listenTime;
	public long timeStamp;

	public BeaconDist(Integer d, Integer dn, Long lt, long curTime)
	{
	    dist = d;
	    duration = dn;
	    listenTime = lt;
	    timeStamp = curTime + BeaconStatModule.window;
	}

	public String toString()
	{
	    return dist+" dur="+duration+" listTime="+listenTime+" t="+timeStamp;
	}
    }

    public BeaconDistStat(String k, Beacon b, String s, int i)
    {
	key = k;
	beacon = b; 
	space = s; 
	id = i; 
	distances   = new ArrayList();
    }

    public BeaconDistStat(String k, Beacon b, String s, int i, Integer dist,
			  Integer duration, Long listenTime)
			  
    {
	this(k, b, s, i);
	insert(b, s, i, dist, duration, listenTime);
    }

    public void insert(Beacon b, String s, int i, Integer dist, 
		       Integer duration, Long listenTime)
    {
	space  = s; 
	id     = i; 
	beacon = b;

	BeaconDist r = new BeaconDist(dist, duration, listenTime, b.timeStamp);
	distances.add(r);

	// ewma update
	ewmaDist = ewmaA*dist.intValue() + (1-ewmaA)*ewmaDist;
    }

    //public void recalibrate() { return; }

    public void refresh(long curTime)
    {
	refresh(curTime, distances);
    }

    private void refresh(long curTime, ArrayList buffer)
    {
	while(buffer.size() > 0) {
	    BeaconDist d = (BeaconDist) buffer.get(0);
	    if((d.timeStamp) > curTime)
		break;
	    buffer.remove(0);
	}
    }

    public boolean shouldRemove()
    {
	return distances.size() <= 0; 
    }

    private double ewmaDist = 0;
    private static final double ewmaA = 0.4;

    public BeaconDistStat cloneAndAnalyze(int statSelect, int enoughSample, int binSize)
    {
	if(shouldRemove())
	    return null;

	/* 
	 * received fewer than enoughSamples over the last 
	 * BeaconStatModule.window
	 */
	if(distances.size() < enoughSample)
	    return null;

	BeaconDistStat result = new BeaconDistStat(key, beacon, space, id);

	int[] dists = new int[distances.size()];
	for(int i = 0; i < distances.size(); i++) {
	    BeaconDist d = (BeaconDist) distances.get(i);
	    dists[i] = d.dist.intValue();
	    result.distances.add(d);         // a shallow copy
	}

	result.stat = new Stat(dists, binSize);

	/*
	result.lastUpdate = 
	    ((BeaconDist) distances.get(distances.size()-1)).timeStamp -
	    BeaconStatModule.window;
	*/

	result.lastUpdate = 
	    ((BeaconDist) distances.get(distances.size()-1)).listenTime.longValue();

	/*
	// FIXME: define threshold value, etc to tune this!
	System.out.println(result.stat.stddev);
	if(result.stat.stddev > 10)
	    return null;
	*/

	/* FIXME: need to implement an algorithm to find distance */	
	switch(statSelect) {
	case 0:
	    result.distanceEst = result.stat.mode; break;
	case 1:
	    result.distanceEst = result.stat.mean; break;
	case 2:
	    result.distanceEst = result.stat.median; break;
	case 3:
	    result.distanceEst = result.stat.max; break;
	case 4:
	    result.distanceEst = result.stat.min; break;
	case 5:
	    result.distanceEst = ewmaDist; break;
	}

	//result.distanceEst = ((BeaconDist) distances.get(distances.size()-1)).dist.intValue();

	return result;
    }

    public String toString() 
    {
	String result = "[" + space+"-"+id+" "+lastUpdate+" "+stat;
	if(distances != null) {
	    result += " { ";
	    for(int i = 0; i < distances.size(); i++)
		result += distances.get(i) +" ";
	    result += " } ";
	}
	result += "]";
	return result;
    }
}
