/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.beacontable;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.*;
import cricketdaemon.reader.parse.ParseMap;
import java.io.*;
import java.util.*;

/**
 * Reads Beacon data objects from a Reader Module and
 * packages each Beacon object with additional hard-coded information.
 *
 * @author Allen Miu
 */
public class BeaconTableModule extends ModuleThread 
    implements Beacon.BeaconHandler
{
    private Hashtable table;

    public BeaconTableModule(ModuleThread mt, String fileName)
    {
	super("BeaconTableModule",0);

	BitSet fieldMask = new BitSet();

	fieldMask.set(ParseMap.SPACE); 
	fieldMask.set(ParseMap.BEACONID);
	fieldMask.set(ParseMap.DISTANCE);
	mt.register(this, fieldMask);

	table = new Hashtable();

	if(fileName != null)
	    initTable(fileName);
    }

    private String createKey(String sp, String id)
    {
	return sp+"%"+id;
    }

    // protected void process(); // using the inherited process method

    // Hard-coded settings overrides information from the actual beacon.
    public void processBeacon(Beacon newData, BitSet fieldMask)
    {
	String key = createKey((String) newData.getField(ParseMap.SPACE),
			       ((Integer) newData.getField(ParseMap.BEACONID)).toString());
	
	Beacon mapping = null;

	synchronized (this) {
	    mapping = (Beacon) table.get(key);
	}

	/*
	  Util.assert(mapping != null,
	  "BeaconTableModule::processBeacon() no mapping found for "+
	  key);
	*/


	if(mapping != null) {
	    /*
	    for(int i = 0; i < ParseMap.NUM_TYPES; i++) {
		// copy everything in the mapping except for the distance
		if(i != ParseMap.DISTANCE)
		    newData.insertValue(mapping.getFieldRaw(i), i);
		if(newData.getFieldRaw(i) != null)
		    fieldMask.set(i);
	    }
	    */
	    newData.insertValue(mapping.getFieldRaw(ParseMap.SPACE), 
				ParseMap.SPACE);
	    fieldMask.set(ParseMap.SPACE);
	    newData.insertValue(mapping.getFieldRaw(ParseMap.BEACONID), 
				ParseMap.BEACONID);
	    fieldMask.set(ParseMap.BEACONID);
	    newData.insertValue(mapping.getFieldRaw(ParseMap.CARTESIAN_COORD), 
				ParseMap.CARTESIAN_COORD);
	    fieldMask.set(ParseMap.CARTESIAN_COORD);
	}

	/*	
	  else {
	  Util.warn("Cannot find translation for received beacon: "+key);
	  }

	  return; 
	  // silently drop -- we don't want to mix translated and 
	  // untranslated beacons for now...
	  }
	*/

	transfer(newData, fieldMask);
    }

    protected void moduleCleanUp() { return; }

    private synchronized void initTable(String configFile) 
    {
	InputStream i = ClassLoader.getSystemResourceAsStream(configFile);
	Util.assert(i != null, "BeaconTableModule::initBeacons() "+
		    configFile+" not found");
	Properties p = new Properties();
	
	try {
	    p.load(i);
	}
	catch (IOException io) {
	    io.printStackTrace();
	    Util.assert(false, "BeaconTableModule::initBeacons() "+configFile);
	}
	Enumeration entries = p.propertyNames();
	while(entries.hasMoreElements()) {
	    String key = (String) entries.nextElement();
	    System.out.println("fetching key "+key);
	    StringTokenizer entry = 
		new StringTokenizer(p.getProperty(key));
	    
	    String spaceAndId = entry.nextToken();
	    double x  = Double.parseDouble(entry.nextToken());
	    double y  = Double.parseDouble(entry.nextToken());
	    double z  = Double.parseDouble(entry.nextToken());

	    int delim = spaceAndId.indexOf("%");
	    String space = spaceAndId.substring(0,delim);
	    String id = spaceAndId.substring(delim+1);

	    System.out.println(id+" "+x+" "+y+" "+z+" "+space);

	    newEntry(key, space, Integer.valueOf(id), 
		     new CartesianCoord(x, y, z));
	}
    }


    private void newEntry(String key, String space, Integer id, 
			 CartesianCoord c)
    {
	// The key's space%id values may be different from space and/or id
	//String key = space+"%"+id; 
	
	Beacon newBeacon = new Beacon(0);
	
	newBeacon.insertInv(space, ParseMap.SPACE);
	newBeacon.insertInv(id, ParseMap.BEACONID);
	newBeacon.insertInv(c, ParseMap.CARTESIAN_COORD);
	
	table.put(key, newBeacon);
    }

    private void clearEntry(String space, String id)
    {
	String key = createKey(space, id);
	table.remove(key);
    }

    private void clearAll()
    {
	table.clear();
    }

    // cmd value0 value1 ...
    // clearall
    // clear kspace kid
    // new kspace kid space id x y z
    public synchronized void parseCmd(StringTokenizer indata) 
    {
	int numTokens = indata.countTokens();

	if(numTokens < 1)
	    return;

	String cmd = indata.nextToken();

	//System.out.println("foobar "+cmd+" numTok "+numTokens);

	if(cmd.equals("clearall")) {
	    clearAll();
	    return;
	}

	if(cmd.equals("clear")) {
	    if(numTokens == 3) {
		String ksp = indata.nextToken();
		String kid = indata.nextToken();
		clearEntry(ksp, kid);
	    }
	    return;
	}

	if(cmd.equals("set")) {
	    if(numTokens == 8) {
		String key = createKey(indata.nextToken(), indata.nextToken());
		String sp  = indata.nextToken();
		Integer id = Integer.valueOf(indata.nextToken());
		CartesianCoord c = CartesianCoord.parseCoord(indata);

		Beacon mapping = (Beacon) table.get(key);

		if(mapping == null)
		    newEntry(key, sp, id, c);
		else {
		    mapping.insertInv(sp, ParseMap.SPACE);
		    mapping.insertInv(id, ParseMap.BEACONID);
		    mapping.insertInv(c, ParseMap.CARTESIAN_COORD);
		}
	    }
	    return;
	}
    }
}


