/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.clientlib;

import cricketdaemon.clientlib.data.CricketData;
import cricketdaemon.clientlib.parse.*;
import cricketdaemon.clientlib.sim.*;
import cricketdaemon.reader.*;
import cricketdaemon.reader.sim.*;
import cricketdaemon.server.*;
import cricketdaemon.Util;
import java.util.*;
import java.net.*;
import java.io.*;

/**
 * A Broker that retrieves updates from the <code>cricketdaemon</code>.
 *
 * @author Allen Miu
 */
public class ServerBroker extends Broker
{
    //public static final int PORT = ServerModule.PORT;
    
    /**
     * Sets the period of sending registration messages to the 
     * cricketdaemon.Server
     */
    public static final int TIMEOUT = 5000;
    public static final int BUFSIZE = 5000;

    public static final String REGISTER_TOKEN = ServerThread.REGISTER_TOKEN;
    public static final String SERVER_VERSION = ServerModule.CRICKETSERVER_VER;

    //private byte[] requestBuf;
    private Socket socket;
    //private OutputStream netOut;
    private OutputStreamWriter netOut;
    private InputStreamReader netInReader;
    private String address;
    private int port;

    // attempt connection to CricketDaemon whenever true
    private boolean reconnect = true;


    //byte[] inBuf = new byte[BUFSIZE];

    /**
     * Sends a registration message to the cricketdaemon.server.Server.
     */
    private synchronized boolean connect()
    {
	try {
	    System.out.println("ServerBroker::connect()");
	    if(socket != null)
		socket.close();

	    socket = new Socket(InetAddress.getByName(address), port);
	    socket.setTcpNoDelay(true);
	    socket.setSoTimeout((int)timeout);
	    netOut = new OutputStreamWriter(socket.getOutputStream(),
					    ServerModule.DEFAULT_ENCODING);
	    netInReader = new InputStreamReader(socket.getInputStream());
	    
	    //netOut.write(requestBuf);
	    write(REGISTER_TOKEN);
	}
	catch(IOException e) {
	    System.out.println("ServerBroker::connect() timeout");
	    return false;
	}
	return true;
    }

    public ServerBroker() 
    {
	this("127.0.0.1", ServerModule.PORT);
    }

    public ServerBroker(String addr, int portNum) 
    {
	super(TIMEOUT); // registration period
	setPriority(Thread.MIN_PRIORITY);
	/*
	try {
	    //requestPacket = 
		// InetAddress.getLocalHost() not the same as
		// InetAddress.getByName("localhost") under Linux java:
		// former returns InetAddress object that binds to the 
		// DNS name of localhost while
		// the later returns InetAddress object that binds to the
		// loopback interface.
		//createRequestPacket(InetAddress.getByName("127.0.0.1"), PORT);
	    requestBuf = 
		REGISTER_TOKEN.getBytes(ServerModule.DEFAULT_ENCODING);	       
	}
	catch(UnsupportedEncodingException e) {
	    e.printStackTrace();
	    Util.assert(false, "ServerBroker() init failed");
	}
	*/
	address = addr;
	port    = portNum;

	// hack to support ViewFinder
	// this will most likely a call to a service discovery service (or INS)
	SimBroker.initServiceCoords(serviceCoords); 
	reconnect = !connect();
    }

    /* info from service discovery system such as INS; 
     * the exact data structure used for this has yet to be fixed.
     * Right now, just use information provided by the BeaconConfigFile.
     */
    /**** This has be deprecated.  Use the getBeaconsHeard() method
	  on CricketData instead. ****/
    /*
      public ArrayList requestBeaconCoords()
      {
      ArrayList beaconCoords = new ArrayList();
      SimReader.initBeacons(beaconCoords, 
      cricketdaemon.CricketDaemon.BEACONCONFIGFILE);
      return beaconCoords;
      }
    */

    private DatagramPacket createRequestPacket(InetAddress addr, int port) 
    {
	try {
	    byte[] buf = 
		REGISTER_TOKEN.getBytes(ServerModule.DEFAULT_ENCODING);
	
	    return new DatagramPacket(buf, buf.length, addr, port);
	}
	catch (UnsupportedEncodingException e) {
	    Util.assert(false, "ServerBroker::createRequestPacket() "+
			"System does not support "+
			ServerModule.DEFAULT_ENCODING +" encoding");
	}
	return null;
    }

    /**
     * Periodically sends registration messages to the Server.
     * Receives, parse the incoming data, and dispatch the data
     * to the intereted application.
     */
    public void run()
    {
	int numWait        = 0;
	CricketData curData = null;
	final long BLOCK_PERIOD = 100; //ms

	int foo = 0;

	while(true) {
	    try {
		synchronized(this) {
		    // Hack: Reader does not block properly in some Java VM.
		    // Hence, we test netInReady.
		    while(reconnect || !netInReader.ready()) {
			if(reconnect) {
			    reconnect = !connect();
			    numWait   = 0;
			}

			//System.out.println("wassup "+numWait+" "+(numWait*BLOCK_PERIOD)+" "+reconnect);

			numWait++;
			if(numWait*BLOCK_PERIOD > 3000)  // 3 seconds
			    reconnect = true;

			try {
			    if(reconnect)
				wait(BLOCK_PERIOD*10);
			    else
				wait(BLOCK_PERIOD);
			}
			catch (InterruptedException e) {
			}
		    }
		    numWait = 0;
		}
		long timeStamp = Field.checkValidPacketVersion(netInReader);
		if(timeStamp > 0) {
		    boolean fragment = Field.popFragmentField(netInReader).equals("1");
		    CricketData d = parsePacket(netInReader, timeStamp);
		    if(curData == null)
			curData = d;
		    else
			curData.merge(d);

		    //System.out.println("parse "+fragment+" "+curData);
		    
		    if(!fragment) {
			//System.out.println("transfer");
			transfer(curData);
			curData = null;
		    }
		}
		
		//System.out.println("foo "+(foo++)+" reconnect="+reconnect+
		//" netInReader.ready="+netInReader.ready());
	    }
	    catch (IOException io) {
		reconnect = true;
		System.out.println("ServerBroker::run() parse reset");
	    }
	}
    }

    // dumps the content of a packet
    private void debugPrint(byte[] b, int length)
    {
	System.out.println("debugPrint");
	try {
	    String result = new String(b, 0, length,
				       ServerModule.DEFAULT_ENCODING);
	    System.out.println(result);
	}
	catch (UnsupportedEncodingException e) {
	    Util.assert(false, "ServerBroker::debugPrint()");
	}
    }

    /**
     * Bootstraps the parser.  Returns a CricketData object containing
     * the parsed data.
     */
    private static CricketData parsePacket(Reader data, long timeStamp) 
	throws IOException
    {
	CricketData dataStructure = new CricketData(timeStamp);
	
	// parse packet
	while(Field.newField(data)) {
	    Field.parseField(dataStructure,
			     data, 
			     PacketParseHandlerMap.instance);
	}

	return dataStructure;
    }


    /* CricketDaemon write API */
    public boolean clearBeaconTable()
    {
	return write(ServerThread.BEACONTABLE_TOKEN+" clearall");
    }

    public boolean clearBeaconTableEntry(String space, int id)
    {
	return write(ServerThread.BEACONTABLE_TOKEN+" clear "+space+" "+id);
    }

    public boolean setBeaconTableEntry(String space, int id, 
				       double x, double y, double z)
    {
	// new keyspace keyid space id x y z
	return write(ServerThread.BEACONTABLE_TOKEN+
		     " set "+space+" "+id+" "+space+" "+id+" "+x+" "+y+" "+z);
    }

    public boolean setPositionAggressive(boolean v) 
    {
	return write(ServerThread.POSITION_TOKEN+" aggressive "+v);
    }

    public boolean setPositionStddevThresh(double v) 
    {
	return write(ServerThread.POSITION_TOKEN+" stddevthresh "+v);
    }

    private synchronized boolean write(String s)
    {
	boolean result = true;

	if(netOut == null) 
	    return false;

	try {
	    netOut.write(s+"\n");
	    netOut.flush();
	}
	catch(IOException e) {
	    reconnect = true;
	    result = false;
	}

	return result;
    }



    /** TESTING **/
    static class TestCallback implements Callback
    {
	public void callback(CricketData data, BitSet mask) {
	    System.out.println("TestCallback: "+data.getCurrentSpace());
	}
    }

    static public void main(String[] args) 
    {
	Util.init("ServerBroker.log");
	ServerBroker foo = new ServerBroker();
	BitSet mask = new BitSet();
	mask.set(SPACE);
	foo.register(new TestCallback(), mask);
	foo.start();
	try {
	    foo.join();
	}
	catch(InterruptedException io) {
	}
	Util.exit();
    }
}

