/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.clientlib.data;

import cricketdaemon.clientlib.*;
import java.util.*;

/**
 * Contains information update dispatched by Broker.  Application uses
 * the public methods here to extract the available information.  A mask
 * is provided to indicate which Broker.TYPE of information is
 * available.
 *
 * Beware when multiple threads have registered callbacks with the
 * Broker: the same <code>CricketData</code> gets passed to every
 * thread. As a result, any modifications made to data may become
 * visible to another thread.  In such case, each thread is responsible
 * for deep cloning the relevant data structure.  Be advised that deep
 * cloning is not implemented for any data strucuture contained within
 * <code>CricketData</code>.  
 *
 * @author Allen Miu
 */
public class CricketData {
    private int seqNum;

    private BeaconRecord currentSpaceObject;
    private ArrayList beaconsHeard;

    /* array list of BeaconRecords containing distance estimates to each
       beacon.  The values are generated as part of the least squares
       computation used for solving the listener's position */
    private ArrayList calcBeaconDist;
    private Position devicePosition;

    private BeaconRecord anchorBeacon;
    private double orient;

    private BitSet mask;
    public BitSet getMask() { return mask; }

    // time that ServerBroker receives data from CricketDaemon
    private long timeStamp;

    public void merge(CricketData other)
    {
	// FIXME: what should be done with seqNum?
	if(currentSpaceObject == null)
	    currentSpaceObject = other.currentSpaceObject;
	if(devicePosition == null)
	    devicePosition = other.devicePosition;
	if(anchorBeacon == null)
	    anchorBeacon = other.anchorBeacon;
	if(orient == 0)
	    orient = other.orient;
	mask.or(other.mask);

	if(beaconsHeard == null)
	    beaconsHeard = other.beaconsHeard;
	else if(other.beaconsHeard != null) {
	    beaconsHeard.addAll(other.beaconsHeard);
	}

	if(calcBeaconDist == null)
	    calcBeaconDist = other.calcBeaconDist;
	else if(other.calcBeaconDist != null) {
	    calcBeaconDist.addAll(other.calcBeaconDist);
	}
    }

    public long getTimeStamp() { return timeStamp; }

    
    public CricketData(long time) 
    {
	mask = new BitSet();
	timeStamp = time;
	orient = 0;
    }

    // FIXME: should deprecate these
    public String getCurrentSpace()
    {
	BeaconRecord nearest = getCurrentSpaceObject();
	if(nearest != null)
	    return nearest.space;
	return null;
    }

    public int getSeqNum() { return seqNum; }
    public BeaconRecord getCurrentSpaceObject() 
    { 
	return currentSpaceObject; 
    }
    public ArrayList getBeaconsHeard() { return beaconsHeard; }
    public ArrayList getCalcBeaconDist() { return calcBeaconDist; }
    public Position getDevicePosition() { return devicePosition; }
    public BeaconRecord getAnchorBeacon() { return anchorBeacon; }
    public double getOrient() { return orient; }


    /* FIXME: seqNum not really used now */
    public void setSeqNum(int c) 
    { 
	mask.set(Broker.SEQNUM);
	seqNum = c; 
    }

    public void setCurrentSpaceObject(BeaconRecord c) 
    { 
	if(c != null) {
	    mask.set(Broker.SPACE);
	    currentSpaceObject = c; 
	}
    }
    public void setBeaconsHeard(ArrayList c) 
    { 
	if(c != null) {
	    mask.set(Broker.BEACONSHEARDLIST);
	    beaconsHeard = c; 
	    //setCurrentSpaceObject((BeaconRecord) beaconsHeard.get(0));
	}
    }

    public void setCalcBeaconDist(ArrayList a) 
    {
	if(a != null) {
	    mask.set(Broker.CALCBEACONDIST);
	    calcBeaconDist = a;
	}
    }

    public void setDevicePosition(Position c) 
    { 
	if(c != null) {
	    if(c.invalid())
		mask.set(Broker.POSERR);
	    else
		mask.set(Broker.DEVICECOORD);
	    devicePosition = c; 
	}
    }

    public void setOrient(double c) 
    { 
	mask.set(Broker.ORIENTATION);
	orient = c; 
    }

    public void setAnchorBeacon(BeaconRecord c) 
    { 
	if(c != null) {
	    mask.set(Broker.ANCHORBEACON);
	    anchorBeacon = c; 
	}
    }

    /* Currently, if the beaconsheard list is available, beaconstat is
       automatically available */
    public void setBeaconStat() {
	mask.set(Broker.BEACONSTAT); 
    }

    public String toString()
    {
	String result = "CricketData: "+seqNum+" "+mask+"\n"+
	    "  currentSpaceObject: "+currentSpaceObject+"\n";
	if(beaconsHeard != null) {
	    result += "  beaconsHeard: [ ";
	    for(int i = 0; i < beaconsHeard.size(); i++) 
		result += beaconsHeard.get(i)+" ::: ";
	    result += " ]\n";
	}
	result += "  devicePosition: "+devicePosition;
	if(calcBeaconDist != null) {
	    result += "  calcBeaconDist: [ ";
	    for(int i = 0; i < calcBeaconDist.size(); i++) 
		result += calcBeaconDist.get(i)+" ::: ";
	    result += " ]\n";
	}
	result += "  anchorBeacon:   "+anchorBeacon;
	result += "  orientation:    "+orient;
	return result;
    }
}
