/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.clientlib.parse;

import cricketdaemon.Util;
import cricketdaemon.clientlib.ServerBroker;
import java.io.*;

/**
 * A Field in the packet.  Contains routines to create, parse, and test
 * a Field.  Assumes that only one thread accesses a Field object; 
 * one of the methods implements synchronization to save some overhead.
 *
 * @author Allen Miu
 */
public class Field 
{
    public String type;
    public int    length;
    public String value;

    public Field(String t, int l, String v) 
    {
	type = t; length = l; value = v;
    }

    /* Below are helpful routines for processing a Field */

    /**
     * Test if the packet contains a valid SERVER_VERSION token
     */

    /// Quest:   WHY NOT JUST USE A BUFFERED READER or something
    ///          simpler??
    /// Answer:  popSubField requires an API that reads N byte
    ///          Java does not offer such an API that reads N
    ///          bytes and return a variable size array (i.e.
    ///          you need to preallocate a buffer and read N bytes
    ///          into it).
    ///          Conceivably, you can read N bytes in a while loop
    ///          but that seems inefficient.  Moreover, blocking
    ///          behavior of read() seems to be unpredictable (i.e.
    ///          it does not block and return old data when
    ///          reader.ready() is false.

    // returns a timestamp
    public static long checkValidPacketVersion(Reader data) 
	throws IOException
    {
	String version = popVersionField(data);
	String timeS   = popTimeStampField(data);
	boolean check =
	    ServerBroker.SERVER_VERSION.compareTo(version) == 0;
	if(check)
	    return Long.parseLong(timeS);
	else
	    return 0;
    }

    /**
     * Pops the TIME_STAMP field from <code>data</code>.
     *
     * Expects the state of <code>data</code> byte stream to be
     * "field|..."
     */
    public static String popTimeStampField(Reader data)
	throws IOException
    {
	return popSubField(data);
    }

    /**
     * Pops the FRAGMENT field from <code>data</code>.
     *
     * Expects the state of <code>data</code> byte stream to be
     * "field|..."
     */
    public static String popFragmentField(Reader data)
	throws IOException
    {
	return popSubField(data);
    }

    /**
     * Pops the SERVER_VERSION field from <code>data</code>.
     *
     * Expects the state of <code>data</code> byte stream to be
     * "field|..."
     */
    public static String popVersionField(Reader data)
	throws IOException
    {
	return popSubField(data);
    }

    /**
     * Parses a field from <code>data</code> and inserts the
     * parsed information into <code>dataStructure</code>, using
     * the type checker and parsing routines contained in <code>map</code>.
     *
     * This method expects the state of the <code>data</code> byte 
     * stream to be "<code>type|length|value|type2|...</code>"
     *
     * The resulting state of <code>data</code> will be 
     * "<code>|type2|...</code>"
     *
     * When the type checking on <code>dataStructure</code> fails, this
     * method causes an assertion fault to indicate an error in the
     * implementation of the given <code>map</code>.
     *
     * @exception IOException - when there is error reading <code>data</code>.
     */
    public static void parseField(Object dataStructure,
				  Reader data, 
				  ParseHandlerMap map)
	throws IOException
    {
	String type = popSubField(data);
	if(type == null)
	    return;  // parse error

	//System.out.println("parseField: "+type);

	if(!map.understands(type)) {
	    //	    System.out.println("skip");
	    skipSubField(data);
	}
	else {
	    Util.assert(map.structCheck(type, dataStructure),
			"Field::parseField() invalid "+
			"data structure passed");
	    int length = getFieldLength(data);

	    //System.out.println("length: "+length);

	    if(length < 0)
		return;

	    String value = popSubField(data, length);

	    //System.out.println("value: "+value+" "+map);

	    map.parse(type, dataStructure, value);
	}
	//	System.out.println("ParseField: "+dataStructure);
    }

    /**
     * Reads the next byte from <code>data</code>.  
     *
     * @return <code>true</code> if the byte read is a field separator.
     */
    public static boolean popSeparator(Reader data)
	throws IOException
    {
	boolean result = data.read() == '|';
	//	System.out.println("popSeparator: "+result);
	return result;
    }

    /**
     * Determines if <code>data</code> contains a new field.
     *
     * Expects the state of the <code>data</code> byte 
     * stream to be either "<code>|field|...</code>" or "" (EOF).
     *
     * The resulting state of <code>data</code> will be 
     * "<code>field|...</code>" or "" (EOF).
     *
     * @return <code>true</code> if <code>data</code> 
     *         contains a new field.
     */
    public static boolean newField(Reader data)
	throws IOException
    {
	return popSeparator(data);
    }
    
    public static char[] fieldBuf = new char[ServerBroker.BUFSIZE];

    /**
     * Pops the next subfield in <code>data</code>.
     *
     * Expects the state of the <code>data</code> byte 
     * stream to be "<code>subfield|subfield2|...</code>".
     *
     * The resulting state of <code>data</code> will be 
     * "<code>subfield2|...</code>".
     *
     * @return the subfield in ASCII
     */
    public static String popSubField(Reader data)
	throws IOException
    {
	for(int i = 0; i < ServerBroker.BUFSIZE; i++) {
	    data.read(fieldBuf, i, 1);
	    if(fieldBuf[i] < 0) {
		System.err.println("Field::popVersionField() invalid: "+
				   new String(fieldBuf, 0, i));
		return null;  // should always find a "|" at the end
	    }
	    if(fieldBuf[i] == '|') {
		String result = new String(fieldBuf, 0, i);
		//System.out.println("popSubField: "+result);
		return result;
	    }
	}
	return null;  // buffer overflow
    }

    /**
     * Pops the next subfield of the specified <code>length</code>
     * in <code>data</code>.
     *
     * @return the subfield in ASCII
     */
    public static String popSubField(Reader data, int length)
	throws IOException
    {
	if(data.read(fieldBuf, 0, length) != length)
	    return null;
	if(!popSeparator(data))
	    return null;
	String result = new String(fieldBuf, 0, length);
	//	System.out.println("popSubField: "+result);
	return result;
    }

    /**
     * Extracts the length field.
     *
     * Expects the state of the <code>data</code> byte 
     * stream to be either "<code>length|value|...</code>"
     *
     * The resulting state of <code>data</code> will be 
     * "<code>value|...</code>".
     *
     * @return length of the value field
     */
    public static int getFieldLength(Reader data)
	throws IOException
    {
	String lengthS = popSubField(data);
	if(lengthS == null)
	    return -1;  // parse error
	
	try {
	    return Integer.parseInt(lengthS);
	}
	catch (NumberFormatException e) {
	    return -1 ; // invalid number
	}
    }

    /**
     * Skip the rest of the field.  Expects the state of the
     * <code>data</code> byte stream to be 
     * "<code>length|value||type2|...</code>"
     *
     * The resulting state of <code>data</code> will be 
     * "<code>|type2|...</code>"
     */
    public static void skipSubField(Reader data)
	throws IOException
    {
	int length = getFieldLength(data);
	if(length >= 0)
	    data.skip(length+1); 
	// skip, including the end field separator
    }

    /**
     * Constructs a field of the format 
     * "<code>|type|length|value|</code>", where the length is the
     * length of the value String.
     */
    public static String toField(String type, String value)
    {
	return toField(type, new StringBuffer(value));
    }

    // FIXME: should optimize ServerModule to use StringBuffers instead
    // of string concatenation.
    public static String toField(String type, StringBuffer value)
    {
	Util.assert(value != null, "ServerModule::toField() null");
	int length = value.length();
	return 
	    value.insert(0, "|").insert(0, length).insert(0, "|").
	    insert(0, type).insert(0, "|").append("|").toString();
    }

    // FIXME: should optimize ServerModule to use StringBuffers instead
    // of string concatenation.
    public static StringBuffer toFieldAppend(StringBuffer dst, String type, 
				       String value)
    {
	Util.assert(value != null, "ServerModule::toField() null");
	int length = value.length();
	return 
	    dst.append("|").append(type).append("|").append(length).
	    append("|").append(value).append("|");
    }


}



