/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.clientlib.sim;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;
import viewfinder.Coord;
import java.util.*;

/**
 * A Broker that generates simulated data from a GUI.  Does not
 * interact with the cricketdaemon.server.Server to obtain data.
 *
 * @author Allen Miu
 */
public class SimBroker extends Broker implements SimControl
{

    private static final int num_beacons = 4;

    private static final String[] services = 
    { 
	"Fridge", "Coffee", "Printer", "Printer",
	"Photocopier",
	"Camera", "Fax",
	"Phone", "Server", "TV", "Water Fountain", 
    };
    private static final int num_service = services.length;

    private static final int[] bx =
    { 
	270, 100, 245, 70
    };

    private static final int[] by =
    { 
	40, 130, 220, 275
    };

    private static final int[] x =
    { 
	110, 110, 110, 110, 110, 255, 50, 225, 200, 170, 200
    };

    private static final int[] y =
    {
	70,  90, 120, 130, 150, 75, 250, 200, 170, 115, 50
    };


    private boolean periodicUpdate = false;

    protected Coord beaconCoord;
    protected Coord deviceCoord;

    protected double absOrient;

    private int minX, minY, maxX, maxY;  // max GUI Coordinates
    private static final int UPDATE_PERIOD = 500;
    private static final BitSet dataMask = new BitSet();
    private static int seqNum = 0;

    public SimBroker(int min_X, int min_Y, int max_X, int max_Y) 
    {
	super(UPDATE_PERIOD);

	dataMask.set(NULL);
	dataMask.set(SPACE);

	minX = min_X;
	minY = min_Y;
	maxX = max_X;
	maxY = max_Y;
	initServiceCoords(serviceCoords);
	initBeaconCoords(beaconCoords);
	deviceCoord = new Coord(max_X/2, max_Y/2, "Device");
	beaconCoord = (Coord) beaconCoords.get(0);
    }

    public static void initServiceCoords(ArrayList v) 
    {
	for(int i = 0; i < num_service; i++) {
	    Coord newp = 
		new Coord(x[i], y[i], services[i]);
	    v.add(newp);
	}
    }

    public static void initBeaconCoords(ArrayList v) 
    {
	for(int i = 0; i < num_beacons; i++) {	
	    Coord newp = 
		new Coord(bx[i], by[i], "B"+String.valueOf(i));
	    v.add(newp);	
	}
    }

    public ArrayList requestBeaconCoords()
    {
	return beaconCoords;
    }

    private CricketData newData()
    {
	double relOrient = 
	    Coord.absToRelOrient(beaconCoord, deviceCoord, absOrient);

	CricketData c = new CricketData(System.currentTimeMillis());

	Position devicePos = new Position(deviceCoord.x, deviceCoord.y, 0);
	Position beaconPos = new Position(beaconCoord.x, beaconCoord.y, 0);
	BeaconRecord br = new BeaconRecord();
	br.pos = beaconPos;

	c.setSeqNum(seqNum++);
	c.setDevicePosition(devicePos);  // device's own position
	c.setAnchorBeacon(br);
	c.setOrient(relOrient);       // orientation wrt anchoring beacon
	return c;
    }

    synchronized public void run()
    {
	if(!periodicUpdate)
	    timeout = 0;

	while(true) {
	    try {
		wait(timeout);
	    }
	    catch (InterruptedException ie) {
		System.out.println("SimBroker::run() interrupted");
	    }
	    
	    transfer(newData());
	}
    }

    /* FIXME: need to implement z axis movement */
    public double moveDeviceUp() 
    {
	return 0;
    }

    public double moveDeviceDown() 
    {
	return 0;
    }

    public double moveDeviceForward() 
    {
	deviceCoord.y -= 1;
	if(deviceCoord.y < minY)
	    deviceCoord.y = minY;
	return deviceCoord.y;
    }

    public double moveDeviceBackward() 
    {
	deviceCoord.y += 1;
	if(deviceCoord.y >= maxY)
	    deviceCoord.y = maxY;
	return deviceCoord.y;
    }

    public double moveDeviceLeft()
    {
	deviceCoord.x -= 1;
	if(deviceCoord.x < minX)
	    deviceCoord.x = minX;
	return deviceCoord.x;
    }

    public double moveDeviceRight()
    {
	deviceCoord.x += 1;
	if(deviceCoord.x >= maxX)
	    deviceCoord.x = maxX;
	return deviceCoord.x;
    }

    public double incrDeviceRotate()
    {
	absOrient += 0.1;
	absOrient %= Coord.fullCircleInRadians;
	return absOrient;
    }

    public double decrDeviceRotate()
    {
	absOrient -= 0.1;
	if(absOrient < 0)
	    absOrient = Coord.fullCircleInRadians + absOrient;
	return absOrient;
    }

    synchronized public void notice()
    {
	if(!periodicUpdate)
	    notifyAll();
    }
}
