/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.module;

import cricketdaemon.CricketDaemon;
import cricketdaemon.Util;
import java.util.*;

/**
 * All Module implementations should extend this base class.  The relevant
 * module interfaces are
 *
 * protected void transfer(Data value, BitSet fieldMask);
 * protected long timeout; 
 * protected Vector inputBuf; // vector of Data
 * protected BitSet registeredMask;
 * public void deRegister(ModuleThread module, BitSet fieldMask);
 * protected void moduleCleanUp();
 *
 *   and the set of data type constants such as
 *
 * protected static final int NULL = 0; 
 * 
 *   No module should define a data type value same as NULL (i.e. 0).
 * 
 * Note:
 *   The main processing routine is
 *
 * abstract protected void process();
 *
 *   MODULES SHOULD DEFINE THIS METHOD TO HANDLE INCOMING DATA FROM
 *   THE inputBuf.  The callback() function should not be used by
 *   the module developer as it is called internally by ModuleThread.
 *
 *   Alternatively, the module can implement a "process handler" for
 *   each type of Data.  A module can implement DataType.Handler,
 *   which defines the "process handler" that would be dispatched to 
 *   handle an object of DataType.  See 
 *   <code>beaconstat.BeaconStatModule</code> for an example.
 *
 *   Finally, if a module is going to modify the data object, 
 *   the module should ALWAYS clone the data object.  Otherwise, the
 *   modification will affect other modules that are accessing the
 *   same data object.
 *
 * Assumes that Vector is synchronized, per Java 1.2 API documentation.
 *
 * @author Allen Miu
 */
public abstract class ModuleThread extends Thread
{
    /** Max number of QueueItem allowed to be queued for every module. */
    public static final int BUFFER_THRESH = 8; // FIXME: may need to 
                                               // parameterize this

    // used to get around the single input queue blocking problem
    public boolean callbackMutex; 

    /**
     * Input Buffer -- used by the producers to input data
     * The module should always clone the data object, if the module is 
     * going to modify it!!!
     */
    // This has to be a Vector because the ModuleThread may access this
    // object in critical sections.
    protected Vector inputBuf; 

    /**
     * Contains a list of modules (RegistryEntry) that registered 
     * with this modules. // unsynchronized
     */
    private ArrayList registry;

    /**
     * Contains an OR of the requested field masks by the registered modules
     */
    protected BitSet registeredMask;

    /**
     * Timeout period to call process(), when inputBuf is empty.  "0"
     * causes thread to wait indefinitely until notify() or notifyAll()
     * is called.
     */
    protected long timeout;



    /*--- Each module should define its own set of available data types. ---*/

    /**
     * Any consumer ModuleThread that registers with the "NULL" bit
     * set in the fieldMask may get a callback from the producer
     * ModuleThread any time.  It is up to the producer ModuleThread
     * to define the semantics of this callback.  
     * The callback "value" may be non-null.
     */
    protected static final int NULL = 0; 

    protected static BitSet CLEARMASK;

    static {
	CLEARMASK = new BitSet();
	CLEARMASK.set(NULL);
    }

    public interface Data
    {
	public void process(Object handler, BitSet fieldMask);
	public String toString();
    }

    public class QueueItem
    {
	public BitSet fieldMask;
	public Data value; // pitfall: value may be null

	public QueueItem(Data v, BitSet t)
	{
	    value = v; fieldMask = t;
	}

	public String toString()
	{
	    return fieldMask+" "+value;
	}
    }

    private class RegistryEntry
    {
	ModuleThread module;
	BitSet fieldMask;

	private RegistryEntry(ModuleThread m, BitSet t)
	{
	    module = m; fieldMask = t;
	}

	public boolean isEqual(RegistryEntry e)
	{
	    // Should not use "module.isEqual(e.module)"
	    // because we want to ensure the references points
	    // to the same module
	    return (module == e.module) && (fieldMask.equals(e.fieldMask));
	}

	// returns false if the target input Buffer is blocked by someone else
	// if this is the case, then, should try to resend later.
	// fixme: this sort of optimization may be an overkill	
	public boolean matchAndFireCallback(Data value, BitSet t)
	{
	    BitSet test = (BitSet) t.clone();
	    test.and(fieldMask);
	    
	    //Util.log("RegistryEntry::matchAndFireCallback "+module.getName());
	    if(test.length() > 0) {
		// someone else accessing the module
		if(module.callbackMutex)
		    return false;
		//Util.log("RegistryEntry::matchAndFireCallback calling callback");		
		module.callback(value, t); 
		//Util.log("RegistryEntry::matchAndFireCallback callback done");
	    }
	    return true;
	}

	public String toString()
	{
	    return module+" "+fieldMask;
	}
    }

    public ModuleThread()
    {
	return;
    }

    public ModuleThread(String name, long timeoutPeriod)
    {
	super(name);
	timeout = timeoutPeriod;
	registry = new ArrayList();
	inputBuf = new Vector();
	callbackMutex  = false;
	// the "NULL" data type is always available
	registeredMask = (BitSet) CLEARMASK.clone();
    }

    // from another producer
    public void callback(Data value, BitSet fieldMask)
    {
	callbackMutex = true;
	// pitfall: implementations should be wary of "null" value
	inputBuf.add(new QueueItem(value, fieldMask));
	if(inputBuf.size() > BUFFER_THRESH)
	    //Util.warn("ModuleThread::callback() "+
	    System.out.println("WARNING: ModuleThread::callback() "+
			       this.getName()+" queue full");

	// notify() is redundant (?) because mutex already protects this method
	//if(!CricketDaemon.singleThread)
	//notify();
	callbackMutex = false;
	if(CricketDaemon.singleThread)
	    process();
    }

    // by a consumer
    public synchronized void register(ModuleThread module, BitSet fieldMask)
    {
	RegistryEntry newEntry = new RegistryEntry(module, fieldMask);
	if(registry.contains(newEntry)) {
	    Util.warn("ModuleThread::register "+module+" multiple entries "+
		      newEntry);
	    return; // silently ignore
	}
	registeredMask.or(fieldMask);
	registry.add(newEntry);
    }

    public synchronized void deRegister(ModuleThread module, BitSet fieldMask)
    {
	registeredMask = (BitSet) CLEARMASK.clone();
	for(int i = 0; i < registry.size(); i++) {
	    RegistryEntry entry = (RegistryEntry) registry.get(i);
	    if(entry.module == module && 
	       (fieldMask == null || entry.fieldMask.equals(fieldMask)))
		registry.remove(i);
	    else // save the fieldMask from the remaining entries
		registeredMask.or(entry.fieldMask); 
	}
    }

    // to another consumer
    protected void transfer(Data value, BitSet fieldMask)
    {
	// fixme: this sort of optimization may be an overkill
	ArrayList consumers = (ArrayList) registry.clone();
	if(fieldMask == null)
	    fieldMask = (BitSet) CLEARMASK.clone();

	fieldMask.set(NULL); // to enable NULL data type

	while(consumers.size() > 0) {
	    RegistryEntry e = (RegistryEntry) consumers.remove(0);
	    if(!e.matchAndFireCallback(value, fieldMask))
		consumers.add(e); // target blocked, resend later
	}
    }

    // should not be modified
    public void run()
    {
	try {
	    while(true) {
		synchronized(this) {
		    if(inputBuf.size() == 0)
			wait(timeout);
		}
		process();
	    }    
	} catch (InterruptedException e) {
	    Util.log(getName() +" interrupted");
	}
	moduleCleanUp();
	Util.log(getName() +" exits");
    }

    /**
     * Called when the thread execution is about to end.  Modules
     * should implement this method to clean up any outstanding resources
     * (e.g. to close opened files).
     */
    abstract protected void moduleCleanUp();

    /**
     * Dequeues and returns the head of the inputBuf.
     */
    protected QueueItem dequeue() 
    {
	Object result = inputBuf.elementAt(0);
	inputBuf.removeElementAt(0);
	return (QueueItem) result;
    }

    protected void dequeueDataAndDispatchHandler(Object handler) 
    {
	Util.assert(inputBuf.size() > 0, 
		    "ModuleThread::dequeueDataAndDispatchHandler() "+
		    "empty buffer");
	QueueItem qI = dequeue();
	Data newData = qI.value;
	newData.process(handler, qI.fieldMask);
    }

    /**
     * Implements the routines for processing input data.
     * Here, the only relevant ModuleThread intefaces are: 
     *
     * protected void transfer(Data value, BitSet fieldMask);
     * protected long timeout; 
     * protected Vector inputBuf; // vector of Data
     * protected BitSet registeredMask;
     * public void deRegister(ModuleThread module, BitSet fieldMask);
     * protected void moduleCleanUp();
     *
     *   and the set of data type constants such as
     *
     * protected static final int NULL = 0; 
     * 
     *   No module should define a data type value same as NULL (i.e. 0).
     * 
     * Note:
     *   The module should always clone the data object from the 
     *   inputBuf, if the module is going to modify it.
     */
    protected void process()
    {
	while(inputBuf.size() > 0) 
	    dequeueDataAndDispatchHandler(this); 	    
    }
}
