/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.module;

import cricketdaemon.Util;
/**
 * A subclass of <code>ModuleThread</code> that differs in the way it
 * interprets the <code>timeoutPeriod</code> paramerter in the
 * constructor.
 *
 * In <code>ModuleThread</code>, the time out is the argument to the
 * <code>wait()</code> call within the default <code>run()</code>
 * method.  Hence, the default <code>run()</code> method will call
 * <code>process()</code> if <bold>no</bold> event occurs for
 * <code>timeout</code> period since the last iteration of the
 * <code>while</code> loop inside the default <code>run()</code> method.
 *
 * The <code>PeriodicModuleThread</code> class calls <code>process()</code>
 * in regular periods, despite the number of <code>process()</code> calls 
 * made in the <code>while</code> within the <code>timeoutPeriod</code>.
 *
 * <code>periodicallyProcess()</code> is called once every 
 * <code>timeoutPeriod</code>, while <code>process()</code> can be called
 * asynchronously.
 *
 * @author Allen Miu
 */
abstract public class PeriodicModuleThread extends ModuleThread
{
    private boolean timeExpired;
    private long expiryTime;

    public PeriodicModuleThread()
    {
	expiryTime = System.currentTimeMillis();
    }

    public PeriodicModuleThread(String name, long timeoutPeriod)
    {
	super(name, timeoutPeriod);
	expiryTime = System.currentTimeMillis();
    }

    public boolean isTimerExpired() { return timeExpired; }
    public long getExpiryTime() { 
	if(isTimerExpired())
	    return System.currentTimeMillis() + timeout; // estimate
	else
	    return expiryTime;
    }

    public void run()
    {
	try {
	    while(true) {
		long waitTime = expiryTime - System.currentTimeMillis();
		timeExpired = waitTime <= 0;
		if(!timeExpired) {
		    synchronized(this) {
			wait(waitTime);
		    }
		}
		process();
		if(timeExpired)
		    expiryTime = timeout + System.currentTimeMillis();
	    }    
	} catch (InterruptedException e) {
	    Util.log(getName() +" interrupted");
	}
	moduleCleanUp();
	Util.log(getName() +" exits");
    }
    
    protected void process() 
    {
	super.process();
	if(isTimerExpired())
	    periodicallyProcess();
    }

    abstract protected void periodicallyProcess();
}
