/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.position;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.*;
import cricketdaemon.space.DistanceEstimate;
import java.util.*;

/**
 * A Data object that holds the coordinate and the derived distances
 * for each of the beacons used in the Matrix solver.
 *
 * @author Allen Miu
 */
public class PositionData implements ModuleThread.Data
{
    public long timeStamp;
    private CartesianCoord coord;
    private DistanceEstimate[] distances;
    private double speedOfSoundFactor;

    public final static PositionData invalidInstance;

    static {
	double[] di = new double[1];
	int[]    ii = new int[1];
	String[] si = new String[1];

	si[0] = "Invalid";
	di[0] = Double.NaN;
	ii[0] = Integer.MIN_VALUE;

	invalidInstance = 
	    new PositionData(Long.MIN_VALUE, Double.NaN, Double.NaN, Double.NaN,
			     di, di, di, si, ii, di, Double.NaN);
    }

    // FIXME: add sequence number and timeStamp
    // FIXME: add beacon coordinates for the respective d_bar values
    public PositionData(long curTime,
			double x, double y, double z,
			double[] bx, double[] by, double[] bz,
			String[] space, int[] id, 
			double[] d_bar, 
			double k)
    {
	this(curTime, x, y, z, bx, by, bz, space, id, d_bar, k, space.length);
    }

    public PositionData(long curTime,
			double x, double y, double z,
			double[] bx, double[] by, double[] bz,
			String[] space, int[] id, double[] d_bar, 
			double k, int length)
    {
	timeStamp = curTime;
	speedOfSoundFactor = k;
	coord = new CartesianCoord(x, y, z);
	distances = new DistanceEstimate[length];
	for(int i = 0; i < length; i++)
	    distances[i] = new DistanceEstimate(space[i], id[i], 
						new CartesianCoord(bx[i], by[i], bz[i]), 
						d_bar[i]);
    }

    public interface PositionDataHandler {
	public void processPositionData(PositionData newData, 
					BitSet fieldMask);
    }

    public void process(Object handler, BitSet fieldMask) 
    {
	Util.assert(handler instanceof PositionDataHandler,
		    "PositionData::process() incorrect handler type\nMake sure the offending module 'implements' PositionData.");
	PositionDataHandler h = (PositionDataHandler) handler;
	h.processPositionData(this, fieldMask);
    }

    public CartesianCoord getCoord()
    {
	return coord;
    }

    public DistanceEstimate[] getDistances()
    {
	return distances;
    }

    public double getSpeedOfSoundFactor()
    {
	return speedOfSoundFactor;
    }

    public String toString()
    {
	String result = "";
	result += timeStamp;
	result += " "+coord.toString();
	if(distances != null) {
	    for(int i = 0; i < distances.length; i++)
		result += " "+ distances[i];
	}
	result += " "+speedOfSoundFactor;
	return result;
    }
}

