/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.position;


import cricketdaemon.beaconstat.*;
import cricketdaemon.beaconstat.BeaconDistStat.BeaconDist;
import cricketdaemon.module.*;
import cricketdaemon.reader.CartesianCoord;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.reader.sim.SimReader;
import cricketdaemon.space.*;
import cricketdaemon.space2.*;
import cricketdaemon.ServiceTime;
import cricketdaemon.UnitConv;
import cricketdaemon.Util;
import java.util.*;

/**
 * A module that implements a least-square matrix solver package to 
 * find the position of the device.
 *
 * As we will phase out the Space* modules, the PositionModule will
 * be modified to accept BeaconStatData instead of SpaceData.
 *
 * @author Allen Miu
 */
public class PositionModule extends ModuleThread 
    implements 
	BeaconStatData.BeaconStatDataHandler,
	SpaceData.SpaceDataHandler
{
    private static BitSet spaceFieldMask;



    /* PositionModule field types */
    public static final int CARTESIAN_COORD    = 1;
    public static final int DISTANCE_ESTIMATES = 2;
    public static final int POSERR             = 3;

    //public static final double K_NOMINAL = 
    //Util.SPEED_OF_SOUND/(Util.CLOCK_FREQ_HZ * 2.54);
    public static final double K_NOMINAL = UnitConv.cricketDistToCM(1);
    public static final int NUMBEACONS_REQUIRED = 3;

    private PositionSolver posSolver;

    private double stddevThresh = 15.0;
    private boolean aggressive = false;
    
    protected PositionModule(double kthresh) {
	posSolver = new PositionSolver(K_NOMINAL, kthresh);
    }

    public PositionModule(ModuleThread mt, double kthresh)
    {
	super("PositionModule", 0);
	prevBeacons = null;

	/*
	  Util.assert(mt instanceof Space2Module,
	  "PositionModule() only works with the Space2Module");
	*/

	BitSet desiredTypeMask = new BitSet();

	/*
	  if(mt instanceof Space2Module)
	  desiredTypeMask.set(SpaceModule.MODE);
	  else if(mt instanceof BeaconStatModule)
	  desiredTypeMask.set(BeaconStatModule.STAT);
	*/
	desiredTypeMask.set(BeaconStatModule.STAT);

	mt.register(this, desiredTypeMask);
	posSolver = new PositionSolver(K_NOMINAL, kthresh);
    }

    protected void moduleCleanUp() { return; }

    private ArrayList prevBeacons;

    protected boolean shouldUpdate(ArrayList a) { return true; }

    /* returns true iff all four distances have changed */
    /*
    protected boolean shouldUpdate(ArrayList a)
    {
	if(prevBeacons == null) {
	    prevBeacons = a;
	    return true;
	}

	int numSame = 0;
	for(int i = 0; i < a.size(); i++) {
	    for(int j = 0; j < prevBeacons.size(); j++) {
		if(a.get(i).equals(prevBeacons.get(j))) {
		    BeaconDistStat cur  = (BeaconDistStat) a.get(i);
		    BeaconDistStat prev = (BeaconDistStat) prevBeacons.get(j);
		    if(cur.distanceEst == prev.distanceEst)
			numSame++;
		}
	    }
	}
	if(numSame == 0) {
	    prevBeacons = a;
	    return true;
	}
	return false;
    }
    */

    /*
    private void pickBeacons(ArrayList beacons, int numBeacons)
    {
	int oldest = 0;
	long oldestTime = ((BeaconDistStat) beacons.get(0)).lastUpdate;
	for(int i = 1; i < numBeacons && i < beacons.size(); i++) {
	    BeaconDistStat b = (BeaconDistStat) beacons.get(i);
	    if(b.lastUpdate < oldestTime) {
		oldest = i;
		oldestTime = b.lastUpdate;
	    }
	}
	
        while(beacons.size() > numBeacons) {
	    BeaconDistStat b = 
		(BeaconDistStat) beacons.remove(numBeacons);
	    BeaconDistStat old = (BeaconDistStat) beacons.get(oldest);
	    if(b.lastUpdate > oldestTime) {
		beacons.remove(oldest);
		beacons.add(oldest, b);
		
		oldest = 0;
		oldestTime = ((BeaconDistStat) beacons.get(0)).lastUpdate;
		for(int i = 1; i < numBeacons && i < beacons.size(); i++) {
		    BeaconDistStat bb = (BeaconDistStat) beacons.get(i);
		    if(bb.lastUpdate < oldestTime) {
			oldest = i;
			oldestTime = bb.lastUpdate;
		    }
		}
	    }
	}
    }
    */

    private void printBeacons(ArrayList beacons)
    {
	for(int i = 0; i < beacons.size(); i++) {
	    System.out.println(beacons.get(i));
	}
    }

    private boolean filterBeacons(ArrayList beacons)
    {
	Iterator it = beacons.iterator();
	while(it.hasNext()) {
	    BeaconDistStat b = (BeaconDistStat) it.next();
	    if (K_NOMINAL*b.stat.stddev > 3*stddevThresh ||
		b.beacon.getField(ParseMap.CARTESIAN_COORD) == null) {
		//System.out.println("FILTERING:::: "+b);
		it.remove();
	    }
	}
	return beacons.size() >= NUMBEACONS_REQUIRED;
    }



    public void processBeaconStatData(BeaconStatData newData, BitSet fieldMask)
    {

	// values array needs to cloned as it is shared by multiple modules
	ArrayList beacons = (ArrayList) newData.values.clone();
	
	if(beacons == null) {
	    Util.warn("PositionModule::processBeaconStatData() beacons == null");
	    return;
	}

	if(!filterBeacons(beacons)) {
	    Util.warn("PositionModule: Too few valid beacons \n\t("+
		      "stddev too high or coord not available)");
	    sendErr();
	    return;
	}

	if(!shouldUpdate(beacons))
	    return;

	//int size = (beacons.size() < 4) ? beacons.size() : 4;
	int size = beacons.size();
	
	double[] x     = new double[size];
	double[] y     = new double[size];
	double[] z     = new double[size];
	String[] space = new String[size];
	int[]    id    = new int[size];
	double[] dist  = new double[size];

	Iterator it = beacons.iterator();

	/*
	  System.out.println("********* before ********* ");
	  printBeacons(beacons);
	  pickBeacons(beacons);
	  System.out.println("****** after ****** ");
	  printBeacons(beacons);
	*/

	//pickBeacons(beacons, size);

	int i = 0;
	while(it.hasNext()) {
	    BeaconDistStat b = (BeaconDistStat) it.next();
	    CartesianCoord c = (CartesianCoord) 
		b.beacon.getField(ParseMap.CARTESIAN_COORD);

	    if(c != null) {
		x[i]     = c.x;
		y[i]     = c.y;
		z[i]     = c.z;
		space[i] = b.space;
		id[i]    = b.id;

		//dist[i]  = b.distanceEst;


		if(aggressive)
		    dist[i] = ((BeaconDist) b.distances.get(b.distances.size()-1)).dist.doubleValue();
		else
		    dist[i]  = b.distanceEst;


		//Util.log(x[i]+" "+y[i]+" "+z[i]+" "+space[i]+
		//" "+id[i]+" "+dist[i]);
		// System.out.println(x[i]+" "+y[i]+" "+z[i]+" "+
		// space[i]+" "+id[i]+" "+dist[i]);
		i++;
	    }
	}
	
	PositionData newEstimate = null;
	/*
	  for(int combo = 0; combo < comboIndexA.length; combo++) {
	      	    System.out.println("combo: "+combo);
	  if(combo == 8)
	  newEstimate = positionEstimate(x, y, z, space, id, dist, combo);
	  else
	  positionEstimate(x, y, z, space, id, dist, combo);
	  }
	*/

	newEstimate = 
	    posSolver.positionEstimateVKnown(x, y, z, space, id, dist);

	/*
	if(size <= 4) 
	    newEstimate = positionEstimateVKnown(x, y, z, space, id, dist);
	else {
	    newEstimate = positionEstimateVUnknown(x, y, z, space, id, dist);
	}
	*/

	/* send position data */
	if(newEstimate != null) 
	    send(newEstimate);
	else
	    sendErr();
    }

    public void processSpaceData(SpaceData newData, BitSet fieldMask)
    {
	return;
    }

    /*
    // fixme:  may want to discard old data that has been
    // queued up and return the position from the last queued item
    public void processSpaceData(SpaceData newData, BitSet fieldMask)
    {
	//int seqNum = newData.seqNum;
	DistanceEstimate[] estimates = newData.getMode();
	SpaceBeaconTuple[] tuples = convertToTuples(estimates);

	if(tuples == null || tuples.length < NUMBEACONS_REQUIRED) {
	    Util.log("PositionModule::processSpaceData() not enough beacons "+
		     "that have coordinate infomration");
	    return;
	}

	double[] x     = new double[NUMBEACONS_REQUIRED];
	double[] y     = new double[NUMBEACONS_REQUIRED];
	double[] z     = new double[NUMBEACONS_REQUIRED];
	String[] space = new String[NUMBEACONS_REQUIRED];
	int[]    id    = new int[NUMBEACONS_REQUIRED];
	double[] dist  = new double[NUMBEACONS_REQUIRED];


	// FIXME: need strategy to pick a subset of required beacons 
	for(int i = 0; i < NUMBEACONS_REQUIRED; i++) {
	    Util.assert(tuples[i].coord != null,
			"PositionModule::processSpaceData() need to run with"+
			" beacon coordinate mapping on");
	    x[i]     = tuples[i].coord.x;
	    y[i]     = tuples[i].coord.y;
	    z[i]     = tuples[i].coord.z;
	    space[i] = tuples[i].space;
	    id[i]    = tuples[i].beaconID;
	    dist[i]  = tuples[i].distance;
	    
	    Util.log("Beacon ["+i+"]: "+x[i]+" "+y[i]+" "+dist[i]);
	}

	PositionData newEstimate = positionEstimate(x, y, z, 
						    space, id, dist, 
						    8);
	
	// send position data 
	send(newEstimate);
    }
    */

    private void send(PositionData newEstimate) 
    {
	BitSet positionFieldMask = (BitSet) CLEARMASK.clone();
	positionFieldMask.set(CARTESIAN_COORD);
	positionFieldMask.set(DISTANCE_ESTIMATES);
	//	    System.out.println("PositionModule::process() "+newEstimate);
	transfer(newEstimate, positionFieldMask);
	
    }

    private void sendErr()
    {
	BitSet positionFieldMask = (BitSet) CLEARMASK.clone();
	positionFieldMask.set(POSERR);
	transfer(PositionData.invalidInstance, positionFieldMask);
    }

    public SpaceBeaconTuple[] convertToTuples(DistanceEstimate[] d) 
    {
	SpaceBeaconTuple[] result = new SpaceBeaconTuple[d.length];
	for(int i = 0; i < result.length; i++)
	    result[i] = (SpaceBeaconTuple) d[i];
	return result;
    }

    // cmd value0 value1 ...
    // aggressive [true|false]
    // stddevthresh double
    // for now, no need synchronized lock
    public void parseCmd(StringTokenizer indata) 
    {
	int numTokens = indata.countTokens();

	if(numTokens < 1)
	    return;

	String cmd = indata.nextToken();
	if(cmd.equals("aggressive")) {
	    String s = null;
	    if(numTokens == 2) {
		s = indata.nextToken();
		aggressive = Boolean.valueOf(s).booleanValue();
	    }
	    System.out.println("foobar aggressive "+aggressive+" str="+s+"E");
	    return;
	}

	if(cmd.equals("stddevthresh")) {
	    if(numTokens == 2) {
		stddevThresh = Double.parseDouble(indata.nextToken());
	    }
	    System.out.println("foobar stddevthresh "+stddevThresh);

	    return;
	}

    }


}




