/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.position;

import Jama.*;
import cricketdaemon.beaconstat.*;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.ServiceTime;
import java.util.*;

public class PositionSolver
{
    

    //public static double ZSQR_THRESH = 10000;
    private double K_NOMINAL;
    private double K_THRESH;
    private double K_NOMINALSQR;

    public PositionSolver(double knominal, double kthresh) {
	K_NOMINAL = knominal;
	K_THRESH = kthresh;
	K_NOMINALSQR = K_NOMINAL*K_NOMINAL;
    }

    // more accurate way to calculate a^2 - b^2
    protected static double diffsqr(double a, double b) 
    {
	return (a+b)*(a-b);
    }

    protected static double sqr(double x) { return x*x; }

    protected static long curTime()
    {
	return System.currentTimeMillis();
    }

    public PositionData positionEstimateVKnown(double[] x, 
					       double[] y,
					       double[] z,
					       String[] space,
					       int[] id,
					       double[] dist)
    {
	int numBeacons = x.length;

	double[] d_bar = new double[numBeacons];

	double[][] amatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    amatrix[i] = new double[2];
	    amatrix[i][0] = 2*(x[i+1]-x[0]);
	    amatrix[i][1] = 2*(y[i+1]-y[0]);
	}
	double[][] bmatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    bmatrix[i] = new double[1];
	    bmatrix[i][0] = diffsqr(x[i+1], x[0]) + diffsqr(y[i+1], y[0]) - 
		K_NOMINALSQR*diffsqr(dist[i+1],dist[0]);
	}

	
	/*
	for(int c = 0; c < numBeacons; c++) 
	System.out.println("foobar: "+x[c]+" "+y[c]+
	" "+z[c]+" "+space[c]+" "+dist[c]);
	*/

	Matrix A = new Matrix(amatrix);
	Matrix B = new Matrix(bmatrix);

	/*
	  System.out.println("new v");
	  A.print(3,numBeacons-1);
	  B.print(1,numBeacons-1);
	*/

	if(numBeacons == 3) {
	    // do a determinate check to see if a unique solution exists!!!
	    double determinate = A.det();
	    if(Math.abs(determinate) < 0.000000001) {
		System.out.println("PositionModule: det == 0");
		return  null; // throw away sample.
	    }
	}

	Matrix X = (new QRDecomposition(A)).solve(B);

	double[][] xmatrix = X.getArray();
	    
	double resultX = xmatrix[0][0];
	double resultY = xmatrix[1][0];
 	double resultZ = -1;

	for(int i = 0; resultZ < 0 && i < numBeacons; i++) {
	    resultZ = sqr(K_NOMINAL*dist[i]) - sqr(resultX - x[i]) - 
		sqr(resultY - y[i]);
	    i++;
	}

	//resultZ = Math.abs(resultZ);
	if(resultZ < 0) {
	    System.out.println("PositionModule: imaginary z [x="+
			       resultX+" y="+resultY+" z="+resultZ+"]");
	    return null;
	}
	else 
	    resultZ = Math.sqrt(resultZ);

	// the d_bars are really meaningless here
	for(int i = 0; i < numBeacons; i++)
	    d_bar[i] = K_NOMINAL*dist[i];

	PositionData result = 
	    new PositionData(curTime(), 
			     resultX, resultY, resultZ, x, y, z,
			     space, id, d_bar, K_NOMINAL);

	//ServiceTime.printElapsedTime("time: ");
	//System.out.println(resultX+" "+" "+resultY+" "+resultZ);

	return result;
    }


    public PositionData positionEstimateVUnknown(double[] x, 
						 double[] y,
						 double[] z,
						 String[] space,
						 int[] id,
						 double[] dist)
    {
	int numBeacons = x.length;

	double[] d_bar = new double[numBeacons];
	
	double[][] amatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    amatrix[i] = new double[3];
	    amatrix[i][0] = 2*(x[i+1]-x[0]);
	    amatrix[i][1] = 2*(y[i+1]-y[0]);
	    amatrix[i][2] = diffsqr(dist[i+1], dist[0]);
	    
	}
	double[][] bmatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    bmatrix[i] = new double[1];
	    bmatrix[i][0] = diffsqr(x[i+1], x[0]) + diffsqr(y[i+1], y[0]);
	}

	Matrix A = new Matrix(amatrix);
	Matrix B = new Matrix(bmatrix);

	/*
	  System.out.println("new no v");
	  A.print(2,numBeacons-1);
	  B.print(1,numBeacons-1);
	*/
	
	//A.print(10, 3);
	if(numBeacons == 4) {
	    double determinate = A.det();
	    //System.out.println("determinate: "+determinate);
	    
	    if(Math.abs(determinate) < 0.000000001) {
		System.out.println("PositionModule::positionEstimate() determinate is zero");
		return null;
	    }
	}


	Matrix X = (new QRDecomposition(A)).solve(B);

	double[][] xmatrix = X.getArray();
	    
	double resultX = xmatrix[0][0];
	double resultY = xmatrix[1][0];
	double resultK = xmatrix[2][0];

	// since this method is used for numBeacons >= 5, we can afford
	// to be aggressive!
	resultK = Math.abs(resultK);

	if(resultK >= 0)
	    resultK = Math.sqrt(resultK);
	else {
	    //System.out.println("PositionModule::positionEstimate() "+
	    //"Invalid sample: k squared is negative "+resultK);
	    return  null; // should just throw away the sample
	}

	/*
	  if(Math.abs(resultK - K_NOMINAL) > K_THRESH) { 
	    System.out.println("PositionModule::positionEstimate() K threshhold exceeded");
	    //System.out.println("\t"+resultK+" "+K_THRESH);
	    return null;
	    }
	*/

 	double resultZ = -1;
	for(int i = 0; resultZ < 0 && i < numBeacons; i++) {
	    resultZ = sqr(resultK*dist[i]) - sqr(resultX - x[i]) - 
		sqr(resultY - y[i]);
	    i++;
	}

	//System.out.println(resultK+" "+dist[0]+" "+resultX+" "+x[0]+" "+resultY+" "+y[0]);
	resultZ = Math.abs(resultZ);
	if(resultZ >= 0)
	    resultZ = Math.sqrt(resultZ);
	else {
	    //System.out.println("PositionModule::positionEstimate() "+
	    //"negZ: x="+resultX+" y="+resultY+" z="+resultZ+" k="+resultK);
	    return null; // should just throw away the sample
	}

	for(int i = 0; i < numBeacons; i++)
	    d_bar[i] = resultK*dist[i];

	PositionData result = 
	    new PositionData(curTime(), 
			     resultX, resultY, resultZ, x, y, z,
			     space, id, d_bar, resultK);

	System.out.println(resultK+" "+resultX+" "+" "+resultY+" "+resultZ);
	
	return result;
    }

    public PositionData positionEstimate2DVKnown(double[] x, 
						 double[] y,
						 String[] space,
						 int[] id,
						 double[] dist)
    {
	int numBeacons = x.length;

	double[] d_bar   = new double[numBeacons];
	double[] Z_ZEROS = new double[numBeacons];
	
	double[][] amatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    amatrix[i] = new double[2];
	    amatrix[i][0] = 2*(x[i+1]-x[0]);
	    amatrix[i][1] = 2*(y[i+1]-y[0]);
	}
	double[][] bmatrix = new double[numBeacons-1][];
	for(int i = 0; i < numBeacons-1; i++) {
	    bmatrix[i] = new double[1];
	    bmatrix[i][0] = K_NOMINALSQR*diffsqr(dist[0], dist[i+1])
		+ diffsqr(x[i+1], x[0]) + diffsqr(y[i+1], y[0]);
	}

	Matrix A = new Matrix(amatrix);
	Matrix B = new Matrix(bmatrix);

	if(numBeacons == 3) {
	    double determinate = A.det();
	    //System.out.println("determinate: "+determinate);
	    
	    if(Math.abs(determinate) < 0.000000001) {
		System.out.println("PositionModule::positionEstimate() determinate is zero");
		return null;
	    }
	}

	Matrix X = (new QRDecomposition(A)).solve(B);

	double[][] xmatrix = X.getArray();
	    
	double resultX = xmatrix[0][0];
	double resultY = xmatrix[1][0];
	//double resultK = xmatrix[2][0];

	// since this method is used for numBeacons >= 5, we can afford
	// to be aggressive!
	//resultK = Math.abs(resultK);

	/*
	  if(resultK >= 0)
	  resultK = Math.sqrt(resultK);
	  else {
	  //System.out.println("PositionModule::positionEstimate() "+
	  //"Invalid sample: k squared is negative "+resultK);
	  return  null; // should just throw away the sample
	  }
	*/

	/*
	  if(Math.abs(resultK - K_NOMINAL) > K_THRESH) { 
	    System.out.println("PositionModule::positionEstimate() K threshhold exceeded");
	    //System.out.println("\t"+resultK+" "+K_THRESH);
	    return null;
	    }
	*/

	for(int i = 0; i < numBeacons; i++)
	    d_bar[i] = K_NOMINAL*dist[i];

	PositionData result = 
	    new PositionData(curTime(), 
			     resultX, resultY, 0, x, y, Z_ZEROS, space, id, d_bar, K_NOMINAL);

	System.out.println(resultX+" "+" "+resultY);
	
	return result;
    }
	
    public static void main(String[] args) {
	PositionSolver ps = new PositionSolver(1, 10000000);
	int numBeacons = 3;
	double[] x = new double[numBeacons];
	double[] y = new double[numBeacons];
	double[] z = new double[numBeacons];
	double[] dist = new double[numBeacons];
	String[] space = new String[numBeacons];
	int[] id    = new int[numBeacons];

	x[0] = 50;
	y[0] = 50;
	z[0] = 0;

	x[1] = 200;
	y[1] = 50;
	z[1] = 0;

	x[2] = 50;
	y[2] = 180;
	z[2] = 0;

	dist[0] = 0;	
	dist[1] = Math.sqrt(sqr(x[1] - x[0]) + sqr(y[1] - x[0]));
	dist[2] = Math.sqrt(sqr(x[2] - x[0]) + sqr(y[2] - x[0]));

	PositionData a = 
	    ps.positionEstimate2DVKnown(x, y, space, id, dist);

	PositionData b = 
	    ps.positionEstimateVKnown(x, y, z, space, id, dist);
	
	System.out.println("a: "+a);
	System.out.println("b: "+b);
    }
}


    // solves system of equations here.

/*
01 *
02
03
12 *
13
23 *
*/

    /* what should be done for an over specified system? */
    /* is it better to fix speed of sound? and use the extra beacon
       to refine estimate? */

    /* select equation combos */

/*
    private static int[] comboIndexA = 
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 3 };
    private static int[] comboIndexB = 
    { 1, 1, 1, 2, 2, 2, 3, 3, 4, 2, 2, 2, 3, 3, 4, 3, 3, 4, 4 };
    private static int[] comboIndexC = 
    { 2, 4, 5, 3, 4, 5, 4, 5, 5, 3, 4, 5, 4, 5, 5, 4, 5, 5, 5 };
*/
    //        2  3  4     6  7  8  9  10 11 12 13 14 15 16 17 18 19

    /* equation combos */
/*  
private static int[] indexI = { 0, 0, 0, 1, 1, 2 };
    private static int[] indexJ = { 1, 2, 3, 2, 3, 3 };
    

    public PositionData positionEstimate(double[] x, 
					 double[] y,
					 double[] z,
					 String[] space,
					 int[] id,
					 double[] dist,
					 int combo)
    {
	double[] d_bar = new double[x.length];

	int j0 = indexJ[comboIndexA[combo]]; int i0 = indexI[comboIndexA[combo]];
	int j1 = indexJ[comboIndexB[combo]]; int i1 = indexI[comboIndexB[combo]];
	int j2 = indexJ[comboIndexC[combo]]; int i2 = indexI[comboIndexC[combo]];

	for(int i = 0; i < x.length; i++) {
	    System.out.println("input "+i+": "+x[i]+" "+y[i]+" "+space[i]+" "+dist[i]);
	}

	double[][] amatrix = 
	{{2*(x[j0]-x[i0]), 2*(y[j0]-y[i0]), diffsqr(dist[j0],dist[i0])},
	 {2*(x[j1]-x[i1]), 2*(y[j1]-y[i1]), diffsqr(dist[j1],dist[i1])},
	 {2*(x[j2]-x[i2]), 2*(y[j2]-y[i2]), diffsqr(dist[j2],dist[i2])}};
	  
	double[][] bmatrix =
	{ { diffsqr(x[j0], x[i0]) + diffsqr(y[j0], y[i0]) },
	  { diffsqr(x[j1], x[i1]) + diffsqr(y[j1], y[i1]) },
	  { diffsqr(x[j2], x[i2]) + diffsqr(y[j2], y[i2]) } };	    

	Matrix A = new Matrix(amatrix);
	Matrix B = new Matrix(bmatrix);

	//A.print(10, 3);
	double determinate = A.det();

	// do a determinate check to see if a unique solution exists!!!
	//	System.out.println("A: "+determinate);

	System.out.println("determinate: "+determinate);

	if(Math.abs(determinate) < 0.000000001) {
	    Util.log("PositionModule::positionEstimate() determinate is zero");
	    System.out.println("PositionModule::positionEstimate() determinate is zero");
	    return null; // throw away sample.
	}
	Matrix X = (new QRDecomposition(A)).solve(B);

	double[][] xmatrix = X.getArray();
	    
	double resultX = xmatrix[0][0];
	double resultY = xmatrix[1][0];
	double resultK = xmatrix[2][0];

	//resultK = Math.abs(resultK);

	if(resultK >= 0)
	    resultK = Math.sqrt(resultK);
	else {
	    Util.log("PositionModule::positionEstimate() "+
		     "Invalid sample: k squared is negative "+resultK);
	    System.out.println("PositionModule::positionEstimate() "+
			       "Invalid sample: k squared is negative "+resultK);
	    return null; // should just throw away the sample
	}
	Util.log("K = "+resultK);
	
	//resultK = 0.868; 
	
	if(Math.abs(resultK - K_NOMINAL) > K_THRESH) { 
	    Util.log("PositionModule::positionEstimate() K threshhold exceeded");
	    Util.log("\t"+resultK+" "+K_THRESH);
	    System.out.println("PositionModule::positionEstimate() K threshhold exceeded");
	    //System.out.println("\t"+resultK+" "+K_THRESH);
	    return null;
	}
	
 	double resultZ = 
	    sqr(resultK*dist[0]) - sqr(resultX - x[0]) - sqr(resultY - y[0]);

	System.out.println(resultK+" "+dist[0]+" "+resultX+" "+x[0]+" "+resultY+" "+y[0]);
	//resultZ = Math.abs(resultZ);

	if(resultZ >= 0)
	    resultZ = Math.sqrt(resultZ);
	else {
	    Util.log("PositionModule::positionEstimate() "+
		     "Invalid sample: z squared is negative "+resultZ);
	    System.out.println("PositionModule::positionEstimate() "+
			       "negZ: x="+resultX+" y="+resultY+" z="+resultZ+" k="+resultK);

	    return null; // should just throw away the sample
	}

	// find all the d_bars
	for(int i = 0; i < x.length; i++)
	    d_bar[i] = resultK*dist[i];

	PositionData result = 
	    new PositionData(curTime(), 
			     resultX, resultY, resultZ, 
			     space, id, d_bar, resultK);
	double kdiff = Math.abs(resultK-K_NOMINAL);
	System.out.println("time="+System.currentTimeMillis()+" x="+resultX+" y="+resultY+" z="+resultZ+" kdiff="+kdiff);
	return result;
    }

    public PositionData positionEstimate3Beacons(double[] x, 
						 double[] y,
						 double[] z,
						 String[] space,
						 int[] id,
						 double[] dist,
						 int combo)
    {
	double[] d_bar = new double[3];

	int j0 = indexJ[comboIndexA[combo]]; int i0 = indexI[comboIndexA[combo]];
	int j1 = indexJ[comboIndexB[combo]]; int i1 = indexI[comboIndexB[combo]];

	for(int i = 0; i < 3; i++) {
	    System.out.println("input "+i+": "+x[i]+" "+y[i]+" "+space[i]+" "+dist[i]);
	}

	double[][] amatrix = 
	{{2*(x[j0]-x[i0]), 2*(y[j0]-y[i0])},
	 {2*(x[j1]-x[i1]), 2*(y[j1]-y[i1])}};
	  
	double[][] bmatrix =
	{ { diffsqr(x[j0], x[i0]) + diffsqr(y[j0], y[i0]) - K_NOMINAL*K_NOMINAL*diffsqr(dist[j0],dist[i0])},
	  { diffsqr(x[j1], x[i1]) + diffsqr(y[j1], y[i1]) - K_NOMINAL*K_NOMINAL*diffsqr(dist[j1],dist[i1])} };

	Matrix A = new Matrix(amatrix);
	Matrix B = new Matrix(bmatrix);

	double determinate = A.det();

	// do a determinate check to see if a unique solution exists!!!
	System.out.println("determinate: "+determinate);

	if(Math.abs(determinate) < 0.000000001) {
	    Util.log("PositionModule::positionEstimate() determinate is zero");
	    System.out.println("PositionModule::positionEstimate() determinate is zero");
	    return null; // throw away sample.
	}
	Matrix X = (new QRDecomposition(A)).solve(B);

	double[][] xmatrix = X.getArray();
	    
	double resultX = xmatrix[0][0];
	double resultY = xmatrix[1][0];

 	double resultZ =  diffsqr(K_NOMINAL*dist[0], resultX - x[0]) - sqr(resultY - y[0]);

	// resultZ is resultZ squared now
	System.out.println(K_NOMINAL+" "+dist[0]+" "+resultX+" "+x[0]+" "+resultY+" "+y[0]+" "+resultZ);
	//resultZ = Math.abs(resultZ);

	
	//  if(resultZ < 0) {
	//  Util.log("PositionModule::positionEstimate() "+
	//  "Invalid sample: z squared is negative "+resultZ);
	//  System.out.println("PositionModule::positionEstimate() "+
	//  "negZ: x="+resultX+" y="+resultY+" z="+resultZ+" k="+K_NOMINAL);
	//  return null; // should just throw away the sample
	//  }

	double absZ = Math.abs(resultZ);
	
	
	//if(absZ > ZSQR_THRESH)
	//    return null;

	resultZ = Math.sqrt(absZ);


	// find all the d_bars
 	for(int i = 0; i < 3; i++)
	     d_bar[i] = K_NOMINAL*dist[i];

	PositionData result = 
	     new PositionData(curTime(), resultX, resultY, resultZ, space, id, d_bar, K_NOMINAL, 3);
	System.out.println("3 time="+System.currentTimeMillis()+" x="+resultX+" y="+resultY+" z="+resultZ);
	return result;
    } 


    protected long curTime()
     {
	 return System.currentTimeMillis();
    } 

    public static double sqr(double x) { return x*x; }

} 

*/ 
