/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.reader;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.beacontable.*;
import cricketdaemon.reader.parse.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Retrieves and parses beacon data from the gpsd.
 *
 * @author Allen Miu
 */
public class GPSDReader extends ModuleThread
{
    private BufferedReader in = null;
    private OutputStream out = null;

    private static final String REGISTRATION_COMMAND = "r\n";
    private static final String ENCODING = "UTF-8";
    private static final String MAGIC_TOKEN = "$Cricket2";
    private static final String TOKEN_DELIM = ",";
    private static final String VALUE_DELIM = "=";

    private String addr;
    private int port;

    public GPSDReader(String addr, int port)
    {
	super("GPSDReader", 0);
	this.addr = addr;
	this.port = port;
	connect();
    }

    private synchronized void connect()
    {
	while(true) {
	    try {
		if(in != null)
		    in.close();
		if(out != null)
		    out.close();
		
		Socket s = new Socket(InetAddress.getByName(addr), port);
		InputStreamReader isr  = 
		    new InputStreamReader(s.getInputStream(), ENCODING);
		in  = new BufferedReader(isr);
		out = s.getOutputStream();
		
		sendRegistration(out);
		
		return; // success
	    }
	    catch (UnknownHostException e) {
		Util.warn("GPSDReader() invalid address "+addr);
	    }
	    catch (IOException e) {
		Util.warn("GPSDReader() error contacting gpsd at "+
			  addr+" port "+port);
	    }

	    try {
		wait(500);
	    }
	    catch(InterruptedException ie) {
	    }
	}
    }

    private void sendRegistration(OutputStream out) 
	throws IOException
    {
	try {
	    byte[] r = REGISTRATION_COMMAND.getBytes(ENCODING);
	    out.write(r);
	}
	catch (UnsupportedEncodingException e) {
	    Util.assert(false, "GPSDReader::sendRegistration() system does "+
			"not support "+ENCODING+" encoding");
	}
    }

    public void run()
    {
	while(true)
	    process();
    }

    protected void moduleCleanUp()
    {  // never really gets called because run() is overridden
	return;
    }

    int foo = 0;

    protected void process()
    {
	String inputStr = null;
	try {
	    inputStr = in.readLine();

	    System.out.println("GPSD "+foo+": "+inputStr);
	    foo++;

	    if(inputStr == null) {
		throw new IOException("GPSD returns null");
	    }

	    StringTokenizer s = 
		new StringTokenizer(inputStr, TOKEN_DELIM);

	    boolean validDistance = false;

	    if(MAGIC_TOKEN.equals(s.nextToken())) {
		Beacon beacon = new Beacon(System.currentTimeMillis());
		BitSet mask   = (BitSet) CLEARMASK.clone();
		while(s.hasMoreTokens()) {
		    String t = s.nextToken();
		    int i = t.indexOf(VALUE_DELIM);
		    if(i <= 0) {
			Util.log("GPSDReader::process() invalid token: "+t);
			return; // silently drops
		    }
		    String type  = t.substring(0, i);
		    int fieldID  = ParseMap.typeToFieldID(type);
		    
		    if(fieldID < 0) {
			Util.log("GPSDReader::process() invalid type: "+t);
			return; // silently drops
		    }

		    /* GPSD v3 build 1 sometimes omit the distance field
		       if it only hears the radio but not the
		       ultrasound.  Here, we're filtering out all
		       readings without a distance because some
		       CricketDaemon may assume the distance field is
		       always available.  */
		    if(fieldID == ParseMap.DISTANCE)
			validDistance = true;

		    String value = t.substring(i+1, t.length());
		    FieldParser f = ParseMap.map[fieldID];
		    Object o = f.GPSDStringToReturnType(value);
		    if(o == null) {
			Util.log("GPSDReader::process() invalid value: "+t);
			return;
		    }
		    beacon.insertValue(f.unparse(o), fieldID);
		    mask.set(fieldID);
		}
		//		System.out.println("GPSD: "+beacon.toString2());
		if(validDistance)
		    transfer(beacon, mask);
		// else silently drop the reading
	    }
	}
	catch (IOException e) {
	    //Util.assert(false, "GPSDReader::process() read error");
	    Util.warn("GPSDReader::process() read error: "+inputStr);
	    connect();
	}
    }

    public static void main(String[] args) 
    {
	try {
	    GPSDReader g = 
		new GPSDReader("18.24.6.238", 2947);
	    DebugModule d = new DebugModule(g);

	    /* comment out the above and uncomment this for translation
	       GPSDReader g = 
	       new GPSDReader(InetAddress.getByName("18.24.6.238"), 2947);
	       BeaconTableModule b = new BeaconTableModule(g, "translate.conf");
	       DebugModule d = new DebugModule(b);
	       b.start();
	    */ 

	    g.start();
	    d.start();
	    g.join();
	}
	catch (InterruptedException e) {
	}
    }
}

