/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.reader;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.logging.LoggingModule;
import java.util.*;
import java.io.*;

/**
 * Simulates a PortReader by feeding test beacon data from a file
 * generated by the Logging module.
 *
 * @author Allen Miu
 */
public class PlaybackReader extends ModuleThread
{
    BufferedReader in;  // data file stream
    long curTimeStamp, prevTimeStamp, breakPoint;
    boolean moreLines, realTimeMode;

    private static final int INIT_VAL = -1;

    public PlaybackReader(String inputFile, boolean rt, long breakP)
    {
	super("PlaybackReader", 0);
	moreLines = true;
	curTimeStamp = INIT_VAL;
	prevTimeStamp = INIT_VAL;
	realTimeMode = rt;
	breakPoint   = breakP;
	try {
	    in = new BufferedReader(new FileReader(inputFile));
	}
	catch (IOException io) {
	    Util.assert(false, "PlaybackReader() cannot open file: "
			+inputFile);
	}
    }

    synchronized public void run()
    {
	long delay = 0;
	try {
	    while(moreLines) {
		prevTimeStamp = curTimeStamp;
		process();
		if(realTimeMode)
		    delay = curTimeStamp - prevTimeStamp;
		if(delay > 0)  // prevents blocking
		    wait(delay);
	    }
	} catch (InterruptedException e) {
	    Util.log(getName() +" interrupted");
	}
	moduleCleanUp();
	Util.log(getName() +" exits");
    }

    protected void moduleCleanUp()
    {
	try {
	    in.close();
	}
	catch (IOException e) {
	    Util.warn("PlaybackReader::moduleCleanUp() error closing file");
	}
    }

    protected void process()
    {
	String s = null;
	try {
	    s = in.readLine();
	    System.out.println("read: "+s);
	}
	catch (IOException e) {
	    Util.assert(false, "PlaybackReader::process() error reading file");
	}

	if(s == null) {
	    moreLines = false;
	    curTimeStamp = prevTimeStamp + 1;
	    Util.log("PlaybackReader::process() end of file");
	    return;
	}

	StringTokenizer st = new StringTokenizer(s);
	PlaybackBeacon p = LoggingModule.parsePlayback(st);

	if( p != null ) {
	    if(p.beacon.seqNum > breakPoint)
		Util.assert(false, "Breakpoint "+breakPoint+" reached");

	    curTimeStamp = p.timeStamp;
	    //	    System.out.println(curTimeStamp);
	    if(prevTimeStamp == INIT_VAL)
		prevTimeStamp = curTimeStamp - 1;

	    /* send beacons */
	    transfer(p.beacon, p.mask);
	}
	else
	    curTimeStamp = prevTimeStamp + 1;
    }

}
