/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.reader;

import cricketdaemon.*;
import cricketdaemon.module.*;
import cricketdaemon.reader.parse.*;
import cricketdaemon.reader.serial.*;
import java.util.*;
import javax.comm.*;
import java.io.*;


/**
 * Retrieves and parses beacon data from the SECOND VERSION
 * of the cricket listeners.
 *
 * @author Bodhi Priyantha, Allen Miu
 */
public class SerialPortReader extends ModuleThread 
    implements SerialPortEventListener
{
    /**
     * Start of Packet markers and field seperators
     */
    private final byte ZERO  = 0;
    private final byte ONE   = 1;
    private final byte TWO   = 2;
    private final byte THREE = 3;

    /** 
     * Stream from which data is read
    */
    protected InputStream in;
    
    /**
     * Input buffer
     */
    protected static final int FIELD_ID_LENGTH = 1;
    protected static final int MAX_FIELD_LENGTH = 256;
    protected byte[] fieldData =
	new byte[MAX_FIELD_LENGTH+FIELD_ID_LENGTH+6]; 

    protected BitSet packetFieldMask = new BitSet();
    protected static FieldParser parser = FieldParser_Integer.instance;

    /* stuff to support dynamic class loading */
    static class SerialPortReaderFactory implements PortReaderFactory
    {
	public ModuleThread create(String configFile)
	{
	    return new SerialPortReader(configFile);
	}
    }

    static 
    {
	PortReaderCreator.
	    portReaderFactories.put(PortReaderCreator.PATH+
				    "SerialPortReader", 
				    new SerialPortReaderFactory());
    }

    /** Creates new SerialPortReader that reads from serial port */
    public SerialPortReader(String configFile)
    {
	super("SerialPortReader", 0);
	CommPortIdentifier portId;
	SerialPort sp = null;
	Properties p  = getProperties(configFile);
	String port   = p.getProperty("port").trim();

	// plus 6 bytes padding for end of field and end of beacon markers
	try{
	    int baudrate  = Integer.parseInt(p.getProperty("baudrate"));
            Util.log("Trying to open port "+port);
            portId = CommPortIdentifier.getPortIdentifier(port);
	    Util.log("Opening port "+ portId.getName());
	    sp = addPort(portId);
	    configPort(sp, port, baudrate);
        } 
	catch (NumberFormatException e) {
	    Util.assert(false, "Invalid baudrate specified in config file");
	}
	catch (NoSuchPortException e) {
            e.printStackTrace();
            Util.assert(false, "Port " + port + " not found!");
	}
	// get I/O streams
	try{
            in = sp.getInputStream();
	}
	catch (IOException e) {
            e.printStackTrace();
            Util.assert(false, "Could not get serial port I/O streams");
        }

	Util.log("SerialPortReader initialization done");
    }

    public SerialPortReader(InputStream i) 
    {
	super("SerialPortReader", 0);
	Util.assert(i != null, "SerialPortReader() InputStream null");
	in = i;
    }

    private Properties getProperties(String fileName) 
    {
	Properties result    = new Properties();
	InputStream file = null;
	try{
	    file = ClassLoader.getSystemResourceAsStream(fileName);
	    Util.assert(file != null, 
			"SerialPortReader::getProperties() "+
			fileName+" not found");
	    result.load(file);
	} catch (IOException e){
	    Util.assert(false, "SerialPortReader::getProperties() "+
			"Could not load "+fileName);
	}
	String platform = System.getProperty("os.name");
	//serach for key words to determine type
	
	String port = result.getProperty("port").trim();

	if(port.length() == 0) {
	    if(platform.indexOf("Windows") >= 0)
		port = "COM1";
	    else if( platform.indexOf("Linux") >= 0)
		port = "/dev/ttyS0";
	}
	System.out.println("Running "+CricketDaemon.VERSION+" on "
			   +platform+" using port: "+port);
	result.setProperty("port", port);

	return result;
    }

    /**
     * Opens the port and assign to serial port.
    */
    private SerialPort addPort(CommPortIdentifier portId)
    {
	SerialPort sp = null;

        // is this a serial port ?
	Util.assert(portId.getPortType() == CommPortIdentifier.PORT_SERIAL,
		    portId.getName() + " is not a serial port!");

	Util.assert(!portId.isCurrentlyOwned(),
		    "Detected " + portId.getName() + " in use by "
		    + portId.getCurrentOwner());

	// try to open the port
	try {
	    sp = (SerialPort) portId.open("cricket", 2000);
	    Util.assert(sp != null, 
			"Error opening port " + portId.getName());
	} catch (PortInUseException e) {
	    Util.assert(false, 
			portId.getName()+" is in use by "+ e.currentOwner);
	}
	return sp;
    }

    /**
     * Configure the serail port accordingly
     */
    private void configPort(SerialPort sp, String port, int baudrate) 
    {

	/*
	 *  Setup an event listener for the port
	 */
	/*
	try {
	    sp.addEventListener(this);
	} catch (TooManyListenersException tmle) {
	    tmle.printStackTrace();
	}
	*/

        // set serial port parameters
	try{
            sp.setSerialPortParams(baudrate,
				   SerialPort.DATABITS_8,
				   SerialPort.STOPBITS_1,
				   SerialPort.PARITY_NONE);
            sp.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);
	    /*
	    sp.notifyOnBreakInterrupt(true);
	    sp.notifyOnCarrierDetect(true);
	    sp.notifyOnCTS(true);
	    sp.notifyOnDataAvailable(true);
	    sp.notifyOnDSR(true);
	    sp.notifyOnFramingError(true);
	    sp.notifyOnOutputEmpty(true);
	    sp.notifyOnOverrunError(true);
	    sp.notifyOnParityError(true);
	    sp.notifyOnRingIndicator(true);
	    */
	    Util.log("configPort "+baudrate+" "+port+" "+sp);
	} catch (UnsupportedCommOperationException e) {
            Util.assert(false, "Couldn't set serial port parameters");
	}
    }

    public void run()
    {
	while(true)
	    process();
	//Util.log(getName() +" exits");
    }

    protected void moduleCleanUp()
    {
	// never really gets called because run() is overridden
	return;
    }

    protected Beacon readData(BitSet mask) throws IOException 
    {
        boolean unstuffed; //tells if a leading 0 was removed
        byte char1, char2; // detects header
        int subFields;     // running sum of sub fields 
        int fieldPos;      // next empty position of the subfield byte array
        
	Beacon beacon = new Beacon(System.currentTimeMillis());

	// synchronization
	char2 = (byte) in.read();
	do {
	    do{    
		char1 = char2;
		char2 = (byte) in.read();
		//wait till we synch up with header (01)
	    } while ( (char1 != ZERO) || (char2 != ONE)); 
	    char1 = (byte) in.read();
	    char2 = (byte) in.read(); 
	} while ( (char1 != ZERO) || (char2 != ONE ) ); //check
	// for next 01 (we must get atleast one more 01 )  

	do{
	    char1 = (byte) in.read();
	    char2 = (byte) in.read(); 
	    //get rid of any further 01s
	} while ( (char1 == ZERO) && (char2 == ONE) ); 
	//now char1 contains 1st byte and char2 contains the2nd byte
	//of 1st subfield;


	// initialization to read fields
	fieldData[0] = char1; // filling parse window
	fieldData[1] = char2; // filling parse window
	fieldPos     = 2;

	subFields    = 0;
	unstuffed = false;

	// Packet Format
	// |01010101|field id low|field id high|data|202|...
	// field id low|field id high|101|

	while (true) {    
	    if(fieldPos >= fieldData.length) {
		// buffer overrun
		Util.log("SerialPortReader::readData() buffer overrun");
		return null;
	    }
		
	    fieldData[fieldPos] = (byte) in.read();
	    fieldPos++; 
	    
	    //check for end of packet
	    if(!unstuffed) {
		if(( fieldData[fieldPos - 3] == ONE ) && 
		   ( fieldData[fieldPos - 2] == ZERO) &&
		   ( fieldData[fieldPos - 1] == ONE ))
		    {
			// get the last field
			parseField(fieldData, fieldPos, beacon, mask);
			break;                  
		    }
	    
		//check for new subfield
		if(( fieldData[fieldPos - 3] == TWO ) && 
		   ( fieldData[fieldPos - 2] == ZERO) &&
		   ( fieldData[fieldPos - 1] == TWO )) { 

		    parseField(fieldData, fieldPos, beacon, mask);

		    //reset
		    fieldData[0] = (byte) in.read(); // filling parse window
		    fieldData[1] = (byte) in.read(); // filling parse window
		    fieldPos     = 2;
		    subFields ++;
		}
		else if ( fieldData[fieldPos-3] == 0 ) { 
		    //do byte unstuffing for the byte at fieldPos - 3
		    //check for start of 0 string 
		    //remove stuffed 0
		    fieldData[fieldPos-3] = fieldData[fieldPos-2];
		    fieldData[fieldPos-2] = fieldData[fieldPos-1];          
		    fieldPos--; //remove the stuffed 0
		    unstuffed = true;
		} 
	    }
	    else if( fieldData[fieldPos - 3] != 0 ) { 
		// check for end of 0 string 
		unstuffed = false;
	    }
			
            // reached end of packet

            // check validity of data (to check the validity, allow for
            // an external module to register with this specifically for
            // error detection/correction by default if there is no
            // registered component, do not do any test
	}          

	return beacon;
    }    

    private void parseField(byte[] buf, int pos, Beacon beacon, BitSet mask) 
    {
	//Util.log("parseField "+pos);
	if(pos > 3) {
	    int fieldLength = pos - 3 - FIELD_ID_LENGTH;
	    if(fieldLength < 0)
		Util.assert(false, "SerialPortReader::readData fieldID short");
	    Integer fieldID = (Integer) 
		parser.parse(fieldData, 0, FIELD_ID_LENGTH);
	    byte[] fieldContent = new byte[fieldLength];
	    System.arraycopy(fieldData, FIELD_ID_LENGTH, fieldContent, 
			     0, fieldLength);
	    beacon.insertValue(fieldContent, fieldID.intValue());
	    mask.set(fieldID.intValue());
	}
	else
	    Util.warn("SerialPortReader::readData() null field");
	/*
	  System.out.println("fid "+fieldID+" flength "+
	  fieldLength);
	*/
    }


    protected void process()
    {
	Beacon newBeacon = null;
	packetFieldMask = (BitSet) CLEARMASK.clone();

	try {
	    newBeacon = readData(packetFieldMask);
	} catch(IOException e){
            e.printStackTrace();
            Util.assert(false, "Error in reading serial port");
	}

	/* send beacons */
	if(newBeacon != null)
	    transfer(newBeacon, packetFieldMask);
    }

    public static class Test extends ModuleThread
    {

	public Test(SerialPortReader r) {
	    super("SerialPortReader::TestModule",0);
	    BitSet fieldMask = new BitSet();
	    fieldMask.set(ParseMap.BEACONID);
	    fieldMask.set(ParseMap.DISTANCE);
	    r.register(this, fieldMask);
	}
	
	protected void process()
	{
	    Beacon newData = (Beacon) ((QueueItem) dequeue()).value;
	    System.out.println(newData);
	}

	protected void moduleCleanUp() { return; }
    }

    public static class TestStream extends ByteArrayInputStream
    {
	private int timeout;
	public TestStream(byte[] buf, int to) 
	{
	    super(buf);
	    timeout = to;
	}

	public TestStream(byte[] buf, int offset, int length, int to)
	{
	    super(buf, offset, length);
	    timeout = to;
	}

	public int read()
	{
	    int next = super.read();
	    if(next >= 0)
		return next;
	    else {
		reset();
		try {
		    synchronized(this) {
			wait(timeout);
		    }
		}
		catch(InterruptedException e) {
		}
		return super.read(); //this.read() could produce inf loop
	    }
	}
    }

    public static void main(String[] args)
    {
	// beaconid 0x11, distance = 0xa
	byte[] testBytes = 
	{ 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 2, 0, 2, 2, 10, 1, 0, 1 };
	TestStream ts = new TestStream(testBytes, 250);
	SerialPortReader pr = new SerialPortReader(ts);
	Test t = new Test(pr);
	
	t.start();
	pr.start();
	try {
	    pr.join();
	}
	catch(InterruptedException e) {
	}
    }

    /**
     * Implementation of the event listener
     */
    public void serialEvent( SerialPortEvent e) {
	int i = e.getEventType();
	switch(i){
	case SerialPortEvent.DATA_AVAILABLE:
	    System.out.println("Data Available");
	    break;
	case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
	    System.out.println("Output Buffer Empty");
	    break;
	case SerialPortEvent.CTS:
	    System.out.println("Clear to Send");
	    break;
	case SerialPortEvent.DSR:
	    System.out.println("Data Set Ready");
	    break;
	case SerialPortEvent.RI:
	    System.out.println("Ring Indicator");
	    break;
	case SerialPortEvent.CD:
	    System.out.println("Carrier Detect");
	    break;
	case SerialPortEvent.OE:
	    System.out.println("Overrun Error");
	    break;
	case SerialPortEvent.PE:
	    System.out.println("Parity Error");
	    break;
	case SerialPortEvent.FE:
	    System.out.println("Framing Error");
	    break;
	case SerialPortEvent.BI:
	    System.out.println("Break Interrupt");
	    break;
	default: 
	    System.out.println("Unknown event");
	}
    }
}
