/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.server;

import cricketdaemon.clientlib.parse.Field;
import cricketdaemon.reader.CartesianCoord;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.space.*;
import cricketdaemon.space2.*;
import cricketdaemon.position.*;
import cricketdaemon.beacontable.*;
import cricketdaemon.beaconstat.*;
import cricketdaemon.beaconstat.BeaconDistStat.*;
import cricketdaemon.orient.*;
import cricketdaemon.Util;
import cricketdaemon.module.*;
import java.util.*;
import java.net.*;
import java.io.*;

/**
 * A module that implements a UDP server for the CricketDaemon.
 *
 * @author Allen Miu
 */
public class ServerModule extends PeriodicModuleThread
    implements 
	OrientData.OrientDataHandler,
	BeaconStatData.BeaconStatDataHandler,
	PositionData.PositionDataHandler,
	SpaceData.SpaceDataHandler,
	ServerConnListener
{
    public static final int MAX_PAYLOAD_LENGTH = 1500;

    public static final String DEFAULT_ENCODING = "UTF-8";

    public static final String CRICKETSERVER_VER = "CricketServer1.0";

    public static final int PORT = 5001;

    // period to refresh registration with server
    public static final int REFRESH_PERIOD = 500; 
    
    private ArrayList registeredClients;

    //    private DatagramSocket out;

    private BeaconTableModule beaconTableModule = null;
    private PositionModule positionModule = null;

    public ServerModule(int port, BeaconTableModule btm)
    {
	super("ServerModule", REFRESH_PERIOD);
	registeredClients = new ArrayList();
	ServerThread st = new ServerThread(this, port);
	beaconTableModule = btm;
	st.start();
    }

    // FIXME: this should be part of the "XXXHandler" abstraction...
    public void registerModuleThread(ModuleThread m)
    {
	BitSet fieldMask = new BitSet();
	if(m instanceof SpaceModule || m instanceof Space2Module) {
	    fieldMask.set(SpaceModule.MEAN);
	    fieldMask.set(SpaceModule.MODE);
	}
	else if(m instanceof PositionModule) {
	    fieldMask.set(PositionModule.CARTESIAN_COORD);
	    fieldMask.set(PositionModule.DISTANCE_ESTIMATES);
	    fieldMask.set(PositionModule.POSERR);
	    positionModule = (PositionModule) m;
	}
	else if(m instanceof BeaconStatModule) {
	    fieldMask.set(BeaconStatModule.STAT);
	}
	else if(m instanceof OrientModule) {
	    fieldMask.set(OrientModule.ANGLE);
	    fieldMask.set(OrientModule.ANCHOR);
	}
	m.register(this, fieldMask);
    }
    
    // poll each TCP socket to see if there is incoming data
    byte[] pollBuf = new byte[64];
    synchronized void pollSockets()
    {
	for(int i = 0; i < registeredClients.size(); i++) {
	    ClientRecord rec = (ClientRecord) registeredClients.get(i);
	    try {
		BufferedReader in = rec.getBufferedReader();

		if(!in.ready()) 
		    continue;

		// Java handles buffer overflows here...
		/*
		int bytesRead = in.read(pollBuf); 
		String magicToken = 
		new String(pollBuf, 0, 
		ServerThread.REGISTER_TOKEN.length(), 
		DEFAULT_ENCODING);
		*/

		StringTokenizer indata = new StringTokenizer(in.readLine());
		if(!indata.hasMoreTokens())
		    continue;

		String type = indata.nextToken();

		System.out.println("foobar pollSockets "+type);

		// add hooks to handle different commands here!
		// format: type tokens\n
		// where '\n' delimits different commands

		if(type.equals(ServerThread.REGISTER_TOKEN)) {
		    /* preserve order */
		    registeredClients.remove(i);
		    registeredClients.add(rec);
		    rec.state = ClientRecord.ESTABLISHED;
		}
		if(type.equals(ServerThread.BEACONTABLE_TOKEN) && (beaconTableModule != null)) {
		    beaconTableModule.parseCmd(indata);
		}

		if(type.equals(ServerThread.POSITION_TOKEN) && (positionModule != null)) {
		    positionModule.parseCmd(indata);
		}

	    }
	    catch (NoSuchElementException nsee) {
		Util.warn("ServerModule::pollSocket() invalid command from "+rec);
	    }
	    catch (IOException e) {
		Util.warn("ServerModule::pollSocket() kill socket "+rec);
		registeredClients.remove(i);
	    }
	}
    }

    public synchronized void registerClient(Socket s) 
    {
	ClientRecord newRec = new ClientRecord(s, 0);
	int index = ClientRecord.indexOf(newRec, registeredClients);
	if(index >= 0)
	    registeredClients.remove(index);  // remove old entry
	registeredClients.add(newRec);
    }

    synchronized private void notifyClients(byte[] payload) 
    {
	// Should check for this in UDP
	/*
	Util.assert(payload.length <= MAX_PAYLOAD_LENGTH,
		    "ServerModule::notifyClients() payload length="
		    + payload.length);
	*/

	//	System.out.println("send "+System.currentTimeMillis());

	// may need to modify this to send only to the interested clients

	//System.out.println("notifyClients");

	for(int i = 0; i < registeredClients.size(); i++) {
	    ClientRecord item = (ClientRecord) registeredClients.get(i);
	    if(item.state == ClientRecord.ESTABLISHED) {
		try {
		    //System.out.println("notifyClients 2");
		    // if this hangs, it means one of the clients hung
		    item.getSocket().getOutputStream().write(payload);
		    //System.out.println("notifyClients 3");
		}
		catch(IOException io) {
		    //item.getSocket().close();
		    //Util.log("ServerModule::notifyClient() kills socket "+item);
		    System.out.println("ServerModule::notifyClient() kills socket "+item);
		    registeredClients.remove(i);
		}
	    }
	}

	//System.out.println("notifyClients returns");
    }

    protected void periodicallyProcess()
    {
	pollSockets();
    }

    protected void moduleCleanUp()
    {
	// no need to shut down all connections as it is UDP
	return;
    }

    private String createPosType(CartesianCoord c) 
    {
	StringBuffer coord = new StringBuffer();
	Field.toFieldAppend(coord, "x", String.valueOf(c.x));
	Field.toFieldAppend(coord, "y", String.valueOf(c.y));
	Field.toFieldAppend(coord, "z", String.valueOf(c.z));
	return Field.toField("pos", coord);
    }

    public void processSpaceData(SpaceData newData, BitSet mask)
    {
	if(!mask.get(SpaceModule.MODE)) {
	    Util.log("ServerModule::processSpaceData() drops "+newData);
	    return; // silently drop
	}

	DistanceEstimate[] est = newData.getMode();
	int i = 0;
	while(i < est.length) {
	    StringBuffer s = new StringBuffer();
	    do {
		StringBuffer record = new StringBuffer();
		
		// FIXME: timestamps and/or seq num
		String space      = est[i].space;
		String beaconID   = String.valueOf(est[i].beaconID);
		String distance   = String.valueOf(est[i].distance);

		Field.toFieldAppend(record, "space", space);
		Field.toFieldAppend(record, "id", beaconID);
		Field.toFieldAppend(record, "dist", distance);
		
		Field.toFieldAppend(s, "dist_est", record.toString());  // nest
		i++;

		// to chop up the output into 1500 chunks (if necessary)
		// 300 is arbitrary (give space for other headers just in case)
	    } while ( est.length < (MAX_PAYLOAD_LENGTH - 300) &&
		      i < est.length );
	    send(Field.toField("array:dist_est", s), i < est.length);
	}
    }

    public void processPositionData(PositionData newData, BitSet mask)
    {
	DistanceEstimate[] est = newData.getDistances();
	int i = 0;
	while( i < est.length ) {
	    StringBuffer s = new StringBuffer();
	    do {
		StringBuffer record = new StringBuffer();
		String space      = est[i].space;
		String beaconID   = String.valueOf(est[i].beaconID);
		String distance   = String.valueOf(est[i].distance);
		
		Field.toFieldAppend(record, "space", space);
		Field.toFieldAppend(record, "id", beaconID);
		Field.toFieldAppend(record, "dist", distance);
		
		Field.toFieldAppend(s, "dist_est", record.toString());  // nest
		i++;
	    } while ( est.length < (MAX_PAYLOAD_LENGTH - 300) &&
		      i < est.length );
	    String st = createPosType(newData.getCoord()) + 
		Field.toField("array:dist_est",s);
	    send(Field.toField("device_pos",st), i < est.length);
	}

    }

    public void processBeaconStatData(BeaconStatData newData, BitSet mask)
    {
	if(!mask.get(BeaconStatModule.STAT)) {
	    Util.log("ServerModule::processBeaconStatData() drops "+newData);
	    return; // silently drop
	}

	// first, send the current space object
	send(Field.toField("cur_space", 
	     parseRecord((BeaconDistStat) newData.values.get(0))), false);

	// send the array 
	int i = 0;
	while(i < newData.values.size()) {
	    StringBuffer s = new StringBuffer();
	    do {
		BeaconDistStat b = (BeaconDistStat) newData.values.get(i);

		String record = parseRecord(b);
		s.append(record);

		i++;
		// 600 is determined by looking at the max record length*2

		// still could fragment improperly if record.length()
		// is greater than max payload length via one call to
		// parseRecord.
	    } while (s.length() < (MAX_PAYLOAD_LENGTH - 600) && 
		     i < newData.values.size());
		//}
	    //while (i < newData.values.size());

	    send(Field.toField("array:stat_est", s), i < newData.values.size());
	    //send(Field.toField("array:stat_est", s), false);
	} 

	    //send(Field.toField("array:stat_est", s));
    }

    private String parseRecord(BeaconDistStat b) 
    {
	StringBuffer record = new StringBuffer();
	Field.toFieldAppend(record, "space", b.space);
	Field.toFieldAppend(record, "id", String.valueOf(b.id));
	Field.toFieldAppend(record, "last_update", 
			    String.valueOf(b.lastUpdate));
	//	    System.out.println("lastUPdate: "+b.lastUpdate);
	StringBuffer stat = new StringBuffer();
	Field.toFieldAppend(stat, "dist", String.valueOf(b.distanceEst));
	Field.toFieldAppend(stat, "median", String.valueOf(b.stat.median));
	Field.toFieldAppend(stat, "mean", String.valueOf(b.stat.mean));
	Field.toFieldAppend(stat, "mode", String.valueOf(b.stat.mode));
	Field.toFieldAppend(stat, "max", String.valueOf(b.stat.max));
	Field.toFieldAppend(stat, "min", String.valueOf(b.stat.min));
	Field.toFieldAppend(stat, "stddev", String.valueOf(b.stat.stddev));
	Field.toFieldAppend(stat, "var", String.valueOf(b.stat.variance));

	Iterator it     = b.distances.iterator();
	StringBuffer as = new StringBuffer(); // array:sample
	while(it.hasNext()) {
	    BeaconDist d   = (BeaconDist) it.next();
	    StringBuffer s = new StringBuffer();
	    Field.toFieldAppend(s, "dist", d.dist.toString());
	    Field.toFieldAppend(s, "time", d.listenTime.toString());
	    Field.toFieldAppend(as, "sample", s.toString());
	}

	Field.toFieldAppend(stat, "array:sample", as.toString());
	Field.toFieldAppend(record, "dist_stat", stat.toString());
	
	Object o = b.beacon.getField(ParseMap.CARTESIAN_COORD);
	if(o != null)
	    record.append(createPosType((CartesianCoord) o));
	
	return Field.toField("dist_est", record);
    }

    public void processOrientData(OrientData newData, BitSet mask)
    {
	send(Field.toField("orient", newData.value.toString()), false);
    }

    private void send(String r, boolean fragment) 
    {
	String frag = "0";
	if(fragment)
	    frag = "1";

	r = CRICKETSERVER_VER+"|"+String.valueOf(System.currentTimeMillis())
	    +"|"+frag+"|"+r+"\n";
	try {
	    byte[] result = r.getBytes(DEFAULT_ENCODING);

	    //System.out.println(result.length);

	    if(result != null)
		notifyClients(result);		
	}
	catch (UnsupportedEncodingException e) {
	    Util.assert(false, "ServerModule::createSpaceByteArray() "+
			"System does not support "+DEFAULT_ENCODING+
			" encoding");
	}
    }
}



