/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.space;

import cricketdaemon.Util;
import java.util.*;

/** 
 * Holds the distance distributions from a single beacon. 
 * Implements the mean, mode methods.
 *
 * @author Allen Miu, Bodhi Priyantha
 */
public class BeaconDistanceEstimator {

    protected String name;      // Beacon unique identify
    protected String space;
    protected int beaconID;
    protected Hashtable distanceDistribution;

    // caches the distance value after analysis
    private double meanDistance = -1;
    private double modeDistance = -1;

    public double compValue;

    /**
     * Constructor, Called with at least one distance value !!
     */
    BeaconDistanceEstimator(String s, int bID, int value)
    {
	distanceDistribution = new Hashtable();
	space = s;
	bID   = beaconID;
	name  = space+"-"+beaconID;
	insert(value);
    }

    /**
     * Inserts a distance value the distance to the beacon is rounded
     * acoording to PRECISION.  If the nuber of data points for a given
     * distance reaches ENOUGHPOINTS, then return value is true.  
     */
    public boolean insert(int dist)
    {
	// Tells if at least one distance has reached threshold frequency
	boolean enough_data = false;

	dist /= SpaceModule.PRECISION; // rounds down 
	dist *= SpaceModule.PRECISION; // the value

	Integer distance  = new Integer(dist);
	Object  frequency = distanceDistribution.get(distance);
	Integer newFreq   = new Integer(1);
	if(frequency != null) {  
	    // increment frequency
	    newFreq = 
		new Integer(((Integer) frequency).intValue() + 1);
	    
	    distanceDistribution.put(distance, newFreq);
	} 
	else { // insert a new distance and a frequency
	    distanceDistribution.put(distance, newFreq);
	}
	
	//count++; // adjusts total data points
	enough_data = newFreq.intValue() >= SpaceModule.ENOUGHPOINTS;

	return enough_data;
    }

    public DistanceEstimate distance(int meanMode)
    {
	double result = -1.0;
	if(meanMode == SpaceModule.MEAN)
	    result = meanDistance = mean();
	else
	    result = modeDistance = mode();
	
	if(result >= 0.0)
	    return new DistanceEstimate( space, beaconID, null, result );
	else 
	    return null;
    }
	
    /**
     * Returns the weighted mean of distance distribution.
     */
    private double mean()
    {
	Util.log("SpaceBeaconReading::mean() "+name);
	int total_dis    = 0;
	int total_sample = 0;

	// compute weighted average
	Enumeration keys = distanceDistribution.keys();
	while(keys.hasMoreElements()) {
	    Integer Dist = (Integer) keys.nextElement();
	    int dist = Dist.intValue();
	    int freq = ((Integer) distanceDistribution.get(Dist)).intValue();
	    total_dis    += freq*dist;
	    total_sample += freq;
	}
	//System.out.println(total_dis+" "+total_sample);
	if(total_sample >= SpaceModule.MEANTHRESH)
	    return (double) total_dis/total_sample;
	else
	    return -1.0;
    }

    /**
     * Returns the mode of distance distribution.
     */
    private double mode() 
    {
	Util.log("SpaceBeaconReading::mode() "+name);
	int mode_freq = 0;
	int mode_dist = 0; // running sum of modes (in case of multiple modes) 
	int mode_count= 0; // used to get the mean in case of multiple modes
	
	Enumeration keys = distanceDistribution.keys();
	while(keys.hasMoreElements()) {
	    Integer Dist = (Integer) keys.nextElement();
	    int dist = Dist.intValue();
	    int freq = ((Integer) distanceDistribution.get(Dist)).intValue();
	    if ( mode_freq < freq ) {  // a new mode
		mode_freq = freq;
		mode_dist = dist;
		mode_count = 1;
	    } 
	    else if ( mode_freq == freq) { // mutiple modes
		mode_dist += dist;
		mode_count++;
	    } // else do nothing
	}

	if (mode_freq >= SpaceModule.MODETHRESH)
	    // mode_count cannot be zero since mode_freq test
	    return (double) mode_dist/mode_count;
	else 
	    return -1.0;
    }

    public String toString() {
	return name;
    }
}

