/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.space;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import java.util.*;

/**
 * The data type that is passed to higher layer module that registers a
 * callback with the Space module.
 *
 * @author Allen Miu
 */
public class SpaceData implements ModuleThread.Data
{
    private static final int MAX_SEQNUM = 65536;
    private static int globalSeqNum = 0;
    public int seqNum;

    public long timeStamp;
    DistanceEstimate[][] values;
    
    public SpaceData()
    {
	seqNum = globalSeqNum;
	globalSeqNum++;             // ensures unique sequence number
	globalSeqNum %= MAX_SEQNUM; // wrap around
	timeStamp = System.currentTimeMillis();
	values = new DistanceEstimate[3][]; // zero is null
    }

    public interface SpaceDataHandler {
	public void processSpaceData(SpaceData newData, BitSet fieldMask);
    }

    public void process(Object handler, BitSet fieldMask) 
    {
	Util.assert(handler instanceof SpaceDataHandler,
		    "SpaceData::process() incorrect handler type");
	SpaceDataHandler h = (SpaceDataHandler) handler;
	h.processSpaceData(this, fieldMask);
    }
    
    public void insert(DistanceEstimate[] value, int field)
    {
	if(field < 0 || field >= values.length || values[field] != null) {
	    // if value already exists, do a null operation to
	    // prevent modules from tampering with beacon data.
	    Util.warn("SpaceData::insertValue() seqNum="+seqNum+
		      " values["+field+"] silently ignored");
	    return;
	}
	values[field] = value;
    }

    public DistanceEstimate[] getMean()
    {
	return values[SpaceModule.MEAN];
    }
    
    public DistanceEstimate[] getMode()
    {
	return values[SpaceModule.MODE];
    }
    
    public String toString()
    {
	String result = "SpaceData:\n";
	if(values[SpaceModule.MEAN] != null) {
	    result += "  List of sorted mean Beacon Distances:  ";
	    for(int i = 0; i < values[SpaceModule.MEAN].length; i++)
		result += "("+values[SpaceModule.MEAN][i]+") ";
	}
	result += "\n";
	if(values[SpaceModule.MODE] != null) {
	    result += "  List of sorted mode Beacon Distances:  ";
	    for(int i = 0; i < values[SpaceModule.MODE].length; i++)
		result += "("+values[SpaceModule.MODE][i]+") ";
	}
	return result;
    }
}

