/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.space2;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.*;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.reader.parse.FieldParser_CartesianCoord.*;
import cricketdaemon.space.*;
import java.util.*;

/**
 * A module that implements an incremental inference algorithm to identify the
 * set of nearest beacons.
 *
 * A sliding window version of this algorithm is being implemented
 * as BeaconStatModule.  We will phase out the Space* modules soon.
 *
 * @author Allen Miu, Bodhi Priyantha
 */
public class Space2Module extends ModuleThread implements Beacon.BeaconHandler
{
    private static BitSet spaceFieldMask;

    /* Space field types */
    // see Space field types

    /* space inference parameters */

    // distance measurement precision in inches
    public final static int PRECISION = 5;

    private Hashtable tuplesMap;   // used to increment counter

    public Space2Module(ModuleThread reader)
    {
	super("Space2Module", 0);

	/* registering with the "port" reader */
	BitSet desiredTypeMask = new BitSet();
	desiredTypeMask.set(ParseMap.SPACE);
	desiredTypeMask.set(ParseMap.BEACONID);
	desiredTypeMask.set(ParseMap.DISTANCE);
	desiredTypeMask.set(ParseMap.CARTESIAN_COORD);

	reader.register(this, desiredTypeMask);

	tuplesMap = new Hashtable();
    }

    // protected void process(); // using the inherited process method

    protected void moduleCleanUp() { return; }

    public void processBeacon(Beacon newData, BitSet fieldMask)
    {
	ArrayList newEstimates = null;

	if(!fieldMask.get(ParseMap.SPACE)) {
	    Util.log("Space2Module::process() space is null");
	    return; // silently drops
	}
	if(!fieldMask.get(ParseMap.DISTANCE)) {
	    Util.log("Space2Module::process() distance is null");
	    return; // silently drops
	}
	String space      = (String)  newData.getField(ParseMap.SPACE);
	Integer distance  = (Integer) newData.getField(ParseMap.DISTANCE);

	Integer beaconID = null;
	if(fieldMask.get(ParseMap.BEACONID))
	    beaconID  = (Integer) newData.getField(ParseMap.BEACONID);
	else
	    beaconID = new Integer(0);

	CartesianCoord coord = null;
	if(fieldMask.get(ParseMap.CARTESIAN_COORD))
	    coord =
		(CartesianCoord) newData.getField(ParseMap.CARTESIAN_COORD);
	else
	    Util.log("Space2Module::process() coord is null");
	
	newEstimates =
	    insert(newData.timeStamp, 
		   space, 
		   beaconID.intValue(), 
		   distance.intValue(), 
		   coord);
	
	/* send space data */
	if(newEstimates != null) {
	    spaceFieldMask = (BitSet) CLEARMASK.clone();
	    spaceFieldMask.set(SpaceModule.MODE);
	    SpaceData newOutData = new SpaceData();
	    newOutData.insert(SpaceModule.sort(newEstimates),
			      SpaceModule.MODE);
	    transfer(newOutData, spaceFieldMask);
	}
	
	SpaceBeaconTuple.garbageCollect();
    }

    // return !null iff a tuple migrates to a count that is currently the
    // maximum

   /**
     * Inserts a new distance value read from the PortReader.
     * Returns true if there are enough full data points.
     */
    public ArrayList insert(long timeStamp,
			    String space, int id, int dist, 
			    CartesianCoord coord)
    {
	double remainder = dist % PRECISION;

	dist /= PRECISION; // rounds down 
	dist *= PRECISION; // the value

	// round up by using simple convention, ignore 
	// special case when remainder == PRECISION/2.0
	// and decide to round up or down depending
	// if the second digit is even or odd
	if(remainder >= (double) PRECISION/2.0)
	    dist += PRECISION;

	// FIXME: coord and timeStamp are extraneous to algorithm?
	SpaceBeaconTuple newTuple = 
	    new SpaceBeaconTuple(timeStamp, space, id, dist, coord);

	SpaceBeaconTuple existingTuple = 
	    (SpaceBeaconTuple) tuplesMap.get(newTuple);

	boolean shouldUpdate = false;
	
	if(existingTuple != null) {
	    // if in table, just increment
	    shouldUpdate = existingTuple.incr();
	    existingTuple.timeStamp = timeStamp; // update timestamp
	    if(coord != null)
		Util.assert(coord.equals(existingTuple.coord),
			    "Space2Module::insert() coords not equal");
	    else
		Util.assert(existingTuple.coord == null,
			    "Space2Module::insert() coords not equal");
	}
	else { // if not in table, insert the new SpaceBeaconTuple
	    tuplesMap.put(newTuple, newTuple);
	    shouldUpdate = newTuple.incr();
	}

	if(shouldUpdate) {
	    return SpaceBeaconTuple.getHighestCountEstimates();
	}
	else
	    return null;
    }
}
