/*
 * Copyright (c) 2000, 2001 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CRICKETAUTHORS which should have accompanied 
 * this software for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CRICKETAUTHORS file.
 */
package cricketdaemon.space2;

import cricketdaemon.Util;
import cricketdaemon.space.*;
import cricketdaemon.reader.*;
import java.util.*;

public class SpaceBeaconTuple extends DistanceEstimate
{
    // The minimum frequency of occurence for a valied mode
    final static int THRESH = 2;
    final static int TIMEOUT = 2500;

    public int count;
    public long timeout;

    public static int highestCount;

    static private ArrayList tuplesQueue; // used to timeout tuples
    static private ArrayList[] tuplesDistribution; // distribution

    public CartesianCoord coord;
    public long timeStamp;

    static {
	tuplesQueue = new ArrayList();
	tuplesDistribution = new ArrayList[THRESH];
	for(int i = 0; i < THRESH; i++)
	    tuplesDistribution[i] = new ArrayList();
	highestCount = 0;
    }    

    public SpaceBeaconTuple(long t, String space, int id, int dist, 
			    CartesianCoord c)
    {
	super(space, id, null, dist);
	timeStamp = t;
	coord = c;
    }

    public static void dumpDataStructures()
    {
	System.out.println("\nhighestCount: "+highestCount);
	System.out.print("tuplesQueue: ");
	for(int i = 0; i < tuplesQueue.size(); i++)
	    System.out.print(tuplesQueue.get(i)+" ");
	System.out.println("\ntuplesDistribution");
	for(int i = 0; i < THRESH; i++) {
	    System.out.print("  "+(i+1)+":");
	    for(int j = 0; j < tuplesDistribution[i].size(); j++)
		System.out.print(tuplesDistribution[i].get(j)+" ");
	    System.out.print("\n");
	}
	System.out.println("");
    }


    public String toString() 
    {
	return super.toString()+"["+count+", "+timeout+"]";
    }
    
    public SpaceBeaconTuple(String s, int b, int d)
    {
	super(s, b, null, d);
	count = 0;
    }
    
    public boolean incr()
    {
	boolean result = false;
	// fixme: inefficient!
	// here, we deQueue and enQueue all the time to update
	// the timeout value, even when count == THRESH
	deQueue();

	// This is added based on Hari's suggestion: Put the new beacon
	// in the highest level.
	//if( count == 0 )
	//count = THRESH;
	//	else 
	if( count < THRESH )
	    count++;
	result = enQueue();
	Util.assert(count <= THRESH,
		    "SpaceBeaconTuple::incr count is too high");
	return result; //count >= Space2Module.THRESH;
    }
    
    private boolean decr()
    {
	if( count > 0 ) {
	    deQueue();
	    count--;
	    enQueue();
	}
	return count <= 0;
    }
    
    private boolean enQueue()
    {
	if(count <= 0)
	    return false;
	
	Util.assert(!tuplesQueue.contains(this),
		    "SpaceBeaconTuple::enQueue() tuplesQueue contains this");
	timeout = System.currentTimeMillis() + TIMEOUT;
	tuplesQueue.add(this);
	Util.assert(!tuplesDistribution[count-1].contains(this),
		    "SpaceBeaconTuple::enQueue() tuplesQueue contains this");
	tuplesDistribution[count-1].add(this);

	if(count >= highestCount) {// need equality part for indicating new
	    highestCount = count;  // members among the highest count group
	    return true;
	}
	return false;
    }
    
    private void deQueue()
    {
	int i = tuplesQueue.indexOf(this);
	if(i >= 0)
	    tuplesQueue.remove(i);
	removeFromDistribution();
    }

    private void removeFromDistribution()
    {
	if(count <= 0)
	    return;
	int i = tuplesDistribution[count-1].indexOf(this);
	if(i >= 0)
	    tuplesDistribution[count-1].remove(i);
	// recalibrate
	for(int j = THRESH-1; j >= 0 ; j--) {
	    if(tuplesDistribution[j].size() > 0) {
		highestCount = j+1;
		break;
	    }
	}
    }
    
    public static void garbageCollect() 
    {
	long curTime = System.currentTimeMillis();
	//System.out.println("GC:::: "+curTime);
	//dumpDataStructures();
	while( tuplesQueue.size() > 0 ) {
	    SpaceBeaconTuple t = (SpaceBeaconTuple) tuplesQueue.get(0);
	    if(t.timeout <= curTime)
		t.decr(); // either removes or reQueues t
	    else
		break;
	}
	//	dumpDataStructures();
    }

    public static ArrayList getHighestCountEstimates() 
    {
	//for(int i = 0; i < tuplesDistribution.length; i++)
	//    System.out.println(i+": "+tuplesDistribution[i]);
	ArrayList result = uniqueFilter(tuplesDistribution[highestCount-1]);
	return result;
    }

    private static class BeaconTupleHashWrapper 
    {
	SpaceBeaconTuple item;

	BeaconTupleHashWrapper(SpaceBeaconTuple t) { item = t; }
	
	public int hashCode() 
	{
	    int result = item.space.hashCode()+item.beaconIDInteger.hashCode();
	    return result;
	}

	public boolean equals(Object o) 
	{
	    return hashCode() == o.hashCode();
	}
    }


    // FIXME: if we are to decouple space2 and "filter", then we need
    // to figure out where the first "sorting" (distance) should take place.
    private static ArrayList uniqueFilter(ArrayList list) 
    {
	// filter out estimates with null coordinates
	// filter out older tuples
	Hashtable table = new Hashtable();
	
	for(int i = 0; i < list.size(); i++) {
	    SpaceBeaconTuple s = (SpaceBeaconTuple) list.get(i);
	    BeaconTupleHashWrapper newKey = new BeaconTupleHashWrapper(s);
	    SpaceBeaconTuple existingItem = 
		(SpaceBeaconTuple) table.get(newKey);
	    if(existingItem != null) {
		// compare timestamps

		// Hari's suggestion:  keep the shortest distance
		if(existingItem.timeStamp > s.timeStamp) {
		//if(existingItem.distance > s.distance) {
		    table.remove(newKey);
		    table.put(newKey, s);
		}
	    }
	    else
		table.put(newKey, s);
	}
	    
	if(table.size() <= 0)
	    return null;

	// now, "table" contains a list of unique and latest tuples from
	// each beacon within listening range.  Copy the contents into
	// a SpaceBeaconTuple[]
	ArrayList result = new ArrayList(table.size());
	Enumeration tuples = table.elements();
	for(int i = 0; i < table.size(); i++) {
	    SpaceBeaconTuple t = (SpaceBeaconTuple) tuples.nextElement();
	    result.add(t);
	}

	// FIXME: sort before returning?
	return result;
    }    
}
//
// get standard deviation
//   reflects "consistency" in distance measurement
// get timestamp
// get frequency
