/**
 * SinkWidget.java
 *
 * Widget called by ControllerGUI for viewing and selecting sink
 * proxies.  
 */

package streamredirector.gui;

import streamredirector.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 *
 * @author  Kalpak Kothari
 */

public class SinkWidget extends JFrame implements ActionListener {

    /** this variable sets whether or not debug println's are printed */
    protected static final boolean DEBUG = Controller.DEBUG;

    // Variables declaration
    private Controller controller;
    private Vector sinks;
    private INSTCPSinkProxyDescriptor currentSinkProxy = null;
    private INSTCPSinkProxyDescriptor selectedSinkProxy = null;
    private JCheckBox otherSinkCheckBox = null;
    private int currentRow = -1;
    private int selectedRow = -1;
    
    // GUI variables
    private JScrollPane jScrollPane1;
    private JTable sinkViewTable;
    private JScrollPane jScrollPane2;
    private JTextArea sinkDetailTextArea;
    private JPanel buttonPanel;
    private JButton applyButton;
    private JButton closeButton;
    private JButton refreshButton;
    private final static Color selectedRowColor = new Color(135,109,255);
    // End of variables declaration

    /** Creates new form SinkWidget */
    public SinkWidget(Controller conn) {
	this.controller = conn;
        initComponents();
    }

    /* Constructor that takes a ControllerGUI checkbox as
     * parameter. Its label and tooltip text are modified when the
     * "Apply" button is pressed.
     * */
    public SinkWidget(Controller conn, JCheckBox checkBox) {
	this.controller = conn;
	this.otherSinkCheckBox = checkBox;
        initComponents();
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     */
    private void initComponents() {
        setTitle("Sink Selector");
        getContentPane().setLayout(new GridBagLayout());

        jScrollPane1 = new JScrollPane();	
        jScrollPane2 = new JScrollPane();
        sinkDetailTextArea = new JTextArea();
	buttonPanel = new JPanel();
	applyButton = new JButton("Apply");
	closeButton = new JButton("Close");
	refreshButton = new JButton("Refresh");

        GridBagConstraints gridBagConstraints5;
	
	// create Sink Viewer Table
	SinkTableModel sinkModel = new SinkTableModel();
        sinkViewTable = new JTable(sinkModel);
        
	jScrollPane1.setPreferredSize(new Dimension(200, 160));
        jScrollPane1.setBorder(new TitledBorder(null, "Sink Viewer", 
						TitledBorder.CENTER, 
						TitledBorder.DEFAULT_POSITION));
        jScrollPane1.setToolTipText("View information of available sinks");
        jScrollPane1.setAutoscrolls(true);

        sinkViewTable.setFont(new Font ("Dialog", 0, 10));
        sinkViewTable.setBackground(Color.lightGray);
        sinkViewTable.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
	sinkViewTable.setPreferredScrollableViewportSize(new Dimension(200, 180));
	sinkViewTable.setRowSelectionAllowed(true);

	// set Table in single row select mode.
	sinkViewTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	// add selection listener to get row selection
	ListSelectionModel rowSM = sinkViewTable.getSelectionModel();
	rowSM.addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    //Ignore extra messages.
                    if (e.getValueIsAdjusting()) return;
                    
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (lsm.isSelectionEmpty()) {
			if(DEBUG)
			    System.out.println("SinkWidget::No rows are selected.");
                    } else {
			// get row index and assign currentSinkProxy with new selection
                        int selRow = lsm.getMinSelectionIndex();
			if(DEBUG) {
			    System.out.println("SinkWidget::Row " + selRow
					       + " is now selected.");
			}
			currentSinkProxy = (INSTCPSinkProxyDescriptor)sinks.elementAt(selRow);
			currentRow = selRow;
			printSinkDetails();
			applyButton.setEnabled(true);
                    }
                }
            });

	// mouselistener for receiving double-click events.
	MouseAdapter sinkViewTableMouseListener = new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
		    if (e.getClickCount() == 2) {
			applyButtonActionPerformed();
		    }
		}
	    };
	
	sinkViewTable.addMouseListener(sinkViewTableMouseListener);
	
        jScrollPane1.getViewport().add(sinkViewTable, BorderLayout.CENTER);

	// Set up column sizes
        initColumnSizes(sinkViewTable);
        
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 0;
        gridBagConstraints5.anchor = GridBagConstraints.NORTH;
        getContentPane().add(jScrollPane1, gridBagConstraints5);

        jScrollPane2.setPreferredSize(new Dimension(200, 110));
        jScrollPane2.setMinimumSize(new Dimension(200, 110));
        jScrollPane2.setBorder(new TitledBorder(null, "Sink Details", TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));
        jScrollPane2.setViewportBorder(new EtchedBorder());
        jScrollPane2.setAutoscrolls(true);

        sinkDetailTextArea.setLineWrap(true);
	sinkDetailTextArea.setWrapStyleWord(true);
        sinkDetailTextArea.setEditable(false);
        sinkDetailTextArea.setFont(new Font ("Dialog", 0, 10));
        // sinkDetailTextArea.setText("Sink Details here...");
        jScrollPane2.setViewportView(sinkDetailTextArea);
        
        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 1;
        getContentPane().add(jScrollPane2, gridBagConstraints5);

	refreshButton.setFont(new Font ("Dialog", 0, 10));
        refreshButton.setToolTipText("Refresh sink list");
	refreshButton.setMnemonic(KeyEvent.VK_R);
	refreshButton.setActionCommand("refreshbutton");
	refreshButton.addActionListener(this);
	buttonPanel.add(refreshButton);

	applyButton.setFont(new Font ("Dialog", 0, 10));
        applyButton.setToolTipText("Apply selection");
	applyButton.setMnemonic(KeyEvent.VK_A);
	applyButton.setEnabled(false);
	applyButton.setActionCommand("applybutton");
	applyButton.addActionListener(this);
	buttonPanel.add(applyButton);

	closeButton.setFont(new Font ("Dialog", 0, 10));
        closeButton.setToolTipText("Close Window");
	closeButton.setMnemonic(KeyEvent.VK_C);
	closeButton.setActionCommand("closebutton");
	closeButton.addActionListener(this);
	buttonPanel.add(closeButton);

        gridBagConstraints5 = new GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 2;
	getContentPane().add(buttonPanel, gridBagConstraints5);

        pack();
    }

    private void applyButtonActionPerformed() {
	// set selected sink in controllerGUI
	if(currentSinkProxy != null) {
	    this.selectedSinkProxy = currentSinkProxy;
	    selectedRow = currentRow;
	}
	else {
	    // default to first sink
	    // this.selectedSinkProxy = (INSTCPSinkProxyDescriptor)sinks.elementAt(0); 
	    // selectedRow = 0;
	    return;
	}
	if(otherSinkCheckBox != null)
	    setControllerGUIOtherSinkCheckBox();
	applyButton.setEnabled(false);
    }

    public void actionPerformed(ActionEvent e) {
	// GUI event handler
	if(e.getActionCommand().equals("applybutton")) {
	    applyButtonActionPerformed();
        }
	else if(e.getActionCommand().equals("refreshbutton")) {
	    // refresh list of sinks in table
	    SinkTableModel sinkModel = new SinkTableModel();
	    sinkViewTable.setModel(sinkModel);
	    // Set up column sizes
	    initColumnSizes(sinkViewTable);
	}
	else if(e.getActionCommand().equals("closebutton")) {
	    this.hide();
	}
    }

    public INSTCPSinkProxyDescriptor getSelectedSinkProxy() {
	// returns the selected sink's descriptor
	return selectedSinkProxy;
    }
    
    public void setSelectedSinkProxy(INSTCPSinkProxyDescriptor setProxy) {
	// sets <setProxy> as the selected sink <selectedSinkProxy>
	int selRow = sinks.indexOf(setProxy);
	if(selRow == -1) {
	    if(DEBUG)
		System.out.println("SinkWidget::setSelectedSinkProxy - Error: Sink Proxy does not exist.\n");
	    return;
	}
	selectedSinkProxy = setProxy;
	sinkViewTable.setRowSelectionInterval(selRow, selRow);
    }

    public void setControllerGUIOtherSinkCheckBox() {
        otherSinkCheckBox.setToolTipText("Play resource at location: " + 
					 selectedSinkProxy.getLocation());
        otherSinkCheckBox.setText("Other: " + selectedSinkProxy.getLocation());
    }

    public void printSinkDetails() {
	// print sink proxy descriptor details in sinkDetailTextArea
	sinkDetailTextArea.setText("Location  : " + currentSinkProxy.getLocation() + "\n");
	sinkDetailTextArea.append( "Hostname: " + currentSinkProxy.getHostName() + "\n");
	sinkDetailTextArea.append( "Port          : " + currentSinkProxy.getPort() + "\n");
	sinkDetailTextArea.append( "MIME Types: ");

	try {
	    String[] mimeTypes = currentSinkProxy.getAllMIMETypes();
	    for (int i = 0; i < mimeTypes.length; i++) {
		sinkDetailTextArea.append(mimeTypes[i] + "; ");
	    }
	}
	catch(Exception ex) {
	    ex.printStackTrace();
	}	
    }


    /*
     * This method picks good column sizes.
     */
    private void initColumnSizes(JTable table) {
        TableColumn column = null;


	// set row height to allow two lines of text
	table.setRowHeight(32);

	// render Location column cells as JTextArea using LocationRenderer
	table.setDefaultRenderer(String.class,
                                 new LocationRenderer());
	
	// location column
	column = table.getColumnModel().getColumn(0);
	column.setPreferredWidth(120);
	
	// audio column
	column = table.getColumnModel().getColumn(1);
	column.setPreferredWidth(35);
	
	// video column
	column = table.getColumnModel().getColumn(2);
	column.setPreferredWidth(35);

    }
    
    /** 
     * Inner Class: SinkTableModel
     * Description: Table model used by SinkViewTable.
     */
    class SinkTableModel extends AbstractTableModel {
	private final String columnNames[] = { "Location", "Audio", "Video" };
	private Object dataValues[][];

	/** this variable sets whether or not debug println's are printed */
	protected static final boolean DEBUG = Controller.DEBUG;

	public SinkTableModel() {
	    super();
	    // create data values
	    createData();
	}
	
	private void createData() {
	    // create data values by querying Controller for Sinks, asking it to refresh the listing of sink proxies
	    sinks = controller.getAllSinks(NetworkResourceManager.REFRESH);
	    if (sinks.size() < 0)
		dataValues = new Object[][] { {"Unknown", 
					       new Boolean(false), 
					       new Boolean(false)} };
	    else {
		dataValues = new Object[sinks.size()][3];
		int row = 0;
		for (Enumeration e = sinks.elements(); e.hasMoreElements(); ) {
		    INSTCPSinkProxyDescriptor s = (INSTCPSinkProxyDescriptor)(e.nextElement());
		    // add location data
		    dataValues[row][0] = s.getLocation();
		    dataValues[row][1] = new Boolean(false);
		    dataValues[row][2] = new Boolean(false);
		    
		    try {
			// check-mark "audio"/"video" columns if relevant MIME types present
			String[] mimeTypes = s.getAllMIMETypes();
			for (int i = 0; i < mimeTypes.length; i++) {
			    if(mimeTypes[i].startsWith("audio"))
				dataValues[row][1] = new Boolean(true);
			    else if(mimeTypes[i].startsWith("video"))
				dataValues[row][2] = new Boolean(true);
			}
		    }
		    catch(Exception ex) {
			ex.printStackTrace();
		    }
		    row++;
		}
	    }
	}
	

        public int getColumnCount() {
            return columnNames.length;
        }
        
        public int getRowCount() {
            return dataValues.length;
        }

        public String getColumnName(int col) {
            return columnNames[col];
        }

        public Object getValueAt(int row, int col) {
            return dataValues[row][col];
        }



        /*
         * JTable uses this method to determine the default renderer/
         * editor for each cell.  
         */
        public Class getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }

        /*
         * This method specifies all cells as uneditable.
         */
        public boolean isCellEditable(int row, int col) {
	    return false;
        }

        /*
         * This method allows table's data to be changed.
         */
        public void setValueAt(Object value, int row, int col) {
            if (DEBUG) {
                System.out.println("Setting value at " + row + "," + col
                                   + " to " + value
                                   + " (an instance of " 
                                   + value.getClass() + ")");
            }

	    dataValues[row][col] = value;
	    fireTableCellUpdated(row, col);
            if (DEBUG) {
                System.out.println("New value of data:");
                printDebugData();
            }
        }
	
        private void printDebugData() {
            int numRows = getRowCount();
            int numCols = getColumnCount();
	    
            for (int i=0; i < numRows; i++) {
                System.out.print("    row " + i + ":");
                for (int j=0; j < numCols; j++) {
                    System.out.print("  " + dataValues[i][j]);
                }
                System.out.println();
            }
            System.out.println("--------------------------");
        }
    }


    /** 
     * Inner Class: LocationRenderer
     * Description: Cell renderer used to render the Location column
     *              as a JTextArea. 
     */
    class LocationRenderer extends JTextArea
                        implements TableCellRenderer {

        public LocationRenderer() {
	    super();
	    setLineWrap(true);
	    setEditable(false);
	    setFont(new Font ("Dialog", 0, 10));
	    setBackground(Color.lightGray);
	}

        public Component getTableCellRendererComponent(JTable table, Object str, 
						       boolean isSelected, boolean hasFocus,
						       int row, int column) {
	    setText((String)str);

	    //	    int numrows = (int)Math.ceil(((double)((String)str).length()) / 19);
	    //	    if(numrows == 0)
	    //		numrows = 1;
	    //	    table.setRowHeight(row, 16 * numrows);

	    if(row == selectedRow) {
		setFont(new Font ("Dialog", Font.BOLD, 10));
		setForeground(selectedRowColor);
	    }
	    else {
		setFont(new Font ("Dialog", 0, 10));
		setForeground(table.getForeground());
	    }
	    
	    if(isSelected)
		setBackground(table.getSelectionBackground());
	    else
		setBackground(table.getBackground());
	    
	    return this;
	}
    }
}


